@@ -20,12 +20,14 @@ package org.apache.tez.client;
 
 import java.io.File;
 import java.io.IOException;
+import java.io.InputStream;
 import java.net.InetSocketAddress;
 import java.net.URI;
 import java.net.URISyntaxException;
 import java.nio.ByteBuffer;
 import java.security.PrivilegedExceptionAction;
 import java.util.ArrayList;
+import java.util.Arrays;
 import java.util.Collection;
 import java.util.HashMap;
 import java.util.HashSet;
@@ -38,7 +40,9 @@ import java.util.Vector;
 import java.util.Map.Entry;
 
 import com.google.common.base.Strings;
+import org.apache.commons.codec.digest.DigestUtils;
 import org.apache.commons.lang.StringUtils;
+import org.apache.commons.math3.util.Precision;
 import org.apache.tez.common.JavaOptsChecker;
 import org.apache.tez.dag.api.records.DAGProtos.AMPluginDescriptorProto;
 import org.apache.tez.serviceplugins.api.ServicePluginsDescriptor;
@@ -73,6 +77,7 @@ import org.apache.hadoop.yarn.api.records.LocalResourceType;
 import org.apache.hadoop.yarn.api.records.LocalResourceVisibility;
 import org.apache.hadoop.yarn.api.records.Priority;
 import org.apache.hadoop.yarn.api.records.Resource;
+import org.apache.hadoop.yarn.api.records.URL;
 import org.apache.hadoop.yarn.api.records.YarnApplicationState;
 import org.apache.hadoop.yarn.conf.YarnConfiguration;
 import org.apache.hadoop.yarn.exceptions.ApplicationNotFoundException;
@@ -85,8 +90,6 @@ import org.apache.tez.common.TezCommonUtils;
 import org.apache.tez.common.TezYARNUtils;
 import org.apache.tez.common.VersionInfo;
 import org.apache.tez.common.security.ACLManager;
-import org.apache.tez.common.security.HistoryACLPolicyManager;
-import org.apache.tez.common.security.HistoryACLPolicyException;
 import org.apache.tez.common.security.JobTokenIdentifier;
 import org.apache.tez.common.security.JobTokenSecretManager;
 import org.apache.tez.common.security.TokenCache;
@@ -132,7 +135,8 @@ public class TezClientUtils {
 
     Path p = new Path(uri);
     FileSystem fs = p.getFileSystem(conf);
-    p = fs.resolvePath(p);
+    p = fs.resolvePath(p.makeQualified(fs.getUri(),
+        fs.getWorkingDirectory()));
     FileSystem targetFS = p.getFileSystem(conf); 
     if (targetFS.isDirectory(p)) {
       return targetFS.listStatus(p);
@@ -175,37 +179,12 @@ public class TezClientUtils {
 
       LOG.info("Using tez.lib.uris value from configuration: "
           + conf.get(TezConfiguration.TEZ_LIB_URIS));
+      LOG.info("Using tez.lib.uris.classpath value from configuration: "
+          + conf.get(TezConfiguration.TEZ_LIB_URIS_CLASSPATH));
 
-      if (tezJarUris.length == 1 && (
-              tezJarUris[0].endsWith(".tar.gz") ||
-              tezJarUris[0].endsWith(".tgz") ||
-              tezJarUris[0].endsWith(".zip") ||
-              tezJarUris[0].endsWith(".tar"))) {
-        String fileName = tezJarUris[0];
+      usingTezArchive = addLocalResources(conf, tezJarUris,
+                         tezJarResources, credentials);
 
-        FileStatus fStatus = getLRFileStatus(fileName, conf)[0];
-        LocalResourceVisibility lrVisibility;
-        if (checkAncestorPermissionsForAllUsers(conf, fileName, FsAction.EXECUTE) &&
-            fStatus.getPermission().getOtherAction().implies(FsAction.READ)) {
-          lrVisibility = LocalResourceVisibility.PUBLIC;
-        } else {
-          lrVisibility = LocalResourceVisibility.PRIVATE;
-        }
-        tezJarResources.put(TezConstants.TEZ_TAR_LR_NAME,
-            LocalResource.newInstance(
-                ConverterUtils.getYarnUrlFromPath(fStatus.getPath()),
-                LocalResourceType.ARCHIVE,
-                lrVisibility,
-                fStatus.getLen(),
-                fStatus.getModificationTime()));
-        Path[] tezJarPaths = { fStatus.getPath() };
-        // obtain credentials
-        TokenCache.obtainTokensForFileSystems(credentials, tezJarPaths, conf);
-        usingTezArchive = true;
-      } else { // Treat as non-archives
-        addLocalResources(conf, tezJarUris, tezJarResources, credentials);
-      }
-      
       if (tezJarResources.isEmpty()) {
         throw new TezUncheckedException(
             "No files found in locations specified in "
@@ -221,41 +200,101 @@ public class TezClientUtils {
     return usingTezArchive;
   }
 
-  private static void addLocalResources(Configuration conf, String[] configUris,
-      Map<String, LocalResource> tezJarResources, Credentials credentials) throws IOException {
+  private static boolean addLocalResources(Configuration conf,
+      String[] configUris, Map<String, LocalResource> tezJarResources,
+      Credentials credentials) throws IOException {
+    boolean usingTezArchive = false;
     if (configUris == null || configUris.length == 0) {
-      return;
+      return usingTezArchive;
     }
     List<Path> configuredPaths = Lists.newArrayListWithCapacity(configUris.length);
     for (String configUri : configUris) {
-      boolean ancestorsHavePermission = checkAncestorPermissionsForAllUsers(conf, configUri,
-          FsAction.EXECUTE);
+      URI u = null;
+      try {
+        u = new URI(configUri);
+      } catch (URISyntaxException e) {
+        throw new IOException("Unable to convert " + configUri + "to URI", e);
+      }
+      Path p = new Path(u);
+      FileSystem remoteFS = p.getFileSystem(conf);
+      p = remoteFS.resolvePath(p.makeQualified(remoteFS.getUri(),
+          remoteFS.getWorkingDirectory()));
+
+      LocalResourceType type = null;
+
+      //Check if path is an archive
+      if(p.getName().endsWith(".tar.gz") ||
+           p.getName().endsWith(".tgz") ||
+           p.getName().endsWith(".zip") ||
+           p.getName().endsWith(".tar")) {
+          type = LocalResourceType.ARCHIVE;
+        } else {
+          type = LocalResourceType.FILE;
+        }
+
       FileStatus [] fileStatuses = getLRFileStatus(configUri, conf);
+
       for (FileStatus fStatus : fileStatuses) {
+        String linkName;
         if (fStatus.isDirectory()) {
           // Skip directories - no recursive search support.
           continue;
         }
+        // If the resource is an archive, we've already done this work
+        if(type != LocalResourceType.ARCHIVE) {
+          u = fStatus.getPath().toUri();
+          p = new Path(u);
+          remoteFS = p.getFileSystem(conf);
+          p = remoteFS.resolvePath(p.makeQualified(remoteFS.getUri(),
+              remoteFS.getWorkingDirectory()));
+          if(null != u.getFragment()) {
+            LOG.warn("Fragment set for link being interpreted as a file," +
+                "URI: " + u.toString());
+          }
+        }
+
+        // Add URI fragment or just the filename
+        Path name = new Path((null == u.getFragment())
+          ? p.getName()
+          : u.getFragment());
+        if (name.isAbsolute()) {
+          throw new IllegalArgumentException("Resource name must be "
+              + "relative, not absolute: " + name
+              + " in URI: " + u.toString());
+        }
+
+        URL url = ConverterUtils.getYarnUrlFromURI(p.toUri());
+        linkName = name.toUri().getPath();
+        // For legacy reasons, set archive to tezlib if there is
+        // only a single archive and no fragment
+        if(type == LocalResourceType.ARCHIVE &&
+            configUris.length == 1 && null == u.getFragment()) {
+          linkName = TezConstants.TEZ_TAR_LR_NAME;
+          usingTezArchive = true;
+        }
+
         LocalResourceVisibility lrVisibility;
-        if (ancestorsHavePermission &&
+        if (checkAncestorPermissionsForAllUsers(conf, p,
+            FsAction.EXECUTE) &&
             fStatus.getPermission().getOtherAction().implies(FsAction.READ)) {
           lrVisibility = LocalResourceVisibility.PUBLIC;
         } else {
           lrVisibility = LocalResourceVisibility.PRIVATE;
         }
-        String rsrcName = fStatus.getPath().getName();
-        if (tezJarResources.containsKey(rsrcName)) {
+
+        if (tezJarResources.containsKey(linkName)) {
           String message = "Duplicate resource found"
-              + ", resourceName=" + rsrcName
+              + ", resourceName=" + linkName
               + ", existingPath=" +
-              tezJarResources.get(rsrcName).getResource().toString()
+              tezJarResources.get(linkName).getResource().toString()
               + ", newPath=" + fStatus.getPath();
           LOG.warn(message);
         }
-        tezJarResources.put(rsrcName,
+
+        tezJarResources.put(linkName,
             LocalResource.newInstance(
-                ConverterUtils.getYarnUrlFromPath(fStatus.getPath()),
-                LocalResourceType.FILE,
+                url,
+                type,
                 lrVisibility,
                 fStatus.getLen(),
                 fStatus.getModificationTime()));
@@ -267,6 +306,7 @@ public class TezClientUtils {
       TokenCache.obtainTokensForFileSystems(credentials,
           configuredPaths.toArray(new Path[configuredPaths.size()]), conf);
     }
+    return usingTezArchive;
   }
 
   static void processTezLocalCredentialsFile(Credentials credentials, Configuration conf)
@@ -407,7 +447,6 @@ public class TezClientUtils {
    * @param amConfig AM Configuration
    * @param tezJarResources Resources to be used by the AM
    * @param sessionCreds the credential object which will be populated with session specific
-   * @param historyACLPolicyManager
    * @param servicePluginsDescriptor descriptor for services which may be running in the AM
    * @return an ApplicationSubmissionContext to launch a Tez AM
    * @throws IOException
@@ -419,7 +458,7 @@ public class TezClientUtils {
       ApplicationId appId, DAG dag, String amName,
       AMConfiguration amConfig, Map<String, LocalResource> tezJarResources,
       Credentials sessionCreds, boolean tezLrsAsArchive,
-      TezApiVersionInfo apiVersionInfo, HistoryACLPolicyManager historyACLPolicyManager,
+      TezApiVersionInfo apiVersionInfo,
       ServicePluginsDescriptor servicePluginsDescriptor, JavaOptsChecker javaOptsChecker)
       throws IOException, YarnException {
 
@@ -457,6 +496,8 @@ public class TezClientUtils {
     // Add Staging dir creds to the list of session credentials.
     TokenCache.obtainTokensForFileSystems(sessionCreds, new Path[]{binaryConfPath}, conf);
 
+    populateTokenCache(conf, sessionCreds);
+
     // Add session specific credentials to the AM credentials.
     amLaunchCredentials.mergeAll(sessionCreds);
 
@@ -527,48 +568,29 @@ public class TezClientUtils {
     }
     amLocalResources.putAll(tezJarResources);
 
-    // Setup Session ACLs and update conf as needed
-    Map<String, String> aclConfigs = null;
-    if (historyACLPolicyManager != null) {
-      if (dag == null) {
-        try{
-          aclConfigs = historyACLPolicyManager.setupSessionACLs(amConfig.getTezConfiguration(),
-              appId);
-        } catch (HistoryACLPolicyException e) {
-          LOG.warn("Disabling history logging for session " + strAppId +
-                   " due to error in setting up history acls " + e);
-          amConfig.getTezConfiguration().setBoolean(TezConfiguration.TEZ_AM_HISTORY_LOGGING_ENABLED,
-              false);
-        }
-      } else {
-        try{
-          // Non-session mode
-          // As only a single DAG is support, we should combine AM and DAG ACLs under the same
-          // acl management layer
-          aclConfigs = historyACLPolicyManager.setupNonSessionACLs(amConfig.getTezConfiguration(),
-              appId, dag.getDagAccessControls());
-        } catch (HistoryACLPolicyException e) {
-          LOG.warn("Disabling history logging for dag " +
-              dag.getName() + " due to error in setting up history acls " + e);
-          dag.setConf(TezConfiguration.TEZ_DAG_HISTORY_LOGGING_ENABLED, "false");
-          // This is non-session mode so disable logging for whole AM
-          amConfig.getTezConfiguration().setBoolean(TezConfiguration.TEZ_AM_HISTORY_LOGGING_ENABLED,
-              false);
-        }
-      }
+    TezConfiguration tezConf = amConfig.getTezConfiguration();
+    // Merge the dag access controls into tez am config.
+    if (dag != null && dag.getDagAccessControls() != null) {
+      // Merge updates the conf object passed. In non session mode, same client object can be used
+      // to submit multiple dags, copying this prevents ACL of one DAG from being used in another.
+      tezConf = new TezConfiguration(amConfig.getTezConfiguration());
+      dag.getDagAccessControls().mergeIntoAmAcls(tezConf);
     }
 
     // emit conf as PB file
-    ConfigurationProto finalConfProto = createFinalConfProtoForApp(amConfig.getTezConfiguration(),
-        aclConfigs, servicePluginsDescriptor);
-    
-    FSDataOutputStream amConfPBOutBinaryStream = null;
-    try {
-      amConfPBOutBinaryStream = TezCommonUtils.createFileForAM(fs, binaryConfPath);
-      finalConfProto.writeTo(amConfPBOutBinaryStream);
-    } finally {
-      if(amConfPBOutBinaryStream != null){
-        amConfPBOutBinaryStream.close();
+    // don't overwrite existing conf, needed for TezClient.getClient() so existing containers have stable resource fingerprints
+    if(!binaryConfPath.getFileSystem(tezConf).exists(binaryConfPath)) {
+      ConfigurationProto finalConfProto = createFinalConfProtoForApp(tezConf,
+              servicePluginsDescriptor);
+
+      FSDataOutputStream amConfPBOutBinaryStream = null;
+      try {
+        amConfPBOutBinaryStream = TezCommonUtils.createFileForAM(fs, binaryConfPath);
+        finalConfProto.writeTo(amConfPBOutBinaryStream);
+      } finally {
+        if (amConfPBOutBinaryStream != null) {
+          amConfPBOutBinaryStream.close();
+        }
       }
     }
 
@@ -654,7 +676,9 @@ public class TezClientUtils {
     // provide this to AuxServices running on the AM node - in case tasks run within the AM,
     // and no other task runs on this node.
     Map<String, ByteBuffer> serviceData = new HashMap<String, ByteBuffer>();
-    serviceData.put(TezConstants.TEZ_SHUFFLE_HANDLER_SERVICE_ID,
+    String auxiliaryService = conf.get(TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID,
+        TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID_DEFAULT);
+    serviceData.put(auxiliaryService,
         TezCommonUtils.serializeServiceData(TokenCache.getSessionToken(amLaunchCredentials)));
 
     // Setup ContainerLaunchContext for AM container
@@ -666,10 +690,18 @@ public class TezClientUtils {
     ApplicationSubmissionContext appContext = Records
         .newRecord(ApplicationSubmissionContext.class);
 
+    Collection<String> tagsFromConf =
+        amConfig.getTezConfiguration().getTrimmedStringCollection(
+        TezConfiguration.TEZ_APPLICATION_TAGS);
+
     appContext.setApplicationType(TezConstants.TEZ_APPLICATION_TYPE);
+    if (tagsFromConf != null && !tagsFromConf.isEmpty()) {
+      appContext.setApplicationTags(new HashSet<String>(tagsFromConf));
+    }
     appContext.setApplicationId(appId);
     appContext.setResource(capability);
-    if (amConfig.getQueueName() != null) {
+    String queueName = amConfig.getQueueName();
+    if (queueName != null && !queueName.isEmpty()) {
       appContext.setQueue(amConfig.getQueueName());
     }
     // set the application priority
@@ -687,25 +719,31 @@ public class TezClientUtils {
     return appContext;
 
   }
+
+  //get secret keys and tokens and store them into TokenCache
+  private static void populateTokenCache(TezConfiguration conf, Credentials credentials)
+          throws IOException{
+    // add the delegation tokens from configuration
+    String[] nameNodes = conf.getStrings(TezConfiguration.TEZ_JOB_FS_SERVERS);
+    LOG.debug("adding the following namenodes' delegation tokens:" +
+            Arrays.toString(nameNodes));
+    if(nameNodes != null) {
+      Path[] ps = new Path[nameNodes.length];
+      for(int i = 0; i < nameNodes.length; i++) {
+        ps[i] = new Path(nameNodes[i]);
+      }
+      TokenCache.obtainTokensForFileSystems(credentials, ps, conf);
+    }
+  }
   
   static DAGPlan prepareAndCreateDAGPlan(DAG dag, AMConfiguration amConfig,
       Map<String, LocalResource> tezJarResources, boolean tezLrsAsArchive,
       Credentials credentials, ServicePluginsDescriptor servicePluginsDescriptor,
       JavaOptsChecker javaOptsChecker) throws IOException {
-    return prepareAndCreateDAGPlan(dag, amConfig, tezJarResources, tezLrsAsArchive, credentials,
-        null, servicePluginsDescriptor, javaOptsChecker);
-  }
-
-  static DAGPlan prepareAndCreateDAGPlan(DAG dag, AMConfiguration amConfig,
-      Map<String, LocalResource> tezJarResources, boolean tezLrsAsArchive,
-      Credentials credentials, Map<String, String> additionalDAGConfigs,
-      ServicePluginsDescriptor servicePluginsDescriptor,
-      JavaOptsChecker javaOptsChecker) throws IOException {
     Credentials dagCredentials = setupDAGCredentials(dag, credentials,
         amConfig.getTezConfiguration());
     return dag.createDag(amConfig.getTezConfiguration(), dagCredentials, tezJarResources,
-        amConfig.getBinaryConfLR(), tezLrsAsArchive, additionalDAGConfigs, servicePluginsDescriptor,
-        javaOptsChecker);
+        amConfig.getBinaryConfLR(), tezLrsAsArchive, servicePluginsDescriptor, javaOptsChecker);
   }
   
   static void maybeAddDefaultLoggingJavaOpts(String logLevel, List<String> vargs) {
@@ -784,21 +822,19 @@ public class TezClientUtils {
   }
 
   static ConfigurationProto createFinalConfProtoForApp(Configuration amConf,
-    Map<String, String> additionalConfigs, ServicePluginsDescriptor servicePluginsDescriptor) {
+    ServicePluginsDescriptor servicePluginsDescriptor) {
     assert amConf != null;
     ConfigurationProto.Builder builder = ConfigurationProto.newBuilder();
     for (Entry<String, String> entry : amConf) {
-      PlanKeyValuePair.Builder kvp = PlanKeyValuePair.newBuilder();
-      kvp.setKey(entry.getKey());
-      kvp.setValue(amConf.get(entry.getKey()));
-      builder.addConfKeyValues(kvp);
-    }
-    if (additionalConfigs != null && !additionalConfigs.isEmpty()) {
-      for (Entry<String, String> entry : additionalConfigs.entrySet()) {
+      String key = entry.getKey();
+      String val = amConf.get(key);
+      if(val != null) {
         PlanKeyValuePair.Builder kvp = PlanKeyValuePair.newBuilder();
-        kvp.setKey(entry.getKey());
-        kvp.setValue(entry.getValue());
+        kvp.setKey(key);
+        kvp.setValue(val);
         builder.addConfKeyValues(kvp);
+      } else {
+        LOG.debug("null value in Configuration after replacement for key={}. Skipping.", key);
       }
     }
 
@@ -960,9 +996,16 @@ public class TezClientUtils {
         || (resource.getMemory() <= 0)) {
       return javaOpts;
     }
-    if (maxHeapFactor <= 0 || maxHeapFactor >= 1) {
+
+    if ((maxHeapFactor <= 0 && !Precision.equals(maxHeapFactor, -1, 0.01)) || maxHeapFactor >= 1) {
       return javaOpts;
     }
+
+    if (Precision.equals(maxHeapFactor, -1, 0.01)) {
+      maxHeapFactor = resource.getMemory() < TezConstants.TEZ_CONTAINER_SMALL_SLAB_BOUND_MB
+        ? TezConstants.TEZ_CONTAINER_MAX_JAVA_HEAP_FRACTION_SMALL_SLAB
+        : TezConstants.TEZ_CONTAINER_MAX_JAVA_HEAP_FRACTION_LARGE_SLAB;
+    }
     int maxMemory = (int)(resource.getMemory() * maxHeapFactor);
     maxMemory = maxMemory <= 0 ? 1 : maxMemory;
 
@@ -970,9 +1013,8 @@ public class TezClientUtils {
         + ( javaOpts != null ? javaOpts : "");
   }
 
-  private static boolean checkAncestorPermissionsForAllUsers(Configuration conf, String uri,
+  private static boolean checkAncestorPermissionsForAllUsers(Configuration conf, Path pathComponent,
                                                              FsAction permission) throws IOException {
-    Path pathComponent = new Path(uri);
     FileSystem fs = pathComponent.getFileSystem(conf);
 
     if (Shell.WINDOWS && fs instanceof LocalFileSystem) {
@@ -1067,4 +1109,28 @@ public class TezClientUtils {
     }
   }
 
+
+  public static byte[] getLocalSha(Path path, Configuration conf) throws IOException {
+    InputStream is = null;
+    try {
+      is = FileSystem.getLocal(conf).open(path);
+      return DigestUtils.sha256(is);
+    } finally {
+      if (is != null) {
+        is.close();
+      }
+    }
+  }
+
+  public static byte[] getResourceSha(URI uri, Configuration conf) throws IOException {
+    InputStream is = null;
+    try {
+      is = FileSystem.get(uri, conf).open(new Path(uri));
+      return DigestUtils.sha256(is);
+    } finally {
+      if (is != null) {
+        is.close();
+      }
+    }
+  }
 }
