@@ -29,6 +29,7 @@ import java.util.List;
 import java.util.Map;
 import java.util.Set;
 import java.util.TreeSet;
+import java.util.zip.Inflater;
 import java.util.Map.Entry;
 
 import javax.annotation.Nullable;
@@ -50,6 +51,7 @@ import org.apache.tez.common.TezCommonUtils;
 import org.apache.tez.common.counters.CounterGroup;
 import org.apache.tez.common.counters.TezCounter;
 import org.apache.tez.common.counters.TezCounters;
+import org.apache.tez.common.security.DAGAccessControls;
 import org.apache.tez.dag.api.EdgeProperty.DataMovementType;
 import org.apache.tez.dag.api.EdgeProperty.DataSourceType;
 import org.apache.tez.dag.api.EdgeProperty.SchedulingType;
@@ -57,6 +59,7 @@ import org.apache.tez.dag.api.Vertex.VertexExecutionContext;
 import org.apache.tez.dag.api.client.StatusGetOpts;
 import org.apache.tez.dag.api.client.rpc.DAGClientAMProtocolRPC.TezAppMasterStatusProto;
 import org.apache.tez.dag.api.records.DAGProtos;
+import org.apache.tez.dag.api.records.DAGProtos.ACLInfo;
 import org.apache.tez.dag.api.records.DAGProtos.AMPluginDescriptorProto;
 import org.apache.tez.dag.api.records.DAGProtos.CallerContextProto;
 import org.apache.tez.dag.api.records.DAGProtos.ConfigurationProto;
@@ -367,12 +370,12 @@ public class DagTypeConverters {
     return builder.build();
   }
 
-  public static String getHistoryTextFromProto(TezEntityDescriptorProto proto) {
+  public static String getHistoryTextFromProto(TezEntityDescriptorProto proto, Inflater inflater) {
     if (!proto.hasHistoryText()) {
       return null;
     }
     try {
-      return new String(TezCommonUtils.decompressByteStringToByteArray(proto.getHistoryText()),
+      return new String(TezCommonUtils.decompressByteStringToByteArray(proto.getHistoryText(), inflater),
           "UTF-8");
     } catch (IOException e) {
       throw new TezUncheckedException(e);
@@ -732,7 +735,7 @@ public class DagTypeConverters {
     if (payload == null) {
       return null;
     }
-    return payload.getPayload();
+    return payload.getRawPayload();
   }
 
   public static VertexExecutionContextProto convertToProto(
@@ -863,4 +866,27 @@ public class DagTypeConverters {
     return callerContext;
   }
 
+  public static ACLInfo convertDAGAccessControlsToProto(DAGAccessControls dagAccessControls) {
+    if (dagAccessControls == null) {
+      return null;
+    }
+    ACLInfo.Builder builder = ACLInfo.newBuilder();
+    builder.addAllUsersWithViewAccess(dagAccessControls.getUsersWithViewACLs());
+    builder.addAllUsersWithModifyAccess(dagAccessControls.getUsersWithModifyACLs());
+    builder.addAllGroupsWithViewAccess(dagAccessControls.getGroupsWithViewACLs());
+    builder.addAllGroupsWithModifyAccess(dagAccessControls.getGroupsWithModifyACLs());
+    return builder.build();
+  }
+
+  public static DAGAccessControls convertDAGAccessControlsFromProto(ACLInfo aclInfo) {
+    if (aclInfo == null) {
+      return null;
+    }
+    DAGAccessControls dagAccessControls = new DAGAccessControls();
+    dagAccessControls.setUsersWithViewACLs(aclInfo.getUsersWithViewAccessList());
+    dagAccessControls.setUsersWithModifyACLs(aclInfo.getUsersWithModifyAccessList());
+    dagAccessControls.setGroupsWithViewACLs(aclInfo.getGroupsWithViewAccessList());
+    dagAccessControls.setGroupsWithModifyACLs(aclInfo.getGroupsWithModifyAccessList());
+    return dagAccessControls;
+  }
 }
