@@ -19,6 +19,7 @@
 package org.apache.tez.dag.api.client;
 
 import javax.annotation.Nullable;
+
 import java.io.IOException;
 import java.text.DecimalFormat;
 import java.util.Collections;
@@ -28,8 +29,9 @@ import java.util.Map;
 import java.util.Set;
 
 import com.google.common.annotations.VisibleForTesting;
-
 import com.google.common.base.Preconditions;
+
+import org.apache.hadoop.yarn.exceptions.ApplicationNotFoundException;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 import org.apache.hadoop.classification.InterfaceAudience.Private;
@@ -58,7 +60,7 @@ public class DAGClientImpl extends DAGClient {
   private final FrameworkClient frameworkClient;
 
   @VisibleForTesting
-  protected DAGClient realClient;
+  protected DAGClientInternal realClient;
   private boolean dagCompleted = false;
   @VisibleForTesting
   protected boolean isATSEnabled = false;
@@ -74,27 +76,28 @@ public class DAGClientImpl extends DAGClient {
       VertexStatus.State.ERROR);
   private long statusPollInterval;
   private long diagnoticsWaitTimeout;
+  private boolean cleanupFrameworkClient;
 
   public DAGClientImpl(ApplicationId appId, String dagId, TezConfiguration conf,
-                       @Nullable FrameworkClient frameworkClient) {
+      YarnConfiguration yarnConf, @Nullable FrameworkClient frameworkClient) {
     this.appId = appId;
     this.dagId = dagId;
     this.conf = conf;
-    if (frameworkClient != null &&
-        conf.getBoolean(TezConfiguration.TEZ_LOCAL_MODE, TezConfiguration.TEZ_LOCAL_MODE_DEFAULT)) {
+    if (frameworkClient != null) {
       this.frameworkClient = frameworkClient;
     } else {
       this.frameworkClient = FrameworkClient.createFrameworkClient(conf);
-      this.frameworkClient.init(conf, new YarnConfiguration(conf));
+      this.frameworkClient.init(conf, yarnConf);
       this.frameworkClient.start();
+      cleanupFrameworkClient = true;
     }
     isATSEnabled = conf.get(TezConfiguration.TEZ_HISTORY_LOGGING_SERVICE_CLASS, "")
-            .equals("org.apache.tez.dag.history.logging.ats.ATSHistoryLoggingService") &&
-            conf.getBoolean(TezConfiguration.TEZ_DAG_HISTORY_LOGGING_ENABLED,
-                 TezConfiguration.TEZ_DAG_HISTORY_LOGGING_ENABLED_DEFAULT) &&
-            conf.getBoolean(TezConfiguration.TEZ_AM_HISTORY_LOGGING_ENABLED,
-                 TezConfiguration.TEZ_AM_HISTORY_LOGGING_ENABLED_DEFAULT) &&
-            DAGClientTimelineImpl.isSupported();
+        .equals("org.apache.tez.dag.history.logging.ats.ATSHistoryLoggingService") &&
+        conf.getBoolean(TezConfiguration.TEZ_DAG_HISTORY_LOGGING_ENABLED,
+            TezConfiguration.TEZ_DAG_HISTORY_LOGGING_ENABLED_DEFAULT) &&
+        conf.getBoolean(TezConfiguration.TEZ_AM_HISTORY_LOGGING_ENABLED,
+            TezConfiguration.TEZ_AM_HISTORY_LOGGING_ENABLED_DEFAULT) &&
+        DAGClientTimelineImpl.isSupported();
 
     realClient = new DAGClientRPCImpl(appId, dagId, conf, this.frameworkClient);
     statusPollInterval = conf.getLong(
@@ -240,6 +243,9 @@ public class DAGClientImpl extends DAGClient {
         if (dagStatus.isCompleted()) {
           return dagStatus;
         }
+      } catch (ApplicationNotFoundException e) {
+        LOG.info("Failed to fetch DAG data for completed DAG from YARN Timeline"
+            + " - Application not found by YARN", e);
       } catch (TezException e) {
         if (LOG.isDebugEnabled()) {
           LOG.info("DAGStatus fetch failed." + e.getMessage());
@@ -291,6 +297,10 @@ public class DAGClientImpl extends DAGClient {
         if (vertexCompletionStates.contains(vertexStatus.getState())) {
           return vertexStatus;
         }
+      } catch (ApplicationNotFoundException e) {
+        LOG.info("Failed to fetch Vertex data for completed DAG from YARN Timeline"
+            + " - Application not found by YARN", e);
+        return null;
       } catch (TezException e) {
         if (LOG.isDebugEnabled()) {
           LOG.debug("ERROR fetching vertex data from Yarn Timeline. " + e.getMessage());
@@ -308,6 +318,16 @@ public class DAGClientImpl extends DAGClient {
     return null;
   }
 
+  @Override
+  public String getDagIdentifierString() {
+    return realClient.getDagIdentifierString();
+  }
+
+  @Override
+  public String getSessionIdentifierString() {
+    return realClient.getSessionIdentifierString();
+  }
+
   @Override
   public void tryKillDAG() throws IOException, TezException {
     if (!dagCompleted) {
@@ -317,22 +337,27 @@ public class DAGClientImpl extends DAGClient {
     }
   }
 
+  @Override
+  public DAGStatus waitForCompletion(long timeMs) throws IOException, TezException, InterruptedException {
+    return _waitForCompletionWithStatusUpdates(timeMs, false, EnumSet.noneOf(StatusGetOpts.class));
+  }
+
   @Override
   public DAGStatus waitForCompletion() throws IOException, TezException, InterruptedException {
-    return _waitForCompletionWithStatusUpdates(false, EnumSet.noneOf(StatusGetOpts.class));
+    return _waitForCompletionWithStatusUpdates(-1, false, EnumSet.noneOf(StatusGetOpts.class));
   }
 
   @Override
   public DAGStatus waitForCompletionWithStatusUpdates(
       @Nullable Set<StatusGetOpts> statusGetOpts) throws IOException, TezException,
       InterruptedException {
-    return _waitForCompletionWithStatusUpdates(true, statusGetOpts);
+    return _waitForCompletionWithStatusUpdates(-1, true, statusGetOpts);
   }
 
   @Override
   public void close() throws IOException {
     realClient.close();
-    if (frameworkClient != null) {
+    if (frameworkClient != null && cleanupFrameworkClient) {
       frameworkClient.stop();
     }
   }
@@ -350,6 +375,10 @@ public class DAGClientImpl extends DAGClient {
     try {
       dagStatus = realClient.getDAGStatus(statusOptions, timeout);
     } catch (DAGNotRunningException e) {
+      LOG.info("DAG is no longer running", e);
+      dagCompleted = true;
+    } catch (ApplicationNotFoundException e) {
+      LOG.info("DAG is no longer running - application not found by YARN", e);
       dagCompleted = true;
     } catch (TezException e) {
       // can be either due to a n/w issue of due to AM completed.
@@ -370,6 +399,10 @@ public class DAGClientImpl extends DAGClient {
     try {
       vertexStatus = realClient.getVertexStatus(vertexName, statusOptions);
     } catch (DAGNotRunningException e) {
+      LOG.info("DAG is no longer running", e);
+      dagCompleted = true;
+    } catch (ApplicationNotFoundException e) {
+      LOG.info("DAG is no longer running - application not found by YARN", e);
       dagCompleted = true;
     } catch (TezException e) {
       // can be either due to a n/w issue of due to AM completed.
@@ -398,6 +431,9 @@ public class DAGClientImpl extends DAGClient {
     ApplicationReport appReport;
     try {
       appReport = frameworkClient.getApplicationReport(appId);
+    } catch (ApplicationNotFoundException e) {
+      LOG.info("DAG is no longer running - application not found by YARN", e);
+      throw new DAGNotRunningException(e);
     } catch (YarnException e) {
       throw new TezException(e);
     }
@@ -473,15 +509,21 @@ public class DAGClientImpl extends DAGClient {
     return dagStatus;
   }
 
-  private DAGStatus _waitForCompletionWithStatusUpdates(boolean vertexUpdates,
+  private DAGStatus _waitForCompletionWithStatusUpdates(long timeMs,
+                                                        boolean vertexUpdates,
                                                         @Nullable Set<StatusGetOpts> statusGetOpts) throws IOException, TezException, InterruptedException {
     DAGStatus dagStatus;
     boolean initPrinted = false;
     boolean runningPrinted = false;
     double dagProgress = -1.0; // Print the first one
     // monitoring
+    Long maxNs = timeMs >= 0 ? (System.nanoTime() + (timeMs * 1000000L)) : null;
     while (true) {
-      dagStatus = getDAGStatus(statusGetOpts, SLEEP_FOR_COMPLETION);
+      try {
+        dagStatus = getDAGStatus(statusGetOpts, SLEEP_FOR_COMPLETION);
+      } catch (DAGNotRunningException ex) {
+        return null;
+      }
       if (!initPrinted
           && (dagStatus.getState() == DAGStatus.State.INITING || dagStatus.getState() == DAGStatus.State.SUBMITTED)) {
         initPrinted = true; // Print once
@@ -494,6 +536,9 @@ public class DAGClientImpl extends DAGClient {
           || dagStatus.getState() == DAGStatus.State.ERROR) {
         break;
       }
+      if (maxNs != null && System.nanoTime() > maxNs) {
+        return null;
+      }
     }// End of while(true)
 
     Set<String> vertexNames = Collections.emptySet();
@@ -506,7 +551,14 @@ public class DAGClientImpl extends DAGClient {
         vertexNames = getDAGStatus(statusGetOpts).getVertexProgress().keySet();
       }
       dagProgress = monitorProgress(vertexNames, dagProgress, null, dagStatus);
-      dagStatus = getDAGStatus(statusGetOpts, SLEEP_FOR_COMPLETION);
+      try {
+        dagStatus = getDAGStatus(statusGetOpts, SLEEP_FOR_COMPLETION);
+      } catch (DAGNotRunningException ex) {
+        return null;
+      }
+      if (maxNs != null && System.nanoTime() > maxNs) {
+        return null;
+      }
     }// end of while
     // Always print the last status irrespective of progress change
     monitorProgress(vertexNames, -1.0, statusGetOpts, dagStatus);
@@ -592,7 +644,7 @@ public class DAGClientImpl extends DAGClient {
   }
 
   @VisibleForTesting
-  public DAGClient getRealClient() {
+  public DAGClientInternal getRealClient() {
     return realClient;
   }
 
