@@ -37,6 +37,7 @@ import com.sun.jersey.api.client.ClientHandlerException;
 import com.sun.jersey.api.client.ClientResponse;
 import com.sun.jersey.api.client.UniformInterfaceException;
 import com.sun.jersey.api.client.WebResource;
+
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 import org.apache.hadoop.classification.InterfaceAudience.Private;
@@ -63,7 +64,7 @@ import org.codehaus.jettison.json.JSONObject;
 
 
 @Private
-public class DAGClientTimelineImpl extends DAGClient {
+public class DAGClientTimelineImpl extends DAGClientInternal {
   private static final Logger LOG = LoggerFactory.getLogger(DAGClientTimelineImpl.class);
 
   private static final String FILTER_BY_FIELDS = "primaryfilters,otherinfo";
@@ -179,6 +180,16 @@ public class DAGClientTimelineImpl extends DAGClient {
     }
   }
 
+  @Override
+  public String getDagIdentifierString() {
+    return dagId;
+  }
+
+  @Override
+  public String getSessionIdentifierString() {
+    return appId.toString();
+  }
+
   @Override
   public void tryKillDAG() throws IOException, TezException {
     throw new TezException("tryKillDAG is unsupported for DAGClientTimelineImpl");
