@@ -22,7 +22,6 @@ import static org.junit.Assert.assertFalse;
 import static org.junit.Assert.assertNotNull;
 import static org.junit.Assert.assertNull;
 import static org.junit.Assert.assertTrue;
-import static org.mockito.Mockito.mock;
 
 import java.io.File;
 import java.io.FileNotFoundException;
@@ -32,7 +31,9 @@ import java.net.URLClassLoader;
 import java.nio.ByteBuffer;
 import java.util.ArrayList;
 import java.util.Arrays;
+import java.util.Collection;
 import java.util.HashMap;
+import java.util.HashSet;
 import java.util.LinkedList;
 import java.util.List;
 import java.util.Map;
@@ -59,7 +60,6 @@ import org.apache.hadoop.yarn.api.records.Resource;
 import org.apache.hadoop.yarn.conf.YarnConfiguration;
 import org.apache.hadoop.yarn.exceptions.YarnException;
 import org.apache.hadoop.yarn.util.Records;
-import org.apache.tez.common.security.HistoryACLPolicyManager;
 import org.apache.tez.common.security.JobTokenIdentifier;
 import org.apache.tez.common.security.JobTokenSecretManager;
 import org.apache.tez.common.security.TokenCache;
@@ -131,7 +131,7 @@ public class TestTezClientUtils {
   /**
    *
    */
-  @Test (timeout=5000)
+  @Test (timeout=10000)
   public void validateSetTezJarLocalResourcesDefinedExistingDirectory() throws Exception {
     URL[] cp = ((URLClassLoader)ClassLoader.getSystemClassLoader()).getURLs();
     StringBuffer buffer = new StringBuffer();
@@ -207,6 +207,122 @@ public class TestTezClientUtils {
     assertFalse(localizedMap.isEmpty());
   }
 
+
+    /**
+   *
+   */
+  @Test (timeout=5000)
+  public void testTezDefaultFS() throws Exception {
+    FileSystem localFs = FileSystem.getLocal(new Configuration());
+    StringBuilder tezLibUris = new StringBuilder();
+
+    Path topDir = new Path(TEST_ROOT_DIR, "testTezDefaultFS");
+    if (localFs.exists(topDir)) {
+      localFs.delete(topDir, true);
+    }
+    localFs.mkdirs(topDir);
+
+    Path file = new Path(topDir, "file.jar");
+
+    Assert.assertTrue(localFs.createNewFile(file));
+    tezLibUris.append(localFs.makeQualified(file).toString());
+
+    TezConfiguration conf = new TezConfiguration();
+    conf.set(TezConfiguration.TEZ_LIB_URIS, tezLibUris.toString());
+    conf.set("fs.defaultFS", "hdfs:///localhost:1234");
+    Credentials credentials = new Credentials();
+    Map<String, LocalResource> localizedMap = new HashMap<String, LocalResource>();
+    TezClientUtils.setupTezJarsLocalResources(conf, credentials, localizedMap);
+
+    Assert.assertTrue(localFs.delete(file, true));
+    Assert.assertTrue(localFs.delete(topDir, true));
+  }
+
+    /**
+   *
+   */
+  @Test (timeout=5000)
+  public void validateSetTezJarLocalResourcesMultipleTarballs() throws Exception {
+    FileSystem localFs = FileSystem.getLocal(new Configuration());
+    StringBuilder tezLibUris = new StringBuilder();
+
+    // Create 2 files
+    Path topDir = new Path(TEST_ROOT_DIR, "validatemultipletarballs");
+    if (localFs.exists(topDir)) {
+      localFs.delete(topDir, true);
+    }
+    localFs.mkdirs(topDir);
+
+    Path tarFile1 = new Path(topDir, "f1.tar.gz");
+    Path tarFile2 = new Path(topDir, "f2.tar.gz");
+
+    Assert.assertTrue(localFs.createNewFile(tarFile1));
+    Assert.assertTrue(localFs.createNewFile(tarFile2));
+    tezLibUris.append(localFs.makeQualified(tarFile1).toString()).append("#tar1").append(",");
+    tezLibUris.append(localFs.makeQualified(tarFile2).toString()).append("#tar2").append(",");
+
+    TezConfiguration conf = new TezConfiguration();
+    conf.set(TezConfiguration.TEZ_LIB_URIS, tezLibUris.toString());
+    Credentials credentials = new Credentials();
+    Map<String, LocalResource> localizedMap = new HashMap<String, LocalResource>();
+    TezClientUtils.setupTezJarsLocalResources(conf, credentials, localizedMap);
+    Set<String> resourceNames = localizedMap.keySet();
+    Assert.assertEquals(2, resourceNames.size());
+    Assert.assertTrue(resourceNames.contains("tar1"));
+    Assert.assertTrue(resourceNames.contains("tar2"));
+    Assert.assertFalse(resourceNames.contains("f1.tar.gz"));
+    Assert.assertFalse(resourceNames.contains("f2.tar.gz"));
+
+
+    Assert.assertTrue(localFs.delete(tarFile1, true));
+    Assert.assertTrue(localFs.delete(tarFile2, true));
+    Assert.assertTrue(localFs.delete(topDir, true));
+  }
+
+    /**
+   *
+   */
+  @Test (timeout=5000)
+  public void validateSetTezJarLocalResourcesMixTarballAndJar() throws Exception {
+    FileSystem localFs = FileSystem.getLocal(new Configuration());
+    StringBuilder tezLibUris = new StringBuilder();
+
+    // Create 2 jars and 1 archive
+    Path topDir = new Path(TEST_ROOT_DIR, "validatetarballandjar");
+    if (localFs.exists(topDir)) {
+      localFs.delete(topDir, true);
+    }
+    localFs.mkdirs(topDir);
+
+    Path tarFile1 = new Path(topDir, "f1.tar.gz");
+    Path jarFile2 = new Path(topDir, "f2.jar");
+    Path jarFile3 = new Path(topDir, "f3.jar");
+
+    Assert.assertTrue(localFs.createNewFile(tarFile1));
+    Assert.assertTrue(localFs.createNewFile(jarFile2));
+    Assert.assertTrue(localFs.createNewFile(jarFile3));
+
+    tezLibUris.append(localFs.makeQualified(topDir).toString()).append(",");
+    tezLibUris.append(localFs.makeQualified(tarFile1).toString()).append("#tar1").append(",");
+
+    TezConfiguration conf = new TezConfiguration();
+    conf.set(TezConfiguration.TEZ_LIB_URIS, tezLibUris.toString());
+    Credentials credentials = new Credentials();
+    Map<String, LocalResource> localizedMap = new HashMap<String, LocalResource>();
+    TezClientUtils.setupTezJarsLocalResources(conf, credentials, localizedMap);
+    Set<String> resourceNames = localizedMap.keySet();
+    Assert.assertEquals(4, resourceNames.size());
+    Assert.assertTrue(resourceNames.contains("tar1"));
+    Assert.assertTrue(resourceNames.contains("f1.tar.gz"));
+    Assert.assertTrue(resourceNames.contains("f2.jar"));
+    Assert.assertTrue(resourceNames.contains("f3.jar"));
+
+    Assert.assertTrue(localFs.delete(tarFile1, true));
+    Assert.assertTrue(localFs.delete(jarFile2, true));
+    Assert.assertTrue(localFs.delete(jarFile3, true));
+    Assert.assertTrue(localFs.delete(topDir, true));
+  }
+
   @Test(timeout = 2000)
   // this test checks if the priority field is set properly in the
   // ApplicationSubmissionContext
@@ -227,7 +343,7 @@ public class TestTezClientUtils {
         appId, null, "dagname",
         amConf, m,
         credentials, false,
-        new TezApiVersionInfo(), null, null, null);
+        new TezApiVersionInfo(), null, null);
     assertEquals(testpriority, appcontext.getPriority().getPriority());
   }
 
@@ -243,6 +359,21 @@ public class TestTezClientUtils {
     assertNull(appContext.getPriority());
   }
 
+  @Test(timeout=1000)
+  public void testSetApplicationTags() {
+    TezConfiguration conf = new TezConfiguration(false);
+    conf.set(TezConfiguration.TEZ_APPLICATION_TAGS, "foo,bar");
+    AMConfiguration amconfig = new AMConfiguration(conf, null, null);
+    ApplicationSubmissionContext appContext = Records
+        .newRecord(ApplicationSubmissionContext.class);
+    Collection<String> tagsFromConf =
+        amconfig.getTezConfiguration().getTrimmedStringCollection(
+        TezConfiguration.TEZ_APPLICATION_TAGS);
+    appContext.setApplicationTags(new HashSet<String>(tagsFromConf));
+    assertTrue(appContext.getApplicationTags().contains("foo"));
+    assertTrue(appContext.getApplicationTags().contains("bar"));
+  }
+
   @Test(timeout = 5000)
   public void testSessionTokenInAmClc() throws IOException, YarnException {
 
@@ -264,7 +395,7 @@ public class TestTezClientUtils {
     ApplicationSubmissionContext appSubmissionContext =
         TezClientUtils.createApplicationSubmissionContext(appId, dag, "amName", amConf,
             new HashMap<String, LocalResource>(), credentials, false, new TezApiVersionInfo(),
-            mock(HistoryACLPolicyManager.class), null, null);
+            null, null);
 
     ContainerLaunchContext amClc = appSubmissionContext.getAMContainerSpec();
     Map<String, ByteBuffer> amServiceData = amClc.getServiceData();
@@ -297,7 +428,7 @@ public class TestTezClientUtils {
     ApplicationSubmissionContext appSubmissionContext =
         TezClientUtils.createApplicationSubmissionContext(appId, dag, "amName", amConf,
             new HashMap<String, LocalResource>(), credentials, false, new TezApiVersionInfo(),
-            mock(HistoryACLPolicyManager.class), null, null);
+            null, null);
 
     List<String> expectedCommands = new LinkedList<String>();
     expectedCommands.add("-Dlog4j.configuratorClass=org.apache.tez.common.TezLog4jConfigurator");
@@ -337,7 +468,7 @@ public class TestTezClientUtils {
     ApplicationSubmissionContext appSubmissionContext =
         TezClientUtils.createApplicationSubmissionContext(appId, dag, "amName", amConf,
             new HashMap<String, LocalResource>(), credentials, false, new TezApiVersionInfo(),
-            mock(HistoryACLPolicyManager.class), null, null);
+            null, null);
 
     List<String> expectedCommands = new LinkedList<String>();
     expectedCommands.add("-Dlog4j.configuratorClass=org.apache.tez.common.TezLog4jConfigurator");
@@ -498,6 +629,16 @@ public class TestTezClientUtils {
     javaOpts = TezClientUtils.maybeAddDefaultMemoryJavaOpts(origJavaOpts,
         Resource.newInstance(355, 1), 100);
     Assert.assertEquals(origJavaOpts, javaOpts);
+
+    origJavaOpts = "";
+    javaOpts = TezClientUtils.maybeAddDefaultMemoryJavaOpts(origJavaOpts,
+        Resource.newInstance(1000, 1), -1);
+    Assert.assertTrue(javaOpts.contains("-Xmx700m"));
+
+    origJavaOpts = "";
+    javaOpts = TezClientUtils.maybeAddDefaultMemoryJavaOpts(origJavaOpts,
+        Resource.newInstance(5000, 1), -1);
+    Assert.assertTrue(javaOpts.contains("-Xmx4000m"));
   }
 
   @Test (timeout=5000)
@@ -524,7 +665,7 @@ public class TestTezClientUtils {
     expected.put("property1", val1);
     expected.put("property2", expVal2);
 
-    ConfigurationProto confProto = TezClientUtils.createFinalConfProtoForApp(conf, null, null);
+    ConfigurationProto confProto = TezClientUtils.createFinalConfProtoForApp(conf, null);
 
     for (PlanKeyValuePair kvPair : confProto.getConfKeyValuesList()) {
       String v = expected.remove(kvPair.getKey());
@@ -628,7 +769,7 @@ public class TestTezClientUtils {
       srcConf.set(entry.getKey(), entry.getValue());
     }
 
-    ConfigurationProto confProto = TezClientUtils.createFinalConfProtoForApp(srcConf, null, null);
+    ConfigurationProto confProto = TezClientUtils.createFinalConfProtoForApp(srcConf, null);
 
     for (PlanKeyValuePair kvPair : confProto.getConfKeyValuesList()) {
       String val = confMap.remove(kvPair.getKey());
@@ -690,7 +831,7 @@ public class TestTezClientUtils {
     Configuration conf = new Configuration(false);
     ServicePluginsDescriptor servicePluginsDescriptor = ServicePluginsDescriptor.create(true);
 
-    ConfigurationProto confProto = TezClientUtils.createFinalConfProtoForApp(conf, null,
+    ConfigurationProto confProto = TezClientUtils.createFinalConfProtoForApp(conf,
         servicePluginsDescriptor);
 
     assertTrue(confProto.hasAmPluginDescriptor());
