@@ -33,6 +33,11 @@ import org.apache.commons.collections4.ListUtils;
 import org.apache.hadoop.yarn.event.Event;
 import org.apache.tez.Utils;
 import org.apache.tez.dag.api.NamedEntityDescriptor;
+import org.apache.tez.dag.app.dag.event.TaskAttemptEventSubmitted;
+import org.apache.tez.runtime.api.TaskFailureType;
+import org.apache.tez.runtime.api.events.TaskAttemptKilledEvent;
+import org.apache.tez.serviceplugins.api.DagInfo;
+import org.apache.tez.serviceplugins.api.ServicePluginError;
 import org.apache.tez.serviceplugins.api.TaskCommunicator;
 import org.apache.tez.dag.api.TezConstants;
 import org.apache.tez.dag.api.UserPayload;
@@ -135,9 +140,9 @@ public class TaskCommunicatorManager extends AbstractService implements
     this.context = context;
     this.taskHeartbeatHandler = thh;
     this.containerHeartbeatHandler = chh;
-    Preconditions.checkArgument(
-        taskCommunicatorDescriptors != null && !taskCommunicatorDescriptors.isEmpty(),
-        "TaskCommunicators must be specified");
+    if (taskCommunicatorDescriptors == null || taskCommunicatorDescriptors.isEmpty()) {
+      throw new IllegalArgumentException("TaskCommunicators must be specified");
+    }
     this.taskCommunicators = new TaskCommunicatorWrapper[taskCommunicatorDescriptors.size()];
     this.taskCommunicatorContexts = new TaskCommunicatorContext[taskCommunicatorDescriptors.size()];
     this.taskCommunicatorServiceWrappers = new ServicePluginLifecycleAbstractService[taskCommunicatorDescriptors.size()];
@@ -275,7 +280,8 @@ public class TaskCommunicatorManager extends AbstractService implements
           taskAttemptEvent = new TaskAttemptEventStatusUpdate(taskAttemptID,
               (TaskStatusUpdateEvent) tezEvent.getEvent());
         } else if (eventType == EventType.TASK_ATTEMPT_COMPLETED_EVENT
-           || eventType == EventType.TASK_ATTEMPT_FAILED_EVENT) {
+           || eventType == EventType.TASK_ATTEMPT_FAILED_EVENT
+           || eventType == EventType.TASK_ATTEMPT_KILLED_EVENT) {
           taFinishedEvents.add(tezEvent);
         } else {
           if (eventType == EventType.INPUT_READ_ERROR_EVENT) {
@@ -304,6 +310,7 @@ public class TaskCommunicatorManager extends AbstractService implements
         EventMetaData sourceMeta = e.getSourceInfo();
         switch (e.getEventType()) {
         case TASK_ATTEMPT_FAILED_EVENT:
+        case TASK_ATTEMPT_KILLED_EVENT:
           TaskAttemptTerminationCause errCause = null;
           switch (sourceMeta.getEventGenerator()) {
           case INPUT:
@@ -322,12 +329,19 @@ public class TaskCommunicatorManager extends AbstractService implements
             throw new TezUncheckedException("Unknown EventProducerConsumerType: " +
                 sourceMeta.getEventGenerator());
           }
-          TaskAttemptFailedEvent taskFailedEvent =(TaskAttemptFailedEvent) e.getEvent();
-          sendEvent(
-               new TaskAttemptEventAttemptFailed(sourceMeta.getTaskAttemptID(),
-                   TaskAttemptEventType.TA_FAILED,
-                  "Error: " + taskFailedEvent.getDiagnostics(),
+          if (e.getEventType() == EventType.TASK_ATTEMPT_FAILED_EVENT) {
+            TaskAttemptFailedEvent taskFailedEvent = (TaskAttemptFailedEvent) e.getEvent();
+            sendEvent(
+                new TaskAttemptEventAttemptFailed(sourceMeta.getTaskAttemptID(),
+                    TaskAttemptEventType.TA_FAILED, taskFailedEvent.getTaskFailureType(),
+                    "Error: " + taskFailedEvent.getDiagnostics(),
                     errCause));
+          } else { // Killed
+            TaskAttemptKilledEvent taskKilledEvent = (TaskAttemptKilledEvent) e.getEvent();
+            sendEvent(
+                new TaskAttemptEventAttemptKilled(sourceMeta.getTaskAttemptID(),
+                    "Error: " + taskKilledEvent.getDiagnostics(), errCause));
+          }
           break;
         case TASK_ATTEMPT_COMPLETED_EVENT:
           sendEvent(
@@ -361,11 +375,15 @@ public class TaskCommunicatorManager extends AbstractService implements
     pingContainerHeartbeatHandler(containerId);
   }
 
-  public void taskStartedRemotely(TezTaskAttemptID taskAttemptID, ContainerId containerId) {
-    sendEvent(new TaskAttemptEventStartedRemotely(taskAttemptID, containerId, null));
+  public void taskSubmitted(TezTaskAttemptID taskAttemptId, ContainerId containerId) {
+    sendEvent(new TaskAttemptEventSubmitted(taskAttemptId, containerId));
     pingContainerHeartbeatHandler(containerId);
   }
 
+  public void taskStartedRemotely(TezTaskAttemptID taskAttemptID) {
+    sendEvent(new TaskAttemptEventStartedRemotely(taskAttemptID));
+  }
+
   public void taskKilled(TezTaskAttemptID taskAttemptId, TaskAttemptEndReason taskAttemptEndReason,
                          String diagnostics) {
     // Regular flow via TaskAttempt will take care of un-registering from the heartbeat handler,
@@ -378,15 +396,17 @@ public class TaskCommunicatorManager extends AbstractService implements
         taskAttemptEndReason)));
   }
 
-  public void taskFailed(TezTaskAttemptID taskAttemptId, TaskAttemptEndReason taskAttemptEndReason,
+  public void taskFailed(TezTaskAttemptID taskAttemptId, TaskFailureType taskFailureType,
+                         TaskAttemptEndReason taskAttemptEndReason,
                          String diagnostics) {
     // Regular flow via TaskAttempt will take care of un-registering from the heartbeat handler,
     // and messages from the scheduler will release the container.
     // TODO TEZ-2003 (post) TEZ-2671 Maybe consider un-registering here itself, since the task is not active anymore,
     // instead of waiting for the unregister to flow through the Container.
     // Fix along the same lines as TEZ-2124 by introducing an explict context.
+    //TODO-3183. Allow the FailureType to be specified
     sendEvent(new TaskAttemptEventAttemptFailed(taskAttemptId,
-        TaskAttemptEventType.TA_FAILED, diagnostics, TezUtilsInternal.fromTaskAttemptEndReason(
+        TaskAttemptEventType.TA_FAILED, taskFailureType, diagnostics, TezUtilsInternal.fromTaskAttemptEndReason(
         taskAttemptEndReason)));
   }
 
@@ -593,6 +613,30 @@ public class TaskCommunicatorManager extends AbstractService implements
     return taskCommunicators[taskCommIndex];
   }
 
+  @Override
+  public void reportError(int taskCommIndex, ServicePluginError servicePluginError,
+                          String diagnostics,
+                          DagInfo dagInfo) {
+    if (servicePluginError.getErrorType() == ServicePluginError.ErrorType.PERMANENT) {
+      String msg = "Fatal Error reported by TaskCommunicator"
+          + ", communicator=" + Utils.getTaskCommIdentifierString(taskCommIndex, context)
+          + ", servicePluginError=" + servicePluginError
+          + ", diagnostics= " + (diagnostics == null ? "" : diagnostics);
+      LOG.error(msg + ", Diagnostics=" + diagnostics);
+      sendEvent(
+          new DAGAppMasterEventUserServiceFatalError(
+              DAGAppMasterEventType.TASK_COMMUNICATOR_SERVICE_FATAL_ERROR,
+              msg, null));
+    } else {
+      Utils
+          .processNonFatalServiceErrorReport(
+              Utils.getTaskCommIdentifierString(taskCommIndex, context), servicePluginError,
+              diagnostics,
+              dagInfo, context,
+              "TaskCommunicator");
+    }
+  }
+
   private void pingContainerHeartbeatHandler(ContainerId containerId) {
     containerHeartbeatHandler.pinged(containerId);
   }
@@ -611,4 +655,34 @@ public class TaskCommunicatorManager extends AbstractService implements
   private void sendEvent(Event<?> event) {
     context.getEventHandler().handle(event);
   }
+
+  @Override
+  public String getTaskCommunicatorClassName(int taskCommId) {
+    return taskCommunicators[taskCommId].getTaskCommunicator().getClass().getName();
+  }
+
+  @Override
+  public String getInProgressLogsUrl(int taskCommId, TezTaskAttemptID attemptID, NodeId containerNodeId) {
+    try {
+      return taskCommunicators[taskCommId].getInProgressLogsUrl(attemptID, containerNodeId);
+    } catch (Exception e) {
+      LOG.warn("Failed to retrieve InProgressLogsUrl from TaskCommunicator,"
+          + ", communicator=" + Utils.getTaskCommIdentifierString(taskCommId, context)
+          + ", attemptId=" + attemptID, e);
+    }
+    return null;
+  }
+
+  @Override
+  public String getCompletedLogsUrl(int taskCommId, TezTaskAttemptID attemptID, NodeId containerNodeId) {
+    try {
+      return taskCommunicators[taskCommId].getCompletedLogsUrl(attemptID, containerNodeId);
+    } catch (Exception e) {
+      LOG.warn("Failed to retrieve CompletedLogsUrl from TaskCommunicator,"
+          + ", communicator=" + Utils.getTaskCommIdentifierString(taskCommId, context)
+          + ", attemptId=" + attemptID, e);
+    }
+    return null;
+  }
+
 }
