@@ -123,7 +123,7 @@ public class TezTaskCommunicatorImpl extends TaskCommunicator {
     super(taskCommunicatorContext);
     this.taskUmbilical = new TezTaskUmbilicalProtocolImpl();
     this.tokenIdentifier = taskCommunicatorContext.getApplicationAttemptId().getApplicationId().toString();
-    this.sessionToken = TokenCache.getSessionToken(taskCommunicatorContext.getCredentials());
+    this.sessionToken = TokenCache.getSessionToken(taskCommunicatorContext.getAMCredentials());
     try {
       conf = TezUtils.createConfFromUserPayload(getContext().getInitialUserPayload());
     } catch (IOException e) {
@@ -221,9 +221,8 @@ public class TezTaskCommunicatorImpl extends TaskCommunicator {
                                          int priority) {
 
     ContainerInfo containerInfo = registeredContainers.get(containerId);
-    Preconditions.checkNotNull(containerInfo,
-        "Cannot register task attempt: " + taskSpec.getTaskAttemptID() + " to unknown container: " +
-            containerId);
+    Preconditions.checkNotNull(containerInfo, "Cannot register task attempt %s to unknown container %s",
+        taskSpec.getTaskAttemptID(), containerId);
     synchronized (containerInfo) {
       if (containerInfo.taskSpec != null) {
         throw new TezUncheckedException(
@@ -315,8 +314,8 @@ public class TezTaskCommunicatorImpl extends TaskCommunicator {
         }
         task = getContainerTask(containerId);
         if (task != null && !task.shouldDie()) {
-          getContext()
-              .taskStartedRemotely(task.getTaskSpec().getTaskAttemptID(), containerId);
+          getContext().taskSubmitted(task.getTaskSpec().getTaskAttemptID(), containerId);
+          getContext().taskStartedRemotely(task.getTaskSpec().getTaskAttemptID());
         }
       }
       if (LOG.isDebugEnabled()) {
@@ -410,7 +409,7 @@ public class TezTaskCommunicatorImpl extends TaskCommunicator {
 
   private ContainerTask getContainerTask(ContainerId containerId) throws IOException {
     ContainerInfo containerInfo = registeredContainers.get(containerId);
-    ContainerTask task = null;
+    ContainerTask task;
     if (containerInfo == null) {
       if (getContext().isKnownContainer(containerId)) {
         LOG.info("Container with id: " + containerId
@@ -422,6 +421,7 @@ public class TezTaskCommunicatorImpl extends TaskCommunicator {
       task = TASK_FOR_INVALID_JVM;
     } else {
       synchronized (containerInfo) {
+        getContext().containerAlive(containerId);
         if (containerInfo.taskSpec != null) {
           if (!containerInfo.taskPulled) {
             containerInfo.taskPulled = true;
