@@ -22,6 +22,7 @@ import java.io.IOException;
 import java.net.URL;
 import java.security.PrivilegedExceptionAction;
 import java.util.ArrayList;
+import java.util.BitSet;
 import java.util.Collection;
 import java.util.Collections;
 import java.util.EnumSet;
@@ -43,7 +44,7 @@ import org.apache.commons.lang.StringUtils;
 import org.apache.commons.lang.exception.ExceptionUtils;
 import org.apache.tez.common.TezUtilsInternal;
 import org.apache.tez.common.counters.LimitExceededException;
-import org.apache.tez.dag.app.dag.event.DAGEventInternalError;
+import org.apache.tez.dag.app.dag.event.DAGEventTerminateDag;
 import org.apache.tez.dag.app.dag.event.DiagnosableEvent;
 import org.apache.tez.state.OnStateChangedCallback;
 import org.apache.tez.state.StateMachineTez;
@@ -190,6 +191,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
   volatile Map<TezVertexID, Vertex> vertices = new HashMap<TezVertexID, Vertex>();
   @VisibleForTesting
   Map<String, Edge> edges = new HashMap<String, Edge>();
+  ArrayList<BitSet> vertexDescendants;
   private TezCounters dagCounters = new TezCounters();
   private Object fullCountersLock = new Object();
   @VisibleForTesting
@@ -198,7 +200,12 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
   private long cachedCountersTimestamp = 0;
   private Set<TezVertexID> reRunningVertices = new HashSet<TezVertexID>();
 
+  // Combined configs for the DAG
   private final Configuration dagConf;
+  // DAG specific configs only
+  // Useful when trying to serialize only the diff from global configs
+  private final Configuration dagOnlyConf;
+
   private final DAGPlan jobPlan;
 
   private final AtomicBoolean internalErrorTriggered = new AtomicBoolean(false);
@@ -253,8 +260,8 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
               EnumSet.of(DAGState.INITED, DAGState.FAILED),
               DAGEventType.DAG_INIT,
               new InitTransition())
-          .addTransition(DAGState.NEW, DAGState.KILLED,
-              DAGEventType.DAG_KILL,
+          .addTransition(DAGState.NEW, EnumSet.of(DAGState.KILLED, DAGState.FAILED),
+              DAGEventType.DAG_TERMINATE,
               new KillNewJobTransition())
           .addTransition(DAGState.NEW, DAGState.ERROR,
               DAGEventType.INTERNAL_ERROR,
@@ -269,8 +276,8 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
           .addTransition(DAGState.INITED, DAGState.RUNNING,
               DAGEventType.DAG_START,
               new StartTransition())
-          .addTransition(DAGState.INITED, DAGState.KILLED,
-              DAGEventType.DAG_KILL,
+          .addTransition(DAGState.INITED, EnumSet.of(DAGState.KILLED, DAGState.FAILED),
+              DAGEventType.DAG_TERMINATE,
               new KillInitedJobTransition())
           .addTransition(DAGState.INITED, DAGState.ERROR,
               DAGEventType.INTERNAL_ERROR,
@@ -287,7 +294,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
               DAGEventType.DAG_VERTEX_RERUNNING,
               new VertexReRunningTransition())
           .addTransition(DAGState.RUNNING, DAGState.TERMINATING,
-              DAGEventType.DAG_KILL, new DAGKilledTransition())
+              DAGEventType.DAG_TERMINATE, new DAGKilledTransition())
           .addTransition(DAGState.RUNNING, DAGState.RUNNING,
               DAGEventType.DAG_DIAGNOSTIC_UPDATE,
               DIAGNOSTIC_UPDATE_TRANSITION)
@@ -311,7 +318,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
               DAGEventType.DAG_COMMIT_COMPLETED,
               COMMIT_COMPLETED_TRANSITION)
           .addTransition(DAGState.COMMITTING, DAGState.TERMINATING, 
-              DAGEventType.DAG_KILL,
+              DAGEventType.DAG_TERMINATE,
               new DAGKilledWhileCommittingTransition())
           .addTransition(
               DAGState.COMMITTING,
@@ -354,7 +361,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
 
               // Ignore-able events
           .addTransition(DAGState.TERMINATING, DAGState.TERMINATING,
-              EnumSet.of(DAGEventType.DAG_KILL,
+              EnumSet.of(DAGEventType.DAG_TERMINATE,
                          DAGEventType.DAG_VERTEX_RERUNNING,
                          DAGEventType.DAG_SCHEDULER_UPDATE))
 
@@ -370,7 +377,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
               INTERNAL_ERROR_TRANSITION)
           // Ignore-able events
           .addTransition(DAGState.SUCCEEDED, DAGState.SUCCEEDED,
-              EnumSet.of(DAGEventType.DAG_KILL,
+              EnumSet.of(DAGEventType.DAG_TERMINATE,
                   DAGEventType.DAG_SCHEDULER_UPDATE,
                   DAGEventType.DAG_VERTEX_COMPLETED))
 
@@ -386,7 +393,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
               INTERNAL_ERROR_TRANSITION)
           // Ignore-able events
           .addTransition(DAGState.FAILED, DAGState.FAILED,
-              EnumSet.of(DAGEventType.DAG_KILL,
+              EnumSet.of(DAGEventType.DAG_TERMINATE,
                   DAGEventType.DAG_START,
                   DAGEventType.DAG_VERTEX_RERUNNING,
                   DAGEventType.DAG_SCHEDULER_UPDATE,
@@ -404,7 +411,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
               INTERNAL_ERROR_TRANSITION)
           // Ignore-able events
           .addTransition(DAGState.KILLED, DAGState.KILLED,
-              EnumSet.of(DAGEventType.DAG_KILL,
+              EnumSet.of(DAGEventType.DAG_TERMINATE,
                   DAGEventType.DAG_START,
                   DAGEventType.DAG_VERTEX_RERUNNING,
                   DAGEventType.DAG_SCHEDULER_UPDATE,
@@ -415,7 +422,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
               DAGState.ERROR,
               DAGState.ERROR,
               EnumSet.of(
-                  DAGEventType.DAG_KILL,
+                  DAGEventType.DAG_TERMINATE,
                   DAGEventType.DAG_INIT,
                   DAGEventType.DAG_START,
                   DAGEventType.DAG_VERTEX_COMPLETED,
@@ -499,6 +506,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
     this.dagId = dagId;
     this.jobPlan = jobPlan;
     this.dagConf = new Configuration(amConf);
+    this.dagOnlyConf = new Configuration(false);
     Iterator<PlanKeyValuePair> iter =
         jobPlan.getDagConf().getConfKeyValuesList().iterator();
     // override the amConf by using DAG level configuration
@@ -506,6 +514,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
       PlanKeyValuePair keyValPair = iter.next();
       TezConfiguration.validateProperty(keyValPair.getKey(), Scope.DAG);
       this.dagConf.set(keyValPair.getKey(), keyValPair.getValue());
+      this.dagOnlyConf.set(keyValPair.getKey(), keyValPair.getValue());
     }
     this.dagName = (jobPlan.getName() != null) ? jobPlan.getName() : "<missing app name>";
     this.userName = appUserName;
@@ -535,7 +544,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
     }
 
     this.aclManager = new ACLManager(appContext.getAMACLManager(), dagUGI.getShortUserName(),
-        this.dagConf);
+        this.jobPlan.getAclInfo());
     // this is only for recovery in case it does not call the init transition
     this.startDAGCpuTime = appContext.getCumulativeCPUTime();
     this.startDAGGCTime = appContext.getCumulativeGCTime();
@@ -795,9 +804,17 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
     try {
       float progress = 0.0f;
       for (Vertex v : getVertices().values()) {
-        progress += v.getProgress();
+        float vertexProgress = v.getProgress();
+        if (vertexProgress >= 0.0f && vertexProgress <= 1.0f) {
+          progress += vertexProgress;
+        }
+      }
+      float dagProgress = progress / getTotalVertices();
+      if (dagProgress >= 0.0f && progress <= 1.0f) {
+        return dagProgress;
+      } else {
+        return 0.0f;
       }
-      return progress / getTotalVertices();
     } finally {
       this.readLock.unlock();
     }
@@ -860,6 +877,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
     int totalKilledTaskCount = 0;
     int totalFailedTaskAttemptCount = 0;
     int totalKilledTaskAttemptCount = 0;
+    int totalRejectedTaskAttemptCount = 0;
     readLock.lock();
     try {
       for(Map.Entry<String, Vertex> entry : vertexMap.entrySet()) {
@@ -872,6 +890,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
         totalKilledTaskCount += progress.getKilledTaskCount();
         totalFailedTaskAttemptCount += progress.getFailedTaskAttemptCount();
         totalKilledTaskAttemptCount += progress.getKilledTaskAttemptCount();
+        totalRejectedTaskAttemptCount += progress.getRejectedTaskAttemptCount();
       }
       ProgressBuilder dagProgress = new ProgressBuilder();
       dagProgress.setTotalTaskCount(totalTaskCount);
@@ -881,6 +900,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
       dagProgress.setKilledTaskCount(totalKilledTaskCount);
       dagProgress.setFailedTaskAttemptCount(totalFailedTaskAttemptCount);
       dagProgress.setKilledTaskAttemptCount(totalKilledTaskAttemptCount);
+      dagProgress.setRejectedTaskAttemptCount(totalRejectedTaskAttemptCount);
       status.setState(getState());
       status.setDiagnostics(diagnostics);
       status.setDAGProgress(dagProgress);
@@ -935,6 +955,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
     int totalKilledTaskCount = 0;
     int totalFailedTaskAttemptCount = 0;
     int totalKilledTaskAttemptCount = 0;
+    int totalRejectedTaskAttemptCount = 0;
     readLock.lock();
     try {
       for(Map.Entry<String, Vertex> entry : vertexMap.entrySet()) {
@@ -946,6 +967,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
         totalKilledTaskCount += progress.getKilledTaskCount();
         totalFailedTaskAttemptCount += progress.getFailedTaskAttemptCount();
         totalKilledTaskAttemptCount += progress.getKilledTaskAttemptCount();
+        totalRejectedTaskAttemptCount += progress.getRejectedTaskAttemptCount();
       }
       ProgressBuilder dagProgress = new ProgressBuilder();
       dagProgress.setTotalTaskCount(totalTaskCount);
@@ -955,6 +977,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
       dagProgress.setKilledTaskCount(totalKilledTaskCount);
       dagProgress.setFailedTaskAttemptCount(totalFailedTaskAttemptCount);
       dagProgress.setKilledTaskAttemptCount(totalKilledTaskAttemptCount);
+      dagProgress.setRejectedTaskAttemptCount(totalRejectedTaskAttemptCount);
       return dagProgress;
     } finally {
       readLock.unlock();
@@ -1222,7 +1245,7 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
     if (recoveryData == null
         || recoveryData.getDAGStartedEvent() == null) {
       DAGStartedEvent startEvt = new DAGStartedEvent(this.dagId,
-          clock.getTime(), this.userName, this.dagName);
+          this.startTime, this.userName, this.dagName);
       this.appContext.getHistoryHandler().handle(
           new DAGHistoryEvent(dagId, startEvt));
     }
@@ -1232,6 +1255,12 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
     if (recoveryData == null
         || recoveryData.getDAGFinishedEvent() == null) {
       Map<String, Integer> taskStats = constructTaskStats(getDAGProgress());
+      if (finishTime < startTime) {
+        LOG.warn("DAG finish time is smaller than start time. "
+            + "startTime=" + startTime
+            + ", finishTime=" + finishTime
+        );
+      }
       DAGFinishedEvent finishEvt = new DAGFinishedEvent(dagId, startTime,
           finishTime, DAGState.SUCCEEDED, "", counters,
           this.userName, this.dagName, taskStats, this.appContext.getApplicationAttemptId(),
@@ -1245,9 +1274,14 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
     if (recoveryData == null
         || recoveryData.getDAGFinishedEvent() == null) {
       Map<String, Integer> taskStats = constructTaskStats(getDAGProgress());
-
+      if (finishTime < startTime) {
+        LOG.warn("DAG finish time is smaller than start time. "
+            + "startTime=" + startTime
+            + ", finishTime=" + finishTime
+        );
+      }
       DAGFinishedEvent finishEvt = new DAGFinishedEvent(dagId, startTime,
-          clock.getTime(), state,
+          finishTime, state,
           StringUtils.join(getDiagnostics(), LINE_SEPARATOR),
           counters, this.userName, this.dagName, taskStats,
           this.appContext.getApplicationAttemptId(), this.jobPlan);
@@ -1362,41 +1396,44 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
   }
   
   private DAGState finished(DAGState finalState) {
-    if (finishTime == 0) {
-      setFinishTime();
-    }
-    entityUpdateTracker.stop();
-
-    boolean recoveryError = false;
-
-    // update cpu time counters before finishing the dag
-    updateCpuCounters();
-    TezCounters counters = null;
+    boolean dagError = false;
     try {
-      counters = getAllCounters();
-    } catch (LimitExceededException e) {
-      addDiagnostic("Counters limit exceeded: " + e.getMessage());
-      finalState = DAGState.FAILED;
-    }
-
-    try {
-      if (finalState == DAGState.SUCCEEDED) {
-        logJobHistoryFinishedEvent(counters);
-      } else {
-        logJobHistoryUnsuccesfulEvent(finalState, counters);
+      if (finishTime == 0) {
+        setFinishTime();
       }
-    } catch (IOException e) {
-      LOG.warn("Failed to persist recovery event for DAG completion"
-          + ", dagId=" + dagId
-          + ", finalState=" + finalState);
-      recoveryError = true;
-    }
+      entityUpdateTracker.stop();
 
-    if (finalState != DAGState.SUCCEEDED) {
-      abortOutputs();
-    }
+      // update cpu time counters before finishing the dag
+      updateCpuCounters();
+      TezCounters counters = null;
+      try {
+        counters = getAllCounters();
+      } catch (LimitExceededException e) {
+        addDiagnostic("Counters limit exceeded: " + e.getMessage());
+        finalState = DAGState.FAILED;
+      }
 
-    if (recoveryError) {
+      try {
+        if (finalState == DAGState.SUCCEEDED) {
+          logJobHistoryFinishedEvent(counters);
+        } else {
+          logJobHistoryUnsuccesfulEvent(finalState, counters);
+        }
+      } catch (IOException e) {
+        LOG.warn("Failed to persist recovery event for DAG completion"
+            + ", dagId=" + dagId
+            + ", finalState=" + finalState, e);
+        dagError = true;
+      }
+
+      if (finalState != DAGState.SUCCEEDED) {
+        abortOutputs();
+      }
+    } catch (Exception e) {
+      dagError = true;
+      LOG.warn("Encountered exception while DAG finish", e);
+    }
+    if (dagError) {
       eventHandler.handle(new DAGAppMasterEventDAGFinished(getID(), DAGState.ERROR));
     } else {
       eventHandler.handle(new DAGAppMasterEventDAGFinished(getID(), finalState));
@@ -1412,6 +1449,11 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
     return userName;
   }
 
+  @Override
+  public int getIndex() {
+    return dagId.getId();
+  }
+
   @Override
   public String getName() {
     return dagName;
@@ -1428,6 +1470,16 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
 
   }
 
+  @Override
+  public BitSet getVertexDescendants(int vertexIndex) {
+    readLock.lock();
+    try {
+      return vertexDescendants.get(vertexIndex);
+    } finally {
+      readLock.unlock();
+    }
+  }
+
   @Override
   public int getSuccessfulVertices() {
     readLock.lock();
@@ -1531,6 +1583,8 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
       parseVertexEdges(this, edgePlans, v);
     }
 
+    computeVertexDescendants();
+
     // Initialize the edges, now that the payload and vertices have been set.
     for (Edge e : edges.values()) {
       try {
@@ -1587,6 +1641,31 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
     }
   }
 
+  private void computeVertexDescendants() {
+    vertexDescendants = new ArrayList<>(numVertices);
+    for (int i = 0; i < numVertices; ++i) {
+      vertexDescendants.add(new BitSet(numVertices));
+    }
+    BitSet verticesVisited = new BitSet(numVertices);
+    for (Vertex v : vertices.values()) {
+      computeVertexDescendants(verticesVisited, v);
+    }
+  }
+
+  private BitSet computeVertexDescendants(BitSet verticesVisited, Vertex v) {
+    int vertexIndex = v.getVertexId().getId();
+    BitSet descendants = vertexDescendants.get(vertexIndex);
+    if (!verticesVisited.get(vertexIndex)) {
+      for (Vertex child : v.getOutputVertices().keySet()) {
+        descendants.set(child.getVertexId().getId());
+        BitSet childDescendants = computeVertexDescendants(verticesVisited, child);
+        descendants.or(childDescendants);
+      }
+      verticesVisited.set(vertexIndex);
+    }
+    return descendants;
+  }
+
   private static void assignDAGScheduler(DAGImpl dag) throws TezException {
     String dagSchedulerClassName = dag.dagConf.get(TezConfiguration.TEZ_AM_DAG_SCHEDULER_CLASS,
         TezConfiguration.TEZ_AM_DAG_SCHEDULER_CLASS_DEFAULT);
@@ -1610,7 +1689,8 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
         dag.eventHandler, dag.taskCommunicatorManagerInterface,
         dag.clock, dag.taskHeartbeatHandler,
         !dag.commitAllOutputsOnSuccess, dag.appContext, vertexLocationHint,
-        dag.vertexGroups, dag.taskSpecificLaunchCmdOption, dag.entityUpdateTracker);
+        dag.vertexGroups, dag.taskSpecificLaunchCmdOption, dag.entityUpdateTracker,
+        dag.dagOnlyConf);
     return v;
   }
 
@@ -1825,28 +1905,41 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
     }
   }
 
+  private void addDiagnostics(DiagnosableEvent event) {
+    if (event.getDiagnosticInfo() != null && !event.getDiagnosticInfo().isEmpty()) {
+      addDiagnostic(event.getDiagnosticInfo());
+    }
+  }
+
   // Task-start has been moved out of InitTransition, so this arc simply
   // hardcodes 0 for both map and reduce finished tasks.
-  private static class KillNewJobTransition
-      implements SingleArcTransition<DAGImpl, DAGEvent> {
+  private static class KillNewJobTransition implements
+      MultipleArcTransition<DAGImpl, DAGEvent, DAGState> {
 
     @Override
-    public void transition(DAGImpl dag, DAGEvent dagEvent) {
+    public DAGState transition(DAGImpl dag, DAGEvent dagEvent) {
+      DAGEventTerminateDag event = (DAGEventTerminateDag) dagEvent;
       dag.setFinishTime();
-      dag.trySetTerminationCause(DAGTerminationCause.DAG_KILL);
-      dag.finished(DAGState.KILLED);
+      dag.trySetTerminationCause(event.getTerminationCause());
+      dag.addDiagnostic("Dag received [" + event.getType() + ", " + event.getTerminationCause() +
+          "] in NEW state.");
+      dag.addDiagnostics(event);
+      return dag.finished(event.getTerminationCause().getFinishedState());
     }
 
   }
 
-  private static class KillInitedJobTransition
-      implements SingleArcTransition<DAGImpl, DAGEvent> {
+  private static class KillInitedJobTransition implements
+      MultipleArcTransition<DAGImpl, DAGEvent, DAGState> {
 
     @Override
-    public void transition(DAGImpl dag, DAGEvent dagEvent) {
-      dag.trySetTerminationCause(DAGTerminationCause.DAG_KILL);
-      dag.addDiagnostic("Job received Kill in INITED state.");
-      dag.finished(DAGState.KILLED);
+    public DAGState transition(DAGImpl dag, DAGEvent dagEvent) {
+      DAGEventTerminateDag event = (DAGEventTerminateDag) dagEvent;
+      dag.trySetTerminationCause(event.getTerminationCause());
+      dag.addDiagnostic("Dag received [" + event.getType() + ", " + event.getTerminationCause() +
+          "] in INITED state.");
+      dag.addDiagnostics(event);
+      return dag.finished(event.getTerminationCause().getFinishedState());
     }
 
   }
@@ -1854,11 +1947,14 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
   private static class DAGKilledTransition
       implements SingleArcTransition<DAGImpl, DAGEvent> {
     @Override
-    public void transition(DAGImpl job, DAGEvent event) {
-      String msg = "Job received Kill while in RUNNING state.";
+    public void transition(DAGImpl job, DAGEvent dagEvent) {
+      DAGEventTerminateDag event = (DAGEventTerminateDag) dagEvent;
+      String msg = "Dag received [" + event.getType() + ", " + event.getTerminationCause() +
+          "] in RUNNING state.";
       LOG.info(msg);
       job.addDiagnostic(msg);
-      job.enactKill(DAGTerminationCause.DAG_KILL, VertexTerminationCause.DAG_KILL);
+      job.addDiagnostics(event);
+      job.enactKill(event.getTerminationCause(), VertexTerminationCause.DAG_TERMINATED);
       // Commit may happen when dag is still in RUNNING (vertex group commit)
       job.cancelCommits();
       // TODO Metrics
@@ -1872,12 +1968,15 @@ public class DAGImpl implements org.apache.tez.dag.app.dag.DAG,
     implements SingleArcTransition<DAGImpl, DAGEvent> {
 
     @Override
-    public void transition(DAGImpl dag, DAGEvent event) {
-      String diag = "DAG received Kill while in COMMITTING state.";
+    public void transition(DAGImpl dag, DAGEvent dagEvent) {
+      DAGEventTerminateDag event = (DAGEventTerminateDag) dagEvent;
+      String diag = "Dag received [" + event.getType() + ", " + event.getTerminationCause() +
+          "] in COMMITTING state.";
       LOG.info(diag);
       dag.addDiagnostic(diag);
+      dag.addDiagnostics(event);
       dag.cancelCommits();
-      dag.trySetTerminationCause(DAGTerminationCause.DAG_KILL);
+      dag.trySetTerminationCause(event.getTerminationCause());
     }
   }
 
