@@ -23,9 +23,13 @@ import java.util.List;
 import java.util.Map;
 import java.util.concurrent.ConcurrentMap;
 import java.util.concurrent.atomic.AtomicBoolean;
+import java.util.zip.Deflater;
 
 import org.apache.hadoop.conf.Configuration;
+import org.apache.tez.common.TezCommonUtils;
 import org.apache.tez.dag.api.TezConfiguration;
+import org.apache.tez.runtime.api.impl.GroupInputSpec;
+import org.apache.tez.runtime.library.common.shuffle.ShuffleUtils;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 import org.apache.hadoop.yarn.event.EventHandler;
@@ -38,6 +42,7 @@ import org.apache.tez.dag.api.EdgeProperty;
 import org.apache.tez.dag.api.TezException;
 import org.apache.tez.dag.api.TezUncheckedException;
 import org.apache.tez.dag.api.UserPayload;
+import org.apache.tez.dag.api.EdgeManagerPluginOnDemand.CompositeEventRouteMetadata;
 import org.apache.tez.dag.api.EdgeManagerPluginOnDemand.EventRouteMetadata;
 import org.apache.tez.dag.app.dag.Task;
 import org.apache.tez.dag.app.dag.Vertex;
@@ -49,6 +54,7 @@ import org.apache.tez.dag.records.TezTaskID;
 import org.apache.tez.runtime.api.Event;
 import org.apache.tez.runtime.api.events.CompositeDataMovementEvent;
 import org.apache.tez.runtime.api.events.DataMovementEvent;
+import org.apache.tez.runtime.api.events.CompositeRoutedDataMovementEvent;
 import org.apache.tez.runtime.api.events.InputFailedEvent;
 import org.apache.tez.runtime.api.events.InputReadErrorEvent;
 import org.apache.tez.runtime.api.impl.EventMetaData;
@@ -66,6 +72,31 @@ public class Edge {
 
   private static final Logger LOG = LoggerFactory.getLogger(Edge.class);
 
+  public List<TezEvent> generateEmptyEventsForAttempt(TezTaskAttemptID attempt) throws Exception {
+
+    if (!edgeProperty.getEdgeSource().getClassName().startsWith("org.apache.tez")) {
+      throw new TezException("Only org.apache.tez outputs are allowed for max percent failure feature. Disallowed Output: "
+          + edgeProperty.getEdgeSource().getClassName());
+    }
+    List<Event> events = new ArrayList<>();
+    Deflater deflater = TezCommonUtils.newBestCompressionDeflater();
+    try {
+      ShuffleUtils.generateEventsForNonStartedOutput(events,
+          edgeManager.getNumDestinationConsumerTasks(attempt.getTaskID().getId()), null, false, true, deflater);
+    } catch (Exception e) {
+      throw new TezException(e);
+    }
+    EventMetaData sourceInfo = new EventMetaData(EventMetaData.EventProducerConsumerType.INPUT,
+        sourceVertex.getName(), getDestinationVertexName(), attempt);
+
+    List<TezEvent> tezEvents = new ArrayList<>(events.size());
+    for (Event e : events) {
+      TezEvent tezEvent = new TezEvent(e, sourceInfo);
+      tezEvents.add(tezEvent);
+    }
+    return tezEvents;
+  }
+
   class EdgeManagerPluginContextImpl implements EdgeManagerPluginContext {
 
     private final UserPayload userPayload;
@@ -99,6 +130,17 @@ public class Edge {
       return destinationVertex.getTotalTasks();
     }
 
+    @Override
+    public String getVertexGroupName() {
+      if (destinationVertex.getGroupInputSpecList() != null) {
+        for (GroupInputSpec group : destinationVertex.getGroupInputSpecList()) {
+          if (group.getGroupVertices().contains(getSourceVertexName())) {
+            return group.getGroupName();
+          }
+        }
+      }
+      return null;
+    }
   }
 
   private EdgeProperty edgeProperty;
@@ -174,19 +216,24 @@ public class Edge {
             + getEdgeInfo(), e);
       }
     }
-    destinationMetaInfo = new EventMetaData(EventProducerConsumerType.INPUT, 
-        destinationVertex.getName(), 
-        sourceVertex.getName(), 
-        null);
+    synchronized (this) {
+      destinationMetaInfo = new EventMetaData(EventProducerConsumerType.INPUT,
+              destinationVertex.getName(),
+              sourceVertex.getName(),
+              null);
+    }
   }
 
-  public synchronized void setEdgeProperty(EdgeProperty newEdgeProperty) throws AMUserCodeException {
-    this.edgeProperty = newEdgeProperty;
-    boolean wasUnInitialized = (edgeManager == null);
-    try {
-      createEdgeManager();
-    } catch (TezException e) {
-      throw new AMUserCodeException(Source.EdgeManager, e);
+  public void setEdgeProperty(EdgeProperty newEdgeProperty) throws AMUserCodeException {
+    boolean wasUnInitialized;
+    synchronized (this) {
+      this.edgeProperty = newEdgeProperty;
+      wasUnInitialized = (edgeManager == null);
+      try {
+        createEdgeManager();
+      } catch (TezException e) {
+        throw new AMUserCodeException(Source.EdgeManager, e);
+      }
     }
     initialize();
     if (wasUnInitialized) {
@@ -199,7 +246,7 @@ public class Edge {
   
   // Test only method for creating specific scenarios
   @VisibleForTesting
-  synchronized void setCustomEdgeManager(EdgeManagerPluginDescriptor descriptor)
+  void setCustomEdgeManager(EdgeManagerPluginDescriptor descriptor)
       throws AMUserCodeException {
     EdgeProperty modifiedEdgeProperty =
         EdgeProperty.create(descriptor,
@@ -210,22 +257,28 @@ public class Edge {
     setEdgeProperty(modifiedEdgeProperty);
   }
   
-  public synchronized void routingToBegin() throws AMUserCodeException {
-    if (edgeManagerContext.getDestinationVertexNumTasks() == 0) {
-      routingNeeded = false;
-    } else if (edgeManagerContext.getDestinationVertexNumTasks() < 0) {
-      throw new TezUncheckedException(
-          "Internal error. Not expected to route events to a destination until parallelism is determined" +
-          " sourceVertex=" + sourceVertex.getLogIdentifier() +
-          " edgeManager=" + edgeManager.getClass().getName());
+  public void routingToBegin() throws AMUserCodeException {
+    int numDestTasks = edgeManagerContext.getDestinationVertexNumTasks();
+    synchronized (this) {
+      if (numDestTasks == 0) {
+        routingNeeded = false;
+      } else if (numDestTasks < 0) {
+        throw new TezUncheckedException(
+                "Internal error. Not expected to route events to a destination until parallelism is determined" +
+                        " sourceVertex=" + sourceVertex.getLogIdentifier() +
+                        " edgeManager=" + edgeManager.getClass().getName());
+      }
+      if (edgeManager instanceof EdgeManagerPluginOnDemand) {
+        onDemandRouting = true;
+      }
     }
-    if (edgeManager instanceof EdgeManagerPluginOnDemand) {
-      onDemandRouting = true;
+
+    if (onDemandRouting) {
       try {
-        ((EdgeManagerPluginOnDemand)edgeManager).prepareForRouting();
+        ((EdgeManagerPluginOnDemand) edgeManager).prepareForRouting();
       } catch (Exception e) {
         throw new AMUserCodeException(Source.EdgeManager,
-            "Fail to prepareForRouting " + getEdgeInfo(), e);
+                "Fail to prepareForRouting " + getEdgeInfo(), e);
       }
     }
     
@@ -538,36 +591,15 @@ public class Edge {
         switch (tezEvent.getEventType()) {
         case COMPOSITE_DATA_MOVEMENT_EVENT:
           {
-            CompositeDataMovementEvent compEvent = (CompositeDataMovementEvent) tezEvent.getEvent();        
-            EventRouteMetadata routeMeta;
-            int numEventsDone;
-            if (pendingRoutes != null) {
-              routeMeta = pendingRoutes.getRouteMeta();
-              numEventsDone = pendingRoutes.getNumEventsRouted();
-            } else {
-              routeMeta = edgeManagerOnDemand
+            CompositeDataMovementEvent compEvent = (CompositeDataMovementEvent) tezEvent.getEvent(); 
+            CompositeEventRouteMetadata routeMeta = edgeManagerOnDemand
                   .routeCompositeDataMovementEventToDestination(srcTaskIndex, taskIndex);
-              numEventsDone = 0;
-            }
+
             if (routeMeta != null) {
-              int listSize = listToAdd.size();
-              int numEvents = routeMeta.getNumEvents();
-              int[] sourceIndices = routeMeta.getSourceIndices();
-              int[] targetIndices = routeMeta.getTargetIndices();
-              while (numEventsDone < numEvents && listSize++ < listMaxSize) {
-                DataMovementEvent e = compEvent.expand(sourceIndices[numEventsDone],
-                    targetIndices[numEventsDone]);
-                numEventsDone++;
-                TezEvent tezEventToSend = new TezEvent(e, tezEvent.getSourceInfo(),
-                    tezEvent.getEventReceivedTime());
-                tezEventToSend.setDestinationInfo(destinationMetaInfo);
-                listToAdd.add(tezEventToSend);
-              }
-              if (numEventsDone < numEvents) {
-                pendingEvents.put(attemptID, new PendingEventRouteMetadata(routeMeta, tezEvent,
-                    numEventsDone));
-                return false;
-              }
+              CompositeRoutedDataMovementEvent edme = compEvent.expandRouted(routeMeta);
+              TezEvent tezEventToSend = new TezEvent(edme, tezEvent.getSourceInfo(), tezEvent.getEventReceivedTime());
+              tezEventToSend.setDestinationInfo(destinationMetaInfo);
+              listToAdd.add(tezEventToSend);
             }
           }
           break;
