@@ -26,6 +26,8 @@ import java.util.LinkedHashMap;
 import java.util.List;
 import java.util.Map;
 import java.util.Map.Entry;
+import java.util.Set;
+import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.locks.Lock;
 import java.util.concurrent.locks.ReadWriteLock;
 import java.util.concurrent.locks.ReentrantReadWriteLock;
@@ -35,7 +37,11 @@ import com.google.common.base.Predicate;
 import com.google.common.collect.Maps;
 
 import org.apache.commons.lang.StringUtils;
+import org.apache.commons.lang.exception.ExceptionUtils;
 import org.apache.hadoop.classification.InterfaceAudience.Private;
+import org.apache.hadoop.yarn.api.records.NodeId;
+import org.apache.tez.dag.app.dag.event.TaskEventTAFailed;
+import org.apache.tez.runtime.api.TaskFailureType;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 import org.apache.hadoop.conf.Configuration;
@@ -50,7 +56,6 @@ import org.apache.hadoop.yarn.util.Clock;
 import org.apache.tez.common.counters.TaskCounter;
 import org.apache.tez.common.counters.TezCounters;
 import org.apache.tez.dag.api.TaskLocationHint;
-import org.apache.tez.dag.api.TezConfiguration;
 import org.apache.tez.dag.api.TezUncheckedException;
 import org.apache.tez.dag.api.oldrecords.TaskAttemptState;
 import org.apache.tez.dag.api.oldrecords.TaskReport;
@@ -70,10 +75,13 @@ import org.apache.tez.dag.app.dag.event.DAGEvent;
 import org.apache.tez.dag.app.dag.event.DAGEventDiagnosticsUpdate;
 import org.apache.tez.dag.app.dag.event.DAGEventSchedulerUpdate;
 import org.apache.tez.dag.app.dag.event.DAGEventType;
+import org.apache.tez.dag.app.dag.event.TaskAttemptEventAttemptKilled;
 import org.apache.tez.dag.app.dag.event.TaskAttemptEventKillRequest;
 import org.apache.tez.dag.app.dag.event.TaskAttemptEventOutputFailed;
+import org.apache.tez.dag.app.dag.event.TaskAttemptEventTerminationCauseEvent;
 import org.apache.tez.dag.app.dag.event.TaskEvent;
 import org.apache.tez.dag.app.dag.event.TaskEventScheduleTask;
+import org.apache.tez.dag.app.dag.event.TaskEventTAKilled;
 import org.apache.tez.dag.app.dag.event.TaskEventTAUpdate;
 import org.apache.tez.dag.app.dag.event.TaskEventTermination;
 import org.apache.tez.dag.app.dag.event.TaskEventType;
@@ -89,6 +97,7 @@ import org.apache.tez.dag.records.TaskAttemptTerminationCause;
 import org.apache.tez.dag.records.TezTaskAttemptID;
 import org.apache.tez.dag.records.TezTaskID;
 import org.apache.tez.dag.records.TezVertexID;
+import org.apache.tez.dag.utils.TezBuilderUtils;
 import org.apache.tez.runtime.api.OutputCommitter;
 import org.apache.tez.runtime.api.impl.TaskSpec;
 import org.apache.tez.runtime.api.impl.TaskStatistics;
@@ -143,10 +152,15 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
   // track the status of TaskAttempt (true mean completed, false mean uncompleted)
   private final Map<Integer, Boolean> taskAttemptStatus = new HashMap<Integer,Boolean>();
 
+  private final Set<NodeId> nodesWithRunningAttempts = Collections
+      .newSetFromMap(new ConcurrentHashMap<NodeId, Boolean>());
+
   private static final SingleArcTransition<TaskImpl , TaskEvent>
      ATTEMPT_KILLED_TRANSITION = new AttemptKilledTransition();
   private static final SingleArcTransition<TaskImpl, TaskEvent>
      KILL_TRANSITION = new KillTransition();
+  private static final SingleArcTransition<TaskImpl, TaskEvent>
+      REDUNDANT_COMPLETED_TRANSITION = new AttemptRedundantCompletedTransition();
 
 
   private static final TaskStateChangedCallback STATE_CHANGED_CALLBACK = new TaskStateChangedCallback();
@@ -240,23 +254,30 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
     .addTransition(TaskStateInternal.SUCCEEDED, //only possible for map tasks
         EnumSet.of(TaskStateInternal.SCHEDULED, TaskStateInternal.SUCCEEDED),
         TaskEventType.T_ATTEMPT_KILLED, new TaskRetroactiveKilledTransition())
-    // Ignore-able transitions.
     .addTransition(
         TaskStateInternal.SUCCEEDED, TaskStateInternal.SUCCEEDED,
-        EnumSet.of(TaskEventType.T_ADD_SPEC_ATTEMPT,
+        TaskEventType.T_ATTEMPT_SUCCEEDED, REDUNDANT_COMPLETED_TRANSITION)
+    // Ignore-able transitions.
+    .addTransition(TaskStateInternal.SUCCEEDED, TaskStateInternal.SUCCEEDED,
+        EnumSet.of(
+            TaskEventType.T_ADD_SPEC_ATTEMPT,
             TaskEventType.T_TERMINATE,
-            TaskEventType.T_ATTEMPT_SUCCEEDED, // Maybe track and reuse later
             TaskEventType.T_ATTEMPT_LAUNCHED))
     .addTransition(TaskStateInternal.SUCCEEDED, TaskStateInternal.SUCCEEDED,
         TaskEventType.T_SCHEDULE)
 
     // Transitions from FAILED state
+    .addTransition(TaskStateInternal.FAILED, TaskStateInternal.FAILED,
+        EnumSet.of(
+            TaskEventType.T_ATTEMPT_FAILED,
+            TaskEventType.T_ATTEMPT_KILLED,
+            TaskEventType.T_ATTEMPT_SUCCEEDED), REDUNDANT_COMPLETED_TRANSITION)
     .addTransition(TaskStateInternal.FAILED, TaskStateInternal.FAILED,
         EnumSet.of(
             TaskEventType.T_TERMINATE,
             TaskEventType.T_SCHEDULE,
             TaskEventType.T_ADD_SPEC_ATTEMPT,
-            TaskEventType.T_ATTEMPT_KILLED))
+            TaskEventType.T_ATTEMPT_LAUNCHED))
 
     // Transitions from KILLED state
     // Ignorable event: T_ATTEMPT_KILLED
@@ -277,6 +298,9 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
             TaskEventType.T_TERMINATE,
             TaskEventType.T_SCHEDULE,
             TaskEventType.T_ADD_SPEC_ATTEMPT,
+            TaskEventType.T_ATTEMPT_LAUNCHED,
+            TaskEventType.T_ATTEMPT_SUCCEEDED,
+            TaskEventType.T_ATTEMPT_FAILED,
             TaskEventType.T_ATTEMPT_KILLED))
 
     // create the topology tables
@@ -324,6 +348,9 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
 
   private final boolean leafVertex;
 
+  @VisibleForTesting
+  long finishTime = -1L;
+
   @Override
   public TaskState getState() {
     readLock.lock();
@@ -348,9 +375,7 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
     readLock = readWriteLock.readLock();
     writeLock = readWriteLock.writeLock();
     this.attempts = Collections.emptyMap();
-    // TODO Avoid reading this from configuration for each task.
-    maxFailedAttempts = this.conf.getInt(TezConfiguration.TEZ_AM_TASK_MAX_FAILED_ATTEMPTS,
-                              TezConfiguration.TEZ_AM_TASK_MAX_FAILED_ATTEMPTS_DEFAULT);
+    maxFailedAttempts = vertex.getVertexConfig().getMaxFailedTaskAttempts();
     taskId = TezTaskID.getInstance(vertexId, taskIndex);
     this.taskCommunicatorManagerInterface = taskCommunicatorManagerInterface;
     this.taskHeartbeatHandler = thh;
@@ -423,7 +448,7 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
     try {
       report.setTaskId(taskId);
       report.setStartTime(getLaunchTime());
-      report.setFinishTime(getFinishTime());
+      report.setFinishTime(getLastTaskAttemptFinishTime());
       report.setTaskState(getState());
       report.setProgress(getProgress());
       return report;
@@ -466,11 +491,18 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
   public float getProgress() {
     readLock.lock();
     try {
-      TaskAttempt bestAttempt = selectBestAttempt();
-      if (bestAttempt == null) {
-        return 0f;
+      final TaskStateInternal state = getInternalState();
+      if (state == TaskStateInternal.RUNNING) {
+        TaskAttempt bestAttempt = selectBestAttempt();
+        if (bestAttempt == null) {
+          return 0f;
+        }
+        return bestAttempt.getProgress();
+      } else if (state == TaskStateInternal.SUCCEEDED) {
+        return 1.0f;
+      } else {
+        return 0.0f;
       }
-      return bestAttempt.getProgress();
     } finally {
       readLock.unlock();
     }
@@ -542,6 +574,16 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
     }
   }
 
+  @Override
+  public long getFinishTime() {
+    readLock.lock();
+    try {
+      return finishTime;
+    } finally {
+      readLock.unlock();
+    }
+  }
+
   @VisibleForTesting
   public TaskStateInternal getInternalState() {
     readLock.lock();
@@ -583,7 +625,7 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
 
   //this is always called in read/write lock
   //TODO Verify behaviour is Task is killed (no finished attempt)
-  private long getFinishTime() {
+  private long getLastTaskAttemptFinishTime() {
     if (!isFinished()) {
       return 0;
     }
@@ -699,9 +741,16 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
   }
 
   TaskAttemptImpl createAttempt(int attemptNumber, TezTaskAttemptID schedulingCausalTA) {
-    return new TaskAttemptImpl(getTaskId(), attemptNumber, eventHandler,
+    TezTaskAttemptID attemptId = TezBuilderUtils.newTaskAttemptId(taskId, attemptNumber);
+    TaskSpec taskSpec = new TaskSpec(attemptId,
+        baseTaskSpec.getDAGName(), baseTaskSpec.getVertexName(),
+        baseTaskSpec.getVertexParallelism(), baseTaskSpec.getProcessorDescriptor(),
+        baseTaskSpec.getInputs(), baseTaskSpec.getOutputs(), baseTaskSpec.getGroupInputs(),
+        baseTaskSpec.getTaskConf());
+    return new TaskAttemptImpl(attemptId, eventHandler,
         taskCommunicatorManagerInterface, conf, clock, taskHeartbeatHandler, appContext,
-        (failedAttempts > 0), taskResource, containerContext, leafVertex, this, schedulingCausalTA);
+        (failedAttempts > 0), taskResource, containerContext, leafVertex, getVertex(),
+        locationHint, taskSpec, schedulingCausalTA, nodesWithRunningAttempts);
   }
 
   @Override
@@ -825,9 +874,7 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
   private void handleTaskAttemptCompletion(TezTaskAttemptID attemptId,
       TaskAttemptStateInternal attemptState) {
     this.sendTaskAttemptCompletionEvent(attemptId, attemptState);
-    if (getInternalState() != TaskStateInternal.SUCCEEDED) {
-      sendDAGSchedulerFinishedEvent(attemptId); // not a retro active action
-    }
+    sendDAGSchedulerFinishedEvent(attemptId);
   }
 
   private void sendDAGSchedulerFinishedEvent(TezTaskAttemptID taId) {
@@ -862,8 +909,9 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
   protected void logJobHistoryTaskFinishedEvent() {
     // FIXME need to handle getting finish time as this function
     // is called from within a transition
+    this.finishTime = clock.getTime();
     TaskFinishedEvent finishEvt = new TaskFinishedEvent(taskId,
-        getVertex().getName(), getLaunchTime(), clock.getTime(),
+        getVertex().getName(), getLaunchTime(), this.finishTime,
         successfulAttempt,
         TaskState.SUCCEEDED, "", getCounters(), failedAttempts);
     this.appContext.getHistoryHandler().handle(
@@ -871,8 +919,9 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
   }
 
   protected void logJobHistoryTaskFailedEvent(TaskState finalState) {
+    this.finishTime = clock.getTime();
     TaskFinishedEvent finishEvt = new TaskFinishedEvent(taskId,
-        getVertex().getName(), getLaunchTime(), clock.getTime(), null,
+        getVertex().getName(), getLaunchTime(), this.finishTime, null,
         finalState, 
         StringUtils.join(getDiagnostics(), LINE_SEPARATOR),
         getCounters(), failedAttempts);
@@ -940,6 +989,7 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
       } else {
         task.scheduledTime = task.clock.getTime();
         task.logJobHistoryTaskStartedEvent();
+        task.vertex.reportTaskStartTime(task.getLaunchTime());
       }
       // No matter whether it is in recovery or normal execution, always schedule new task attempt.
       // TaskAttempt will continue the recovery if necessary and send task attempt status
@@ -965,14 +1015,15 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
     public void transition(TaskImpl task, TaskEvent event) {
       LOG.info("Scheduling a redundant attempt for task " + task.taskId);
       task.counters.findCounter(TaskCounter.NUM_SPECULATIONS).increment(1);
-      TezTaskAttemptID earliestUnfinishedAttempt = null;
+      TaskAttempt earliestUnfinishedAttempt = null;
       for (TaskAttempt ta : task.attempts.values()) {
         // find the oldest running attempt
         if (!ta.isFinished()) {
-          earliestUnfinishedAttempt = ta.getID();
+          earliestUnfinishedAttempt = ta;
+          task.nodesWithRunningAttempts.add(ta.getNodeId());
         }
       }
-      task.addAndScheduleAttempt(earliestUnfinishedAttempt);
+      task.addAndScheduleAttempt(earliestUnfinishedAttempt.getID());
     }
   }
 
@@ -980,10 +1031,10 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
   private static class AttemptSucceededTransition
       implements MultipleArcTransition<TaskImpl, TaskEvent, TaskStateInternal> {
 
-    private boolean recoverSuccessTaskAttempt(TaskImpl task) {
+    private String recoverSuccessTaskAttempt(TaskImpl task) {
       // Found successful attempt
       // Recover data
-      boolean recoveredData = true;
+      String errorMsg = null;
       if (task.getVertex().getOutputCommitters() != null
           && !task.getVertex().getOutputCommitters().isEmpty()) {
         for (Entry<String, OutputCommitter> entry
@@ -993,28 +1044,29 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
               + ", output=" + entry.getKey());
           OutputCommitter committer = entry.getValue();
           if (!committer.isTaskRecoverySupported()) {
-            LOG.info("Task recovery not supported by committer"
-                + ", failing task attempt"
+            errorMsg = "Task recovery not supported by committer"
+                + ", failing task attempt";
+            LOG.info(errorMsg
                 + ", taskId=" + task.getTaskId()
                 + ", attemptId=" + task.successfulAttempt
                 + ", output=" + entry.getKey());
-            recoveredData = false;
             break;
           }
           try {
             committer.recoverTask(task.getTaskId().getId(),
                 task.appContext.getApplicationAttemptId().getAttemptId()-1);
           } catch (Exception e) {
+            errorMsg = "Task recovery failed by committer: "
+                + ExceptionUtils.getStackTrace(e);
             LOG.warn("Task recovery failed by committer"
                 + ", taskId=" + task.getTaskId()
                 + ", attemptId=" + task.successfulAttempt
                 + ", output=" + entry.getKey(), e);
-            recoveredData = false;
             break;
           }
         }
       }
-      return recoveredData;
+      return errorMsg;
     }
 
     @Override
@@ -1024,13 +1076,14 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
       // recovery. In that case just reschedule new attempt if numFailedAttempts does not exceeded maxFailedAttempts.
       if (task.recoveryData!= null
           && task.recoveryData.isTaskAttemptSucceeded(successTaId)) {
-        boolean recoveredData = recoverSuccessTaskAttempt(task);
-        if (!recoveredData) {
-          // Move this TA to KILLED (TEZ-2958)
-          LOG.info("Can not recovery the successful task attempt, schedule new task attempt,"
+        String errorMsg = recoverSuccessTaskAttempt(task);
+        if (errorMsg != null) {
+          LOG.info("Can not recover the successful task attempt, schedule new task attempt,"
               + "taskId=" + task.getTaskId());
           task.successfulAttempt = null;
           task.addAndScheduleAttempt(successTaId);
+          task.eventHandler.handle(new TaskAttemptEventAttemptKilled(successTaId,
+              errorMsg, TaskAttemptTerminationCause.TERMINATED_AT_RECOVERY, true));
           return TaskStateInternal.RUNNING;
         } else {
           task.successfulAttempt = successTaId;
@@ -1101,7 +1154,21 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
           TaskAttemptStateInternal.KILLED);
       // we KillWaitAttemptCompletedTransitionready have a spare
       task.taskAttemptStatus.put(castEvent.getTaskAttemptID().getId(), true);
-      task.getVertex().incrementKilledTaskAttemptCount();
+
+      boolean isRejection = false;
+      if (event instanceof TaskEventTAKilled) {
+        TaskEventTAKilled killEvent = (TaskEventTAKilled) event;
+        if (killEvent.getCausalEvent() instanceof TaskAttemptEventTerminationCauseEvent) {
+          TaskAttemptEventTerminationCauseEvent cause =
+              (TaskAttemptEventTerminationCauseEvent)killEvent.getCausalEvent();
+          isRejection = cause.getTerminationCause() == TaskAttemptTerminationCause.SERVICE_BUSY;
+        }
+      }
+      if (isRejection) { // TODO: remove as part of TEZ-3881.
+        task.getVertex().incrementRejectedTaskAttemptCount();
+      } else {
+        task.getVertex().incrementKilledTaskAttemptCount();
+      }
       if (task.shouldScheduleNewAttempt()) {
         task.addAndScheduleAttempt(castEvent.getTaskAttemptID());
       }
@@ -1144,7 +1211,7 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
     public TaskStateInternal transition(TaskImpl task, TaskEvent event) {
       task.failedAttempts++;
       task.getVertex().incrementFailedTaskAttemptCount();
-      TaskEventTAUpdate castEvent = (TaskEventTAUpdate) event;
+      TaskEventTAFailed castEvent = (TaskEventTAFailed) event;
       schedulingCausalTA = castEvent.getTaskAttemptID();
       task.addDiagnosticInfo("TaskAttempt " + castEvent.getTaskAttemptID().getId() + " failed,"
           + " info=" + task.getAttempt(castEvent.getTaskAttemptID()).getDiagnostics());
@@ -1155,7 +1222,8 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
       // The attempt would have informed the scheduler about it's failure
 
       task.taskAttemptStatus.put(castEvent.getTaskAttemptID().getId(), true);
-      if (task.failedAttempts < task.maxFailedAttempts) {
+      if (task.failedAttempts < task.maxFailedAttempts &&
+          castEvent.getTaskFailureType() == TaskFailureType.NON_FATAL) {
         task.handleTaskAttemptCompletion(
             ((TaskEventTAUpdate) event).getTaskAttemptID(),
             TaskAttemptStateInternal.FAILED);
@@ -1167,9 +1235,15 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
           task.addAndScheduleAttempt(getSchedulingCausalTA());
         }
       } else {
-        LOG.info("Failing task: " + task.getTaskId()
-            + ", currentFailedAttempts: " + task.failedAttempts + ", maxFailedAttempts: "
-            + task.maxFailedAttempts);
+        if (castEvent.getTaskFailureType() == TaskFailureType.NON_FATAL) {
+          LOG.info(
+              "Failing task: {} due to too many failed attempts. currentFailedAttempts={}, maxFailedAttempts={}",
+              task.getTaskId(), task.failedAttempts, task.maxFailedAttempts);
+        } else {
+          LOG.info(
+              "Failing task: {} due to {} error reported by TaskAttempt. CurrentFailedAttempts={}",
+              task.getTaskId(), TaskFailureType.FATAL, task.failedAttempts);
+        }
         task.handleTaskAttemptCompletion(
             ((TaskEventTAUpdate) event).getTaskAttemptID(),
             TaskAttemptStateInternal.FAILED);
@@ -1197,13 +1271,7 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
 
     @Override
     public TaskStateInternal transition(TaskImpl task, TaskEvent event) {
-      if (task.leafVertex) {
-        LOG.error("Unexpected event for task of leaf vertex " + event.getType() + ", taskId: "
-            + task.getTaskId());
-        task.internalError(event.getType());
-      }
-
-      TaskEventTAUpdate castEvent = (TaskEventTAUpdate) event;
+      TaskEventTAFailed castEvent = (TaskEventTAFailed) event;
       TezTaskAttemptID failedAttemptId = castEvent.getTaskAttemptID();
       TaskAttempt failedAttempt = task.getAttempt(failedAttemptId);
       ContainerId containerId = failedAttempt.getAssignedContainerID();
@@ -1222,10 +1290,16 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
       if (task.getInternalState() == TaskStateInternal.SUCCEEDED &&
           !failedAttemptId.equals(task.successfulAttempt)) {
         // don't allow a different task attempt to override a previous
-        // succeeded state
+        // succeeded state and mark the attempt status as done
+        task.taskAttemptStatus.put(failedAttemptId.getId(), true);
         return TaskStateInternal.SUCCEEDED;
       }
-      
+
+      if (task.leafVertex) {
+        LOG.error("Unexpected event for task of leaf vertex " + event.getType() + ", taskId: "
+            + task.getTaskId());
+        task.internalError(event.getType());
+      }
       Preconditions.checkState(castEvent.getCausalEvent() != null);
       TaskAttemptEventOutputFailed destinationEvent = 
           (TaskAttemptEventOutputFailed) castEvent.getCausalEvent();
@@ -1266,14 +1340,8 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
 
       TaskEventTAUpdate attemptEvent = (TaskEventTAUpdate) event;
       TezTaskAttemptID attemptId = attemptEvent.getTaskAttemptID();
-      if(task.successfulAttempt == attemptId) {
-        // successful attempt is now killed. reschedule
-        // tell the job about the rescheduling
-        unSucceed(task);
-        task.handleTaskAttemptCompletion(
-            attemptId,
-            TaskAttemptStateInternal.KILLED);
-        task.eventHandler.handle(new VertexEventTaskReschedule(task.taskId));
+      TaskStateInternal resultState = TaskStateInternal.SUCCEEDED;
+      if(task.successfulAttempt.equals(attemptId)) {
         // typically we are here because this map task was run on a bad node and
         // we want to reschedule it on a different node.
         // Depending on whether there are previous failed attempts or not this
@@ -1282,14 +1350,12 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
         // from the map splitInfo. So the bad node might be sent as a location
         // to the RM. But the RM would ignore that just like it would ignore
         // currently pending container requests affinitized to bad nodes.
-        task.addAndScheduleAttempt(attemptId);
-        return TaskStateInternal.SCHEDULED;
-      } else {
-        // nothing to do
-        LOG.info("Ignoring kill of attempt: " + attemptId + " because attempt: " +
-            task.successfulAttempt + " is already successful");
-        return TaskStateInternal.SUCCEEDED;
+        unSucceed(task);
+        task.eventHandler.handle(new VertexEventTaskReschedule(task.taskId));
+        resultState = TaskStateInternal.SCHEDULED;
       }
+      ATTEMPT_KILLED_TRANSITION.transition(task, event);
+      return resultState;
     }
   }
 
@@ -1313,6 +1379,15 @@ public class TaskImpl implements Task, EventHandler<TaskEvent> {
     }
   }
 
+  private static class AttemptRedundantCompletedTransition
+      implements SingleArcTransition<TaskImpl, TaskEvent> {
+    @Override
+    public void transition(TaskImpl task, TaskEvent event) {
+      TezTaskAttemptID successTaId = ((TaskEventTAUpdate)event).getTaskAttemptID();
+      task.taskAttemptStatus.put(successTaId.getId(), true);
+    }
+  }
+
   private static class TaskStateChangedCallback
       implements OnStateChangedCallback<TaskStateInternal, TaskImpl> {
 
