@@ -24,6 +24,7 @@ import java.lang.reflect.UndeclaredThrowableException;
 import java.security.PrivilegedExceptionAction;
 import java.util.ArrayList;
 import java.util.Collection;
+import java.util.Collections;
 import java.util.List;
 import java.util.Map;
 import java.util.Set;
@@ -35,6 +36,10 @@ import javax.annotation.Nullable;
 
 import org.apache.commons.lang.exception.ExceptionUtils;
 import org.apache.tez.dag.app.dag.event.DAGEventInternalError;
+import org.apache.tez.dag.app.dag.event.VertexEventRouteEvent;
+import org.apache.tez.dag.records.TezTaskAttemptID;
+import org.apache.tez.runtime.api.events.CustomProcessorEvent;
+import org.apache.tez.runtime.api.impl.GroupInputSpec;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 import org.apache.hadoop.security.UserGroupInformation;
@@ -270,6 +275,29 @@ public class VertexManager {
       // Recovery handling is taken care of by the Vertex.
     }
 
+    @Override
+    public void sendEventToProcessor(Collection<CustomProcessorEvent> events, int taskId) {
+      checkAndThrowIfDone();
+      Preconditions.checkArgument(taskId >= 0 && taskId < managedVertex.getTotalTasks(),
+        "Invalid taskId " + taskId + "; " + "There are " + managedVertex.getTotalTasks()
+          + " tasks in total.");
+
+      if (events != null && events.size() > 0) {
+        List<TezEvent> tezEvents = new ArrayList<>();
+        for (CustomProcessorEvent event : events) {
+          TezEvent tezEvent = new TezEvent(event, null);
+          // use dummy task attempt id since this is not an task attempt specific event and task
+          // attempt id won't be used anyway
+          EventMetaData destinationMeta = new EventMetaData(EventProducerConsumerType.PROCESSOR,
+            managedVertex.getName(), managedVertex.getName(),
+            TezTaskAttemptID.getInstance(managedVertex.getTask(taskId).getTaskId(), -1));
+          tezEvent.setDestinationInfo(destinationMeta);
+          tezEvents.add(tezEvent);
+        }
+        appContext.getEventHandler().handle(
+          new VertexEventRouteEvent(managedVertex.getVertexId(), tezEvents));
+      }
+    }
 
     @Override
     public synchronized void setVertexLocationHint(VertexLocationHint locationHint) {
@@ -359,6 +387,18 @@ public class VertexManager {
       managedVertex.doneReconfiguringVertex();
     }
 
+    @Override
+    public Map<String, List<String>> getInputVertexGroups() {
+      checkAndThrowIfDone();
+      Map<String, List<String>> inputGroups = Maps.newHashMap();
+      if (managedVertex.getGroupInputSpecList() != null) {
+        for (GroupInputSpec group : managedVertex.getGroupInputSpecList()) {
+          inputGroups.put(group.getGroupName(), Collections.unmodifiableList(group.getGroupVertices()));
+        }
+      }
+      return inputGroups;
+    }
+
     @Override
     public void onStateUpdated(VertexStateUpdate event) {
       // this is not called by the vertex manager plugin. 
@@ -384,13 +424,13 @@ public class VertexManager {
     
     pluginContext = new VertexManagerPluginContextImpl();
     Preconditions.checkArgument(pluginDesc != null);
+    payload = pluginDesc.getUserPayload();
+    pluginFailed = new AtomicBoolean(false);
     plugin = ReflectionUtils.createClazzInstance(pluginDesc.getClassName(),
         new Class[] { VertexManagerPluginContext.class }, new Object[] { pluginContext });
-    payload = pluginDesc.getUserPayload();
     execService = appContext.getExecService();
     eventQueue = new LinkedBlockingQueue<VertexManagerEvent>();
     eventInFlight = new AtomicBoolean(false);
-    pluginFailed = new AtomicBoolean(false);
   }
 
   public VertexManagerPlugin getPlugin() {
