@@ -18,7 +18,6 @@
 
 package org.apache.tez.dag.app.launcher;
 
-
 import java.io.IOException;
 import java.net.InetAddress;
 import java.net.UnknownHostException;
@@ -44,10 +43,14 @@ import com.google.common.util.concurrent.ListeningExecutorService;
 import com.google.common.util.concurrent.MoreExecutors;
 import com.google.common.util.concurrent.ThreadFactoryBuilder;
 
+import org.apache.tez.common.DagContainerLauncher;
+import org.apache.tez.common.ReflectionUtils;
 import org.apache.tez.common.TezUtils;
-import org.apache.tez.hadoop.shim.DefaultHadoopShim;
+import org.apache.tez.dag.records.TezDAGID;
+import org.apache.tez.common.security.JobTokenSecretManager;
+import org.apache.tez.runtime.library.common.TezRuntimeUtils;
+import org.apache.tez.runtime.library.common.shuffle.ShuffleUtils;
 import org.apache.tez.serviceplugins.api.ContainerLaunchRequest;
-import org.apache.tez.serviceplugins.api.ContainerLauncher;
 import org.apache.tez.serviceplugins.api.ContainerLauncherContext;
 import org.apache.tez.serviceplugins.api.ContainerStopRequest;
 import org.apache.tez.serviceplugins.api.TaskAttemptEndReason;
@@ -68,7 +71,6 @@ import org.apache.tez.dag.app.TaskCommunicatorManagerInterface;
 import org.apache.tez.dag.app.TezTaskCommunicatorImpl;
 import org.apache.tez.runtime.api.ExecutionContext;
 import org.apache.tez.runtime.api.impl.ExecutionContextImpl;
-import org.apache.tez.runtime.library.common.shuffle.ShuffleUtils;
 import org.apache.tez.runtime.task.TezChild;
 
 
@@ -77,7 +79,7 @@ import org.apache.tez.runtime.task.TezChild;
  * Since all (sub)tasks share the same local directory, they must be executed
  * sequentially in order to avoid creating/deleting the same files/dirs.
  */
-public class LocalContainerLauncher extends ContainerLauncher {
+public class LocalContainerLauncher extends DagContainerLauncher {
 
   private static final Logger LOG = LoggerFactory.getLogger(LocalContainerLauncher.class);
 
@@ -88,11 +90,16 @@ public class LocalContainerLauncher extends ContainerLauncher {
   private final Map<String, String> localEnv;
   private final ExecutionContext executionContext;
   private final int numExecutors;
-  private final boolean isPureLocalMode;
+  private final boolean isLocalMode;
+  int shufflePort = TezRuntimeUtils.INVALID_PORT;
+  private DeletionTracker deletionTracker;
 
-  private final ConcurrentHashMap<ContainerId, RunningTaskCallback>
+  private final ConcurrentHashMap<ContainerId, ListenableFuture<?>>
       runningContainers =
-      new ConcurrentHashMap<ContainerId, RunningTaskCallback>();
+      new ConcurrentHashMap<>();
+
+  private final ConcurrentHashMap<ContainerId, TezLocalCacheManager>
+          cacheManagers = new ConcurrentHashMap<>();
 
   private final ExecutorService callbackExecutor = Executors.newFixedThreadPool(1,
       new ThreadFactoryBuilder().setDaemon(true).setNameFormat("CallbackExecutor").build());
@@ -108,7 +115,7 @@ public class LocalContainerLauncher extends ContainerLauncher {
                                 AppContext context,
                                 TaskCommunicatorManagerInterface taskCommunicatorManagerInterface,
                                 String workingDirectory,
-                                boolean isPureLocalMode) throws UnknownHostException {
+                                boolean isLocalMode) throws UnknownHostException, TezException {
     // TODO Post TEZ-2003. Most of this information is dynamic and only available after the AM
     // starts up. It's not possible to set these up via a static payload.
     // Will need some kind of mechanism to dynamically crate payloads / bind to parameters
@@ -117,17 +124,10 @@ public class LocalContainerLauncher extends ContainerLauncher {
     this.context = context;
     this.tal = taskCommunicatorManagerInterface;
     this.workingDirectory = workingDirectory;
-    this.isPureLocalMode = isPureLocalMode;
-    if (isPureLocalMode) {
-      localEnv = Maps.newHashMap();
-      AuxiliaryServiceHelper.setServiceDataIntoEnv(
-          ShuffleUtils.SHUFFLE_HANDLER_SERVICE_ID, ByteBuffer.allocate(4).putInt(0), localEnv);
-    } else {
-      localEnv = System.getenv();
-    }
+    this.isLocalMode = isLocalMode;
 
     // Check if the hostname is set in the environment before overriding it.
-    String host = isPureLocalMode ? InetAddress.getLocalHost().getHostName() :
+    String host = isLocalMode ? InetAddress.getLocalHost().getHostName() :
         System.getenv(Environment.NM_HOST.name());
     executionContext = new ExecutionContextImpl(host);
 
@@ -138,6 +138,16 @@ public class LocalContainerLauncher extends ContainerLauncher {
       throw new TezUncheckedException(
           "Failed to parse user payload for " + LocalContainerLauncher.class.getSimpleName(), e);
     }
+    if (isLocalMode) {
+      String auxiliaryService = conf.get(TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID,
+          TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID_DEFAULT);
+      localEnv = Maps.newHashMap();
+      shufflePort = 0;
+      AuxiliaryServiceHelper.setServiceDataIntoEnv(
+          auxiliaryService, ByteBuffer.allocate(4).putInt(shufflePort), localEnv);
+    } else {
+      localEnv = System.getenv();
+    }
     numExecutors = conf.getInt(TezConfiguration.TEZ_AM_INLINE_TASK_EXECUTION_MAX_TASKS,
         TezConfiguration.TEZ_AM_INLINE_TASK_EXECUTION_MAX_TASKS_DEFAULT);
     Preconditions.checkState(numExecutors >=1, "Must have at least 1 executor");
@@ -145,6 +155,15 @@ public class LocalContainerLauncher extends ContainerLauncher {
         new ThreadFactoryBuilder().setDaemon(true).setNameFormat("LocalTaskExecutionThread #%d")
             .build());
     this.taskExecutorService = MoreExecutors.listeningDecorator(rawExecutor);
+    boolean cleanupDagDataOnComplete = ShuffleUtils.isTezShuffleHandler(conf)
+        && conf.getBoolean(TezConfiguration.TEZ_AM_DAG_CLEANUP_ON_COMPLETION,
+        TezConfiguration.TEZ_AM_DAG_CLEANUP_ON_COMPLETION_DEFAULT);
+    if (cleanupDagDataOnComplete) {
+      String deletionTrackerClassName = conf.get(TezConfiguration.TEZ_AM_DELETION_TRACKER_CLASS,
+          TezConfiguration.TEZ_AM_DELETION_TRACKER_CLASS_DEFAULT);
+      deletionTracker = ReflectionUtils.createClazzInstance(
+          deletionTrackerClassName, new Class[]{Configuration.class}, new Object[]{conf});
+    }
   }
 
   @Override
@@ -168,6 +187,9 @@ public class LocalContainerLauncher extends ContainerLauncher {
       taskExecutorService.shutdownNow();
     }
     callbackExecutor.shutdownNow();
+    if (deletionTracker != null) {
+      deletionTracker.shutdown();
+    }
   }
 
 
@@ -208,6 +230,10 @@ public class LocalContainerLauncher extends ContainerLauncher {
 
   private void handleLaunchFailed(Throwable t, ContainerId containerId) {
     String message;
+
+    // clean up distributed cache files
+    cleanupCacheFiles(containerId);
+
     if (t instanceof RejectedExecutionException) {
       message = "Failed to queue container launch for container Id: " + containerId;
     } else {
@@ -223,10 +249,22 @@ public class LocalContainerLauncher extends ContainerLauncher {
     String tokenIdentifier = context.getApplicationID().toString();
     try {
       TezChild tezChild;
+
       try {
         int taskCommId = context.getTaskCommunicatorIdentifier(event.getTaskCommunicatorName());
+
+        Configuration conf = context.getAMConf();
+        if (isLocalMode) {
+          TezLocalCacheManager cacheManager = new TezLocalCacheManager(
+              event.getContainerLaunchContext().getLocalResources(),
+              conf
+          );
+          cacheManagers.put(event.getContainerId(), cacheManager);
+          cacheManager.localize();
+        }
+
         tezChild =
-            createTezChild(context.getAMConf(), event.getContainerId(), tokenIdentifier,
+            createTezChild(conf, event.getContainerId(), tokenIdentifier,
                 context.getApplicationAttemptId().getAttemptId(), context.getLocalDirs(),
                 ((TezTaskCommunicatorImpl)tal.getTaskCommunicator(taskCommId).getTaskCommunicator()).getUmbilical(),
                 TezCommonUtils.parseCredentialsBytes(event.getContainerLaunchContext().getTokens().array()));
@@ -243,8 +281,11 @@ public class LocalContainerLauncher extends ContainerLauncher {
       ListenableFuture<TezChild.ContainerExecutionResult> runningTaskFuture =
           taskExecutorService.submit(createSubTask(tezChild, event.getContainerId()));
       RunningTaskCallback callback = new RunningTaskCallback(event.getContainerId());
-      runningContainers.put(event.getContainerId(), callback);
+      runningContainers.put(event.getContainerId(), runningTaskFuture);
       Futures.addCallback(runningTaskFuture, callback, callbackExecutor);
+      if (deletionTracker != null) {
+        deletionTracker.addNodeShufflePort(event.getNodeId(), shufflePort);
+      }
     } catch (RejectedExecutionException e) {
       handleLaunchFailed(e, event.getContainerId());
     }
@@ -252,19 +293,16 @@ public class LocalContainerLauncher extends ContainerLauncher {
 
   private void stop(ContainerStopRequest event) {
     // A stop_request will come in when a task completes and reports back or a preemption decision
-    // is made. Currently the LocalTaskScheduler does not support preemption. Also preemption
-    // will not work in local mode till Tez supports task preemption instead of container preemption.
-    RunningTaskCallback callback =
+    // is made.
+    ListenableFuture future =
         runningContainers.get(event.getContainerId());
-    if (callback == null) {
+    if (future == null) {
       LOG.info("Ignoring stop request for containerId: " + event.getContainerId());
     } else {
       LOG.info(
-          "Ignoring stop request for containerId {}. Relying on regular task shutdown for it to end",
+          "Stopping containerId: {}",
           event.getContainerId());
-      // Allow the tezChild thread to run it's course. It'll receive a shutdown request from the
-      // AM eventually since the task and container will be unregistered.
-      // This will need to be fixed once interrupting tasks is supported.
+      future.cancel(true);
     }
     // Send this event to maintain regular control flow. This isn't of much use though.
     getContext().containerStopRequested(event.getContainerId());
@@ -298,6 +336,9 @@ public class LocalContainerLauncher extends ContainerLauncher {
                 (result.getThrowable() == null ? null : result.getThrowable().getMessage()) :
                 result.getErrorMessage(), TaskAttemptEndReason.APPLICATION_ERROR);
       }
+
+      // clean up distributed cache files
+      cleanupCacheFiles(containerId);
     }
 
     @Override
@@ -317,6 +358,22 @@ public class LocalContainerLauncher extends ContainerLauncher {
             TezChild.ContainerExecutionResult.ExitStatus.SUCCESS.getExitCode(),
             "CancellationException", TaskAttemptEndReason.COMMUNICATION_ERROR.CONTAINER_EXITED);
       }
+
+      // clean up distributed cache files
+      cleanupCacheFiles(containerId);
+    }
+  }
+
+  private void cleanupCacheFiles(ContainerId container) {
+    if (isLocalMode) {
+      TezLocalCacheManager manager = cacheManagers.remove(container);
+      try {
+        if (manager != null) {
+          manager.cleanup();
+        }
+      } catch (IOException e) {
+        LOG.info("Unable to clean up local cache files: ", e);
+      }
     }
   }
 
@@ -347,7 +404,7 @@ public class LocalContainerLauncher extends ContainerLauncher {
     Map<String, String> containerEnv = new HashMap<String, String>();
     containerEnv.putAll(localEnv);
     // Use the user from env if it's available.
-    String user = isPureLocalMode ? System.getenv(Environment.USER.name()) : context.getUser();
+    String user = isLocalMode ? System.getenv(Environment.USER.name()) : context.getUser();
     containerEnv.put(Environment.USER.name(), user);
 
     long memAvailable;
@@ -381,4 +438,11 @@ public class LocalContainerLauncher extends ContainerLauncher {
     }
   }
 
+  @Override
+  public void dagComplete(TezDAGID dag, JobTokenSecretManager jobTokenSecretManager) {
+    if (deletionTracker != null) {
+      deletionTracker.dagComplete(dag, jobTokenSecretManager);
+    }
+  }
+
 }
