@@ -38,6 +38,8 @@ import org.apache.tez.dag.api.NamedEntityDescriptor;
 import org.apache.tez.dag.api.TezConstants;
 import org.apache.tez.dag.app.ServicePluginLifecycleAbstractService;
 import org.apache.tez.dag.app.dag.event.DAGAppMasterEventUserServiceFatalError;
+import org.apache.tez.serviceplugins.api.DagInfo;
+import org.apache.tez.serviceplugins.api.ServicePluginError;
 import org.apache.tez.serviceplugins.api.TaskScheduler;
 import org.apache.tez.serviceplugins.api.TaskSchedulerContext;
 import org.apache.tez.serviceplugins.api.TaskSchedulerContext.AppFinalStatus;
@@ -87,6 +89,8 @@ import org.apache.tez.dag.app.rm.node.AMNodeEventTaskAttemptEnded;
 import org.apache.tez.dag.app.rm.node.AMNodeEventTaskAttemptSucceeded;
 import org.apache.tez.dag.app.web.WebUIService;
 import org.apache.tez.dag.records.TaskAttemptTerminationCause;
+import org.apache.tez.hadoop.shim.HadoopShim;
+import org.apache.tez.hadoop.shim.HadoopShimsLoader;
 
 import com.google.common.base.Preconditions;
 
@@ -122,7 +126,7 @@ public class TaskSchedulerManager extends AbstractService implements
   @VisibleForTesting
   final ExecutorService appCallbackExecutor;
 
-  private final boolean isPureLocalMode;
+  private final boolean isLocalMode;
   // If running in non local-only mode, the YARN task scheduler will always run to take care of
   // registration with YARN and heartbeats to YARN.
   // Splitting registration and heartbeats is not straight-forward due to the taskScheduler being
@@ -130,10 +134,13 @@ public class TaskSchedulerManager extends AbstractService implements
   // Custom AppIds to avoid container conflicts if there's multiple sources
   private final long SCHEDULER_APP_ID_BASE = 111101111;
   private final long SCHEDULER_APP_ID_INCREMENT = 111111111;
+  private final HadoopShim hadoopShim;
 
   BlockingQueue<AMSchedulerEvent> eventQueue
                               = new LinkedBlockingQueue<AMSchedulerEvent>();
 
+  private final String yarnSchedulerClassName;
+
   // Not tracking container / task to schedulerId. Instead relying on everything flowing through
   // the system and being propagated back via events.
 
@@ -157,7 +164,10 @@ public class TaskSchedulerManager extends AbstractService implements
     this.taskSchedulerDescriptors = null;
     this.webUI = null;
     this.historyUrl = null;
-    this.isPureLocalMode = false;
+    this.isLocalMode = false;
+    this.hadoopShim = new HadoopShimsLoader(appContext.getAMConf()).getHadoopShim();
+    this.yarnSchedulerClassName = appContext.getAMConf().get(TezConfiguration.TEZ_AM_YARN_SCHEDULER_CLASS,
+        TezConfiguration.TEZ_AM_YARN_SCHEDULER_CLASS_DEFAULT);
   }
 
   /**
@@ -169,7 +179,7 @@ public class TaskSchedulerManager extends AbstractService implements
    * @param webUI
    * @param schedulerDescriptors the list of scheduler descriptors. Tez internal classes will not have the class names populated.
    *                         An empty list defaults to using the YarnTaskScheduler as the only source.
-   * @param isPureLocalMode whether the AM is running in local mode
+   * @param isLocalMode whether the AM is running in local mode
    */
   @SuppressWarnings("rawtypes")
   public TaskSchedulerManager(AppContext appContext,
@@ -177,7 +187,8 @@ public class TaskSchedulerManager extends AbstractService implements
                               ContainerSignatureMatcher containerSignatureMatcher,
                               WebUIService webUI,
                               List<NamedEntityDescriptor> schedulerDescriptors,
-                              boolean isPureLocalMode) {
+                              boolean isLocalMode,
+                              HadoopShim hadoopShim) {
     super(TaskSchedulerManager.class.getName());
     Preconditions.checkArgument(schedulerDescriptors != null && !schedulerDescriptors.isEmpty(),
         "TaskSchedulerDescriptors must be specified");
@@ -187,7 +198,10 @@ public class TaskSchedulerManager extends AbstractService implements
     this.containerSignatureMatcher = containerSignatureMatcher;
     this.webUI = webUI;
     this.historyUrl = getHistoryUrl();
-    this.isPureLocalMode = isPureLocalMode;
+    this.isLocalMode = isLocalMode;
+    this.hadoopShim = hadoopShim;
+    this.yarnSchedulerClassName = appContext.getAMConf().get(TezConfiguration.TEZ_AM_YARN_SCHEDULER_CLASS,
+        TezConfiguration.TEZ_AM_YARN_SCHEDULER_CLASS_DEFAULT);
     this.appCallbackExecutor = createAppCallbackExecutorService();
     if (this.webUI != null) {
       this.webUI.setHistoryUrl(this.historyUrl);
@@ -257,6 +271,9 @@ public class TaskSchedulerManager extends AbstractService implements
     case S_TA_LAUNCH_REQUEST:
       handleTaLaunchRequest((AMSchedulerEventTALaunchRequest) sEvent);
       break;
+    case S_TA_STATE_UPDATED:
+      handleTAStateUpdated((AMSchedulerEventTAStateUpdated) sEvent);
+      break;
     case S_TA_ENDED: // TaskAttempt considered complete.
       AMSchedulerEventTAEnded event = (AMSchedulerEventTAEnded)sEvent;
       switch(event.getState()) {
@@ -403,10 +420,14 @@ public class TaskSchedulerManager extends AbstractService implements
       // Inform the Node - the task has asked to be STOPPED / has already
       // stopped.
       // AMNodeImpl blacklisting logic does not account for KILLED attempts.
-      sendEvent(new AMNodeEventTaskAttemptEnded(appContext.getAllContainers().
-          get(attemptContainerId).getContainer().getNodeId(), event.getSchedulerId(),
-          attemptContainerId,
-          attempt.getID(), event.getState() == TaskAttemptState.FAILED));
+      AMContainer amContainer = appContext.getAllContainers().get(attemptContainerId);
+      // DAG can be shutting down so protect against container cleanup race
+      if (amContainer != null) {
+        Container container = amContainer.getContainer();
+        sendEvent(new AMNodeEventTaskAttemptEnded(container.getNodeId(), event.getSchedulerId(),
+            attemptContainerId,
+            attempt.getID(), event.getState() == TaskAttemptState.FAILED));
+      }
     }
   }
 
@@ -419,9 +440,14 @@ public class TaskSchedulerManager extends AbstractService implements
     if (event.getUsedContainerId() != null) {
       sendEvent(new AMContainerEventTASucceeded(usedContainerId,
           event.getAttemptID()));
-      sendEvent(new AMNodeEventTaskAttemptSucceeded(appContext.getAllContainers().
-          get(usedContainerId).getContainer().getNodeId(), event.getSchedulerId(), usedContainerId,
-          event.getAttemptID()));
+      AMContainer amContainer = appContext.getAllContainers().get(usedContainerId);
+      // DAG can be shutting down so protect against container cleanup race
+      if (amContainer != null) {
+        Container container = amContainer.getContainer();
+        sendEvent(new AMNodeEventTaskAttemptSucceeded(container.getNodeId(), event.getSchedulerId(),
+            usedContainerId,
+            event.getAttemptID()));
+      }
     }
 
     boolean wasContainerAllocated = false;
@@ -519,6 +545,23 @@ public class TaskSchedulerManager extends AbstractService implements
     }
   }
 
+  private void handleTAStateUpdated(AMSchedulerEventTAStateUpdated event) {
+    try {
+      taskSchedulers[event.getSchedulerId()].taskStateUpdated(event.getTaskAttempt(), event.getState());
+    } catch (Exception e) {
+      String msg = "Error in TaskScheduler for handling Task State Update"
+          + ", eventType=" + event.getType()
+          + ", scheduler=" + Utils.getTaskSchedulerIdentifierString(event.getSchedulerId(), appContext)
+          + ", taskAttemptId=" + event.getTaskAttempt().getID()
+          + ", state=" + event.getState();
+      LOG.error(msg, e);
+      sendEvent(
+          new DAGAppMasterEventUserServiceFatalError(
+              DAGAppMasterEventType.TASK_SCHEDULER_SERVICE_FATAL_ERROR,
+              msg, e));
+    }
+  }
+
   @VisibleForTesting
   TaskScheduler createTaskScheduler(String host, int port, String trackingUrl,
                                     AppContext appContext,
@@ -546,9 +589,11 @@ public class TaskSchedulerManager extends AbstractService implements
 
   @VisibleForTesting
   TaskScheduler createYarnTaskScheduler(TaskSchedulerContext taskSchedulerContext,
-                                        int schedulerId) {
-    LOG.info("Creating TaskScheduler: YarnTaskSchedulerService");
-    return new YarnTaskSchedulerService(taskSchedulerContext);
+                                        int schedulerId) throws TezException {
+    LOG.info("Creating YARN TaskScheduler: {}", yarnSchedulerClassName);
+    return ReflectionUtils.createClazzInstance(yarnSchedulerClassName,
+        new Class[] { TaskSchedulerContext.class },
+        new Object[] { taskSchedulerContext });
   }
 
   @VisibleForTesting
@@ -577,8 +622,11 @@ public class TaskSchedulerManager extends AbstractService implements
     int j = 0;
     for (int i = 0; i < taskSchedulerDescriptors.length; i++) {
       long customAppIdIdentifier;
-      if (isPureLocalMode || taskSchedulerDescriptors[i].getEntityName().equals(
-          TezConstants.getTezYarnServicePluginName())) { // Use the app identifier from the appId.
+      if ((isLocalMode && taskSchedulerDescriptors[i].getEntityName()
+          .equals(TezConstants.getTezUberServicePluginName()) ||
+          taskSchedulerDescriptors[i].getEntityName()
+              .equals(TezConstants.getTezYarnServicePluginName()))) {
+        // Use the provided appId instead of constructing one for containers.
         customAppIdIdentifier = appContext.getApplicationID().getClusterTimestamp();
       } else {
         customAppIdIdentifier = SCHEDULER_APP_ID_BASE + (j++ * SCHEDULER_APP_ID_INCREMENT);
@@ -653,12 +701,14 @@ public class TaskSchedulerManager extends AbstractService implements
 
   public void initiateStop() {
     for (int i = 0 ; i < taskSchedulers.length ; i++) {
-      try {
-        taskSchedulers[i].getTaskScheduler().initiateStop();
-      } catch (Exception e) {
-        // Ignore for now as scheduler stop invoked on shutdown
-        LOG.error("Failed to do a clean initiateStop for Scheduler: "
-            + Utils.getTaskSchedulerIdentifierString(i, appContext), e);
+      if (taskSchedulers[i] != null) {
+        try {
+          taskSchedulers[i].getTaskScheduler().initiateStop();
+        } catch (Exception e) {
+          // Ignore for now as scheduler stop invoked on shutdown
+          LOG.error("Failed to do a clean initiateStop for Scheduler: "
+              + Utils.getTaskSchedulerIdentifierString(i, appContext), e);
+        }
       }
     }
   }
@@ -701,10 +751,15 @@ public class TaskSchedulerManager extends AbstractService implements
     // because the deallocateTask downcall may have raced with the
     // taskAllocated() upcall
     assert task.equals(taskAttempt);
- 
-    if (appContext.getAllContainers().get(containerId).getState() == AMContainerState.ALLOCATED) {
-      sendEvent(new AMContainerEventLaunchRequest(containerId, taskAttempt.getVertexID(),
-          event.getContainerContext(), event.getLauncherId(), event.getTaskCommId()));
+
+    AMContainer amContainer = appContext.getAllContainers().get(containerId);
+    // Even though we just added this container,
+    // DAG can be shutting down so protect against container cleanup race
+    if (amContainer != null) {
+      if (amContainer.getState() == AMContainerState.ALLOCATED) {
+        sendEvent(new AMContainerEventLaunchRequest(containerId, taskAttempt.getVertexID(),
+            event.getContainerContext(), event.getLauncherId(), event.getTaskCommId()));
+      }
     }
     sendEvent(new AMContainerEventAssignTA(containerId, taskAttempt.getID(),
         event.getRemoteTaskSpec(), event.getContainerContext().getLocalResources(), event
@@ -763,9 +818,11 @@ public class TaskSchedulerManager extends AbstractService implements
       int schedulerId,
       Resource maxContainerCapability,
       Map<ApplicationAccessType, String> appAcls, 
-      ByteBuffer clientAMSecretKey) {
+      ByteBuffer clientAMSecretKey,
+      String queueName) {
     this.appContext.getClusterInfo().setMaxContainerCapability(
         maxContainerCapability);
+    this.appContext.setQueueName(queueName);
     this.appAcls = appAcls;
     this.clientService.setClientAMSecretKey(clientAMSecretKey);
   }
@@ -793,6 +850,8 @@ public class TaskSchedulerManager extends AbstractService implements
       } else {
         finishState = FinalApplicationStatus.UNDEFINED;
       }
+      finishState = hadoopShim.applyFinalApplicationStatusCorrection(finishState,
+          dagAppMaster.isSession(), appMasterState == DAGAppMasterState.ERROR);
       List<String> diagnostics = dagAppMaster.getDiagnostics();
       if(diagnostics != null) {
         for (String s : diagnostics) {
@@ -844,9 +903,35 @@ public class TaskSchedulerManager extends AbstractService implements
     return dagAppMaster.getProgress();
   }
 
-  public void onError(int schedulerId, Throwable t) {
-    LOG.info("Error reported by scheduler {} - {}", schedulerId, t);
-    sendEvent(new DAGAppMasterEventSchedulingServiceError(t));
+  public void reportError(int taskSchedulerIndex, ServicePluginError servicePluginError,
+                          String diagnostics,
+                          DagInfo dagInfo) {
+    if (servicePluginError == YarnTaskSchedulerServiceError.RESOURCEMANAGER_ERROR) {
+      LOG.info("Error reported by scheduler {} - {}",
+          Utils.getTaskSchedulerIdentifierString(taskSchedulerIndex, appContext) + ": " +
+              diagnostics);
+      if (taskSchedulerDescriptors[taskSchedulerIndex].getClassName().equals(yarnSchedulerClassName)) {
+        LOG.warn(
+            "Reporting a SchedulerServiceError to the DAGAppMaster since the error" +
+                " was reported by the YARN task scheduler");
+        sendEvent(new DAGAppMasterEventSchedulingServiceError(diagnostics));
+      }
+    } else if (servicePluginError.getErrorType() == ServicePluginError.ErrorType.PERMANENT) {
+      String msg = "Fatal error reported by TaskScheduler"
+          + ", scheduler=" + Utils.getTaskSchedulerIdentifierString(taskSchedulerIndex, appContext)
+          + ", servicePluginError=" + servicePluginError
+          + ", diagnostics= " + (diagnostics == null ? "" : diagnostics);
+      LOG.error(msg);
+      sendEvent(
+          new DAGAppMasterEventUserServiceFatalError(
+              DAGAppMasterEventType.TASK_SCHEDULER_SERVICE_FATAL_ERROR,
+              msg, null));
+    } else {
+      Utils.processNonFatalServiceErrorReport(
+          Utils.getTaskSchedulerIdentifierString(taskSchedulerIndex, appContext),
+          servicePluginError, diagnostics, dagInfo,
+          appContext, "TaskScheduler");
+    }
   }
 
   public void dagCompleted() {
@@ -870,12 +955,20 @@ public class TaskSchedulerManager extends AbstractService implements
     // the context has updated information.
   }
 
+  public int getVertexIndexForTask(Object task) {
+    TaskAttempt attempt = (TaskAttempt) task;
+    return attempt.getVertexID().getId();
+  }
+
   public void preemptContainer(int schedulerId, ContainerId containerId) {
     // TODO Why is this making a call back into the scheduler, when the call is originating from there.
     // An AMContainer instance should already exist if an attempt is being made to preempt it
     AMContainer amContainer = appContext.getAllContainers().get(containerId);
     try {
-      taskSchedulers[amContainer.getTaskSchedulerIdentifier()].deallocateContainer(containerId);
+      // DAG can be shutting down so protect against container cleanup race
+      if (amContainer != null) {
+        taskSchedulers[amContainer.getTaskSchedulerIdentifier()].deallocateContainer(containerId);
+      }
     } catch (Exception e) {
       String msg = "Error in TaskScheduler when preempting container"
           + ", scheduler=" + Utils.getTaskSchedulerIdentifierString(amContainer.getTaskSchedulerIdentifier(), appContext)
@@ -916,13 +1009,21 @@ public class TaskSchedulerManager extends AbstractService implements
   }
 
   public boolean hasUnregistered() {
+    // Only return true if all task schedulers that were registered successfully unregister
+    if (taskSchedulers.length == 0) {
+      return false;
+    }
     boolean result = true;
-    for (int i = 0 ; i < taskSchedulers.length ; i++) {
+    for (int i = 0; i < taskSchedulers.length; i++) {
       // Explicitly not catching any exceptions around this API
       // No clear route to recover. Better to crash.
+      if (taskSchedulers[i] == null) {
+        return false;
+      }
       try {
         result = result & this.taskSchedulers[i].hasUnregistered();
       } catch (Exception e) {
+        result = false;
         String msg = "Error in TaskScheduler when checking if a scheduler has unregistered"
             + ", scheduler=" + Utils.getTaskSchedulerIdentifierString(i, appContext);
         LOG.error(msg, e);
@@ -943,14 +1044,11 @@ public class TaskSchedulerManager extends AbstractService implements
     Configuration config = this.appContext.getAMConf();
     String historyUrl = "";
 
-    String loggingClass =  config.get(TezConfiguration.TEZ_HISTORY_LOGGING_SERVICE_CLASS, "");
     String historyUrlTemplate = config.get(TezConfiguration.TEZ_AM_TEZ_UI_HISTORY_URL_TEMPLATE,
             TezConfiguration.TEZ_AM_TEZ_UI_HISTORY_URL_TEMPLATE_DEFAULT);
     String historyUrlBase = config.get(TezConfiguration.TEZ_HISTORY_URL_BASE, "");
 
-
-    if (loggingClass.equals("org.apache.tez.dag.history.logging.ats.ATSHistoryLoggingService") &&
-        !historyUrlTemplate.isEmpty() &&
+    if (!historyUrlTemplate.isEmpty() &&
         !historyUrlBase.isEmpty()) {
       // replace the placeholders, while tolerating extra or missing "/" in input. replace all
       // instances of consecutive "/" with single (except for the http(s):// case
@@ -959,13 +1057,25 @@ public class TaskSchedulerManager extends AbstractService implements
           .replaceAll(HISTORY_URL_BASE, historyUrlBase)
           .replaceAll("([^:])/{2,}", "$1/");
 
-      // make sure we have a valid scheme
-      if (!historyUrl.startsWith("http")) {
-        historyUrl = "http://" + historyUrl;
+      // FIXME: we should not need this check in the first place.
+      // Might be better to remove it in the next release and not fix badly configured URLs.
+      boolean checkSchemePrefix = config.getBoolean(
+          TezConfiguration.TEZ_AM_UI_HISTORY_URL_SCHEME_CHECK_ENABLED,
+          TezConfiguration.TEZ_AM_UI_HISTORY_URL_SCHEME_CHECK_ENABLED_DEFAULT);
+
+      if (checkSchemePrefix) {
+        // make sure we have a valid scheme
+        if (!historyUrl.startsWith("http")) {
+          historyUrl = "http://" + historyUrl;
+        }
       }
     }
 
     return historyUrl;
   }
 
+  public String getTaskSchedulerClassName(int taskSchedulerIndex) {
+    return taskSchedulers[taskSchedulerIndex].getTaskScheduler().getClass().getName();
+  }
+
 }
