@@ -35,7 +35,7 @@ import java.util.concurrent.PriorityBlockingQueue;
 import java.util.concurrent.atomic.AtomicBoolean;
 import java.util.concurrent.atomic.AtomicInteger;
 
-import org.apache.hadoop.yarn.api.records.ContainerState;
+import org.apache.hadoop.util.StringUtils;
 import org.apache.tez.common.TezUtils;
 import org.apache.tez.serviceplugins.api.TaskScheduler;
 import org.apache.tez.serviceplugins.api.TaskSchedulerContext;
@@ -47,6 +47,7 @@ import org.apache.commons.lang.exception.ExceptionUtils;
 import org.apache.commons.math3.random.RandomDataGenerator;
 import org.apache.hadoop.classification.InterfaceAudience.Private;
 import org.apache.hadoop.conf.Configuration;
+import org.apache.hadoop.util.ExitUtil;
 import org.apache.hadoop.yarn.api.protocolrecords.RegisterApplicationMasterResponse;
 import org.apache.hadoop.yarn.api.records.Container;
 import org.apache.hadoop.yarn.api.records.ContainerId;
@@ -62,6 +63,7 @@ import org.apache.hadoop.yarn.exceptions.YarnException;
 import org.apache.hadoop.yarn.util.RackResolver;
 import org.apache.hadoop.yarn.util.resource.Resources;
 import org.apache.tez.serviceplugins.api.TaskAttemptEndReason;
+import org.apache.tez.dag.app.dag.TaskAttempt;
 import org.apache.tez.dag.api.TezConfiguration;
 import org.apache.tez.dag.api.TezUncheckedException;
 import org.apache.tez.common.ContainerSignatureMatcher;
@@ -117,7 +119,7 @@ public class YarnTaskSchedulerService extends TaskScheduler
       new HashMap<ContainerId, HeldContainer>();
   
   Set<Priority> priorityHasAffinity = Sets.newHashSet();
-  
+
   Set<NodeId> blacklistedNodes = Collections
       .newSetFromMap(new ConcurrentHashMap<NodeId, Boolean>());
   
@@ -134,9 +136,9 @@ public class YarnTaskSchedulerService extends TaskScheduler
 
   AtomicBoolean isStopStarted = new AtomicBoolean(false);
 
-  private ContainerAssigner NODE_LOCAL_ASSIGNER = new NodeLocalContainerAssigner();
-  private ContainerAssigner RACK_LOCAL_ASSIGNER = new RackLocalContainerAssigner();
-  private ContainerAssigner NON_LOCAL_ASSIGNER = new NonLocalContainerAssigner();
+  private ContainerAssigner NODE_LOCAL_ASSIGNER;
+  private ContainerAssigner RACK_LOCAL_ASSIGNER;
+  private ContainerAssigner NON_LOCAL_ASSIGNER;
 
   DelayedContainerManager delayedContainerManager;
   long localitySchedulingDelay;
@@ -339,6 +341,9 @@ public class YarnTaskSchedulerService extends TaskScheduler
     Preconditions.checkArgument(preemptionMaxWaitTime >=0, "Preemption max wait time must be >=0");
 
     delayedContainerManager = new DelayedContainerManager();
+    NODE_LOCAL_ASSIGNER = new NodeLocalContainerAssigner();
+    RACK_LOCAL_ASSIGNER = new RackLocalContainerAssigner();
+    NON_LOCAL_ASSIGNER = new NonLocalContainerAssigner();
     LOG.info("YarnTaskScheduler initialized with configuration: " +
             "maxRMHeartbeatInterval: " + heartbeatIntervalMax +
             ", containerReuseEnabled: " + shouldReuseContainers +
@@ -367,7 +372,8 @@ public class YarnTaskSchedulerService extends TaskScheduler
       getContext().setApplicationRegistrationData(
           response.getMaximumResourceCapability(),
           response.getApplicationACLs(),
-          response.getClientToAMTokenMasterKey());
+          response.getClientToAMTokenMasterKey(),
+          response.getQueue());
 
       delayedContainerManager.start();
     } catch (YarnException e) {
@@ -470,7 +476,12 @@ public class YarnTaskSchedulerService extends TaskScheduler
         }
 
         // container neither allocated nor released
-        LOG.info("Ignoring unknown container: " + containerStatus.getContainerId());
+        if (delayedContainer != null) {
+          LOG.info("Delayed container {} completed", containerStatus.getContainerId());
+          maybeRescheduleContainerAtPriority(delayedContainer.getContainer().getPriority());
+        } else {
+          LOG.info("Ignoring unknown container: " + containerStatus.getContainerId());
+        }
       }
     }
 
@@ -674,7 +685,7 @@ public class YarnTaskSchedulerService extends TaskScheduler
         heldContainer.resetLocalityMatchLevel();
         delayedContainerManager.addDelayedContainer(
             heldContainer.getContainer(), currentTime
-                + localitySchedulingDelay);        
+                + localitySchedulingDelay);
       }
     } else if (state.equals(AMState.RUNNING_APP)) {
       // clear min held containers since we need to allocate to tasks
@@ -916,7 +927,9 @@ public class YarnTaskSchedulerService extends TaskScheduler
       LOG.error("Got TaskSchedulerError, " + ExceptionUtils.getStackTrace(t));
       return;
     }
-    getContext().onError(t);
+    LOG.error("Got Error from RMClient", t);
+    getContext().reportError(YarnTaskSchedulerServiceError.RESOURCEMANAGER_ERROR, StringUtils.stringifyException(t),
+        null);
   }
 
   @Override
@@ -1308,23 +1321,8 @@ public class YarnTaskSchedulerService extends TaskScheduler
           // to us anymore. So we need to ask for this again. If there is no
           // outstanding request at that priority then its fine to not ask again.
           // See TEZ-915 for more details
-          for (Map.Entry<Object, CookieContainerRequest> entry : taskRequests.entrySet()) {
-            Object task = entry.getKey();
-            CookieContainerRequest request = entry.getValue();
-            if (request.getPriority().equals(lowestPriNewContainer.getPriority())) {
-              LOG.info("Resending request for task again: " + task);
-              deallocateTask(task, true, null, null);
-              allocateTask(task, request.getCapability(), 
-                  (request.getNodes() == null ? null : 
-                    request.getNodes().toArray(new String[request.getNodes().size()])), 
-                    (request.getRacks() == null ? null : 
-                      request.getRacks().toArray(new String[request.getRacks().size()])), 
-                    request.getPriority(), 
-                    request.getCookie().getContainerSignature(),
-                    request.getCookie().getAppCookie());
-              break;
-            }
-          }
+          maybeRescheduleContainerAtPriority(lowestPriNewContainer.getPriority());
+
           // come back and free more new containers if needed
           continue;
         }
@@ -1421,6 +1419,26 @@ public class YarnTaskSchedulerService extends TaskScheduler
     return true;
   }
 
+  private void maybeRescheduleContainerAtPriority(Priority priority) {
+    for (Map.Entry<Object, CookieContainerRequest> entry : taskRequests.entrySet()) {
+      Object task = entry.getKey();
+      CookieContainerRequest request = entry.getValue();
+      if (request.getPriority().equals(priority)) {
+        LOG.info("Resending request for task again: " + task);
+        deallocateTask(task, true, null, null);
+        allocateTask(task, request.getCapability(),
+            (request.getNodes() == null ? null :
+              request.getNodes().toArray(new String[request.getNodes().size()])),
+              (request.getRacks() == null ? null :
+                request.getRacks().toArray(new String[request.getRacks().size()])),
+                request.getPriority(),
+                request.getCookie().getContainerSignature(),
+                request.getCookie().getAppCookie());
+        break;
+      }
+    }
+  }
+
   private boolean fitsIn(Resource toFit, Resource resource) {
     // YARN-893 prevents using correct library code
     //return Resources.fitsIn(toFit, resource);
@@ -1516,6 +1534,11 @@ public class YarnTaskSchedulerService extends TaskScheduler
   private boolean canAssignTaskToContainer(
       CookieContainerRequest cookieContainerRequest, Container container) {
     HeldContainer heldContainer = heldContainers.get(container.getId());
+    Object task = getTask(cookieContainerRequest);
+    if (task instanceof TaskAttempt
+        && ((TaskAttempt) task).getNodesWithSiblingRunningAttempts().contains(container.getNodeId())) {
+      return false;
+    }
     if (heldContainer == null || heldContainer.isNew()) { // New container.
       return true;
     } else {
@@ -1767,9 +1790,9 @@ public class YarnTaskSchedulerService extends TaskScheduler
       Container container = entry.getValue();
       // check for blacklisted nodes. There may be race conditions between
       // setting blacklist and receiving allocations
+      CookieContainerRequest request = entry.getKey();
+      Object task = getTask(request);
       if (blacklistedNodes.contains(container.getNodeId())) {
-        CookieContainerRequest request = entry.getKey();
-        Object task = getTask(request);
         LOG.info("Container: " + container.getId() + 
             " allocated on blacklisted node: " + container.getNodeId() + 
             " for task: " + task);
@@ -1953,6 +1976,14 @@ public class YarnTaskSchedulerService extends TaskScheduler
     
     @Override
     public void run() {
+      try {
+        mainLoop();
+      } catch (Throwable e) {
+        ExitUtil.terminate(1, e);
+      }
+    }
+
+    private void mainLoop() {
       while(running) {
         // Try assigning all containers if there's a request to do so.
         if (tryAssigningAll) {
@@ -2128,6 +2159,11 @@ public class YarnTaskSchedulerService extends TaskScheduler
       boolean added =  false;
       synchronized(this) {
         added = delayedContainers.offer(delayedContainer);
+        if (drainedDelayedContainersForTest != null) {
+          synchronized (drainedDelayedContainersForTest) {
+            drainedDelayedContainersForTest.set(false);
+          }
+        }
         this.notify();
       }
       if (!added) {
@@ -2335,11 +2371,12 @@ public class YarnTaskSchedulerService extends TaskScheduler
       // Merge the container signatures to account for any changes to the container
       // footprint. For example, re-localization of additional resources will
       // cause the held container's signature to change.
-      lastAssignedContainerSignature = taskInfo.getCookie().getContainerSignature();
-      if (lastTaskInfo != null && lastTaskInfo.getCookie().getContainerSignature() != null) {
+      if (lastAssignedContainerSignature != null) {
         lastAssignedContainerSignature = signatureMatcher.union(
-            lastTaskInfo.getCookie().getContainerSignature(),
+            lastAssignedContainerSignature,
             taskInfo.getCookie().getContainerSignature());
+      } else {
+        lastAssignedContainerSignature = taskInfo.getCookie().getContainerSignature();
       }
       lastTaskInfo = taskInfo;
     }
