@@ -32,9 +32,12 @@ import org.apache.tez.Utils;
 import org.apache.tez.common.TezUtilsInternal;
 import org.apache.tez.dag.app.dag.event.DAGAppMasterEventType;
 import org.apache.tez.dag.app.dag.event.DAGAppMasterEventUserServiceFatalError;
+import org.apache.tez.dag.app.rm.node.AMNodeEventContainerCompleted;
 import org.apache.tez.serviceplugins.api.ContainerEndReason;
 import org.apache.tez.common.ContainerSignatureMatcher;
 import org.apache.tez.serviceplugins.api.TaskAttemptEndReason;
+import org.apache.tez.state.OnStateChangedCallback;
+import org.apache.tez.state.StateMachineTez;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 import org.apache.hadoop.security.Credentials;
@@ -48,7 +51,6 @@ import org.apache.hadoop.yarn.event.EventHandler;
 import org.apache.hadoop.yarn.state.InvalidStateTransitonException;
 import org.apache.hadoop.yarn.state.MultipleArcTransition;
 import org.apache.hadoop.yarn.state.SingleArcTransition;
-import org.apache.hadoop.yarn.state.StateMachine;
 import org.apache.hadoop.yarn.state.StateMachineFactory;
 import org.apache.hadoop.yarn.util.Clock;
 import org.apache.tez.dag.app.AppContext;
@@ -91,6 +93,7 @@ public class AMContainerImpl implements AMContainer {
   private final int schedulerId;
   private final int launcherId;
   private final int taskCommId;
+  private String auxiliaryService;
 
   private final List<TezTaskAttemptID> completedAttempts =
       new LinkedList<TezTaskAttemptID>();
@@ -118,6 +121,8 @@ public class AMContainerImpl implements AMContainer {
 
   private Credentials credentials;
   private boolean credentialsChanged = false;
+
+  private boolean completedMessageSent = false;
   
   // TODO Consider registering with the TAL, instead of the TAL pulling.
   // Possibly after splitting TAL and ContainerListener.
@@ -127,8 +132,11 @@ public class AMContainerImpl implements AMContainer {
 
   // TODO Create a generic ERROR state. Container tries informing relevant components in this case.
 
+  private final NonRunningStateEnteredCallback NON_RUNNING_STATE_ENTERED_CALLBACK = new NonRunningStateEnteredCallback();
+
+  private final StateMachineTez<AMContainerState, AMContainerEventType, AMContainerEvent, AMContainerImpl>
+      stateMachine;
 
-  private final StateMachine<AMContainerState, AMContainerEventType, AMContainerEvent> stateMachine;
   private static final StateMachineFactory
       <AMContainerImpl, AMContainerState, AMContainerEventType, AMContainerEvent>
       stateMachineFactory =
@@ -313,7 +321,7 @@ public class AMContainerImpl implements AMContainer {
   // additional change - JvmID, YarnChild, etc depend on TaskType.
   public AMContainerImpl(Container container, ContainerHeartbeatHandler chh,
       TaskCommunicatorManagerInterface tal, ContainerSignatureMatcher signatureMatcher,
-      AppContext appContext, int schedulerId, int launcherId, int taskCommId) {
+      AppContext appContext, int schedulerId, int launcherId, int taskCommId, String auxiliaryService) {
     ReentrantReadWriteLock rwLock = new ReentrantReadWriteLock();
     this.readLock = rwLock.readLock();
     this.writeLock = rwLock.writeLock();
@@ -328,7 +336,20 @@ public class AMContainerImpl implements AMContainer {
     this.schedulerId = schedulerId;
     this.launcherId = launcherId;
     this.taskCommId = taskCommId;
-    this.stateMachine = stateMachineFactory.make(this);
+    this.auxiliaryService = auxiliaryService;
+    this.stateMachine = new StateMachineTez<>(stateMachineFactory.make(this), this);
+    augmentStateMachine();
+  }
+
+
+  private void augmentStateMachine() {
+    stateMachine
+        .registerStateEnteredCallback(AMContainerState.STOP_REQUESTED,
+            NON_RUNNING_STATE_ENTERED_CALLBACK)
+        .registerStateEnteredCallback(AMContainerState.STOPPING,
+            NON_RUNNING_STATE_ENTERED_CALLBACK)
+        .registerStateEnteredCallback(AMContainerState.COMPLETED,
+            NON_RUNNING_STATE_ENTERED_CALLBACK);
   }
 
   @Override
@@ -401,6 +422,7 @@ public class AMContainerImpl implements AMContainer {
     return this.taskCommId;
   }
 
+  @Override
   public boolean isInErrorState() {
     return inError;
   }
@@ -421,7 +443,7 @@ public class AMContainerImpl implements AMContainer {
         LOG.error("Can't handle event " + event.getType()
             + " at current state " + oldState + " for ContainerId "
             + this.containerId, e);
-        inError = true;
+        setError();
         // TODO Can't set state to COMPLETED. Add a default error state.
       }
       if (oldState != getState()) {
@@ -460,10 +482,8 @@ public class AMContainerImpl implements AMContainer {
       container.credentialsChanged = true;
 
       TezDAGID dagId = null;
-      Map<String, LocalResource> dagLocalResources = null;
       if (container.appContext.getCurrentDAG() != null) {
         dagId = container.appContext.getCurrentDAG().getID();
-        dagLocalResources = container.appContext.getCurrentDAG().getLocalResources();
       }
 
       // TODO TEZ-2625 This should ideally be handled inside of user code. Will change once
@@ -483,13 +503,13 @@ public class AMContainerImpl implements AMContainer {
                 msg, e));
         // We have not registered with any of the listeners etc yet. Send out a deallocateContainer
         // message and return. The AM will shutdown shortly.
-        container.inError = true;
+        container.setError();
         container.deAllocate();
         return;
       }
 
       ContainerLaunchContext clc = AMContainerHelpers.createContainerLaunchContext(
-          dagId, dagLocalResources,
+          dagId,
           container.appContext.getApplicationACLs(),
           container.getContainerId(),
           containerContext.getLocalResources(),
@@ -498,7 +518,7 @@ public class AMContainerImpl implements AMContainer {
           cAddress,
           containerContext.getCredentials(),
           container.appContext, container.container.getResource(),
-          container.appContext.getAMConf());
+          container.appContext.getAMConf(), container.auxiliaryService);
 
       // Registering now, so that in case of delayed NM response, the child
       // task is not told to die since the TAL does not know about the container.
@@ -516,7 +536,7 @@ public class AMContainerImpl implements AMContainer {
     @Override
     public void transition(AMContainerImpl container, AMContainerEvent cEvent) {
       AMContainerEventAssignTA event = (AMContainerEventAssignTA) cEvent;
-      container.inError = true;
+      container.setError();
       container.registerFailedAttempt(event.getTaskAttemptId());
       container.maybeSendNodeFailureForFailedAssignment(event
           .getTaskAttemptId());
@@ -962,7 +982,7 @@ public class AMContainerImpl implements AMContainer {
     @Override
     public void transition(AMContainerImpl container, AMContainerEvent cEvent) {
       AMContainerEventAssignTA event = (AMContainerEventAssignTA) cEvent;
-      container.inError = true;
+      container.setError();
       String errorMessage = "AttemptId: " + event.getTaskAttemptId() +
           " cannot be allocated to container: " + container.getContainerId() +
           " in " + container.getState() + " state";
@@ -1033,7 +1053,7 @@ public class AMContainerImpl implements AMContainer {
 
     @Override
     public void transition(AMContainerImpl container, AMContainerEvent cEvent) {
-      container.inError = true;
+      container.setError();
     }
   }
 
@@ -1047,7 +1067,7 @@ public class AMContainerImpl implements AMContainer {
       // think the container is still around and assign a task to it. The task
       // ends up getting a CONTAINER_KILLED message. Task could handle this by
       // asking for a reschedule in this case. Will end up FAILING the task instead of KILLING it.
-      container.inError = true;
+      container.setError();
       AMContainerEventAssignTA event = (AMContainerEventAssignTA) cEvent;
       String errorMessage = "AttemptId: " + event.getTaskAttemptId()
           + " cannot be allocated to container: " + container.getContainerId()
@@ -1059,9 +1079,19 @@ public class AMContainerImpl implements AMContainer {
     }
   }
 
+  private static class NonRunningStateEnteredCallback
+      implements OnStateChangedCallback<AMContainerState, AMContainerImpl> {
+
+    @Override
+    public void onStateChanged(AMContainerImpl amContainer,
+                               AMContainerState amContainerState) {
+      amContainer.handleNonRunningStateEntered();
+    }
+  }
+
   private void handleExtraTAAssign(
       AMContainerEventAssignTA event, TezTaskAttemptID currentTaId) {
-    this.inError = true;
+    setError();
     String errorMessage = "AMScheduler Error: Multiple simultaneous " +
         "taskAttempt allocations to: " + this.getContainerId() +
         ". Attempts: " + currentTaId + ", " + event.getTaskAttemptId() +
@@ -1079,6 +1109,19 @@ public class AMContainerImpl implements AMContainer {
     this.unregisterFromContainerListener();
   }
 
+  private void setError() {
+    this.inError = true;
+    handleNonRunningStateEntered();
+  }
+
+  private void handleNonRunningStateEntered() {
+    if (!completedMessageSent) {
+      completedMessageSent = true;
+      sendEvent(new AMNodeEventContainerCompleted(getContainer().getNodeId(),
+          schedulerId, containerId));
+    }
+  }
+
   protected void registerFailedAttempt(TezTaskAttemptID taId) {
     failedAssignments.add(taId);
   }
@@ -1100,12 +1143,12 @@ public class AMContainerImpl implements AMContainer {
 
   protected void sendTerminatedToTaskAttempt(
       TezTaskAttemptID taId, String message, TaskAttemptTerminationCause errCause) {
-    sendEvent(new TaskAttemptEventContainerTerminated(taId, message, errCause));
+    sendEvent(new TaskAttemptEventContainerTerminated(containerId, taId, message, errCause));
   }
   
   protected void sendContainerTerminatedBySystemToTaskAttempt(
     TezTaskAttemptID taId, String message, TaskAttemptTerminationCause errorCause) {
-      sendEvent(new TaskAttemptEventContainerTerminatedBySystem(taId, message, errorCause));
+      sendEvent(new TaskAttemptEventContainerTerminatedBySystem(containerId, taId, message, errorCause));
   }
 
   protected void sendTerminatingToTaskAttempt(TezTaskAttemptID taId,
