@@ -490,6 +490,9 @@ public class HistoryEventJsonConversion {
       primaryFilters.put(ATSConstants.CALLER_CONTEXT_TYPE,
           event.getDAGPlan().getCallerContext().getCallerType());
     }
+    if (event.getQueueName() != null) {
+      primaryFilters.put(ATSConstants.DAG_QUEUE_NAME, event.getQueueName());
+    }
 
     jsonObject.put(ATSConstants.PRIMARY_FILTERS, primaryFilters);
 
@@ -515,6 +518,9 @@ public class HistoryEventJsonConversion {
       otherInfo.put(ATSConstants.CALLER_CONTEXT_TYPE,
           event.getDAGPlan().getCallerContext().getCallerType());
     }
+    if (event.getQueueName() != null) {
+      otherInfo.put(ATSConstants.DAG_QUEUE_NAME, event.getQueueName());
+    }
     jsonObject.put(ATSConstants.OTHER_INFO, otherInfo);
 
     return jsonObject;
@@ -548,6 +554,9 @@ public class HistoryEventJsonConversion {
     if (event.getTaskAttemptError() != null) {
       otherInfo.put(ATSConstants.TASK_ATTEMPT_ERROR_ENUM, event.getTaskAttemptError().name());
     }
+    if (event.getTaskFailureType() != null) {
+      otherInfo.put(ATSConstants.TASK_FAILURE_TYPE, event.getTaskFailureType().name());
+    }
     otherInfo.put(ATSConstants.DIAGNOSTICS, event.getDiagnostics());
     otherInfo.put(ATSConstants.COUNTERS,
         DAGUtils.convertCountersToJSON(event.getCounters()));
@@ -555,6 +564,22 @@ public class HistoryEventJsonConversion {
       otherInfo.put(ATSConstants.LAST_DATA_EVENTS, 
           DAGUtils.convertDataEventDependencyInfoToJSON(event.getDataEvents()));
     }
+    if (event.getNodeId() != null) {
+      otherInfo.put(ATSConstants.NODE_ID, event.getNodeId().toString());
+    }
+    if (event.getContainerId() != null) {
+      otherInfo.put(ATSConstants.CONTAINER_ID, event.getContainerId().toString());
+    }
+    if (event.getInProgressLogsUrl() != null) {
+      otherInfo.put(ATSConstants.IN_PROGRESS_LOGS_URL, event.getInProgressLogsUrl());
+    }
+    if (event.getCompletedLogsUrl() != null) {
+      otherInfo.put(ATSConstants.COMPLETED_LOGS_URL, event.getCompletedLogsUrl());
+    }
+    if (event.getNodeHttpAddress() != null) {
+      otherInfo.put(ATSConstants.NODE_HTTP_ADDRESS, event.getNodeHttpAddress());
+    }
+
     jsonObject.put(ATSConstants.OTHER_INFO, otherInfo);
 
     return jsonObject;
@@ -697,6 +722,10 @@ public class HistoryEventJsonConversion {
         otherInfo.put(entry.getKey(), entry.getValue().intValue());
       }
     }
+    if (event.getServicePluginInfo() != null) {
+      otherInfo.put(ATSConstants.SERVICE_PLUGIN,
+          DAGUtils.convertServicePluginToJSON(event.getServicePluginInfo()));
+    }
 
     jsonObject.put(ATSConstants.OTHER_INFO, otherInfo);
 
@@ -768,6 +797,10 @@ public class HistoryEventJsonConversion {
     otherInfo.put(ATSConstants.INIT_TIME, event.getInitedTime());
     otherInfo.put(ATSConstants.NUM_TASKS, event.getNumTasks());
     otherInfo.put(ATSConstants.PROCESSOR_CLASS_NAME, event.getProcessorName());
+    if (event.getServicePluginInfo() != null) {
+      otherInfo.put(ATSConstants.SERVICE_PLUGIN,
+          DAGUtils.convertServicePluginToJSON(event.getServicePluginInfo()));
+    }
     jsonObject.put(ATSConstants.OTHER_INFO, otherInfo);
 
     return jsonObject;
