@@ -14,21 +14,30 @@
 
 package org.apache.tez.dag.app;
 
+import org.apache.tez.dag.app.dag.DAGState;
+import org.apache.tez.dag.app.dag.Vertex;
+import org.apache.tez.dag.records.TezVertexID;
 import static org.junit.Assert.assertEquals;
 import static org.junit.Assert.assertNotNull;
 import static org.junit.Assert.assertNull;
 import static org.junit.Assert.assertTrue;
+import static org.mockito.Mockito.mock;
+import static org.mockito.Mockito.spy;
+import static org.mockito.Mockito.verify;
+import static org.mockito.Mockito.when;
 
 import java.io.ByteArrayInputStream;
 import java.io.DataInput;
 import java.io.DataInputStream;
 import java.io.DataOutput;
 import java.io.File;
+import java.io.FileOutputStream;
 import java.io.IOException;
 import java.nio.ByteBuffer;
-import java.util.Collections;
+import java.util.HashMap;
 import java.util.LinkedList;
 import java.util.List;
+import java.util.Map;
 
 import com.google.common.base.Preconditions;
 import com.google.common.collect.BiMap;
@@ -50,7 +59,6 @@ import org.apache.hadoop.security.token.TokenIdentifier;
 import org.apache.hadoop.yarn.api.records.ApplicationAttemptId;
 import org.apache.hadoop.yarn.api.records.ApplicationId;
 import org.apache.hadoop.yarn.api.records.ContainerId;
-import org.apache.hadoop.yarn.api.records.LocalResource;
 import org.apache.hadoop.yarn.util.SystemClock;
 import org.apache.tez.client.TezApiVersionInfo;
 import org.apache.tez.common.TezCommonUtils;
@@ -66,9 +74,11 @@ import org.apache.tez.dag.api.UserPayload;
 import org.apache.tez.dag.api.records.DAGProtos;
 import org.apache.tez.dag.api.records.DAGProtos.AMPluginDescriptorProto;
 import org.apache.tez.dag.api.records.DAGProtos.DAGPlan;
+import org.apache.tez.dag.api.records.DAGProtos.PlanLocalResourcesProto;
 import org.apache.tez.dag.api.records.DAGProtos.TezNamedEntityDescriptorProto;
 import org.apache.tez.dag.api.records.DAGProtos.TezUserPayloadProto;
 import org.apache.tez.dag.app.dag.impl.DAGImpl;
+import org.apache.tez.dag.app.rm.TaskSchedulerManager;
 import org.apache.tez.dag.records.TezDAGID;
 import org.junit.After;
 import org.junit.Assert;
@@ -99,6 +109,32 @@ public class TestDAGAppMaster {
     FileUtil.fullyDelete(TEST_DIR);
   }
 
+  @Test(timeout = 20000)
+  public void testInvalidSession() throws Exception {
+    // AM should fail if not the first attempt and in session mode and
+    // DAG recovery is disabled, otherwise the app can succeed without
+    // finishing an in-progress DAG.
+    ApplicationId appId = ApplicationId.newInstance(1, 1);
+    ApplicationAttemptId attemptId = ApplicationAttemptId.newInstance(appId, 2);
+    DAGAppMasterForTest dam = new DAGAppMasterForTest(attemptId, true);
+    TezConfiguration conf = new TezConfiguration(false);
+    conf.setBoolean(TezConfiguration.DAG_RECOVERY_ENABLED, false);
+    dam.init(conf);
+    dam.start();
+    verify(dam.mockScheduler).setShouldUnregisterFlag();
+    verify(dam.mockShutdown).shutdown();
+    List<String> diags = dam.getDiagnostics();
+    boolean found = false;
+    for (String diag : diags) {
+      if (diag.contains(DAGAppMaster.INVALID_SESSION_ERR_MSG)) {
+        found = true;
+        break;
+      }
+    }
+    Assert.assertTrue("Missing invalid session diagnostics", found);
+    dam.stop();
+  }
+
   @Test(timeout = 5000)
   public void testPluginParsing() throws IOException {
     BiMap<String, Integer> pluginMap = HashBiMap.create();
@@ -358,6 +394,81 @@ public class TestDAGAppMaster {
     testDagCredentials(true);
   }
 
+  @Test
+  public void testBadProgress() throws Exception {
+    TezConfiguration conf = new TezConfiguration();
+    conf.setBoolean(TezConfiguration.TEZ_AM_CREDENTIALS_MERGE, true);
+    conf.setBoolean(TezConfiguration.TEZ_LOCAL_MODE, true);
+    conf.set(TezConfiguration.TEZ_AM_STAGING_DIR, TEST_DIR.toString());
+    ApplicationId appId = ApplicationId.newInstance(1, 1);
+    ApplicationAttemptId attemptId = ApplicationAttemptId.newInstance(appId, 1);
+
+    // create some sample AM credentials
+    Credentials amCreds = new Credentials();
+    JobTokenSecretManager jtsm = new JobTokenSecretManager();
+    JobTokenIdentifier identifier = new JobTokenIdentifier(
+        new Text(appId.toString()));
+    Token<JobTokenIdentifier> sessionToken =
+        new Token<JobTokenIdentifier>(identifier, jtsm);
+    sessionToken.setService(identifier.getJobId());
+    TokenCache.setSessionToken(sessionToken, amCreds);
+    TestTokenSecretManager ttsm = new TestTokenSecretManager();
+    Text tokenAlias1 = new Text("alias1");
+    Token<TestTokenIdentifier> amToken1 = new Token<TestTokenIdentifier>(
+        new TestTokenIdentifier(new Text("amtoken1")), ttsm);
+    amCreds.addToken(tokenAlias1, amToken1);
+
+    FileSystem fs = FileSystem.getLocal(conf);
+    FSDataOutputStream sessionJarsPBOutStream =
+        TezCommonUtils.createFileForAM(fs, new Path(TEST_DIR.toString(),
+            TezConstants.TEZ_AM_LOCAL_RESOURCES_PB_FILE_NAME));
+    DAGProtos.PlanLocalResourcesProto.getDefaultInstance()
+        .writeDelimitedTo(sessionJarsPBOutStream);
+    sessionJarsPBOutStream.close();
+    DAGAppMaster am = spy(new DAGAppMaster(attemptId,
+        ContainerId.newContainerId(attemptId, 1),
+        "127.0.0.1", 0, 0, new SystemClock(), 1, true,
+        TEST_DIR.toString(), new String[] {TEST_DIR.toString()},
+        new String[] {TEST_DIR.toString()},
+        new TezApiVersionInfo().getVersion(), amCreds,
+        "someuser", null));
+    when(am.getState()).thenReturn(DAGAppMasterState.RUNNING);
+    am.init(conf);
+    am.start();
+    Credentials dagCreds = new Credentials();
+    Token<TestTokenIdentifier> dagToken1 = new Token<TestTokenIdentifier>(
+        new TestTokenIdentifier(new Text("dagtoken1")), ttsm);
+    dagCreds.addToken(tokenAlias1, dagToken1);
+    Text tokenAlias3 = new Text("alias3");
+    Token<TestTokenIdentifier> dagToken2 = new Token<TestTokenIdentifier>(
+        new TestTokenIdentifier(new Text("dagtoken2")), ttsm);
+    dagCreds.addToken(tokenAlias3, dagToken2);
+    TezDAGID dagId = TezDAGID.getInstance(appId, 1);
+    DAGPlan dagPlan = DAGPlan.newBuilder()
+        .setName("somedag")
+        .setCredentialsBinary(
+            DagTypeConverters.convertCredentialsToProto(dagCreds))
+        .build();
+    DAGImpl dag = spy(am.createDAG(dagPlan, dagId));
+    am.setCurrentDAG(dag);
+    when(dag.getState()).thenReturn(DAGState.RUNNING);
+    Map<TezVertexID, Vertex> map = new HashMap<TezVertexID, Vertex>();
+    TezVertexID mockVertexID = mock(TezVertexID.class);
+    Vertex mockVertex = mock(Vertex.class);
+    when(mockVertex.getProgress()).thenReturn(Float.NaN);
+    map.put(mockVertexID, mockVertex);
+    when(dag.getVertices()).thenReturn(map);
+    when(dag.getTotalVertices()).thenReturn(1);
+    Assert.assertEquals("Progress was NaN and should be reported as 0",
+        0, am.getProgress(), 0);
+    when(mockVertex.getProgress()).thenReturn(-10f);
+    Assert.assertEquals("Progress was negative and should be reported as 0",
+        0, am.getProgress(), 0);
+    when(mockVertex.getProgress()).thenReturn(10f);
+    Assert.assertEquals("Progress was greater than 1 and should be reported as 0",
+        0, am.getProgress(), 0);
+  }
+
   @SuppressWarnings("deprecation")
   private void testDagCredentials(boolean doMerge) throws IOException {
     TezConfiguration conf = new TezConfiguration();
@@ -398,7 +509,7 @@ public class TestDAGAppMaster {
         "127.0.0.1", 0, 0, new SystemClock(), 1, true,
         TEST_DIR.toString(), new String[] {TEST_DIR.toString()},
         new String[] {TEST_DIR.toString()},
-        new TezApiVersionInfo().getVersion(), 1, amCreds,
+        new TezApiVersionInfo().getVersion(), amCreds,
         "someuser", null);
     am.init(conf);
     am.start();
@@ -510,4 +621,52 @@ public class TestDAGAppMaster {
       return new TestTokenIdentifier();
     }
   }
+
+  private static class DAGAppMasterForTest extends DAGAppMaster {
+    private DAGAppMasterShutdownHandler mockShutdown;
+    private TaskSchedulerManager mockScheduler = mock(TaskSchedulerManager.class);
+
+    public DAGAppMasterForTest(ApplicationAttemptId attemptId, boolean isSession) {
+      super(attemptId, ContainerId.newContainerId(attemptId, 1), "hostname", 12345, 12346,
+          new SystemClock(), 0, isSession, TEST_DIR.getAbsolutePath(),
+          new String[] { TEST_DIR.getAbsolutePath() }, new String[] { TEST_DIR.getAbsolutePath() },
+          new TezDagVersionInfo().getVersion(), createCredentials(), "jobname", null);
+    }
+
+    private static Credentials createCredentials() {
+      Credentials creds = new Credentials();
+      JobTokenSecretManager jtsm = new JobTokenSecretManager();
+      JobTokenIdentifier jtid = new JobTokenIdentifier(new Text());
+      Token<JobTokenIdentifier> token = new Token<JobTokenIdentifier>(jtid, jtsm);
+      TokenCache.setSessionToken(token, creds);
+      return creds;
+    }
+
+    private static void stubSessionResources() throws IOException {
+      FileOutputStream out = new FileOutputStream(
+          new File(TEST_DIR, TezConstants.TEZ_AM_LOCAL_RESOURCES_PB_FILE_NAME));
+      PlanLocalResourcesProto planProto = PlanLocalResourcesProto.getDefaultInstance();
+      planProto.writeDelimitedTo(out);
+      out.close();
+    }
+
+    @Override
+    public synchronized void serviceInit(Configuration conf) throws Exception {
+      stubSessionResources();
+      conf.setBoolean(TezConfiguration.TEZ_AM_WEBSERVICE_ENABLE, false);
+      super.serviceInit(conf);
+    }
+
+    @Override
+    protected DAGAppMasterShutdownHandler createShutdownHandler() {
+      mockShutdown = mock(DAGAppMasterShutdownHandler.class);
+      return mockShutdown;
+    }
+
+    @Override
+    protected TaskSchedulerManager createTaskSchedulerManager(
+        List<NamedEntityDescriptor> taskSchedulerDescriptors) {
+      return mockScheduler;
+    }
+  }
 }
