@@ -19,10 +19,10 @@
 package org.apache.tez.dag.app.dag.impl;
 
 import static org.mockito.Matchers.any;
+import static org.mockito.Matchers.anyInt;
 import static org.mockito.Mockito.doReturn;
+import static org.mockito.Mockito.doThrow;
 import static org.mockito.Mockito.mock;
-import static org.mockito.Mockito.times;
-import static org.mockito.Mockito.verify;
 import static org.mockito.Mockito.when;
 
 import java.io.IOException;
@@ -39,8 +39,11 @@ import java.util.concurrent.locks.Condition;
 import java.util.concurrent.locks.ReentrantLock;
 
 import org.apache.commons.lang.StringUtils;
+import org.apache.tez.common.DrainDispatcher;
 import org.apache.tez.common.counters.Limits;
 import org.apache.tez.common.counters.TezCounters;
+import org.apache.tez.dag.api.TezConstants;
+import org.apache.tez.dag.app.dag.event.DAGEventTerminateDag;
 import org.apache.tez.hadoop.shim.DefaultHadoopShim;
 import org.apache.tez.hadoop.shim.HadoopShim;
 import org.slf4j.Logger;
@@ -50,7 +53,6 @@ import org.apache.hadoop.security.Credentials;
 import org.apache.hadoop.yarn.api.records.ApplicationAttemptId;
 import org.apache.hadoop.yarn.api.records.ApplicationId;
 import org.apache.hadoop.yarn.api.records.Resource;
-import org.apache.hadoop.yarn.event.DrainDispatcher;
 import org.apache.hadoop.yarn.event.EventHandler;
 import org.apache.hadoop.yarn.util.Clock;
 import org.apache.hadoop.yarn.util.SystemClock;
@@ -110,6 +112,7 @@ import org.apache.tez.dag.app.dag.event.DAGAppMasterEventDAGFinished;
 import org.apache.tez.dag.app.dag.event.DAGAppMasterEventType;
 import org.apache.tez.dag.app.dag.event.DAGEvent;
 import org.apache.tez.dag.app.dag.event.DAGEventStartDag;
+import org.apache.tez.dag.app.dag.event.DAGEventCommitCompleted;
 import org.apache.tez.dag.app.dag.event.DAGEventType;
 import org.apache.tez.dag.app.dag.event.DAGEventVertexCompleted;
 import org.apache.tez.dag.app.dag.event.DAGEventVertexReRunning;
@@ -139,6 +142,7 @@ import org.apache.tez.runtime.api.events.InputReadErrorEvent;
 import org.apache.tez.runtime.api.impl.EventMetaData;
 import org.apache.tez.runtime.api.impl.EventMetaData.EventProducerConsumerType;
 import org.apache.tez.runtime.api.impl.TezEvent;
+import org.apache.tez.state.StateMachineTez;
 import org.junit.After;
 import org.junit.Assert;
 import org.junit.Before;
@@ -986,6 +990,8 @@ public class TestDAGImpl {
     doReturn(clusterInfo).when(dagWithCustomEdgeAppContext).getClusterInfo();
     dispatcher.register(TaskAttemptEventType.class, new TaskAttemptEventDisptacher2());
     dispatcher.register(AMSchedulerEventType.class, new AMSchedulerEventHandler());
+    when(dagWithCustomEdgeAppContext.getContainerLauncherName(anyInt())).thenReturn(
+        TezConstants.getTezYarnServicePluginName());
   }
 
   private void initDAG(DAGImpl impl) {
@@ -1641,8 +1647,7 @@ public class TestDAGImpl {
     startDAG(dag);
     dispatcher.await();
 
-    dispatcher.getEventHandler().handle(
-        new DAGEvent(dagId, DAGEventType.DAG_KILL));
+    dispatcher.getEventHandler().handle(new DAGEventTerminateDag(dagId, DAGTerminationCause.DAG_KILL, null));
     dispatcher.await();
 
     Assert.assertEquals(DAGState.KILLED, dag.getState());
@@ -1654,9 +1659,18 @@ public class TestDAGImpl {
 
   }
 
-  @SuppressWarnings("unchecked")
   @Test(timeout = 5000)
   public void testKillRunningDAG() {
+    _testTerminateRunningDAG(DAGTerminationCause.DAG_KILL);
+  }
+
+  @Test(timeout = 5000)
+  public void testServiceErrorRunningDAG() {
+    _testTerminateRunningDAG(DAGTerminationCause.SERVICE_PLUGIN_ERROR);
+  }
+
+  @SuppressWarnings("unchecked")
+  private void _testTerminateRunningDAG(DAGTerminationCause terminationCause) {
     initDAG(dag);
     startDAG(dag);
     dispatcher.await();
@@ -1674,7 +1688,7 @@ public class TestDAGImpl {
     Assert.assertEquals(VertexState.SUCCEEDED, v0.getState());
     Assert.assertEquals(VertexState.RUNNING, v1.getState());
 
-    dispatcher.getEventHandler().handle(new DAGEvent(dagId, DAGEventType.DAG_KILL));
+    dispatcher.getEventHandler().handle(new DAGEventTerminateDag(dagId, terminationCause, null));
     dispatcher.await();
 
     Assert.assertEquals(DAGState.TERMINATING, dag.getState());
@@ -1817,7 +1831,7 @@ public class TestDAGImpl {
       dispatcher.getEventHandler().handle(new DAGEventVertexCompleted(
           TezVertexID.getInstance(dagId, 5), VertexState.FAILED));
     } else if (testState == DAGStatus.State.KILLED) {
-      dispatcher.getEventHandler().handle(new DAGEvent(dagId, DAGEventType.DAG_KILL));
+      dispatcher.getEventHandler().handle(new DAGEventTerminateDag(dagId, DAGTerminationCause.DAG_KILL, null));
     } else if (testState == DAGStatus.State.ERROR) {
       dispatcher.getEventHandler().handle(new DAGEventStartDag(dagId, new LinkedList<URL>()));
     } else {
@@ -1871,11 +1885,21 @@ public class TestDAGImpl {
     }
   }
 
+
+  @Test(timeout = 5000)
+  public void testDAGKill() {
+    _testDAGTerminate(DAGTerminationCause.DAG_KILL);
+  }
+
+  @Test(timeout = 5000)
+  public void testDAGServiceError() {
+    _testDAGTerminate(DAGTerminationCause.SERVICE_PLUGIN_ERROR);
+  }
+
   // Couple of vertices succeed. DAG_KILLED processed, which causes the rest of the vertices to be
   // marked as KILLED.
   @SuppressWarnings("unchecked")
-  @Test(timeout = 5000)
-  public void testDAGKill() {
+  private void _testDAGTerminate(DAGTerminationCause terminationCause) {
     initDAG(dag);
     startDAG(dag);
     dispatcher.await();
@@ -1887,10 +1911,10 @@ public class TestDAGImpl {
 
     dispatcher.getEventHandler().handle(new DAGEventVertexCompleted(
         TezVertexID.getInstance(dagId, 1), VertexState.SUCCEEDED));
-    dispatcher.getEventHandler().handle(new DAGEvent(dagId, DAGEventType.DAG_KILL));
+    dispatcher.getEventHandler().handle(new DAGEventTerminateDag(dagId, terminationCause, null));
     dispatcher.await();
-    Assert.assertEquals(DAGState.KILLED, dag.getState());
-    Assert.assertEquals(DAGTerminationCause.DAG_KILL, dag.getTerminationCause());
+    Assert.assertEquals(terminationCause.getFinishedState(), dag.getState());
+    Assert.assertEquals(terminationCause, dag.getTerminationCause());
     Assert.assertEquals(2, dag.getSuccessfulVertices());
 
     int killedCount = 0;
@@ -1902,16 +1926,70 @@ public class TestDAGImpl {
     Assert.assertEquals(4, killedCount);
 
     for (Vertex v : dag.getVertices().values()) {
-      Assert.assertEquals(VertexTerminationCause.DAG_KILL, v.getTerminationCause());
+      Assert.assertEquals(VertexTerminationCause.DAG_TERMINATED, v.getTerminationCause());
     }
 
     Assert.assertEquals(1, dagFinishEventHandler.dagFinishEvents);
   }
 
+  @Test (timeout = 5000L)
+  @SuppressWarnings("unchecked")
+  public void testDAGHang() throws Exception {
+    conf.setBoolean(
+        TezConfiguration.TEZ_AM_COMMIT_ALL_OUTPUTS_ON_DAG_SUCCESS,
+        false);
+    dag = Mockito.spy(new DAGImpl(dagId, conf, dagPlan,
+        dispatcher.getEventHandler(), taskCommunicatorManagerInterface,
+        fsTokens, clock, "user", thh, appContext));
+    StateMachineTez<DAGState, DAGEventType, DAGEvent, DAGImpl> spyStateMachine =
+        Mockito.spy(new StateMachineTez<DAGState, DAGEventType, DAGEvent, DAGImpl>(
+            dag.stateMachineFactory.make(dag), dag));
+    when(dag.getStateMachine()).thenReturn(spyStateMachine);
+    dag.entityUpdateTracker = new StateChangeNotifierForTest(dag);
+    doReturn(dag).when(appContext).getCurrentDAG();
+    DAGImpl.OutputKey outputKey = Mockito.mock(DAGImpl.OutputKey.class);
+    ListenableFuture future = Mockito.mock(ListenableFuture.class);
+    dag.commitFutures.put(outputKey, future);
+    initDAG(dag);
+    startDAG(dag);
+    dispatcher.await();
+
+    dispatcher.getEventHandler().handle(new DAGEventVertexCompleted(
+        TezVertexID.getInstance(dagId, 0), VertexState.SUCCEEDED));
+    dispatcher.getEventHandler().handle(new DAGEventVertexCompleted(
+        TezVertexID.getInstance(dagId, 1), VertexState.SUCCEEDED));
+    dispatcher.getEventHandler().handle(new DAGEventVertexCompleted(
+        TezVertexID.getInstance(dagId, 2), VertexState.SUCCEEDED));
+    dispatcher.getEventHandler().handle(new DAGEventVertexCompleted(
+        TezVertexID.getInstance(dagId, 3), VertexState.SUCCEEDED));
+    dispatcher.getEventHandler().handle(new DAGEventVertexCompleted(
+        TezVertexID.getInstance(dagId, 4), VertexState.SUCCEEDED));
+    dispatcher.getEventHandler().handle(new DAGEventVertexCompleted(
+        TezVertexID.getInstance(dagId, 5), VertexState.SUCCEEDED));
+    dispatcher.await();
+    Assert.assertEquals(DAGState.COMMITTING, dag.getState());
+    DAGEventCommitCompleted dagEvent = new DAGEventCommitCompleted(
+        dagId, outputKey, false , new RuntimeException("test"));
+    doThrow(new RuntimeException("test")).when(
+        dag).logJobHistoryUnsuccesfulEvent(any(DAGState.class), any(TezCounters.class));
+    dag.handle(dagEvent);
+    dispatcher.await();
+    Assert.assertTrue("DAG did not terminate!", dag.getInternalState() == DAGState.FAILED);
+  }
+
+  @Test(timeout = 5000)
+  public void testDAGKillVertexSuccessAfterTerminated() {
+    _testDAGKillVertexSuccessAfterTerminated(DAGTerminationCause.DAG_KILL);
+  }
+
+  @Test(timeout = 5000)
+  public void testDAGServiceErrorVertexSuccessAfterTerminated() {
+    _testDAGKillVertexSuccessAfterTerminated(DAGTerminationCause.SERVICE_PLUGIN_ERROR);
+  }
+
   // Vertices succeed after a DAG kill has been processed. Should be ignored.
   @SuppressWarnings("unchecked")
-  @Test(timeout = 5000)
-  public void testDAGKillVertexSuccessAfterKill() {
+  private void _testDAGKillVertexSuccessAfterTerminated(DAGTerminationCause terminationCause) {
     initDAG(dag);
     startDAG(dag);
     dispatcher.await();
@@ -1923,10 +2001,10 @@ public class TestDAGImpl {
 
     dispatcher.getEventHandler().handle(new DAGEventVertexCompleted(
         TezVertexID.getInstance(dagId, 1), VertexState.SUCCEEDED));
-    dispatcher.getEventHandler().handle(new DAGEvent(dagId, DAGEventType.DAG_KILL));
+    dispatcher.getEventHandler().handle(new DAGEventTerminateDag(dagId, terminationCause, null));
     dispatcher.await();
 
-    Assert.assertEquals(DAGState.KILLED, dag.getState());
+    Assert.assertEquals(terminationCause.getFinishedState(), dag.getState());
 
     // Vertex SUCCESS gets processed after the DAG has reached the KILLED state. Should be ignored.
     for (int i = 2; i < 6; ++i) {
@@ -1943,18 +2021,27 @@ public class TestDAGImpl {
     }
     Assert.assertEquals(4, killedCount);
 
-    Assert.assertEquals(DAGTerminationCause.DAG_KILL, dag.getTerminationCause());
+    Assert.assertEquals(terminationCause, dag.getTerminationCause());
     Assert.assertEquals(2, dag.getSuccessfulVertices());
     for (Vertex v : dag.getVertices().values()) {
-      Assert.assertEquals(VertexTerminationCause.DAG_KILL, v.getTerminationCause());
+      Assert.assertEquals(VertexTerminationCause.DAG_TERMINATED, v.getTerminationCause());
     }
     Assert.assertEquals(1, dagFinishEventHandler.dagFinishEvents);
   }
 
-  // Vertex KILLED after a DAG_KILLED is issued. Termination reason should be DAG_KILLED
-  @SuppressWarnings("unchecked")
   @Test(timeout = 5000)
   public void testDAGKillPending() {
+    _testDAGKillPending(DAGTerminationCause.DAG_KILL);
+  }
+
+  @Test(timeout = 5000)
+  public void testDAGServiceErrorPending() {
+    _testDAGKillPending(DAGTerminationCause.SERVICE_PLUGIN_ERROR);
+  }
+
+  // Vertex KILLED after a DAG_KILLED is issued. Termination reason should be DAG_KILLED
+  @SuppressWarnings("unchecked")
+  private void _testDAGKillPending(DAGTerminationCause terminationCause) {
     initDAG(dag);
     startDAG(dag);
     dispatcher.await();
@@ -1972,17 +2059,17 @@ public class TestDAGImpl {
           TezVertexID.getInstance(dagId, i), VertexState.SUCCEEDED));
     }
     dispatcher.await();
-    dispatcher.getEventHandler().handle(new DAGEvent(dagId, DAGEventType.DAG_KILL));
+    dispatcher.getEventHandler().handle(new DAGEventTerminateDag(dagId, terminationCause, null));
     dispatcher.await();
-    Assert.assertEquals(DAGState.KILLED, dag.getState());
+    Assert.assertEquals(terminationCause.getFinishedState(), dag.getState());
 
     dispatcher.getEventHandler().handle(new DAGEventVertexCompleted(
         TezVertexID.getInstance(dagId, 5), VertexState.KILLED));
     dispatcher.await();
-    Assert.assertEquals(DAGState.KILLED, dag.getState());
+    Assert.assertEquals(terminationCause.getFinishedState(), dag.getState());
     Assert.assertEquals(5, dag.getSuccessfulVertices());
     Assert.assertEquals(dag.getVertex(TezVertexID.getInstance(dagId, 5)).getTerminationCause(),
-        VertexTerminationCause.DAG_KILL);
+        VertexTerminationCause.DAG_TERMINATED);
     Assert.assertEquals(1, dagFinishEventHandler.dagFinishEvents);
   }
 
@@ -2158,7 +2245,7 @@ public class TestDAGImpl {
     }
 
     @Override
-    public EventRouteMetadata routeCompositeDataMovementEventToDestination(
+    public CompositeEventRouteMetadata routeCompositeDataMovementEventToDestination(
         int sourceTaskIndex, int destinationTaskIndex)
         throws Exception {
       if (exLocation == ExceptionLocation.RouteDataMovementEventToDestination) {
@@ -2238,13 +2325,13 @@ public class TestDAGImpl {
     dispatcher.await();
     for (int i=0; i<3; ++i) {
       Vertex v = mrrDag.getVertex("vertex"+(i+1));
-      dispatcher.getEventHandler().handle(new VertexEventTaskCompleted(
-          TezTaskID.getInstance(v.getVertexId(), 0), TaskState.SUCCEEDED));
       TezCounters ctrs = new TezCounters();
       for (int j = 0; j < 50; ++j) {
         ctrs.findCounter("g", "c" + i + "_" + j).increment(1);
       }
       ((VertexImpl) v).setCounters(ctrs);
+      dispatcher.getEventHandler().handle(new VertexEventTaskCompleted(
+        TezTaskID.getInstance(v.getVertexId(), 0), TaskState.SUCCEEDED));
       dispatcher.await();
       Assert.assertEquals(VertexState.SUCCEEDED, v.getState());
       Assert.assertEquals(i+1, mrrDag.getSuccessfulVertices());
