@@ -39,12 +39,12 @@ import org.apache.hadoop.yarn.api.records.ApplicationId;
 import org.apache.hadoop.yarn.api.records.ContainerId;
 import org.apache.hadoop.yarn.api.records.NodeId;
 import org.apache.hadoop.yarn.api.records.Resource;
-import org.apache.hadoop.yarn.event.DrainDispatcher;
 import org.apache.hadoop.yarn.event.EventHandler;
 import org.apache.hadoop.yarn.util.Clock;
 import org.apache.hadoop.yarn.util.SystemClock;
 import org.apache.log4j.Level;
 import org.apache.log4j.LogManager;
+import org.apache.tez.common.DrainDispatcher;
 import org.apache.tez.common.MockDNSToSwitchMapping;
 import org.apache.tez.common.security.ACLManager;
 import org.apache.tez.dag.api.TezConfiguration;
@@ -87,16 +87,12 @@ import org.apache.tez.dag.app.dag.event.DAGEvent;
 import org.apache.tez.dag.app.dag.event.DAGEventRecoverEvent;
 import org.apache.tez.dag.app.dag.event.DAGEventType;
 import org.apache.tez.dag.app.dag.event.TaskAttemptEvent;
-import org.apache.tez.dag.app.dag.event.TaskAttemptEventSchedule;
 import org.apache.tez.dag.app.dag.event.TaskAttemptEventType;
 import org.apache.tez.dag.app.dag.event.TaskEvent;
-import org.apache.tez.dag.app.dag.event.TaskEventScheduleTask;
 import org.apache.tez.dag.app.dag.event.TaskEventType;
 import org.apache.tez.dag.app.dag.event.VertexEvent;
-import org.apache.tez.dag.app.dag.event.VertexEventRouteEvent;
 import org.apache.tez.dag.app.dag.event.VertexEventType;
 import org.apache.tez.dag.app.dag.impl.TestVertexImpl.CountingOutputCommitter;
-import org.apache.tez.dag.app.dag.impl.TestVertexImpl.EventHandlingRootInputInitializer;
 import org.apache.tez.dag.app.rm.AMSchedulerEvent;
 import org.apache.tez.dag.app.rm.AMSchedulerEventType;
 import org.apache.tez.dag.app.rm.TaskSchedulerManager;
@@ -105,7 +101,6 @@ import org.apache.tez.dag.history.HistoryEvent;
 import org.apache.tez.dag.history.HistoryEventHandler;
 import org.apache.tez.dag.history.HistoryEventType;
 import org.apache.tez.dag.history.events.DAGInitializedEvent;
-import org.apache.tez.dag.history.events.DAGRecoveredEvent;
 import org.apache.tez.dag.history.events.DAGStartedEvent;
 import org.apache.tez.dag.history.events.TaskAttemptFinishedEvent;
 import org.apache.tez.dag.history.events.TaskAttemptStartedEvent;
@@ -121,6 +116,7 @@ import org.apache.tez.dag.records.TezTaskID;
 import org.apache.tez.dag.records.TezVertexID;
 import org.apache.tez.hadoop.shim.DefaultHadoopShim;
 import org.apache.tez.runtime.api.Event;
+import org.apache.tez.runtime.api.TaskFailureType;
 import org.apache.tez.runtime.api.InputInitializer;
 import org.apache.tez.runtime.api.InputInitializerContext;
 import org.apache.tez.runtime.api.InputSpecUpdate;
@@ -130,9 +126,7 @@ import org.apache.tez.runtime.api.events.DataMovementEvent;
 import org.apache.tez.runtime.api.events.InputInitializerEvent;
 import org.apache.tez.runtime.api.impl.EventMetaData;
 import org.apache.tez.runtime.api.impl.EventMetaData.EventProducerConsumerType;
-import org.apache.tez.runtime.api.impl.TaskSpec;
 import org.apache.tez.runtime.api.impl.TezEvent;
-import org.apache.tez.runtime.library.common.InputIdentifier;
 import org.junit.After;
 import org.junit.Assert;
 import org.junit.Before;
@@ -322,6 +316,7 @@ public class TestDAGRecovery {
     final ListenableFuture<Void> mockFuture = mock(ListenableFuture.class);
     when(appContext.getHadoopShim()).thenReturn(new DefaultHadoopShim());
     when(appContext.getApplicationID()).thenReturn(appAttemptId.getApplicationId());
+    when(appContext.getClock()).thenReturn(new SystemClock());
 
     Mockito.doAnswer(new Answer() {
       public ListenableFuture<Void> answer(InvocationOnMock invocation) {
@@ -705,7 +700,7 @@ public class TestDAGRecovery {
     List<TezEvent> inputGeneratedTezEvents = new ArrayList<TezEvent>();
     VertexInitializedEvent v1InitedEvent = new VertexInitializedEvent(v1Id, 
         "vertex1", 0L, v1InitedTime, 
-        v1NumTask, "", null, inputGeneratedTezEvents);
+        v1NumTask, "", null, inputGeneratedTezEvents, null);
     VertexRecoveryData vertexRecoveryData = new VertexRecoveryData(v1InitedEvent,
         null, null, null, null, false);
     doReturn(vertexRecoveryData).when(dagRecoveryData).getVertexRecoveryData(v1Id);
@@ -737,7 +732,7 @@ public class TestDAGRecovery {
     List<TezEvent> inputGeneratedTezEvents = new ArrayList<TezEvent>();
     VertexInitializedEvent v1InitedEvent = new VertexInitializedEvent(v1Id, 
         "vertex1", 0L, v1InitedTime, 
-        v1NumTask, "", null, inputGeneratedTezEvents);
+        v1NumTask, "", null, inputGeneratedTezEvents, null);
     VertexConfigurationDoneEvent v1ReconfigureDoneEvent = new VertexConfigurationDoneEvent(v1Id, 
         0L, v1NumTask, null, null, null, true);
     VertexRecoveryData vertexRecoveryData = new VertexRecoveryData(v1InitedEvent,
@@ -777,7 +772,7 @@ public class TestDAGRecovery {
     List<TezEvent> inputGeneratedTezEvents = new ArrayList<TezEvent>();
     VertexInitializedEvent v1InitedEvent = new VertexInitializedEvent(v1Id, 
         "vertex1", 0L, v1InitedTime, 
-        v1NumTask, "", null, inputGeneratedTezEvents);
+        v1NumTask, "", null, inputGeneratedTezEvents, null);
     VertexConfigurationDoneEvent v1ReconfigureDoneEvent = new VertexConfigurationDoneEvent(v1Id, 
         0L, v1NumTask, null, null, null, true);
     VertexStartedEvent v1StartedEvent = new VertexStartedEvent(v1Id, 0L, v1StartedTime);
@@ -808,7 +803,7 @@ public class TestDAGRecovery {
     List<TezEvent> inputGeneratedTezEvents = new ArrayList<TezEvent>();
     VertexInitializedEvent v1InitedEvent = new VertexInitializedEvent(v1Id, 
         "vertex1", 0L, v1InitedTime, 
-        v1NumTask, "", null, inputGeneratedTezEvents);
+        v1NumTask, "", null, inputGeneratedTezEvents, null);
     Map<String, InputSpecUpdate> rootInputSpecs = new HashMap<String, InputSpecUpdate>();
     VertexConfigurationDoneEvent v1ReconfigureDoneEvent = new VertexConfigurationDoneEvent(v1Id, 
         0L, v1NumTask, null, null, rootInputSpecs, true);
@@ -883,8 +878,8 @@ public class TestDAGRecovery {
     taGeneratedEvents.add(new TezEvent(DataMovementEvent.create(ByteBuffer.wrap(new byte[0])), metadata));
     TaskAttemptFinishedEvent taFinishedEvent = new TaskAttemptFinishedEvent(
         ta1t1v1Id, "v1", 0L, 0L, 
-        TaskAttemptState.SUCCEEDED, null, "", null, 
-        null, taGeneratedEvents, 0L, null, 0L);
+        TaskAttemptState.SUCCEEDED, null, null, "", null,
+        null, taGeneratedEvents, 0L, null, 0L, null, null, null, null, null);
     TaskAttemptRecoveryData taRecoveryData = new TaskAttemptRecoveryData(taStartedEvent, taFinishedEvent);
     Map<TezTaskAttemptID, TaskAttemptRecoveryData> taRecoveryDataMap =
         new HashMap<TezTaskAttemptID, TaskAttemptRecoveryData>();
@@ -916,7 +911,7 @@ public class TestDAGRecovery {
     List<TezEvent> inputGeneratedTezEvents = new ArrayList<TezEvent>();
     VertexInitializedEvent v1InitedEvent = new VertexInitializedEvent(v1Id, 
         "vertex1", 0L, v1InitedTime, 
-        v1NumTask, "", null, inputGeneratedTezEvents);
+        v1NumTask, "", null, inputGeneratedTezEvents, null);
     Map<String, InputSpecUpdate> rootInputSpecs = new HashMap<String, InputSpecUpdate>();
     VertexConfigurationDoneEvent v1ReconfigureDoneEvent = new VertexConfigurationDoneEvent(v1Id, 
         0L, v1NumTask, null, null, rootInputSpecs, true);
@@ -941,8 +936,8 @@ public class TestDAGRecovery {
     initMockDAGRecoveryDataForTaskAttempt();
     TaskAttemptFinishedEvent taFinishedEvent = new TaskAttemptFinishedEvent(
         ta1t1v1Id, "v1", ta1LaunchTime, ta1FinishedTime, 
-        TaskAttemptState.FAILED, TaskAttemptTerminationCause.CONTAINER_LAUNCH_FAILED, "", null, 
-        null, null, 0L, null, 0L);
+        TaskAttemptState.FAILED, TaskFailureType.NON_FATAL, TaskAttemptTerminationCause.CONTAINER_LAUNCH_FAILED, "", null,
+        null, null, 0L, null, 0L, null, null, null, null, null);
     TaskAttemptRecoveryData taRecoveryData = new TaskAttemptRecoveryData(null, taFinishedEvent);
     doReturn(taRecoveryData).when(dagRecoveryData).getTaskAttemptRecoveryData(ta1t1v1Id);
     
@@ -970,8 +965,8 @@ public class TestDAGRecovery {
     initMockDAGRecoveryDataForTaskAttempt();
     TaskAttemptFinishedEvent taFinishedEvent = new TaskAttemptFinishedEvent(
         ta1t1v1Id, "v1", ta1LaunchTime, ta1FinishedTime, 
-        TaskAttemptState.KILLED, TaskAttemptTerminationCause.TERMINATED_BY_CLIENT, "", null, 
-        null, null, 0L, null, 0L);
+        TaskAttemptState.KILLED, null, TaskAttemptTerminationCause.TERMINATED_BY_CLIENT, "", null,
+        null, null, 0L, null, 0L, null, null, null, null, null);
     TaskAttemptRecoveryData taRecoveryData = new TaskAttemptRecoveryData(null, taFinishedEvent);
     doReturn(taRecoveryData).when(dagRecoveryData).getTaskAttemptRecoveryData(ta1t1v1Id);
     
@@ -1030,8 +1025,8 @@ public class TestDAGRecovery {
         sourceInfo));
     TaskAttemptFinishedEvent taFinishedEvent = new TaskAttemptFinishedEvent(
         ta1t1v1Id, "v1", ta1LaunchTime, ta1FinishedTime, 
-        TaskAttemptState.SUCCEEDED, null, "", null, 
-        null, taGeneratedEvents, 0L, null, 0L);
+        TaskAttemptState.SUCCEEDED, null, null, "", null,
+        null, taGeneratedEvents, 0L, null, 0L, null, null, null, null, null);
     TaskAttemptRecoveryData taRecoveryData = new TaskAttemptRecoveryData(taStartedEvent, taFinishedEvent);
     doReturn(taRecoveryData).when(dagRecoveryData).getTaskAttemptRecoveryData(ta1t1v1Id);
 
@@ -1053,7 +1048,7 @@ public class TestDAGRecovery {
    * V2's committer is not recovery supported
    */
   @Test//(timeout=5000)
-  public void testTARecoverFromSucceeded_OutputCommitterRecoveryNotSupported() {
+  public void testTARecoverFromSucceeded_OutputCommitterRecoveryNotSupported() throws Exception{
     initMockDAGRecoveryDataForTaskAttempt();
     // set up v2 recovery data
     // ta1t1v2: TaskAttemptStartedEvent -> TaskAttemptFinishedEvent(SUCCEEDED)
@@ -1068,8 +1063,8 @@ public class TestDAGRecovery {
     taGeneratedEvents.add(new TezEvent(DataMovementEvent.create(ByteBuffer.wrap(new byte[0])), metadata));
     TaskAttemptFinishedEvent taFinishedEvent = new TaskAttemptFinishedEvent(
         ta1t1v2Id, "vertex2", ta1LaunchTime, ta1FinishedTime, 
-        TaskAttemptState.SUCCEEDED, null, "", null, 
-        null, taGeneratedEvents, 0L, null, 0L);
+        TaskAttemptState.SUCCEEDED, null, null, "", null,
+        null, taGeneratedEvents, 0L, null, 0L, null, null, null, null, null);
     TaskAttemptRecoveryData taRecoveryData = new TaskAttemptRecoveryData(taStartedEvent, taFinishedEvent);
     doReturn(taRecoveryData).when(dagRecoveryData).getTaskAttemptRecoveryData(ta1t1v2Id);   
     Map<TezTaskAttemptID, TaskAttemptRecoveryData> taRecoveryDataMap =
@@ -1084,7 +1079,7 @@ public class TestDAGRecovery {
 
     VertexInitializedEvent v2InitedEvent = new VertexInitializedEvent(v2Id, 
         "vertex2", 0L, v1InitedTime, 
-        v1NumTask, "", null, null);
+        v1NumTask, "", null, null, null);
     VertexConfigurationDoneEvent v2ReconfigureDoneEvent = new VertexConfigurationDoneEvent(v2Id, 
         0L, v1NumTask, null, null, null, false);
     VertexStartedEvent v2StartedEvent = new VertexStartedEvent(v2Id, 0L, v1StartedTime);
@@ -1098,8 +1093,8 @@ public class TestDAGRecovery {
     
     TaskImpl task = (TaskImpl)dag.getVertex(v2Id).getTask(t1v2Id);
     TaskAttemptImpl taskAttempt = (TaskAttemptImpl)task.getAttempt(ta1t1v2Id);
-    assertEquals(TaskAttemptStateInternal.SUCCEEDED, taskAttempt.getInternalState());
-    historyEventHandler.verifyHistoryEvent(0, HistoryEventType.TASK_ATTEMPT_FINISHED);
+    assertEquals(TaskAttemptStateInternal.KILLED, taskAttempt.getInternalState());
+    historyEventHandler.verifyHistoryEvent(1, HistoryEventType.TASK_ATTEMPT_FINISHED);
     assertEquals(TaskStateInternal.RUNNING, task.getInternalState());
     // new task attempt is scheduled
     assertEquals(2, task.getAttempts().size());
@@ -1119,8 +1114,8 @@ public class TestDAGRecovery {
         mock(NodeId.class), "", "", "");
     TaskAttemptFinishedEvent taFinishedEvent = new TaskAttemptFinishedEvent(
         ta1t1v1Id, "v1", ta1LaunchTime, ta1FinishedTime, 
-        TaskAttemptState.FAILED, TaskAttemptTerminationCause.INPUT_READ_ERROR, "", null, 
-        null, null, 0L, null, 0L);
+        TaskAttemptState.FAILED, TaskFailureType.NON_FATAL, TaskAttemptTerminationCause.INPUT_READ_ERROR, "", null,
+        null, null, 0L, null, 0L, null, null, null, null, null);
     TaskAttemptRecoveryData taRecoveryData = new TaskAttemptRecoveryData(taStartedEvent, taFinishedEvent);
     doReturn(taRecoveryData).when(dagRecoveryData).getTaskAttemptRecoveryData(ta1t1v1Id);
     
@@ -1150,8 +1145,8 @@ public class TestDAGRecovery {
         mock(NodeId.class), "", "", "");
     TaskAttemptFinishedEvent taFinishedEvent = new TaskAttemptFinishedEvent(
         ta1t1v1Id, "v1", ta1FinishedTime, ta1FinishedTime, 
-        TaskAttemptState.KILLED, TaskAttemptTerminationCause.TERMINATED_BY_CLIENT, "", null, 
-        null, null, 0L, null, 0L);
+        TaskAttemptState.KILLED, null, TaskAttemptTerminationCause.TERMINATED_BY_CLIENT, "", null,
+        null, null, 0L, null, 0L, null, null, null, null, null);
     TaskAttemptRecoveryData taRecoveryData = new TaskAttemptRecoveryData(taStartedEvent, taFinishedEvent);
     doReturn(taRecoveryData).when(dagRecoveryData).getTaskAttemptRecoveryData(ta1t1v1Id);
     
