@@ -38,6 +38,7 @@ import java.util.Set;
 import java.util.concurrent.ExecutionException;
 import java.util.concurrent.atomic.AtomicBoolean;
 
+import org.apache.hadoop.yarn.api.records.LocalResourceType;
 import org.apache.tez.common.TezUtils;
 import org.apache.tez.serviceplugins.api.TaskScheduler;
 import org.mockito.ArgumentCaptor;
@@ -96,6 +97,7 @@ import org.junit.Test;
 import org.mockito.invocation.InvocationOnMock;
 import org.mockito.stubbing.Answer;
 
+import com.google.common.collect.ImmutableMap;
 import com.google.common.collect.Lists;
 import com.google.common.collect.Maps;
 import com.google.common.collect.Sets;
@@ -191,17 +193,12 @@ public class TestContainerReuse {
       createLaunchRequestEvent(taID31, ta31, resource, host1,
         defaultRack, priority);
 
-    drainNotifier.set(false);
     taskSchedulerManager.handleEvent(lrTa11);
-    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
-    drainNotifier.set(false);
     taskSchedulerManager.handleEvent(lrTa21);
-    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
 
     Container containerHost1 = createContainer(1, host1[0], resource, priority);
     Container containerHost2 = createContainer(2, host2[0], resource, priority);
 
-    drainNotifier.set(false);
     taskScheduler.onContainersAllocated(
       Lists.newArrayList(containerHost1, containerHost2));
     TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
@@ -320,17 +317,12 @@ public class TestContainerReuse {
     AMSchedulerEventTALaunchRequest lrTa31 = createLaunchRequestEvent(
       taID31, ta31, resource, host1, defaultRack, priority);
 
-    drainNotifier.set(false);
     taskSchedulerManager.handleEvent(lrTa11);
-    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
-    drainNotifier.set(false);
     taskSchedulerManager.handleEvent(lrTa21);
-    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
 
     Container containerHost1 = createContainer(1, host1[0], resource, priority);
     Container containerHost2 = createContainer(2, host2[0], resource, priority);
 
-    drainNotifier.set(false);
     taskScheduler.onContainersAllocated(Lists.newArrayList(containerHost1, containerHost2));
     TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
     drainableAppCallback.drain();
@@ -432,7 +424,6 @@ public class TestContainerReuse {
     Container container1 = createContainer(1, "host1", resource1, priority1);
 
     // One container allocated.
-    drainNotifier.set(false);
     taskScheduler.onContainersAllocated(Collections.singletonList(container1));
     TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
     drainableAppCallback.drain();
@@ -478,7 +469,6 @@ public class TestContainerReuse {
     Container container2 = createContainer(2, "host2", resource1, priority1);
 
     // Second container allocated. Should be allocated to the last task.
-    drainNotifier.set(false);
     taskScheduler.onContainersAllocated(Collections.singletonList(container2));
     TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
     drainableAppCallback.drain();
@@ -573,17 +563,12 @@ public class TestContainerReuse {
     AMSchedulerEventTALaunchRequest lrEvent2 =
         createLaunchRequestEvent(taID12, ta12, resource1, host1, racks, priority1);
 
-    drainNotifier.set(false);
     taskSchedulerManager.handleEvent(lrEvent1);
-    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
-    drainNotifier.set(false);
     taskSchedulerManager.handleEvent(lrEvent2);
-    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
 
     Container container1 = createContainer(1, "host1", resource1, priority1);
 
     // One container allocated.
-    drainNotifier.set(false);
     taskScheduler.onContainersAllocated(Collections.singletonList(container1));
     TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
     drainableAppCallback.drain();
@@ -622,16 +607,11 @@ public class TestContainerReuse {
         createLaunchRequestEvent(taID14, ta14, resource1, host2, racks,
             priority1, localResources, tsLaunchCmdOpts);
 
-    Container container2 = createContainer(2, "host2", resource1, priority1);
-    drainNotifier.set(false);
     taskSchedulerManager.handleEvent(lrEvent3);
-    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
-    drainNotifier.set(false);
     taskSchedulerManager.handleEvent(lrEvent4);
-    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
 
     // Container started
-    drainNotifier.set(false);
+    Container container2 = createContainer(2, "host2", resource1, priority1);
     taskScheduler.onContainersAllocated(Collections.singletonList(container2));
     TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
     drainableAppCallback.drain();
@@ -675,11 +655,10 @@ public class TestContainerReuse {
           priority1, localResources, taskSpecificLaunchCmdOption.getTaskSpecificOption("", "v1", 6));
 
     // Container started
-    Container container3 = createContainer(2, "host3", resource1, priority1);
+    Container container3 = createContainer(3, "host3", resource1, priority1);
     taskSchedulerManager.handleEvent(lrEvent5);
     taskSchedulerManager.handleEvent(lrEvent6);
 
-    drainNotifier.set(false);
     taskScheduler.onContainersAllocated(Collections.singletonList(container3));
     TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
     drainableAppCallback.drain();
@@ -772,14 +751,11 @@ public class TestContainerReuse {
       taID12, ta12, resource1, emptyHosts, racks, priority);
 
     // Send launch request for task 1 only, deterministic assignment to this task.
-    drainNotifier.set(false);
     taskSchedulerManager.handleEvent(lrEvent11);
-    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
 
     Container container1 = createContainer(1, "randomHost", resource1, priority);
 
     // One container allocated.
-    drainNotifier.set(false);
     taskScheduler.onContainersAllocated(Collections.singletonList(container1));
     TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
     drainableAppCallback.drain();
@@ -803,8 +779,8 @@ public class TestContainerReuse {
     eventHandler.verifyNoInvocations(AMContainerEventStopRequest.class);
     eventHandler.reset();
 
-    LOG.info("Sleeping to ensure that the scheduling loop runs");
-    Thread.sleep(3000l);
+    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
+    drainableAppCallback.drain();
     verify(taskSchedulerManager).taskAllocated(
         eq(0), eq(ta12), any(Object.class), eq(container1));
 
@@ -812,9 +788,10 @@ public class TestContainerReuse {
     taskSchedulerManager.handleEvent(
       new AMSchedulerEventTAEnded(ta12, container1.getId(),
         TaskAttemptState.SUCCEEDED, null, null, 0));
+
+    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
     drainableAppCallback.drain();
-    LOG.info("Sleeping to ensure that the scheduling loop runs");
-    Thread.sleep(3000l);
+
     verify(rmClient).releaseAssignedContainer(eq(container1.getId()));
     eventHandler.verifyInvocation(AMContainerEventStopRequest.class);
 
@@ -831,7 +808,9 @@ public class TestContainerReuse {
     tezConf.setLong(
         TezConfiguration.TEZ_AM_CONTAINER_REUSE_LOCALITY_DELAY_ALLOCATION_MILLIS, 1l);
     tezConf.setLong(
-        TezConfiguration.TEZ_AM_CONTAINER_IDLE_RELEASE_TIMEOUT_MIN_MILLIS, 2000l);
+        TezConfiguration.TEZ_AM_CONTAINER_IDLE_RELEASE_TIMEOUT_MIN_MILLIS, 20l);
+    tezConf.setLong(
+            TezConfiguration.TEZ_AM_CONTAINER_IDLE_RELEASE_TIMEOUT_MAX_MILLIS, 30l);
     tezConf.setInt(
         TezConfiguration.TEZ_AM_SESSION_MIN_HELD_CONTAINERS, 1);
 
@@ -900,14 +879,11 @@ public class TestContainerReuse {
       taID21, ta21, resource1, host1, racks, priority2);
 
     // Send launch request for task 1 onle, deterministic assignment to this task.
-    drainNotifier.set(false);
     taskSchedulerManager.handleEvent(lrEvent11);
-    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
 
     Container container1 = createContainer(1, host1[0], resource1, priority1);
 
     // One container allocated.
-    drainNotifier.set(false);
     taskScheduler.onContainersAllocated(Collections.singletonList(container1));
     TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
     drainableAppCallback.drain();
@@ -934,8 +910,9 @@ public class TestContainerReuse {
             TaskAttemptState.SUCCEEDED, null, null, 0));
     verify(rmClient, times(0)).releaseAssignedContainer(eq(container1.getId()));
 
-    LOG.info("Sleeping to ensure that the scheduling loop runs");
-    Thread.sleep(3000l);
+    LOG.info("Sleeping to ensure that the container has been idled longer " +
+        "than TEZ_AM_CONTAINER_IDLE_RELEASE_TIMEOUT_MAX_MILLIS ");
+    Thread.sleep(50l);
     // container should not get released due to min held containers
     verify(rmClient, times(0)).releaseAssignedContainer(eq(container1.getId()));
 
@@ -1014,17 +991,12 @@ public class TestContainerReuse {
     TaskAttempt ta112 = mock(TaskAttempt.class);
     AMSchedulerEventTALaunchRequest lrEvent12 = createLaunchRequestEvent(taID112, ta112, resource1, host1, racks, priority1, dag1LRs);
 
-    drainNotifier.set(false);
     taskSchedulerManager.handleEvent(lrEvent11);
-    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
-    drainNotifier.set(false);
     taskSchedulerManager.handleEvent(lrEvent12);
-    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
 
     Container container1 = createContainer(1, "host1", resource1, priority1);
 
     // One container allocated.
-    drainNotifier.set(false);
     taskScheduler.onContainersAllocated(Collections.singletonList(container1));
     TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
     drainableAppCallback.drain();
@@ -1078,11 +1050,9 @@ public class TestContainerReuse {
 
     taskSchedulerManager.handleEvent(lrEvent21);
     taskSchedulerManager.handleEvent(lrEvent22);
-    drainableAppCallback.drain();
 
-    // TODO This is terrible, need a better way to ensure the scheduling loop has run
-    LOG.info("Sleeping to ensure that the scheduling loop runs");
-    Thread.sleep(6000l);
+    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
+    drainableAppCallback.drain();
     verify(taskSchedulerManager).taskAllocated(eq(0), eq(ta211), any(Object.class),
         eq(container1));
     verify(rmClient, times(0)).releaseAssignedContainer(eq(container1.getId()));
@@ -1158,8 +1128,11 @@ public class TestContainerReuse {
     String rsrc1 = "rsrc1";
     String rsrc2 = "rsrc2";
     LocalResource lr1 = mock(LocalResource.class);
+    doReturn(LocalResourceType.FILE).when(lr1).getType();
     LocalResource lr2 = mock(LocalResource.class);
+    doReturn(LocalResourceType.FILE).when(lr2).getType();
     LocalResource lr3 = mock(LocalResource.class);
+    doReturn(LocalResourceType.FILE).when(lr3).getType();
 
     AMContainerEventAssignTA assignEvent = null;
 
@@ -1173,7 +1146,8 @@ public class TestContainerReuse {
     TaskAttempt ta111 = mock(TaskAttempt.class);
     doReturn(taID111).when(ta111).getID();
     doReturn("Mock for TA " + taID111.toString()).when(ta111).toString();
-    AMSchedulerEventTALaunchRequest lrEvent11 = createLaunchRequestEvent(taID111, ta111, resource1, host1, racks, priority1, v11LR);
+    AMSchedulerEventTALaunchRequest lrEvent11 = createLaunchRequestEvent(
+        taID111, ta111, resource1, host1, racks, priority1, v11LR);
 
     Map<String, LocalResource> v12LR = Maps.newHashMap();
     v12LR.put(rsrc1, lr1);
@@ -1184,20 +1158,32 @@ public class TestContainerReuse {
     TaskAttempt ta112 = mock(TaskAttempt.class);
     doReturn(taID112).when(ta112).getID();
     doReturn("Mock for TA " + taID112.toString()).when(ta112).toString();
-    AMSchedulerEventTALaunchRequest lrEvent12 = createLaunchRequestEvent(taID112, ta112, resource1, host1, racks, priority1, v12LR);
+    AMSchedulerEventTALaunchRequest lrEvent12 = createLaunchRequestEvent(
+        taID112, ta112, resource1, host1, racks, priority1, v12LR);
+
+    //Vertex 1, Task 3, Attempt 1, host1
+    TezTaskAttemptID taID113 = TezTaskAttemptID.getInstance(TezTaskID.getInstance(vertexID11, 3), 1);
+    TaskAttempt ta113 = mock(TaskAttempt.class);
+    doReturn(taID113).when(ta113).getID();
+    doReturn("Mock for TA " + taID113.toString()).when(ta113).toString();
+    AMSchedulerEventTALaunchRequest lrEvent13 = createLaunchRequestEvent(
+        taID113, ta113, resource1, host1, racks, priority1, new HashMap<String, LocalResource>());
+    //Vertex 1, Task 4, Attempt 1, host1
+    TezTaskAttemptID taID114 = TezTaskAttemptID.getInstance(TezTaskID.getInstance(vertexID11, 4), 1);
+    TaskAttempt ta114 = mock(TaskAttempt.class);
+    doReturn(taID114).when(ta114).getID();
+    doReturn("Mock for TA " + taID114.toString()).when(ta114).toString();
+    AMSchedulerEventTALaunchRequest lrEvent14 = createLaunchRequestEvent(
+        taID114, ta114, resource1, host1, racks, priority1, new HashMap<String, LocalResource>());
+
 
-    drainNotifier.set(false);
     taskSchedulerManager.handleEvent(lrEvent11);
-    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
-    drainNotifier.set(false);
     taskSchedulerManager.handleEvent(lrEvent12);
-    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
 
     Container container1 = createContainer(1, "host1", resource1, priority1);
     Container container2 = createContainer(2, "host1", resource1, priority1);
 
     // One container allocated.
-    drainNotifier.set(false);
     taskScheduler.onContainersAllocated(Collections.singletonList(container1));
     TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
     drainableAppCallback.drain();
@@ -1220,14 +1206,50 @@ public class TestContainerReuse {
 
     // Task assigned to container completed successfully.
     // Verify reuse across hosts.
-    taskSchedulerManager.handleEvent(new AMSchedulerEventTAEnded(ta112, container1.getId(), TaskAttemptState.SUCCEEDED, null, null, 0));
+    taskSchedulerManager.handleEvent(new AMSchedulerEventTAEnded(ta112, container1.getId(),
+        TaskAttemptState.SUCCEEDED, null, null, 0));
     drainableAppCallback.drain();
     verifyDeAllocateTask(taskScheduler, ta112, true, null, null);
     verify(rmClient, times(0)).releaseAssignedContainer(eq(container1.getId()));
     eventHandler.verifyNoInvocations(AMContainerEventStopRequest.class);
     eventHandler.reset();
 
-    // Setup DAG2 with additional resources. Make sure the container, even without all resources, is reused.
+    // Task 3
+    taskSchedulerManager.handleEvent(lrEvent13);
+    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
+    drainableAppCallback.drain();
+    verify(taskSchedulerManager).taskAllocated(eq(0), eq(ta113), any(Object.class), eq(container1));
+    verify(rmClient, times(0)).releaseAssignedContainer(eq(container1.getId()));
+    eventHandler.verifyNoInvocations(AMContainerEventStopRequest.class);
+    eventHandler.reset();
+
+    taskSchedulerManager.handleEvent(new AMSchedulerEventTAEnded(ta113, container1.getId(),
+        TaskAttemptState.SUCCEEDED, null, null, 0));
+    drainableAppCallback.drain();
+    verifyDeAllocateTask(taskScheduler, ta113, true, null, null);
+    verify(rmClient, times(0)).releaseAssignedContainer(eq(container1.getId()));
+    eventHandler.verifyNoInvocations(AMContainerEventStopRequest.class);
+    eventHandler.reset();
+
+    // Task 4
+    taskSchedulerManager.handleEvent(lrEvent14);
+    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
+    drainableAppCallback.drain();
+    verify(taskSchedulerManager).taskAllocated(eq(0), eq(ta114), any(Object.class), eq(container1));
+    verify(rmClient, times(0)).releaseAssignedContainer(eq(container1.getId()));
+    eventHandler.verifyNoInvocations(AMContainerEventStopRequest.class);
+    eventHandler.reset();
+
+    taskSchedulerManager.handleEvent(new AMSchedulerEventTAEnded(ta114, container1.getId(),
+        TaskAttemptState.SUCCEEDED, null, null, 0));
+    drainableAppCallback.drain();
+    verifyDeAllocateTask(taskScheduler, ta114, true, null, null);
+    verify(rmClient, times(0)).releaseAssignedContainer(eq(container1.getId()));
+    eventHandler.verifyNoInvocations(AMContainerEventStopRequest.class);
+    eventHandler.reset();
+
+
+    // Setup DAG2 with different resources.
     TezDAGID dagID2 = TezDAGID.getInstance("0", 2, 0);
     dagIDAnswer.setDAGID(dagID2);
 
@@ -1246,19 +1268,47 @@ public class TestContainerReuse {
         host1, racks, priority1, v21LR);
 
     taskSchedulerManager.handleEvent(lrEvent21);
+
+    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
     drainableAppCallback.drain();
 
-    // TODO This is terrible, need a better way to ensure the scheduling loop has run
-    LOG.info("Sleeping to ensure that the scheduling loop runs");
-    Thread.sleep(6000l);
-    drainNotifier.set(false);
-    taskScheduler.onContainersAllocated(Collections.singletonList(container2));
-
-    Thread.sleep(6000l);
     verify(rmClient, times(1)).releaseAssignedContainer(eq(container1.getId()));
+
+    taskScheduler.onContainersAllocated(Collections.singletonList(container2));
+    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
+    drainableAppCallback.drain();
     verify(taskSchedulerManager).taskAllocated(eq(0), eq(ta211), any(Object.class), eq(container2));
     eventHandler.reset();
 
+    taskSchedulerManager.handleEvent(new AMSchedulerEventTAEnded(ta211, container2.getId(),
+        TaskAttemptState.SUCCEEDED, null, null, 0));
+    drainableAppCallback.drain();
+    verifyDeAllocateTask(taskScheduler, ta211, true, null, null);
+    verify(rmClient, times(0)).releaseAssignedContainer(eq(container2.getId()));
+    eventHandler.reset();
+
+
+    // Setup DAG3 with a subset of resources from DAG2 (container should be reused).
+    TezDAGID dagID3 = TezDAGID.getInstance("0", 3, 0);
+    dagIDAnswer.setDAGID(dagID3);
+
+    Map<String, LocalResource> v31LR = Collections.singletonMap(rsrc1, lr1);
+
+    // dag3, vertex1, task1, ta1
+    TezTaskAttemptID taID311 = TezTaskAttemptID.getInstance(
+        TezTaskID.getInstance(TezVertexID.getInstance(dagID3, 1), 1), 1);
+    TaskAttempt ta311 = mock(TaskAttempt.class);
+    doReturn(taID311).when(ta311).getID();
+    doReturn("Mock for TA " + taID311).when(ta311).toString();
+    AMSchedulerEventTALaunchRequest lrEvent31 = createLaunchRequestEvent(taID311, ta311, resource1,
+        host1, racks, priority1, v31LR);
+
+    taskSchedulerManager.handleEvent(lrEvent31);
+    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
+    drainableAppCallback.drain();
+    verify(taskSchedulerManager).taskAllocated(eq(0), eq(ta311), any(Object.class), eq(container2));
+    eventHandler.reset();
+
     taskScheduler.shutdown();
     taskSchedulerManager.close();
   }
@@ -1308,7 +1358,6 @@ public class TestContainerReuse {
 
     Resource resource1 = Resource.newInstance(1024, 1);
     String[] host1 = {"host1"};
-    String[] host2 = {"host2"};
 
     String []racks = {"/default-rack"};
     Priority priority1 = Priority.newInstance(1);
@@ -1325,7 +1374,6 @@ public class TestContainerReuse {
     Container container1 = createContainer(1, "host1", resource1, priority1);
 
     // One container allocated.
-    drainNotifier.set(false);
     taskScheduler.onContainersAllocated(Collections.singletonList(container1));
     drainableAppCallback.drain();
     verify(taskSchedulerManager, times(0)).taskAllocated(eq(0), eq(ta11),
@@ -1334,6 +1382,289 @@ public class TestContainerReuse {
     taskSchedulerManager.close();
   }
 
+  @Test(timeout=5000)
+  public void testDifferentResourceContainerReuse() throws Exception {
+    Configuration tezConf = new Configuration(new YarnConfiguration());
+    tezConf.setBoolean(TezConfiguration.TEZ_AM_CONTAINER_REUSE_ENABLED, true);
+    tezConf.setBoolean(TezConfiguration.TEZ_AM_CONTAINER_REUSE_RACK_FALLBACK_ENABLED, true);
+    tezConf.setLong(TezConfiguration.TEZ_AM_CONTAINER_REUSE_LOCALITY_DELAY_ALLOCATION_MILLIS, 0);
+    tezConf.setLong(TezConfiguration.TEZ_AM_CONTAINER_IDLE_RELEASE_TIMEOUT_MIN_MILLIS, 0);
+    tezConf.setLong(TezConfiguration.TEZ_AM_CONTAINER_IDLE_RELEASE_TIMEOUT_MAX_MILLIS, 0);
+
+    CapturingEventHandler eventHandler = new CapturingEventHandler();
+    TezDAGID dagID = TezDAGID.getInstance("0", 0, 0);
+
+    AMRMClient<CookieContainerRequest> rmClientCore = new AMRMClientForTest();
+    TezAMRMClientAsync<CookieContainerRequest> rmClient =
+        spy(new AMRMClientAsyncForTest(rmClientCore, 100));
+
+    AppContext appContext = mock(AppContext.class);
+    doReturn(new Configuration(false)).when(appContext).getAMConf();
+    AMContainerMap amContainerMap = new AMContainerMap(mock(ContainerHeartbeatHandler.class),
+        mock(TaskCommunicatorManagerInterface.class), new ContainerContextMatcher(), appContext);
+    AMNodeTracker amNodeTracker = new AMNodeTracker(eventHandler, appContext);
+    doReturn(amContainerMap).when(appContext).getAllContainers();
+    doReturn(amNodeTracker).when(appContext).getNodeTracker();
+    doReturn(DAGAppMasterState.RUNNING).when(appContext).getAMState();
+    doReturn(dagID).when(appContext).getCurrentDAGID();
+    doReturn(mock(ClusterInfo.class)).when(appContext).getClusterInfo();
+
+    TaskSchedulerManager taskSchedulerManagerReal = new TaskSchedulerManagerForTest(appContext,
+        eventHandler, rmClient, new AlwaysMatchesContainerMatcher(),
+        TezUtils.createUserPayloadFromConf(tezConf));
+    TaskSchedulerManager taskSchedulerManager = spy(taskSchedulerManagerReal);
+    taskSchedulerManager.init(tezConf);
+    taskSchedulerManager.start();
+
+    TaskSchedulerWithDrainableContext taskScheduler = (TaskSchedulerWithDrainableContext)
+        ((TaskSchedulerManagerForTest) taskSchedulerManager).getSpyTaskScheduler();
+    TaskSchedulerContextDrainable drainableAppCallback = taskScheduler.getDrainableAppCallback();
+    AtomicBoolean drainNotifier = new AtomicBoolean(false);
+    taskScheduler.delayedContainerManager.drainedDelayedContainersForTest = drainNotifier;
+
+    Resource resource1 = Resource.newInstance(1024, 1);
+    Resource resource2 = Resource.newInstance(2048, 1);
+    String[] host1 = {"host1"};
+    String[] host2 = {"host2"};
+
+    String []racks = {"/default-rack"};
+    Priority priority1 = Priority.newInstance(1);
+
+    TezVertexID vertexID1 = TezVertexID.getInstance(dagID, 1);
+
+    //Vertex 1, Task 1, Attempt 1, host1
+    TezTaskAttemptID taID11 = TezTaskAttemptID.getInstance(TezTaskID.getInstance(vertexID1, 1), 1);
+    TaskAttempt ta11 = mock(TaskAttempt.class);
+    AMSchedulerEventTALaunchRequest lrEvent1 = createLaunchRequestEvent(
+        taID11, ta11, resource1, host1, racks, priority1);
+
+    //Vertex 1, Task 2, Attempt 1, host1
+    TezTaskAttemptID taID12 = TezTaskAttemptID.getInstance(TezTaskID.getInstance(vertexID1, 2), 1);
+    TaskAttempt ta12 = mock(TaskAttempt.class);
+    AMSchedulerEventTALaunchRequest lrEvent2 = createLaunchRequestEvent(
+        taID12, ta12, resource1, host1, racks, priority1);
+
+    //Vertex 1, Task 3, Attempt 1, host2
+    TezTaskAttemptID taID13 = TezTaskAttemptID.getInstance(TezTaskID.getInstance(vertexID1, 3), 1);
+    TaskAttempt ta13 = mock(TaskAttempt.class);
+    AMSchedulerEventTALaunchRequest lrEvent3 = createLaunchRequestEvent(
+        taID13, ta13, resource2, host2, racks, priority1);
+
+    //Vertex 1, Task 4, Attempt 1, host2
+    TezTaskAttemptID taID14 = TezTaskAttemptID.getInstance(TezTaskID.getInstance(vertexID1, 4), 1);
+    TaskAttempt ta14 = mock(TaskAttempt.class);
+    AMSchedulerEventTALaunchRequest lrEvent4 = createLaunchRequestEvent(
+        taID14, ta14, resource2, host2, racks, priority1);
+
+    taskSchedulerManager.handleEvent(lrEvent1);
+    taskSchedulerManager.handleEvent(lrEvent2);
+    taskSchedulerManager.handleEvent(lrEvent3);
+    taskSchedulerManager.handleEvent(lrEvent4);
+
+    Container container1 = createContainer(1, "host1", resource1, priority1);
+    Container container2 = createContainer(2, "host2", resource2, priority1);
+
+    // One container allocated, should start ta11
+    taskScheduler.onContainersAllocated(Collections.singletonList(container1));
+    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
+    drainableAppCallback.drain();
+    verify(taskSchedulerManager).taskAllocated(eq(0), eq(ta11), any(Object.class), eq(container1));
+
+    // Second container allocated, should start ta13
+    taskScheduler.onContainersAllocated(Collections.singletonList(container2));
+    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
+    drainableAppCallback.drain();
+    verify(taskSchedulerManager).taskAllocated(eq(0), eq(ta13), any(Object.class), eq(container2));
+
+    // ta11 finished, should start ta12
+    taskSchedulerManager.handleEvent(new AMSchedulerEventTAEnded(ta11, container1.getId(),
+        TaskAttemptState.SUCCEEDED, null, null, 0));
+    drainableAppCallback.drain();
+    verifyDeAllocateTask(taskScheduler, ta11, true, null, null);
+    verify(taskSchedulerManager).taskAllocated(eq(0), eq(ta12), any(Object.class), eq(container1));
+    verify(rmClient, times(0)).releaseAssignedContainer(eq(container1.getId()));
+    eventHandler.verifyNoInvocations(AMContainerEventStopRequest.class);
+    eventHandler.reset();
+
+    // ta13 finished, should start ta14
+    taskSchedulerManager.handleEvent(new AMSchedulerEventTAEnded(ta13, container2.getId(),
+        TaskAttemptState.SUCCEEDED, null, null, 0));
+    drainableAppCallback.drain();
+    verifyDeAllocateTask(taskScheduler, ta13, true, null, null);
+    verify(taskSchedulerManager).taskAllocated(eq(0), eq(ta14), any(Object.class), eq(container2));
+    verify(rmClient, times(0)).releaseAssignedContainer(eq(container2.getId()));
+    eventHandler.verifyNoInvocations(AMContainerEventStopRequest.class);
+    eventHandler.reset();
+
+    // ta12 finished no pending requests, should release container1
+    taskSchedulerManager.handleEvent(new AMSchedulerEventTAEnded(ta12, container1.getId(),
+        TaskAttemptState.SUCCEEDED, null, null, 0));
+    drainableAppCallback.drain();
+    verifyDeAllocateTask(taskScheduler, ta12, true, null, null);
+    verify(rmClient).releaseAssignedContainer(eq(container1.getId()));
+    eventHandler.verifyInvocation(AMContainerEventStopRequest.class);
+    eventHandler.reset();
+
+    // ta14 finished no pending requests, should release container2.
+    taskSchedulerManager.handleEvent(new AMSchedulerEventTAEnded(ta14, container2.getId(),
+        TaskAttemptState.SUCCEEDED, null, null, 0));
+    drainableAppCallback.drain();
+    verifyDeAllocateTask(taskScheduler, ta14, true, null, null);
+    verify(rmClient).releaseAssignedContainer(eq(container2.getId()));
+    eventHandler.verifyInvocation(AMContainerEventStopRequest.class);
+    eventHandler.reset();
+
+    taskScheduler.shutdown();
+    taskSchedulerManager.close();
+  }
+
+  @Test(timeout=5000)
+  public void testEnvironmentVarsContainerReuse() throws Exception {
+    Configuration tezConf = new Configuration(new YarnConfiguration());
+    tezConf.setBoolean(TezConfiguration.TEZ_AM_CONTAINER_REUSE_ENABLED, true);
+    tezConf.setBoolean(TezConfiguration.TEZ_AM_CONTAINER_REUSE_RACK_FALLBACK_ENABLED, true);
+    tezConf.setLong(TezConfiguration.TEZ_AM_CONTAINER_REUSE_LOCALITY_DELAY_ALLOCATION_MILLIS, 0);
+    tezConf.setLong(TezConfiguration.TEZ_AM_CONTAINER_IDLE_RELEASE_TIMEOUT_MIN_MILLIS, 0);
+    tezConf.setLong(TezConfiguration.TEZ_AM_CONTAINER_IDLE_RELEASE_TIMEOUT_MAX_MILLIS, 0);
+
+    CapturingEventHandler eventHandler = new CapturingEventHandler();
+    TezDAGID dagID = TezDAGID.getInstance("0", 0, 0);
+
+    AMRMClient<CookieContainerRequest> rmClientCore = new AMRMClientForTest();
+    TezAMRMClientAsync<CookieContainerRequest> rmClient =
+        spy(new AMRMClientAsyncForTest(rmClientCore, 100));
+
+    AppContext appContext = mock(AppContext.class);
+    doReturn(new Configuration(false)).when(appContext).getAMConf();
+    AMContainerMap amContainerMap = new AMContainerMap(mock(ContainerHeartbeatHandler.class),
+        mock(TaskCommunicatorManagerInterface.class), new ContainerContextMatcher(), appContext);
+    AMNodeTracker amNodeTracker = new AMNodeTracker(eventHandler, appContext);
+    doReturn(amContainerMap).when(appContext).getAllContainers();
+    doReturn(amNodeTracker).when(appContext).getNodeTracker();
+    doReturn(DAGAppMasterState.RUNNING).when(appContext).getAMState();
+    doReturn(dagID).when(appContext).getCurrentDAGID();
+    doReturn(mock(ClusterInfo.class)).when(appContext).getClusterInfo();
+
+    TaskSchedulerManager taskSchedulerManagerReal = new TaskSchedulerManagerForTest(appContext,
+        eventHandler, rmClient, new ContainerContextMatcher(),
+        TezUtils.createUserPayloadFromConf(tezConf));
+    TaskSchedulerManager taskSchedulerManager = spy(taskSchedulerManagerReal);
+    taskSchedulerManager.init(tezConf);
+    taskSchedulerManager.start();
+
+    TaskSchedulerWithDrainableContext taskScheduler = (TaskSchedulerWithDrainableContext)
+        ((TaskSchedulerManagerForTest) taskSchedulerManager).getSpyTaskScheduler();
+    TaskSchedulerContextDrainable drainableAppCallback = taskScheduler.getDrainableAppCallback();
+    AtomicBoolean drainNotifier = new AtomicBoolean(false);
+    taskScheduler.delayedContainerManager.drainedDelayedContainersForTest = drainNotifier;
+
+    Resource resource1 = Resource.newInstance(1024, 1);
+    String[] host1 = {"host1"};
+
+    String []racks = {"/default-rack"};
+    Priority priority1 = Priority.newInstance(1);
+
+    TezVertexID vertexID1 = TezVertexID.getInstance(dagID, 1);
+
+    // Create containers with same resources and then see how tasks are allocated
+    // when the environments vary
+    Map<String, String> env1 = ImmutableMap.of("env1", "val1", "env2", "val2");
+    Map<String, String> env2 = ImmutableMap.of("env3", "val3", "env4", "val4");
+    Map<String, String> env3 = ImmutableMap.of("env1", "val1");
+    Map<String, String> env4 = ImmutableMap.of("env1", "val1", "env4", "val4");
+
+    //Vertex 1, Task 1, Attempt 1, host1, r1=lr1
+    TezTaskAttemptID taID11 = TezTaskAttemptID.getInstance(TezTaskID.getInstance(vertexID1, 1), 1);
+    TaskAttempt ta11 = mock(TaskAttempt.class);
+    AMSchedulerEventTALaunchRequest lrEvent1 = createLaunchRequestEvent(
+        taID11, ta11, resource1, host1, racks, priority1, "", env1);
+
+    //Vertex 1, Task 2, Attempt 1, host1, r1=lr2
+    TezTaskAttemptID taID12 = TezTaskAttemptID.getInstance(TezTaskID.getInstance(vertexID1, 2), 1);
+    TaskAttempt ta12 = mock(TaskAttempt.class);
+    AMSchedulerEventTALaunchRequest lrEvent2 = createLaunchRequestEvent(
+        taID12, ta12, resource1, host1, racks, priority1, "", env2);
+
+    //Vertex 1, Task 3, Attempt 1, host1, r1=lr1 & r2=lr2
+    TezTaskAttemptID taID13 = TezTaskAttemptID.getInstance(TezTaskID.getInstance(vertexID1, 3), 1);
+    TaskAttempt ta13 = mock(TaskAttempt.class);
+    AMSchedulerEventTALaunchRequest lrEvent3 = createLaunchRequestEvent(
+        taID13, ta13, resource1, host1, racks, priority1, "", env3);
+
+    //Vertex 1, Task 4, Attempt 1, host1, r2=lr2
+    TezTaskAttemptID taID14 = TezTaskAttemptID.getInstance(TezTaskID.getInstance(vertexID1, 4), 1);
+    TaskAttempt ta14 = mock(TaskAttempt.class);
+    AMSchedulerEventTALaunchRequest lrEvent4 = createLaunchRequestEvent(
+        taID14, ta14, resource1, host1, racks, priority1, "", env4);
+
+    taskSchedulerManager.handleEvent(lrEvent1);
+    taskSchedulerManager.handleEvent(lrEvent2);
+    taskSchedulerManager.handleEvent(lrEvent3);
+    taskSchedulerManager.handleEvent(lrEvent4);
+
+    Container container1 = createContainer(1, "host1", resource1, priority1);
+    Container container2 = createContainer(2, "host1", resource1, priority1);
+    Container container3 = createContainer(3, "host1", resource1, priority1);
+
+    // One container allocated, should start ta11 on container1
+    taskScheduler.onContainersAllocated(Collections.singletonList(container1));
+    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
+    drainableAppCallback.drain();
+    verify(taskSchedulerManager).taskAllocated(eq(0), eq(ta11), any(Object.class), eq(container1));
+
+    // finish ta11, should start ta13
+    taskSchedulerManager.handleEvent(new AMSchedulerEventTAEnded(ta11, container1.getId(),
+        TaskAttemptState.SUCCEEDED, null, null, 0));
+    drainableAppCallback.drain();
+    verifyDeAllocateTask(taskScheduler, ta11, true, null, null);
+    verify(taskSchedulerManager).taskAllocated(eq(0), eq(ta13), any(Object.class), eq(container1));
+    verify(rmClient, times(0)).releaseAssignedContainer(eq(container1.getId()));
+    eventHandler.verifyNoInvocations(AMContainerEventStopRequest.class);
+    eventHandler.reset();
+
+    // ta13 finished, cannot reuse container1, should release container1
+    taskSchedulerManager.handleEvent(new AMSchedulerEventTAEnded(ta13, container1.getId(),
+        TaskAttemptState.SUCCEEDED, null, null, 0));
+    drainableAppCallback.drain();
+    verifyDeAllocateTask(taskScheduler, ta13, true, null, null);
+    verify(rmClient).releaseAssignedContainer(eq(container1.getId()));
+    eventHandler.verifyInvocation(AMContainerEventStopRequest.class);
+    eventHandler.reset();
+
+    // Second container allocated, should start ta12.
+    taskScheduler.onContainersAllocated(Collections.singletonList(container2));
+    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
+    drainableAppCallback.drain();
+    verify(taskSchedulerManager).taskAllocated(eq(0), eq(ta12), any(Object.class), eq(container2));
+
+    // ta12 finished, cannot reuse container, should release container2
+    taskSchedulerManager.handleEvent(new AMSchedulerEventTAEnded(ta12, container2.getId(),
+        TaskAttemptState.SUCCEEDED, null, null, 0));
+    drainableAppCallback.drain();
+    verifyDeAllocateTask(taskScheduler, ta12, true, null, null);
+    verify(rmClient).releaseAssignedContainer(eq(container2.getId()));
+    eventHandler.verifyInvocation(AMContainerEventStopRequest.class);
+    eventHandler.reset();
+
+    // Third container allocated, should start ta14.
+    taskScheduler.onContainersAllocated(Collections.singletonList(container3));
+    TestTaskSchedulerHelpers.waitForDelayedDrainNotify(drainNotifier);
+    drainableAppCallback.drain();
+    verify(taskSchedulerManager).taskAllocated(eq(0), eq(ta14), any(Object.class), eq(container3));
+
+    // ta14 finished, should release container3
+    taskSchedulerManager.handleEvent(new AMSchedulerEventTAEnded(ta14, container3.getId(),
+        TaskAttemptState.SUCCEEDED, null, null, 0));
+    drainableAppCallback.drain();
+    verifyDeAllocateTask(taskScheduler, ta14, true, null, null);
+    verify(rmClient).releaseAssignedContainer(eq(container3.getId()));
+    eventHandler.verifyInvocation(AMContainerEventStopRequest.class);
+    eventHandler.reset();
+
+    taskScheduler.shutdown();
+    taskSchedulerManager.close();
+  }
+
   private Container createContainer(int id, String host, Resource resource, Priority priority) {
     @SuppressWarnings("deprecation")
     ContainerId containerID = ContainerId.newInstance(
@@ -1361,7 +1692,7 @@ public class TestContainerReuse {
       Collections.singletonList(new InputSpec("vertexName",
           InputDescriptor.create("inputClassName"), 1)),
       Collections.singletonList(new OutputSpec("vertexName",
-          OutputDescriptor.create("outputClassName"), 1)), null), ta, locationHint,
+          OutputDescriptor.create("outputClassName"), 1)), null, null), ta, locationHint,
       priority.getPriority(), containerContext, 0, 0, 0);
     return lr;
   }
@@ -1386,6 +1717,14 @@ public class TestContainerReuse {
         new ContainerContext(localResources, new Credentials(), new HashMap<String, String>(), jvmOpts));
   }
 
+  private AMSchedulerEventTALaunchRequest createLaunchRequestEvent(TezTaskAttemptID taID,
+      TaskAttempt ta, Resource capability, String[] hosts, String[] racks, Priority priority,
+      String jvmOpts, Map<String, String> environment) {
+    return createLaunchRequestEvent(taID, ta, capability, hosts, racks, priority,
+        new ContainerContext(new HashMap<String, LocalResource>(), new Credentials(), environment,
+            jvmOpts));
+  }
+
   private static class ChangingDAGIDAnswer implements Answer<TezDAGID> {
 
     private TezDAGID dagID;
