@@ -24,14 +24,17 @@ import static org.mockito.Mockito.mock;
 import static org.mockito.Mockito.spy;
 import static org.mockito.Mockito.when;
 
+import javax.annotation.Nonnull;
+import javax.annotation.Nullable;
 import java.io.IOException;
 import java.nio.ByteBuffer;
 import java.util.Collection;
-import java.util.LinkedList;
 import java.util.List;
 import java.util.Map;
+import java.util.Queue;
 import java.util.concurrent.Callable;
 import java.util.concurrent.CompletionService;
+import java.util.concurrent.ConcurrentLinkedQueue;
 import java.util.concurrent.ExecutionException;
 import java.util.concurrent.ExecutorCompletionService;
 import java.util.concurrent.ExecutorService;
@@ -45,7 +48,6 @@ import com.google.common.base.Preconditions;
 import com.google.common.collect.Lists;
 import com.google.common.collect.Maps;
 import org.apache.hadoop.conf.Configuration;
-import org.apache.hadoop.service.AbstractService;
 import org.apache.hadoop.yarn.api.protocolrecords.RegisterApplicationMasterResponse;
 import org.apache.hadoop.yarn.api.records.ApplicationAccessType;
 import org.apache.hadoop.yarn.api.records.ApplicationAttemptId;
@@ -68,6 +70,9 @@ import org.apache.tez.dag.api.UserPayload;
 import org.apache.tez.dag.app.AppContext;
 import org.apache.tez.dag.app.ServicePluginLifecycleAbstractService;
 import org.apache.tez.dag.app.rm.YarnTaskSchedulerService.CookieContainerRequest;
+import org.apache.tez.hadoop.shim.HadoopShimsLoader;
+import org.apache.tez.serviceplugins.api.DagInfo;
+import org.apache.tez.serviceplugins.api.ServicePluginError;
 import org.apache.tez.serviceplugins.api.TaskScheduler;
 import org.apache.tez.serviceplugins.api.TaskSchedulerContext;
 
@@ -75,6 +80,11 @@ class TestTaskSchedulerHelpers {
 
   // Mocking AMRMClientImpl to make use of getMatchingRequest
   static class AMRMClientForTest extends AMRMClientImpl<CookieContainerRequest> {
+    AMRMClientForTest() {
+      super();
+      this.clusterAvailableResources = Resource.newInstance(4000, 4);
+      this.clusterNodeCount = 5;
+    }
 
     @Override
     protected void serviceStart() {
@@ -89,6 +99,7 @@ class TestTaskSchedulerHelpers {
   // Mocking AMRMClientAsyncImpl to make use of getMatchingRequest
   static class AMRMClientAsyncForTest extends
       TezAMRMClientAsync<CookieContainerRequest> {
+    private RegisterApplicationMasterResponse mockRegResponse;
 
     public AMRMClientAsyncForTest(
         AMRMClient<CookieContainerRequest> client,
@@ -101,7 +112,7 @@ class TestTaskSchedulerHelpers {
     @Override
     public RegisterApplicationMasterResponse registerApplicationMaster(
         String appHostName, int appHostPort, String appTrackingUrl) {
-      RegisterApplicationMasterResponse mockRegResponse = mock(RegisterApplicationMasterResponse.class);
+      mockRegResponse = mock(RegisterApplicationMasterResponse.class);
       Resource mockMaxResource = mock(Resource.class);
       Map<ApplicationAccessType, String> mockAcls = mock(Map.class);
       when(mockRegResponse.getMaximumResourceCapability()).thenReturn(
@@ -122,6 +133,10 @@ class TestTaskSchedulerHelpers {
     @Override
     protected void serviceStop() {
     }
+
+    RegisterApplicationMasterResponse getRegistrationResponse() {
+      return mockRegResponse;
+    }
   }
   
   // Overrides start / stop. Will be controlled without the extra event handling thread.
@@ -140,7 +155,7 @@ class TestTaskSchedulerHelpers {
                                        UserPayload defaultPayload) {
       super(appContext, null, eventHandler, containerSignatureMatcher, null,
           Lists.newArrayList(new NamedEntityDescriptor("FakeScheduler", null)),
-          false);
+          false, new HadoopShimsLoader(appContext.getAMConf()).getHadoopShim());
       this.amrmClientAsync = amrmClientAsync;
       this.containerSignatureMatcher = containerSignatureMatcher;
       this.defaultPayload = defaultPayload;
@@ -159,7 +174,7 @@ class TestTaskSchedulerHelpers {
       TaskSchedulerContextDrainable drainable = new TaskSchedulerContextDrainable(wrapper);
 
       taskSchedulers[0] = new TaskSchedulerWrapper(
-          new TaskSchedulerWithDrainableContext(drainable, amrmClientAsync));
+          spy(new TaskSchedulerWithDrainableContext(drainable, amrmClientAsync)));
       taskSchedulerServiceWrappers[0] =
           new ServicePluginLifecycleAbstractService(taskSchedulers[0].getTaskScheduler());
     }
@@ -172,11 +187,8 @@ class TestTaskSchedulerHelpers {
     public void serviceStart() {
       instantiateSchedulers("host", 0, "", appContext);
       // Init the service so that reuse configuration is picked up.
-      ((AbstractService)taskSchedulerServiceWrappers[0]).init(getConfig());
-      ((AbstractService)taskSchedulerServiceWrappers[0]).start();
-      // For some reason, the spy needs to be setup after sertvice startup.
-      taskSchedulers[0] = new TaskSchedulerWrapper(spy(taskSchedulers[0].getTaskScheduler()));
-
+      taskSchedulerServiceWrappers[0].init(getConfig());
+      taskSchedulerServiceWrappers[0].start();
     }
 
     @Override
@@ -187,8 +199,7 @@ class TestTaskSchedulerHelpers {
   @SuppressWarnings("rawtypes")
   static class CapturingEventHandler implements EventHandler {
 
-    private List<Event> events = new LinkedList<Event>();
-
+    private Queue<Event> events = new ConcurrentLinkedQueue<Event>();
 
     public void handle(Event event) {
       events.add(event);
@@ -277,15 +288,16 @@ class TestTaskSchedulerHelpers {
 
     @Override
     public void setApplicationRegistrationData(Resource maxContainerCapability,
-        Map<ApplicationAccessType, String> appAcls, ByteBuffer key) {
+        Map<ApplicationAccessType, String> appAcls, ByteBuffer key, String queueName) {
       invocations++;
-      real.setApplicationRegistrationData(maxContainerCapability, appAcls, key);
+      real.setApplicationRegistrationData(maxContainerCapability, appAcls, key, queueName);
     }
 
     @Override
-    public void onError(Throwable t) {
+    public void reportError(@Nonnull ServicePluginError servicePluginError, String message,
+                            DagInfo dagInfo) {
       invocations++;
-      real.onError(t);
+      real.reportError(servicePluginError, message, dagInfo);
     }
 
     @Override
@@ -327,6 +339,12 @@ class TestTaskSchedulerHelpers {
       return real.getApplicationAttemptId();
     }
 
+    @Nullable
+    @Override
+    public DagInfo getCurrentDagInfo() {
+      return real.getCurrentDagInfo();
+    }
+
     @Override
     public String getAppHostName() {
       return real.getAppHostName();
@@ -347,6 +365,11 @@ class TestTaskSchedulerHelpers {
       return real.getAMState();
     }
 
+    @Override
+    public int getVertexIndexForTask(Object task) {
+      return real.getVertexIndexForTask(task);
+    }
+
     @Override
     public void preemptContainer(ContainerId cId) {
       invocations++;
