@@ -28,11 +28,13 @@ import static org.mockito.Matchers.eq;
 import static org.mockito.Mockito.RETURNS_DEEP_STUBS;
 import static org.mockito.Mockito.doReturn;
 import static org.mockito.Mockito.mock;
+import static org.mockito.Mockito.reset;
 import static org.mockito.Mockito.spy;
 import static org.mockito.Mockito.times;
 import static org.mockito.Mockito.verify;
 import static org.mockito.Mockito.when;
 
+import javax.annotation.Nullable;
 import java.io.IOException;
 import java.lang.reflect.Method;
 import java.net.InetSocketAddress;
@@ -68,32 +70,40 @@ import org.apache.tez.dag.api.TezConstants;
 import org.apache.tez.dag.api.TezException;
 import org.apache.tez.dag.api.UserPayload;
 import org.apache.tez.dag.api.client.DAGClientServer;
+import org.apache.tez.dag.api.oldrecords.TaskAttemptState;
 import org.apache.tez.dag.app.AppContext;
 import org.apache.tez.dag.app.ContainerContext;
 import org.apache.tez.dag.app.ServicePluginLifecycleAbstractService;
+import org.apache.tez.dag.app.dag.DAG;
 import org.apache.tez.dag.app.dag.TaskAttempt;
 import org.apache.tez.dag.app.dag.event.DAGAppMasterEventType;
 import org.apache.tez.dag.app.dag.event.DAGAppMasterEventUserServiceFatalError;
+import org.apache.tez.dag.app.dag.event.DAGEventTerminateDag;
 import org.apache.tez.dag.app.dag.impl.TaskAttemptImpl;
 import org.apache.tez.dag.app.dag.impl.TaskImpl;
 import org.apache.tez.dag.app.dag.impl.VertexImpl;
 import org.apache.tez.dag.app.rm.container.AMContainer;
 import org.apache.tez.dag.app.rm.container.AMContainerEventAssignTA;
 import org.apache.tez.dag.app.rm.container.AMContainerEventCompleted;
+import org.apache.tez.dag.app.rm.container.AMContainerEventStopRequest;
 import org.apache.tez.dag.app.rm.container.AMContainerEventType;
 import org.apache.tez.dag.app.rm.container.AMContainerMap;
 import org.apache.tez.dag.app.rm.container.AMContainerState;
 import org.apache.tez.dag.app.web.WebUIService;
+import org.apache.tez.dag.helpers.DagInfoImplForTest;
 import org.apache.tez.dag.records.TaskAttemptTerminationCause;
 import org.apache.tez.dag.records.TezDAGID;
 import org.apache.tez.dag.records.TezTaskAttemptID;
 import org.apache.tez.dag.records.TezTaskID;
 import org.apache.tez.dag.records.TezVertexID;
+import org.apache.tez.hadoop.shim.HadoopShimsLoader;
 import org.apache.tez.runtime.api.impl.TaskSpec;
+import org.apache.tez.serviceplugins.api.ServicePluginErrorDefaults;
 import org.apache.tez.serviceplugins.api.ServicePluginException;
 import org.apache.tez.serviceplugins.api.TaskAttemptEndReason;
 import org.apache.tez.serviceplugins.api.TaskScheduler;
 import org.apache.tez.serviceplugins.api.TaskSchedulerContext;
+import org.apache.tez.serviceplugins.api.TaskSchedulerDescriptor;
 import org.junit.Assert;
 import org.junit.Before;
 import org.junit.Test;
@@ -123,7 +133,8 @@ public class TestTaskSchedulerManager {
                                     ContainerSignatureMatcher containerSignatureMatcher,
                                     WebUIService webUI) {
       super(appContext, clientService, eventHandler, containerSignatureMatcher, webUI,
-          Lists.newArrayList(new NamedEntityDescriptor("FakeDescriptor", null)), false);
+          Lists.newArrayList(new NamedEntityDescriptor("FakeDescriptor", null)), false,
+          new HadoopShimsLoader(appContext.getAMConf()).getHadoopShim());
     }
 
     @Override
@@ -208,6 +219,79 @@ public class TestTaskSchedulerManager {
     assertEquals(mockAttemptId, assignEvent.getTaskAttemptId());
   }
 
+  @Test(timeout = 5000)
+  public void testTASucceededAfterContainerCleanup() throws Exception {
+    Configuration conf = new Configuration(false);
+    schedulerHandler.init(conf);
+    schedulerHandler.start();
+
+    TaskAttemptImpl mockTaskAttempt = mock(TaskAttemptImpl.class);
+    TezTaskAttemptID mockAttemptId = mock(TezTaskAttemptID.class);
+    when(mockAttemptId.getId()).thenReturn(0);
+    when(mockTaskAttempt.getID()).thenReturn(mockAttemptId);
+    Resource resource = Resource.newInstance(1024, 1);
+    ContainerContext containerContext =
+        new ContainerContext(new HashMap<String, LocalResource>(), new Credentials(),
+            new HashMap<String, String>(), "");
+    int priority = 10;
+    TaskLocationHint locHint = TaskLocationHint.createTaskLocationHint(new HashSet<String>(), null);
+
+    ContainerId mockCId = mock(ContainerId.class);
+    Container container = mock(Container.class);
+    when(container.getId()).thenReturn(mockCId);
+
+    AMContainer mockAMContainer = mock(AMContainer.class);
+    when(mockAMContainer.getContainerId()).thenReturn(mockCId);
+    when(mockAMContainer.getState()).thenReturn(AMContainerState.IDLE);
+
+    // Returning null container will replicate container cleanup scenario
+    when(mockAMContainerMap.get(mockCId)).thenReturn(null);
+
+    AMSchedulerEventTALaunchRequest lr =
+        new AMSchedulerEventTALaunchRequest(mockAttemptId, resource, null, mockTaskAttempt, locHint,
+            priority, containerContext, 0, 0, 0);
+    schedulerHandler.taskAllocated(0, mockTaskAttempt, lr, container);
+    assertEquals(1, mockEventHandler.events.size());
+    assertTrue(mockEventHandler.events.get(0) instanceof AMContainerEventAssignTA);
+    AMContainerEventAssignTA assignEvent =
+        (AMContainerEventAssignTA) mockEventHandler.events.get(0);
+    assertEquals(priority, assignEvent.getPriority());
+    assertEquals(mockAttemptId, assignEvent.getTaskAttemptId());
+  }
+
+  @Test(timeout = 5000)
+  public void testTAUnsuccessfulAfterContainerCleanup() throws Exception {
+    Configuration conf = new Configuration(false);
+    schedulerHandler.init(conf);
+    schedulerHandler.start();
+
+    TaskAttemptImpl mockTaskAttempt = mock(TaskAttemptImpl.class);
+    TezTaskAttemptID mockAttemptId = mock(TezTaskAttemptID.class);
+    when(mockAttemptId.getId()).thenReturn(0);
+    when(mockTaskAttempt.getID()).thenReturn(mockAttemptId);
+
+    ContainerId mockCId = mock(ContainerId.class);
+    Container container = mock(Container.class);
+    when(container.getId()).thenReturn(mockCId);
+
+    AMContainer mockAMContainer = mock(AMContainer.class);
+    when(mockAMContainer.getContainerId()).thenReturn(mockCId);
+    when(mockAMContainer.getState()).thenReturn(AMContainerState.IDLE);
+    when(mockTaskAttempt.getAssignedContainerID()).thenReturn(mockCId);
+
+    // Returning null container will replicate container cleanup scenario
+    when(mockAMContainerMap.get(mockCId)).thenReturn(null);
+
+    schedulerHandler.handleEvent(
+        new AMSchedulerEventTAEnded(
+            mockTaskAttempt, mockCId, TaskAttemptState.KILLED, null, null, 0));
+    assertEquals(1, mockEventHandler.events.size());
+    assertTrue(mockEventHandler.events.get(0) instanceof AMContainerEventStopRequest);
+    AMContainerEventStopRequest stopEvent =
+        (AMContainerEventStopRequest) mockEventHandler.events.get(0);
+    assertEquals(mockCId, stopEvent.getContainerId());
+  }
+
   @Test (timeout = 5000)
   public void testTaskBasedAffinity() throws Exception {
     Configuration conf = new Configuration(false);
@@ -279,7 +363,7 @@ public class TestTaskSchedulerManager {
     schedulerHandler.stop();
     schedulerHandler.close();
   }
-  
+
   @Test (timeout = 5000)
   public void testContainerInternalPreempted() throws IOException, ServicePluginException {
     Configuration conf = new Configuration(false);
@@ -309,6 +393,37 @@ public class TestTaskSchedulerManager {
     schedulerHandler.stop();
     schedulerHandler.close();
   }
+
+  @Test(timeout = 5000)
+  public void testContainerInternalPreemptedAfterContainerCleanup() throws IOException, ServicePluginException {
+    Configuration conf = new Configuration(false);
+    schedulerHandler.init(conf);
+    schedulerHandler.start();
+
+    AMContainer mockAmContainer = mock(AMContainer.class);
+    when(mockAmContainer.getTaskSchedulerIdentifier()).thenReturn(0);
+    when(mockAmContainer.getContainerLauncherIdentifier()).thenReturn(0);
+    when(mockAmContainer.getTaskCommunicatorIdentifier()).thenReturn(0);
+    ContainerId mockCId = mock(ContainerId.class);
+    verify(mockTaskScheduler, times(0)).deallocateContainer((ContainerId) any());
+    // Returning null container will replicate container cleanup scenario
+    when(mockAMContainerMap.get(mockCId)).thenReturn(null);
+    schedulerHandler.preemptContainer(0, mockCId);
+    verify(mockTaskScheduler, times(0)).deallocateContainer(mockCId);
+    assertEquals(1, mockEventHandler.events.size());
+    Event event = mockEventHandler.events.get(0);
+    assertEquals(AMContainerEventType.C_COMPLETED, event.getType());
+    AMContainerEventCompleted completedEvent = (AMContainerEventCompleted) event;
+    assertEquals(mockCId, completedEvent.getContainerId());
+    assertEquals("Container preempted internally", completedEvent.getDiagnostics());
+    assertTrue(completedEvent.isPreempted());
+    Assert.assertFalse(completedEvent.isDiskFailed());
+    assertEquals(TaskAttemptTerminationCause.INTERNAL_PREEMPTION,
+        completedEvent.getTerminationCause());
+
+    schedulerHandler.stop();
+    schedulerHandler.close();
+  }
   
   @Test (timeout = 5000)
   public void testContainerDiskFailed() throws IOException {
@@ -381,44 +496,61 @@ public class TestTaskSchedulerManager {
   @Test (timeout = 5000)
   public void testHistoryUrlConf() throws Exception {
     Configuration conf = schedulerHandler.appContext.getAMConf();
-
-    // ensure history url is empty when timeline server is not the logging class
-    conf.set(TezConfiguration.TEZ_HISTORY_URL_BASE, "http://ui-host:9999");
-    assertTrue("".equals(schedulerHandler.getHistoryUrl()));
-
-    // ensure expansion of url happens
-    conf.set(TezConfiguration.TEZ_HISTORY_LOGGING_SERVICE_CLASS,
-        "org.apache.tez.dag.history.logging.ats.ATSHistoryLoggingService");
     final ApplicationId mockApplicationId = mock(ApplicationId.class);
     doReturn("TEST_APP_ID").when(mockApplicationId).toString();
     doReturn(mockApplicationId).when(mockAppContext).getApplicationID();
-    assertTrue("http://ui-host:9999/#/tez-app/TEST_APP_ID"
-        .equals(schedulerHandler.getHistoryUrl()));
+
+    // ensure history url is empty when timeline server is not the logging class
+    conf.set(TezConfiguration.TEZ_HISTORY_URL_BASE, "http://ui-host:9999");
+    assertEquals("http://ui-host:9999/#/tez-app/TEST_APP_ID",
+        schedulerHandler.getHistoryUrl());
 
     // ensure the trailing / in history url is handled
     conf.set(TezConfiguration.TEZ_HISTORY_URL_BASE, "http://ui-host:9998/");
-    assertTrue("http://ui-host:9998/#/tez-app/TEST_APP_ID"
-        .equals(schedulerHandler.getHistoryUrl()));
+    assertEquals("http://ui-host:9998/#/tez-app/TEST_APP_ID",
+        schedulerHandler.getHistoryUrl());
 
     // ensure missing scheme in history url is handled
     conf.set(TezConfiguration.TEZ_HISTORY_URL_BASE, "ui-host:9998/");
-    Assert.assertTrue("http://ui-host:9998/#/tez-app/TEST_APP_ID"
-        .equals(schedulerHandler.getHistoryUrl()));
+    assertEquals("http://ui-host:9998/#/tez-app/TEST_APP_ID",
+        schedulerHandler.getHistoryUrl());
 
     // handle bad template ex without begining /
     conf.set(TezConfiguration.TEZ_AM_TEZ_UI_HISTORY_URL_TEMPLATE,
         "__HISTORY_URL_BASE__#/somepath");
-    assertTrue("http://ui-host:9998/#/somepath"
-        .equals(schedulerHandler.getHistoryUrl()));
+    assertEquals("http://ui-host:9998/#/somepath",
+        schedulerHandler.getHistoryUrl());
 
     conf.set(TezConfiguration.TEZ_AM_TEZ_UI_HISTORY_URL_TEMPLATE,
         "__HISTORY_URL_BASE__?viewPath=tez-app/__APPLICATION_ID__");
     conf.set(TezConfiguration.TEZ_HISTORY_URL_BASE, "http://localhost/ui/tez");
-    assertTrue("http://localhost/ui/tez?viewPath=tez-app/TEST_APP_ID"
-        .equals(schedulerHandler.getHistoryUrl()));
+    assertEquals("http://localhost/ui/tez?viewPath=tez-app/TEST_APP_ID",
+        schedulerHandler.getHistoryUrl());
 
   }
 
+  @Test (timeout = 5000)
+  public void testHistoryUrlWithoutScheme() throws Exception {
+    Configuration conf = schedulerHandler.appContext.getAMConf();
+    final ApplicationId mockApplicationId = mock(ApplicationId.class);
+    doReturn("TEST_APP_ID").when(mockApplicationId).toString();
+    doReturn(mockApplicationId).when(mockAppContext).getApplicationID();
+
+    conf.set(TezConfiguration.TEZ_HISTORY_URL_BASE, "/foo/bar/");
+    conf.setBoolean(TezConfiguration.TEZ_AM_UI_HISTORY_URL_SCHEME_CHECK_ENABLED, false);
+    assertEquals("/foo/bar/#/tez-app/TEST_APP_ID",
+        schedulerHandler.getHistoryUrl());
+
+    conf.set(TezConfiguration.TEZ_HISTORY_URL_BASE, "ui-host:9998/foo/bar/");
+    assertEquals("ui-host:9998/foo/bar/#/tez-app/TEST_APP_ID",
+        schedulerHandler.getHistoryUrl());
+
+    conf.setBoolean(TezConfiguration.TEZ_AM_UI_HISTORY_URL_SCHEME_CHECK_ENABLED, true);
+    conf.set(TezConfiguration.TEZ_HISTORY_URL_BASE, "ui-host:9998/foo/bar/");
+    assertEquals("http://ui-host:9998/foo/bar/#/tez-app/TEST_APP_ID",
+        schedulerHandler.getHistoryUrl());
+  }
+
   @Test(timeout = 5000)
   public void testNoSchedulerSpecified() throws IOException {
     try {
@@ -542,6 +674,99 @@ public class TestTaskSchedulerManager {
         eq(launchRequest2));
   }
 
+  @SuppressWarnings("unchecked")
+  @Test(timeout = 5000)
+  public void testShutdownBeforeStartTaskScheduler() {
+    Configuration conf = new TezConfiguration();
+    AppContext appContext = mock(AppContext.class, RETURNS_DEEP_STUBS);
+    doReturn(conf).when(appContext).getAMConf();
+
+    List<NamedEntityDescriptor> list = new LinkedList<>();
+    list.add(null);
+
+    TaskSchedulerManager taskSchedulerManager =
+        new TaskSchedulerManager(appContext, null, null,
+            null, null, list, false,null);
+    assertFalse("Should not return true unless actually unregistered successfully",
+        taskSchedulerManager.hasUnregistered());
+  }
+
+  @SuppressWarnings("unchecked")
+  @Test(timeout = 5000)
+  public void testReportFailureFromTaskScheduler() {
+    String dagName = DAG_NAME;
+    Configuration conf = new TezConfiguration();
+    String taskSchedulerName = "testTaskScheduler";
+    String expIdentifier = "[0:" + taskSchedulerName + "]";
+    EventHandler eventHandler = mock(EventHandler.class);
+    AppContext appContext = mock(AppContext.class, RETURNS_DEEP_STUBS);
+    doReturn(taskSchedulerName).when(appContext).getTaskSchedulerName(0);
+    doReturn(eventHandler).when(appContext).getEventHandler();
+    doReturn(conf).when(appContext).getAMConf();
+    InetSocketAddress address = new InetSocketAddress("host", 55000);
+
+    DAGClientServer dagClientServer = mock(DAGClientServer.class);
+    doReturn(address).when(dagClientServer).getBindAddress();
+
+    DAG dag = mock(DAG.class);
+    TezDAGID dagId = TezDAGID.getInstance(ApplicationId.newInstance(1, 0), DAG_INDEX);
+    doReturn(dagName).when(dag).getName();
+    doReturn(dagId).when(dag).getID();
+    doReturn(dag).when(appContext).getCurrentDAG();
+
+    NamedEntityDescriptor<TaskSchedulerDescriptor> namedEntityDescriptor =
+        new NamedEntityDescriptor<>(taskSchedulerName, TaskSchedulerForFailureTest.class.getName());
+    List<NamedEntityDescriptor> list = new LinkedList<>();
+    list.add(namedEntityDescriptor);
+
+    TaskSchedulerManager taskSchedulerManager =
+        new TaskSchedulerManager(appContext, dagClientServer, eventHandler,
+            mock(ContainerSignatureMatcher.class), mock(WebUIService.class), list, false,
+            new HadoopShimsLoader(appContext.getAMConf()).getHadoopShim()) {
+          @Override
+          TaskSchedulerContext wrapTaskSchedulerContext(TaskSchedulerContext rawContext) {
+            // Avoid wrapping in threads
+            return rawContext;
+          }
+        };
+    try {
+      taskSchedulerManager.init(new TezConfiguration());
+      taskSchedulerManager.start();
+
+      taskSchedulerManager.getTotalResources(0);
+      ArgumentCaptor<Event> argumentCaptor = ArgumentCaptor.forClass(Event.class);
+      verify(eventHandler, times(1)).handle(argumentCaptor.capture());
+
+      Event rawEvent = argumentCaptor.getValue();
+      assertTrue(rawEvent instanceof DAGEventTerminateDag);
+      DAGEventTerminateDag killEvent = (DAGEventTerminateDag) rawEvent;
+      assertTrue(killEvent.getDiagnosticInfo().contains("ReportError"));
+      assertTrue(killEvent.getDiagnosticInfo()
+          .contains(ServicePluginErrorDefaults.SERVICE_UNAVAILABLE.name()));
+      assertTrue(killEvent.getDiagnosticInfo().contains(expIdentifier));
+
+
+      reset(eventHandler);
+      taskSchedulerManager.getAvailableResources(0);
+      argumentCaptor = ArgumentCaptor.forClass(Event.class);
+
+      verify(eventHandler, times(1)).handle(argumentCaptor.capture());
+      rawEvent = argumentCaptor.getValue();
+
+      assertTrue(rawEvent instanceof DAGAppMasterEventUserServiceFatalError);
+      DAGAppMasterEventUserServiceFatalError event =
+          (DAGAppMasterEventUserServiceFatalError) rawEvent;
+      assertEquals(DAGAppMasterEventType.TASK_SCHEDULER_SERVICE_FATAL_ERROR, event.getType());
+      assertTrue(event.getDiagnosticInfo().contains("ReportedFatalError"));
+      assertTrue(
+          event.getDiagnosticInfo().contains(ServicePluginErrorDefaults.INCONSISTENT_STATE.name()));
+      assertTrue(event.getDiagnosticInfo().contains(expIdentifier));
+
+    } finally {
+      taskSchedulerManager.stop();
+    }
+  }
+
   @SuppressWarnings("unchecked")
   @Test(timeout = 5000)
   public void testTaskSchedulerUserError() {
@@ -561,8 +786,7 @@ public class TestTaskSchedulerManager {
     doReturn(address).when(mockClientService).getBindAddress();
     TaskSchedulerManager taskSchedulerManager =
         new TaskSchedulerManager(taskScheduler, appContext, mock(ContainerSignatureMatcher.class),
-            mockClientService,
-            Executors.newFixedThreadPool(1)) {
+            mockClientService, Executors.newFixedThreadPool(1)) {
           @Override
           protected void instantiateSchedulers(String host, int port, String trackingUrl,
                                                AppContext appContext) throws TezException {
@@ -649,7 +873,8 @@ public class TestTaskSchedulerManager {
                                          List<NamedEntityDescriptor> schedulerDescriptors,
                                          boolean isPureLocalMode) {
       super(appContext, clientService, eventHandler, containerSignatureMatcher, webUI,
-          schedulerDescriptors, isPureLocalMode);
+          schedulerDescriptors, isPureLocalMode,
+          new HadoopShimsLoader(appContext.getAMConf()).getHadoopShim());
       yarnTaskScheduler = mock(TaskScheduler.class);
       uberTaskScheduler = mock(TaskScheduler.class);
     }
@@ -803,4 +1028,83 @@ public class TestTaskSchedulerManager {
       return false;
     }
   }
+
+  private static final String DAG_NAME = "dagName";
+  private static final int DAG_INDEX = 1;
+  public static class TaskSchedulerForFailureTest extends TaskScheduler {
+
+    public TaskSchedulerForFailureTest(TaskSchedulerContext taskSchedulerContext) {
+      super(taskSchedulerContext);
+    }
+
+    @Override
+    public Resource getAvailableResources() throws ServicePluginException {
+      getContext().reportError(ServicePluginErrorDefaults.INCONSISTENT_STATE, "ReportedFatalError", null);
+      return Resource.newInstance(1024, 1);
+    }
+
+    @Override
+    public Resource getTotalResources() throws ServicePluginException {
+      getContext()
+          .reportError(ServicePluginErrorDefaults.SERVICE_UNAVAILABLE, "ReportError", new DagInfoImplForTest(DAG_INDEX, DAG_NAME));
+      return Resource.newInstance(1024, 1);
+    }
+
+    @Override
+    public int getClusterNodeCount() throws ServicePluginException {
+      return 0;
+    }
+
+    @Override
+    public void blacklistNode(NodeId nodeId) throws ServicePluginException {
+
+    }
+
+    @Override
+    public void unblacklistNode(NodeId nodeId) throws ServicePluginException {
+
+    }
+
+    @Override
+    public void allocateTask(Object task, Resource capability, String[] hosts, String[] racks,
+                             Priority priority, Object containerSignature,
+                             Object clientCookie) throws
+        ServicePluginException {
+
+    }
+
+    @Override
+    public void allocateTask(Object task, Resource capability, ContainerId containerId,
+                             Priority priority, Object containerSignature,
+                             Object clientCookie) throws
+        ServicePluginException {
+
+    }
+
+    @Override
+    public boolean deallocateTask(Object task, boolean taskSucceeded,
+                                  TaskAttemptEndReason endReason,
+                                  @Nullable String diagnostics) throws ServicePluginException {
+      return false;
+    }
+
+    @Override
+    public Object deallocateContainer(ContainerId containerId) throws ServicePluginException {
+      return null;
+    }
+
+    @Override
+    public void setShouldUnregister() throws ServicePluginException {
+
+    }
+
+    @Override
+    public boolean hasUnregistered() throws ServicePluginException {
+      return false;
+    }
+
+    @Override
+    public void dagComplete() throws ServicePluginException {
+    }
+  }
 }
