@@ -92,11 +92,49 @@ public class TestAMWebController {
     mockRequest = mock(HttpServletRequest.class);
   }
 
+  @Test
+  public void testEncodeHeaders() {
+    String validOrigin = "http://localhost:12345";
+    String encodedValidOrigin = AMWebController.encodeHeader(validOrigin);
+    Assert.assertEquals("Valid origin encoding should match exactly",
+        validOrigin, encodedValidOrigin);
+
+    String httpResponseSplitOrigin = validOrigin + " \nSecondHeader: value";
+    String encodedResponseSplitOrigin =
+        AMWebController.encodeHeader(httpResponseSplitOrigin);
+    Assert.assertEquals("Http response split origin should be protected against",
+        validOrigin, encodedResponseSplitOrigin);
+
+    // Test Origin List
+    String validOriginList = "http://foo.example.com:12345 http://bar.example.com:12345";
+    String encodedValidOriginList = AMWebController
+        .encodeHeader(validOriginList);
+    Assert.assertEquals("Valid origin list encoding should match exactly",
+        validOriginList, encodedValidOriginList);
+  }
+
+  @Test(timeout = 5000)
+  public void testCorsHeadersWithOrigin() {
+    AMWebController amWebController = new AMWebController(mockRequestContext, mockAppContext,
+        "TEST_HISTORY_URL");
+    AMWebController spy = spy(amWebController);
+    String originURL = "http://origin.com:8080";
+
+    doReturn(mockResponse).when(spy).response();
+
+    doReturn(mockRequest).when(spy).request();
+    doReturn(originURL).when(mockRequest).getHeader(AMWebController.ORIGIN);
+
+    spy.setCorsHeaders();
+    verify(mockResponse).setHeader("Access-Control-Allow-Origin", originURL);
+  }
+
   @Test(timeout = 5000)
   public void testCorsHeadersAreSet() {
     AMWebController amWebController = new AMWebController(mockRequestContext, mockAppContext,
         "TEST_HISTORY_URL");
     AMWebController spy = spy(amWebController);
+    doReturn(mockRequest).when(spy).request();
     doReturn(mockResponse).when(spy).response();
     spy.setCorsHeaders();
 
@@ -381,6 +419,11 @@ public class TestAMWebController {
     doReturn(status).when(mockVertex).getState();
     doReturn(progress).when(mockVertex).getProgress();
     doReturn(pb).when(mockVertex).getVertexProgress();
+    doReturn(1L).when(mockVertex).getInitTime();
+    doReturn(1L).when(mockVertex).getStartTime();
+    doReturn(2L).when(mockVertex).getFinishTime();
+    doReturn(1L).when(mockVertex).getFirstTaskStartTime();
+    doReturn(2L).when(mockVertex).getLastTaskFinishTime();
 
     TezCounters counters = new TezCounters();
     counters.addGroup("g1", "g1");
@@ -417,6 +460,18 @@ public class TestAMWebController {
         vertex2Result.get("killedTaskAttempts"));
     Assert.assertEquals(Integer.toString(progress.getFailedTaskAttemptCount()),
         vertex2Result.get("failedTaskAttempts"));
+    String str0 = Long.toString(mockVertex2.getInitTime());
+    String str1 = vertex2Result.get("initTime");
+    Assert.assertEquals(Long.toString(mockVertex2.getInitTime()),
+        vertex2Result.get("initTime"));
+    Assert.assertEquals(Long.toString(mockVertex2.getStartTime()),
+        vertex2Result.get("startTime"));
+    Assert.assertEquals(Long.toString(mockVertex2.getFinishTime()),
+        vertex2Result.get("finishTime"));
+    Assert.assertEquals(Long.toString(mockVertex2.getFirstTaskStartTime()),
+        vertex2Result.get("firstTaskStartTime"));
+    Assert.assertEquals(Long.toString(mockVertex2.getLastTaskFinishTime()),
+        vertex2Result.get("lastTaskFinishTime"));
   }
 
   //-- Get Tasks Info Tests -----------------------------------------------------------------------
