@@ -26,11 +26,11 @@ import java.util.concurrent.Callable;
 import java.util.concurrent.ExecutorService;
 import java.util.concurrent.Executors;
 import java.util.concurrent.RejectedExecutionException;
+import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicLong;
 import java.util.concurrent.atomic.AtomicReference;
 
 import com.google.common.base.Preconditions;
-import com.google.common.base.Stopwatch;
 import com.google.common.collect.HashMultimap;
 import com.google.common.collect.Multimap;
 import com.google.common.util.concurrent.FutureCallback;
@@ -53,6 +53,7 @@ import org.apache.hadoop.service.AbstractService;
 import org.apache.hadoop.yarn.api.ApplicationConstants;
 import org.apache.hadoop.yarn.util.AuxiliaryServiceHelper;
 import org.apache.tez.common.TezCommonUtils;
+import org.apache.tez.common.TezExecutors;
 import org.apache.tez.common.TezTaskUmbilicalProtocol;
 import org.apache.tez.common.security.JobTokenIdentifier;
 import org.apache.tez.common.security.TokenCache;
@@ -74,6 +75,7 @@ import org.apache.tez.shufflehandler.ShuffleHandler;
 import org.apache.tez.test.service.rpc.TezTestServiceProtocolProtos.RunContainerRequestProto;
 import org.apache.tez.test.service.rpc.TezTestServiceProtocolProtos.SubmitWorkRequestProto;
 import org.apache.tez.util.ProtoConverters;
+import org.apache.tez.util.StopWatch;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
@@ -89,6 +91,8 @@ public class ContainerRunnerImpl extends AbstractService implements ContainerRun
   private final Map<String, String> localEnv = new HashMap<String, String>();
   private volatile FileSystem localFs;
   private final long memoryPerExecutor;
+
+  private final TezExecutors sharedExecutor;
   // TODO Support for removing queued containers, interrupting / killing specific containers - when preemption is supported
 
 
@@ -96,7 +100,8 @@ public class ContainerRunnerImpl extends AbstractService implements ContainerRun
 
   public ContainerRunnerImpl(int numExecutors, String[] localDirsBase,
                              AtomicReference<InetSocketAddress> localAddress,
-                             long totalMemoryAvailableBytes) {
+                             long totalMemoryAvailableBytes,
+                             TezExecutors sharedExecutor) {
     super("ContainerRunnerImpl");
     Preconditions.checkState(numExecutors > 0,
         "Invalid number of executors: " + numExecutors + ". Must be > 0");
@@ -116,6 +121,7 @@ public class ContainerRunnerImpl extends AbstractService implements ContainerRun
         "memoryPerExecutorDerived=" + memoryPerExecutor +
         ", numExecutors=" + numExecutors
     );
+    this.sharedExecutor = sharedExecutor;
   }
 
   @Override
@@ -131,9 +137,9 @@ public class ContainerRunnerImpl extends AbstractService implements ContainerRun
   public void serviceStart() {
   }
 
-  public void setShufflePort(int shufflePort) {
+  public void setShufflePort(String auxiliaryService, int shufflePort) {
     AuxiliaryServiceHelper.setServiceDataIntoEnv(
-        TezConstants.TEZ_SHUFFLE_HANDLER_SERVICE_ID,
+        auxiliaryService,
         ByteBuffer.allocate(4).putInt(shufflePort), localEnv);
   }
 
@@ -261,7 +267,7 @@ public class ContainerRunnerImpl extends AbstractService implements ContainerRun
     ShuffleHandler.get().registerApplication(request.getApplicationIdString(), jobToken, request.getUser());
     TaskRunnerCallable callable = new TaskRunnerCallable(request, new Configuration(getConfig()),
         new ExecutionContextImpl(localAddress.get().getHostName()), env, localDirs,
-        workingDir, credentials, memoryPerExecutor);
+        workingDir, credentials, memoryPerExecutor, sharedExecutor);
     ListenableFuture<ContainerExecutionResult> future = executorService.submit(callable);
     Futures.addCallback(future, new TaskRunnerCallback(request, callable));
   }
@@ -300,7 +306,7 @@ public class ContainerRunnerImpl extends AbstractService implements ContainerRun
 
     @Override
     public ContainerExecutionResult call() throws Exception {
-      Stopwatch sw = new Stopwatch().start();
+      StopWatch sw = new StopWatch().start();
       tezChild =
           new TezChild(conf, request.getAmHost(), request.getAmPort(),
               request.getContainerIdString(),
@@ -310,7 +316,7 @@ public class ContainerRunnerImpl extends AbstractService implements ContainerRun
               new DefaultHadoopShim());
       ContainerExecutionResult result = tezChild.run();
       LOG.info("ExecutionTime for Container: " + request.getContainerIdString() + "=" +
-          sw.stop().elapsedMillis());
+          sw.stop().now(TimeUnit.MILLISECONDS));
       return result;
     }
 
@@ -384,12 +390,13 @@ public class ContainerRunnerImpl extends AbstractService implements ContainerRun
     private volatile TezTaskRunner2 taskRunner;
     private volatile TaskReporter taskReporter;
     private TezTaskUmbilicalProtocol umbilical;
+    private final TezExecutors sharedExecutor;
 
 
     TaskRunnerCallable(SubmitWorkRequestProto request, Configuration conf,
                              ExecutionContext executionContext, Map<String, String> envMap,
                              String[] localDirs, String workingDir, Credentials credentials,
-                             long memoryAvailable) {
+                             long memoryAvailable, TezExecutors sharedExecutor) {
       this.request = request;
       this.conf = conf;
       this.executionContext = executionContext;
@@ -404,19 +411,22 @@ public class ContainerRunnerImpl extends AbstractService implements ContainerRun
           .setDaemon(true)
           .setNameFormat("TezTaskRunner_" + request.getTaskSpec().getTaskAttemptIdString()).build());
       executor = MoreExecutors.listeningDecorator(executorReal);
+      this.sharedExecutor = sharedExecutor;
     }
 
     @Override
     public ContainerExecutionResult call() throws Exception {
 
       // TODO Consolidate this code with TezChild.
-      Stopwatch sw = new Stopwatch().start();
+      StopWatch sw = new StopWatch().start();
       UserGroupInformation taskUgi = UserGroupInformation.createRemoteUser(request.getUser());
       taskUgi.addCredentials(credentials);
 
       Token<JobTokenIdentifier> jobToken = TokenCache.getSessionToken(credentials);
       Map<String, ByteBuffer> serviceConsumerMetadata = new HashMap<String, ByteBuffer>();
-      serviceConsumerMetadata.put(TezConstants.TEZ_SHUFFLE_HANDLER_SERVICE_ID,
+      String auxiliaryService = conf.get(TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID,
+          TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID_DEFAULT);
+      serviceConsumerMetadata.put(auxiliaryService,
           TezCommonUtils.convertJobTokenToBytes(jobToken));
       Multimap<String, String> startedInputsMap = HashMultimap.create();
 
@@ -451,7 +461,7 @@ public class ContainerRunnerImpl extends AbstractService implements ContainerRun
           request.getAppAttemptNumber(),
           serviceConsumerMetadata, envMap, startedInputsMap, taskReporter, executor, objectRegistry,
           pid,
-          executionContext, memoryAvailable, false, new DefaultHadoopShim());
+          executionContext, memoryAvailable, false, new DefaultHadoopShim(), sharedExecutor);
 
       boolean shouldDie;
       try {
@@ -473,7 +483,7 @@ public class ContainerRunnerImpl extends AbstractService implements ContainerRun
         FileSystem.closeAllForUGI(taskUgi);
       }
       LOG.info("ExecutionTime for Container: " + request.getContainerIdString() + "=" +
-          sw.stop().elapsedMillis());
+          sw.stop().now(TimeUnit.MILLISECONDS));
       return new ContainerExecutionResult(ContainerExecutionResult.ExitStatus.SUCCESS, null,
           null);
     }
@@ -555,4 +565,4 @@ public class ContainerRunnerImpl extends AbstractService implements ContainerRun
           taskSpec.getTaskAttemptID());
     }
   }
-}
\ No newline at end of file
+}
