@@ -50,6 +50,7 @@ import org.apache.hadoop.mapreduce.ClusterMetrics;
 import org.apache.hadoop.mapreduce.Counters;
 import org.apache.hadoop.mapreduce.JobContext;
 import org.apache.hadoop.mapreduce.JobID;
+import org.apache.hadoop.mapreduce.JobPriority;
 import org.apache.hadoop.mapreduce.JobStatus;
 import org.apache.hadoop.mapreduce.QueueAclsInfo;
 import org.apache.hadoop.mapreduce.QueueInfo;
@@ -514,8 +515,9 @@ public class YARNRunner implements ClientProtocol {
                     TezRuntimeConfiguration.TEZ_RUNTIME_KEY_CLASS),
                 stageConfs[i - 1].get(TezRuntimeConfiguration.TEZ_RUNTIME_VALUE_CLASS),
                 MRPartitioner.class.getName(), partitionerConf)
+                .setFromConfigurationUnfiltered(stageConfs[i-1])
                 .configureInput().useLegacyInput().done()
-                .setFromConfiguration(stageConfs[i - 1]).build();
+                .build();
         Edge edge = Edge.create(vertices[i - 1], vertices[i], edgeConf.createDefaultEdgeProperty());
         dag.addEdge(edge);
       }
@@ -667,6 +669,11 @@ public class YARNRunner implements ClientProtocol {
     resMgrDelegate.setJobPriority(arg0, arg1);
   }
 
+  public JobPriority getJobPriority(JobID jobid) throws IOException,
+      InterruptedException {
+    throw new UnsupportedOperationException();
+  }
+
   @Override
   public long getProtocolVersion(String arg0, long arg1) throws IOException {
     return resMgrDelegate.getProtocolVersion(arg0, arg1);
@@ -811,9 +818,10 @@ public class YARNRunner implements ClientProtocol {
   private static class MRInputHelpersInternal extends MRInputHelpers {
 
     protected static UserPayload createMRInputPayload(Configuration conf,
-                                                 MRRuntimeProtos.MRSplitsProto mrSplitsProto) throws
-        IOException {
-      return MRInputHelpers.createMRInputPayload(conf, mrSplitsProto);
+        MRRuntimeProtos.MRSplitsProto mrSplitsProto) throws
+            IOException {
+      return MRInputHelpers.createMRInputPayload(conf, mrSplitsProto, false,
+          true);
     }
   }
 
