@@ -18,7 +18,6 @@
 
 package org.apache.tez.mapreduce.hadoop;
 
-import java.io.DataOutputStream;
 import java.io.IOException;
 import java.util.ArrayList;
 import java.util.Arrays;
@@ -35,6 +34,7 @@ import com.google.common.collect.Iterables;
 import com.google.common.collect.Lists;
 import com.google.protobuf.ByteString;
 
+import org.apache.tez.runtime.api.InputContext;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 import org.apache.hadoop.classification.InterfaceAudience;
@@ -61,6 +61,7 @@ import org.apache.hadoop.yarn.api.records.LocalResourceType;
 import org.apache.hadoop.yarn.api.records.LocalResourceVisibility;
 import org.apache.hadoop.yarn.util.ConverterUtils;
 import org.apache.tez.common.TezUtils;
+import org.apache.tez.common.io.NonSyncDataOutputStream;
 import org.apache.tez.dag.api.DataSourceDescriptor;
 import org.apache.tez.dag.api.InputDescriptor;
 import org.apache.tez.dag.api.TaskLocationHint;
@@ -101,6 +102,7 @@ public class MRInputHelpers {
    * as a data source to a {@link org.apache.tez.dag.api.Vertex}
    */
   @InterfaceStability.Unstable
+  @InterfaceAudience.LimitedPrivate({"hive, pig"})
   public static DataSourceDescriptor configureMRInputWithLegacySplitGeneration(Configuration conf,
                                                                                Path splitsDir,
                                                                                boolean useLegacyInput) {
@@ -110,7 +112,7 @@ public class MRInputHelpers {
 
       InputDescriptor inputDescriptor = InputDescriptor.create(useLegacyInput ? MRInputLegacy.class
           .getName() : MRInput.class.getName())
-          .setUserPayload(createMRInputPayload(conf, null));
+          .setUserPayload(createMRInputPayload(conf, null, false, true));
       Map<String, LocalResource> additionalLocalResources = new HashMap<String, LocalResource>();
       updateLocalResourcesForInputSplits(conf, inputSplitInfo,
           additionalLocalResources);
@@ -139,6 +141,7 @@ public class MRInputHelpers {
    * @throws IOException
    */
   @InterfaceStability.Evolving
+  @InterfaceAudience.LimitedPrivate({"hive, pig"})
   public static MRRuntimeProtos.MRInputUserPayloadProto parseMRInputPayload(UserPayload payload)
       throws IOException {
     return MRRuntimeProtos.MRInputUserPayloadProto.parseFrom(ByteString.copyFrom(payload.getPayload()));
@@ -156,6 +159,7 @@ public class MRInputHelpers {
    */
   @SuppressWarnings("unchecked")
   @InterfaceStability.Evolving
+  @InterfaceAudience.LimitedPrivate({"hive, pig"})
   public static InputSplit createOldFormatSplitFromUserPayload(
       MRRuntimeProtos.MRSplitProto splitProto, SerializationFactory serializationFactory)
       throws IOException {
@@ -238,6 +242,7 @@ public class MRInputHelpers {
   }
 
   @InterfaceStability.Evolving
+  @InterfaceAudience.LimitedPrivate({"hive, pig"})
   public static MRRuntimeProtos.MRSplitProto createSplitProto(
       org.apache.hadoop.mapred.InputSplit oldSplit) throws IOException {
     MRRuntimeProtos.MRSplitProto.Builder builder = MRRuntimeProtos.MRSplitProto.newBuilder();
@@ -246,7 +251,7 @@ public class MRInputHelpers {
 
     ByteString.Output os = ByteString
         .newOutput(SPLIT_SERIALIZED_LENGTH_ESTIMATE);
-    oldSplit.write(new DataOutputStream(os));
+    oldSplit.write(new NonSyncDataOutputStream(os));
     ByteString splitBs = os.toByteString();
     builder.setSplitBytes(splitBs);
 
@@ -277,8 +282,40 @@ public class MRInputHelpers {
    * @throws InterruptedException
    */
   @InterfaceStability.Unstable
-  public static InputSplitInfoMem generateInputSplitsToMem(Configuration conf, boolean groupSplits,
-                                                           int targetTasks)
+  @InterfaceAudience.LimitedPrivate({"hive, pig"})
+  public static InputSplitInfoMem generateInputSplitsToMem(Configuration conf,
+                                                           boolean groupSplits, int targetTasks)
+      throws IOException, ClassNotFoundException, InterruptedException {
+    return generateInputSplitsToMem(conf, groupSplits, true, targetTasks);
+  }
+
+  /**
+   * Generates Input splits and stores them in a {@link org.apache.tez.mapreduce.protos.MRRuntimeProtos.MRSplitsProto} instance.
+   *
+   * Returns an instance of {@link InputSplitInfoMem}
+   *
+   * With grouping enabled, the eventual configuration used by the tasks, will have
+   * the user-specified InputFormat replaced by either {@link org.apache.hadoop.mapred.split.TezGroupedSplitsInputFormat}
+   * or {@link org.apache.hadoop.mapreduce.split.TezGroupedSplitsInputFormat}
+   *
+   * @param conf
+   *          an instance of Configuration which is used to determine whether
+   *          the mapred of mapreduce API is being used. This Configuration
+   *          instance should also contain adequate information to be able to
+   *          generate splits - like the InputFormat being used and related
+   *          configuration.
+   * @param groupSplits whether to group the splits or not
+   * @param sortSplits whether to sort the splits or not
+   * @param targetTasks the number of target tasks if grouping is enabled. Specify as 0 otherwise.
+   * @return an instance of {@link InputSplitInfoMem} which supports a subset of
+   *         the APIs defined on {@link InputSplitInfo}
+   * @throws IOException
+   * @throws ClassNotFoundException
+   * @throws InterruptedException
+   */
+  @InterfaceStability.Unstable
+  public static InputSplitInfoMem generateInputSplitsToMem(Configuration conf,
+      boolean groupSplits, boolean sortSplits, int targetTasks)
       throws IOException, ClassNotFoundException, InterruptedException {
 
     InputSplitInfoMem splitInfoMem = null;
@@ -289,7 +326,7 @@ public class MRInputHelpers {
       }
       Job job = Job.getInstance(conf);
       org.apache.hadoop.mapreduce.InputSplit[] splits =
-          generateNewSplits(job, groupSplits, targetTasks);
+          generateNewSplits(job, groupSplits, sortSplits, targetTasks);
       splitInfoMem = new InputSplitInfoMem(splits, createTaskLocationHintsFromSplits(splits),
           splits.length, job.getCredentials(), job.getConfiguration());
     } else {
@@ -297,14 +334,12 @@ public class MRInputHelpers {
         LOG.debug("Generating mapred api input splits");
       }
       org.apache.hadoop.mapred.InputSplit[] splits =
-          generateOldSplits(jobConf, groupSplits, targetTasks);
+          generateOldSplits(jobConf, groupSplits, sortSplits, targetTasks);
       splitInfoMem = new InputSplitInfoMem(splits, createTaskLocationHintsFromSplits(splits),
           splits.length, jobConf.getCredentials(), jobConf);
     }
-    if (LOG.isDebugEnabled()) {
-      LOG.debug("NumSplits: " + splitInfoMem.getNumTasks() + ", SerializedSize: "
-          + splitInfoMem.getSplitsProto().getSerializedSize());
-    }
+    LOG.info("NumSplits: " + splitInfoMem.getNumTasks() + ", SerializedSize: "
+        + splitInfoMem.getSplitsProto().getSerializedSize());
     return splitInfoMem;
   }
 
@@ -380,8 +415,8 @@ public class MRInputHelpers {
 
   @SuppressWarnings({ "rawtypes", "unchecked" })
   private static org.apache.hadoop.mapreduce.InputSplit[] generateNewSplits(
-      JobContext jobContext, boolean groupSplits, int numTasks)
-      throws ClassNotFoundException, IOException,
+      JobContext jobContext, boolean groupSplits, boolean sortSplits,
+      int numTasks) throws ClassNotFoundException, IOException,
       InterruptedException {
     Configuration conf = jobContext.getConfiguration();
 
@@ -414,15 +449,20 @@ public class MRInputHelpers {
     org.apache.hadoop.mapreduce.InputSplit[] splits = (org.apache.hadoop.mapreduce.InputSplit[]) array
         .toArray(new org.apache.hadoop.mapreduce.InputSplit[array.size()]);
 
-    // sort the splits into order based on size, so that the biggest
-    // go first
-    Arrays.sort(splits, new InputSplitComparator());
+    if (sortSplits) {
+      // sort the splits into order based on size, so that the biggest
+      // go first
+      Arrays.sort(splits, new InputSplitComparator());
+    } else {
+      Collections.shuffle(Arrays.asList(splits));
+    }
     return splits;
   }
 
   @SuppressWarnings({ "rawtypes", "unchecked" })
   private static org.apache.hadoop.mapred.InputSplit[] generateOldSplits(
-      JobConf jobConf, boolean groupSplits, int numTasks) throws IOException {
+      JobConf jobConf, boolean groupSplits, boolean sortSplits, int numTasks)
+      throws IOException {
 
     // This is the real InputFormat
     org.apache.hadoop.mapred.InputFormat inputFormat;
@@ -446,9 +486,11 @@ public class MRInputHelpers {
     }
     org.apache.hadoop.mapred.InputSplit[] splits = finalInputFormat
         .getSplits(jobConf, jobConf.getNumMapTasks());
-    // sort the splits into order based on size, so that the biggest
-    // go first
-    Arrays.sort(splits, new OldInputSplitComparator());
+    if (sortSplits) {
+      // sort the splits into order based on size, so that the biggest
+      // go first
+      Arrays.sort(splits, new OldInputSplitComparator());
+    }
     return splits;
   }
 
@@ -520,7 +562,7 @@ public class MRInputHelpers {
       ClassNotFoundException {
 
     org.apache.hadoop.mapreduce.InputSplit[] splits =
-        generateNewSplits(jobContext, false, 0);
+        generateNewSplits(jobContext, false, true, 0);
 
     Configuration conf = jobContext.getConfiguration();
 
@@ -557,7 +599,7 @@ public class MRInputHelpers {
                                                    Path inputSplitDir) throws IOException {
 
     org.apache.hadoop.mapred.InputSplit[] splits =
-        generateOldSplits(jobConf, false, 0);
+        generateOldSplits(jobConf, false, true, 0);
 
     JobSplitWriter.createSplitFiles(inputSplitDir, jobConf,
         inputSplitDir.getFileSystem(jobConf), splits);
@@ -679,23 +721,41 @@ public class MRInputHelpers {
     Preconditions
         .checkArgument(conf != null, "Configuration must be specified");
     return createMRInputPayload(TezUtils.createByteStringFromConf(conf),
-        null, true);
+        null, true, true);
   }
 
   @InterfaceAudience.Private
   protected static UserPayload createMRInputPayload(Configuration conf,
-                                               MRRuntimeProtos.MRSplitsProto mrSplitsProto) throws
+                                                    MRRuntimeProtos.MRSplitsProto mrSplitsProto) throws
       IOException {
+    return createMRInputPayload(conf, mrSplitsProto, false, true);
+  }
+
+  /**
+   * When isGrouped is true, it specifies that grouping of input splits be
+   * performed by Tez The conf should have the input format class configuration
+   * set to the TezGroupedSplitsInputFormat. The real input format class name
+   * should be passed as an argument to this method.
+   * <p/>
+   * With grouping enabled, the eventual configuration used by the tasks, will have
+   * the user-specified InputFormat replaced by either {@link org.apache.hadoop.mapred.split.TezGroupedSplitsInputFormat}
+   * or {@link org.apache.hadoop.mapreduce.split.TezGroupedSplitsInputFormat}
+   */
+  @InterfaceAudience.Private
+  protected static UserPayload createMRInputPayload(Configuration conf,
+      MRRuntimeProtos.MRSplitsProto mrSplitsProto, boolean isGrouped,
+      boolean isSorted) throws
+          IOException {
     Preconditions
         .checkArgument(conf != null, "Configuration must be specified");
 
     return createMRInputPayload(TezUtils.createByteStringFromConf(conf),
-        mrSplitsProto, false);
+        mrSplitsProto, isGrouped, isSorted);
   }
 
   private static UserPayload createMRInputPayload(ByteString bytes,
-                                             MRRuntimeProtos.MRSplitsProto mrSplitsProto,
-                                             boolean isGrouped) throws IOException {
+    MRRuntimeProtos.MRSplitsProto mrSplitsProto,
+    boolean isGrouped, boolean isSorted) throws IOException {
     MRRuntimeProtos.MRInputUserPayloadProto.Builder userPayloadBuilder =
         MRRuntimeProtos.MRInputUserPayloadProto
             .newBuilder();
@@ -704,7 +764,170 @@ public class MRInputHelpers {
       userPayloadBuilder.setSplits(mrSplitsProto);
     }
     userPayloadBuilder.setGroupingEnabled(isGrouped);
-    return UserPayload.create(userPayloadBuilder.build().toByteString().asReadOnlyByteBuffer());
+    userPayloadBuilder.setSortSplitsEnabled(isSorted);
+    return UserPayload.create(userPayloadBuilder.build().
+        toByteString().asReadOnlyByteBuffer());
+  }
+
+
+  private static String getStringProperty(Configuration conf, String propertyName) {
+    Preconditions.checkNotNull(conf, "Configuration must be provided");
+    Preconditions.checkNotNull(propertyName, "Property name must be provided");
+    return conf.get(propertyName);
+  }
+
+  private static int getIntProperty(Configuration conf, String propertyName) {
+    return Integer.parseInt(getStringProperty(conf, propertyName));
+  }
+
+  /**
+   * @see {@link InputContext#getDagIdentifier}
+   * @param conf configuration instance
+   * @return dag index
+   */
+  @Public
+  public static int getDagIndex(Configuration conf) {
+    return getIntProperty(conf, MRInput.TEZ_MAPREDUCE_DAG_INDEX);
+  }
+
+  /**
+   * Returns string representation of full DAG identifier
+   * @param conf configuration instance
+   * @return dag identifier
+   */
+  @Public
+  public static String getDagIdString(Configuration conf) {
+    return getStringProperty(conf, MRInput.TEZ_MAPREDUCE_DAG_ID);
+  }
+
+  /**
+   * * @see {@link InputContext#getTaskVertexIndex}
+   * @param conf configuration instance
+   * @return vertex index
+   */
+  @Public
+  public static int getVertexIndex(Configuration conf) {
+    return getIntProperty(conf, MRInput.TEZ_MAPREDUCE_VERTEX_INDEX);
+  }
+
+  /**
+   * Returns string representation of full vertex identifier
+   * @param conf configuration instance
+   * @return vertex identifier
+   */
+  @Public
+  public static String getVertexIdString(Configuration conf) {
+    return getStringProperty(conf, MRInput.TEZ_MAPREDUCE_VERTEX_ID);
+  }
+
+  /**
+   * @see {@link InputContext#getTaskIndex}
+   * @param conf configuration instance
+   * @return task index
+   */
+  @Public
+  public static int getTaskIndex(Configuration conf) {
+    return getIntProperty(conf, MRInput.TEZ_MAPREDUCE_TASK_INDEX);
+  }
+
+  /**
+   * Returns string representation of full task identifier
+   * @param conf configuration instance
+   * @return task identifier
+   */
+  @Public
+  public static String getTaskIdString(Configuration conf) {
+    return getStringProperty(conf, MRInput.TEZ_MAPREDUCE_TASK_ID);
+  }
+
+  /**
+   * @see {@link InputContext#getTaskAttemptNumber}
+   * @param conf configuration instance
+   * @return task attempt index
+   */
+  @Public
+  public static int getTaskAttemptIndex(Configuration conf) {
+    return getIntProperty(conf, MRInput.TEZ_MAPREDUCE_TASK_ATTEMPT_INDEX);
+  }
+
+  /**
+   * Returns string representation of full task attempt identifier
+   * @param conf configuration instance
+   * @return task attempt identifier
+   */
+  @Public
+  public static String getTaskAttemptIdString(Configuration conf) {
+    return getStringProperty(conf, MRInput.TEZ_MAPREDUCE_TASK_ATTEMPT_ID);
+  }
+
+  /**
+   * @see {@link InputContext#getInputIndex}
+   * @param conf configuration instance
+   * @return input index
+   */
+  @Public
+  public static int getInputIndex(Configuration conf) {
+    return getIntProperty(conf, MRInput.TEZ_MAPREDUCE_INPUT_INDEX);
+  }
+
+  /**
+   * @see {@link InputContext#getDAGName}
+   * @param conf configuration instance
+   * @return dag name
+   */
+  @Public
+  public static String getDagName(Configuration conf) {
+    return getStringProperty(conf, MRInput.TEZ_MAPREDUCE_DAG_NAME);
+  }
+
+  /**
+   * @see {@link InputContext#getTaskVertexName}
+   * @param conf configuration instance
+   * @return vertex name
+   */
+  @Public
+  public static String getVertexName(Configuration conf) {
+    return getStringProperty(conf, MRInput.TEZ_MAPREDUCE_VERTEX_NAME);
+  }
+
+  /**
+   * @see {@link InputContext#getSourceVertexName}
+   * @param conf configuration instance
+   * @return source name
+   */
+  @Public
+  public static String getInputName(Configuration conf) {
+    return getStringProperty(conf, MRInput.TEZ_MAPREDUCE_INPUT_NAME);
+  }
+
+  /**
+   * @see {@link InputContext#getApplicationId}
+   * @param conf configuration instance
+   * @return applicationId as a string
+   */
+  @Public
+  public static String getApplicationIdString(Configuration conf) {
+    return getStringProperty(conf, MRInput.TEZ_MAPREDUCE_APPLICATION_ID);
+  }
+
+  /**
+   * @see {@link InputContext#getUniqueIdentifier}
+   * @param conf configuration instance
+   * @return unique identifier for the input
+   */
+  @Public
+  public static String getUniqueIdentifier(Configuration conf) {
+    return getStringProperty(conf, MRInput.TEZ_MAPREDUCE_UNIQUE_IDENTIFIER);
+  }
+
+  /**
+   * @see {@link InputContext#getDAGAttemptNumber}
+   * @param conf configuration instance
+   * @return attempt number
+   */
+  @Public
+  public static int getDagAttemptNumber(Configuration conf) {
+    return getIntProperty(conf, MRInput.TEZ_MAPREDUCE_DAG_ATTEMPT_NUMBER);
   }
 
 }
