@@ -28,6 +28,7 @@ import java.util.concurrent.locks.ReentrantLock;
 
 import com.google.protobuf.ByteString;
 
+import org.apache.tez.runtime.api.ProgressFailedException;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 import org.apache.hadoop.classification.InterfaceAudience.Private;
@@ -36,6 +37,7 @@ import org.apache.hadoop.conf.Configuration;
 import org.apache.hadoop.fs.FileSystem;
 import org.apache.hadoop.fs.Path;
 import org.apache.hadoop.mapred.InputFormat;
+import org.apache.hadoop.mapreduce.InputSplit;
 import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
 import org.apache.hadoop.mapreduce.split.JobSplit.TaskSplitIndex;
 import org.apache.hadoop.mapreduce.split.JobSplit.TaskSplitMetaInfo;
@@ -82,7 +84,25 @@ import com.google.common.collect.Lists;
  */
 @Public
 public class MRInput extends MRInputBase {
-  
+
+  @Private public static final String TEZ_MAPREDUCE_DAG_INDEX = "tez.mapreduce.dag.index";
+  @Private public static final String TEZ_MAPREDUCE_DAG_NAME = "tez.mapreduce.dag.name";
+  @Private public static final String TEZ_MAPREDUCE_VERTEX_INDEX = "tez.mapreduce.vertex.index";
+  @Private public static final String TEZ_MAPREDUCE_VERTEX_NAME = "tez.mapreduce.vertex.name";
+  @Private public static final String TEZ_MAPREDUCE_TASK_INDEX = "tez.mapreduce.task.index";
+  @Private public static final String TEZ_MAPREDUCE_TASK_ATTEMPT_INDEX = "tez.mapreduce.task.attempt.index";
+  @Private public static final String TEZ_MAPREDUCE_INPUT_INDEX = "tez.mapreduce.input.index";
+  @Private public static final String TEZ_MAPREDUCE_INPUT_NAME = "tez.mapreduce.input.name";
+  @Private public static final String TEZ_MAPREDUCE_APPLICATION_ID = "tez.mapreduce.application.id";
+  @Private public static final String TEZ_MAPREDUCE_UNIQUE_IDENTIFIER = "tez.mapreduce.unique.identifier";
+  @Private public static final String TEZ_MAPREDUCE_DAG_ATTEMPT_NUMBER = "tez.mapreduce.dag.attempt.number";
+  @Private public static final String TEZ_MAPREDUCE_DAG_ID= "tez.mapreduce.dag.id";
+  @Private public static final String TEZ_MAPREDUCE_VERTEX_ID = "tez.mapreduce.vertex.id";
+  @Private public static final String TEZ_MAPREDUCE_TASK_ID = "tez.mapreduce.task.id";
+  @Private public static final String TEZ_MAPREDUCE_TASK_ATTEMPT_ID = "tez.mapreduce.task.attempt.id";
+
+
+
   /**
    * Helper class to configure {@link MRInput}
    *
@@ -93,6 +113,7 @@ public class MRInput extends MRInputBase {
     final boolean inputFormatProvided;
     boolean useNewApi;
     boolean groupSplitsInAM = true;
+    boolean sortSplitsInAM = true;
     boolean generateSplitsInAM = true;
     String inputClassName = MRInput.class.getName();
     boolean getCredentialsForSourceFilesystem = true;
@@ -171,7 +192,17 @@ public class MRInput extends MRInputBase {
       groupSplitsInAM = value;
       return this;
     }
-    
+
+    /**
+     * Set whether splits should be sorted (default true)
+     * @param value whether to sort splits in the AM or not
+     * @return {@link org.apache.tez.mapreduce.input.MRInput.MRInputConfigBuilder}
+     */
+    public MRInputConfigBuilder sortSplits(boolean value) {
+      sortSplitsInAM = value;
+      return this;
+    }
+
     /**
      * Set whether splits should be generated in the Tez App Master (default true)
      * @param value whether to generate splits in the AM or not
@@ -246,7 +277,7 @@ public class MRInput extends MRInputBase {
       InputSplitInfo inputSplitInfo;
       setupBasicConf(conf);
       try {
-        inputSplitInfo = MRInputHelpers.generateInputSplitsToMem(conf, false, 0);
+        inputSplitInfo = MRInputHelpers.generateInputSplitsToMem(conf, false, true, 0);
       } catch (Exception e) {
         throw new TezUncheckedException(e);
       }
@@ -278,12 +309,8 @@ public class MRInput extends MRInputBase {
 
       Collection<URI> uris = maybeGetURIsForCredentials();
 
-      UserPayload payload = null;
-      if (groupSplitsInAM) {
-        payload = MRInputHelpersInternal.createMRInputPayloadWithGrouping(conf);
-      } else {
-        payload = MRInputHelpersInternal.createMRInputPayload(conf, null);
-      }
+      UserPayload payload = MRInputHelpersInternal.createMRInputPayload(
+          conf, groupSplitsInAM, sortSplitsInAM);
 
       DataSourceDescriptor ds = DataSourceDescriptor
           .create(InputDescriptor.create(inputClassName).setUserPayload(payload),
@@ -306,12 +333,8 @@ public class MRInput extends MRInputBase {
       
       Collection<URI> uris = maybeGetURIsForCredentials();
 
-      UserPayload payload = null;
-      if (groupSplitsInAM) {
-        payload = MRInputHelpersInternal.createMRInputPayloadWithGrouping(conf);
-      } else {
-        payload = MRInputHelpersInternal.createMRInputPayload(conf, null);
-      }
+      UserPayload payload = MRInputHelpersInternal.createMRInputPayload(
+          conf, groupSplitsInAM, sortSplitsInAM);
 
       DataSourceDescriptor ds = DataSourceDescriptor.create(
           InputDescriptor.create(inputClassName).setUserPayload(payload),
@@ -471,10 +494,16 @@ public class MRInput extends MRInputBase {
         TaskSplitMetaInfo thisTaskMetaInfo = allMetaInfo[getContext().getTaskIndex()];
         TaskSplitIndex splitMetaInfo = new TaskSplitIndex(thisTaskMetaInfo.getSplitLocation(),
             thisTaskMetaInfo.getStartOffset());
+        long splitLength = -1;
         if (useNewApi) {
           org.apache.hadoop.mapreduce.InputSplit newInputSplit = MRInputUtils
               .getNewSplitDetailsFromDisk(splitMetaInfo, jobConf, getContext().getCounters()
                   .findCounter(TaskCounter.SPLIT_RAW_BYTES));
+          try {
+            splitLength = newInputSplit.getLength();
+          } catch (InterruptedException e) {
+            LOG.warn("Got interrupted while reading split length: ", e);
+          }
           mrReader = new MRReaderMapReduce(jobConf, newInputSplit, getContext().getCounters(),
               inputRecordCounter, getContext().getApplicationId().getClusterTimestamp(),
               getContext().getTaskVertexIndex(), getContext().getApplicationId().getId(),
@@ -483,10 +512,15 @@ public class MRInput extends MRInputBase {
           org.apache.hadoop.mapred.InputSplit oldInputSplit = MRInputUtils
               .getOldSplitDetailsFromDisk(splitMetaInfo, jobConf, getContext().getCounters()
                   .findCounter(TaskCounter.SPLIT_RAW_BYTES));
+          splitLength = oldInputSplit.getLength();
           mrReader =
               new MRReaderMapred(jobConf, oldInputSplit, getContext().getCounters(),
                   inputRecordCounter, getContext());
         }
+        if (splitLength != -1) {
+          getContext().getCounters().findCounter(TaskCounter.INPUT_SPLIT_LENGTH_BYTES)
+              .increment(splitLength);
+        }
       }
     } finally {
       rrLock.unlock();
@@ -581,8 +615,13 @@ public class MRInput extends MRInputBase {
     }
   }
 
-  public float getProgress() throws IOException, InterruptedException {
-    return mrReader.getProgress();
+  @Override
+  public float getProgress() throws ProgressFailedException, InterruptedException {
+    try {
+      return (mrReader != null) ? mrReader.getProgress() : 0.0f;
+    } catch (IOException e) {
+      throw new ProgressFailedException("getProgress encountered IOException ", e);
+    }
   }
 
   void processSplitEvent(InputDataInformationEvent event)
@@ -632,38 +671,52 @@ public class MRInput extends MRInputBase {
     }
     Preconditions.checkState(initEvent != null, "InitEvent must be specified");
     MRSplitProto splitProto = MRSplitProto.parseFrom(ByteString.copyFrom(initEvent.getUserPayload()));
-    Object split = null;
+    Object splitObj = null;
+    long splitLength = -1;
     if (useNewApi) {
-      split = MRInputUtils.getNewSplitDetailsFromEvent(splitProto, jobConf);
+      InputSplit split = MRInputUtils.getNewSplitDetailsFromEvent(splitProto, jobConf);
+      splitObj = split;
+      try {
+        splitLength = split.getLength();
+      } catch (InterruptedException e) {
+        LOG.warn("Thread interrupted while getting split length: ", e);
+      }
       if (LOG.isDebugEnabled()) {
         LOG.debug(getContext().getSourceVertexName() + " split Details -> SplitClass: " +
-            split.getClass().getName() + ", NewSplit: "
-            + split);
+            split.getClass().getName() + ", NewSplit: " + split + ", length: " + splitLength);
       }
 
     } else {
-      split = MRInputUtils.getOldSplitDetailsFromEvent(splitProto, jobConf);
+      org.apache.hadoop.mapred.InputSplit split =
+          MRInputUtils.getOldSplitDetailsFromEvent(splitProto, jobConf);
+      splitObj = split;
+      splitLength = split.getLength();
       if (LOG.isDebugEnabled()) {
         LOG.debug(getContext().getSourceVertexName() + " split Details -> SplitClass: " +
-            split.getClass().getName() + ", OldSplit: "
-            + split);
+            split.getClass().getName() + ", OldSplit: " + split + ", length: " + splitLength);
       }
     }
-    mrReader.setSplit(split);
+    if (splitLength != -1) {
+      getContext().getCounters().findCounter(TaskCounter.INPUT_SPLIT_LENGTH_BYTES)
+          .increment(splitLength);
+    }
+    mrReader.setSplit(splitObj);
     LOG.info(getContext().getSourceVertexName() + " initialized RecordReader from event");
   }
 
   private static class MRInputHelpersInternal extends MRInputHelpers {
 
-    protected static UserPayload createMRInputPayloadWithGrouping(Configuration conf) throws
-        IOException {
-      return MRInputHelpers.createMRInputPayloadWithGrouping(conf);
+    protected static UserPayload createMRInputPayload(Configuration conf,
+        boolean isGrouped, boolean isSorted) throws IOException {
+      return MRInputHelpers.createMRInputPayload(conf, null, isGrouped,
+          isSorted);
     }
 
     protected static UserPayload createMRInputPayload(Configuration conf,
-                                                 MRRuntimeProtos.MRSplitsProto mrSplitsProto) throws
+        MRRuntimeProtos.MRSplitsProto mrSplitsProto) throws
         IOException {
-      return MRInputHelpers.createMRInputPayload(conf, mrSplitsProto);
+      return MRInputHelpers.createMRInputPayload(conf, mrSplitsProto, false,
+          true);
     }
   }
 
