@@ -484,6 +484,57 @@ public class TestGroupedSplits {
       }
     }
   }
+
+
+  @Test (timeout = 30000)
+  public void testS3Scenario() throws IOException {
+    //There can be multiple nodes in cluster, but locations would be "localhost" in s3
+    String[] locations = {"localhost"};
+    int oriSplits = 52;
+    int desiredSplits = 19;
+    long splitLength = 231958;
+
+    InputSplit[] origSplits = new InputSplit[oriSplits];
+
+    for (int i = 0; i < oriSplits; i++) {
+      String[] splitLoc = locations;
+      origSplits[i] = new TestInputSplit(splitLength, splitLoc, i);
+    }
+
+    TezMapredSplitsGrouper grouper = new TezMapredSplitsGrouper();
+    JobConf conf = new JobConf(defaultConf);
+    conf = (JobConf) TezSplitGrouper.newConfigBuilder(conf).build();
+
+    //Create splits now
+    InputSplit[] groupedSplits =
+        grouper.getGroupedSplits(conf, origSplits, desiredSplits, "SampleFormat");
+
+    //Verify
+    int splitsInGroup = oriSplits / desiredSplits;
+    int totalSplits = (int) Math.ceil(oriSplits * 1.0 / splitsInGroup);
+    assertEquals(totalSplits, groupedSplits.length);
+
+
+    // min split optimization should not be invoked if any location is not localhost
+    String[] nonLocalLocations = { "EmptyLocation", "localhost" };
+
+    origSplits = new InputSplit[oriSplits];
+
+    for (int i = 0; i < oriSplits; i++) {
+      String[] splitLoc = nonLocalLocations;
+      origSplits[i] = new TestInputSplit(splitLength, splitLoc, i);
+    }
+
+    grouper = new TezMapredSplitsGrouper();
+    conf = new JobConf(defaultConf);
+    conf = (JobConf) TezSplitGrouper.newConfigBuilder(conf).build();
+
+    //Create splits now
+    groupedSplits = grouper.getGroupedSplits(conf, origSplits, desiredSplits, "SampleFormat");
+
+    //splits should be 1
+    assertEquals(1, groupedSplits.length);
+  }
   
   @SuppressWarnings({ "rawtypes", "unchecked" })
   @Test(timeout=10000)
