@@ -28,8 +28,11 @@ import com.google.common.collect.Multimap;
 import com.google.common.collect.Multimaps;
 import com.google.common.collect.Ordering;
 
+import org.apache.commons.logging.Log;
+import org.apache.commons.logging.LogFactory;
 import org.apache.hadoop.util.StringInterner;
 import org.apache.tez.dag.api.oldrecords.TaskState;
+import org.apache.tez.dag.history.HistoryEventType;
 import org.codehaus.jettison.json.JSONException;
 import org.codehaus.jettison.json.JSONObject;
 
@@ -46,6 +49,8 @@ import static org.apache.hadoop.classification.InterfaceStability.Evolving;
 @Evolving
 public class VertexInfo extends BaseInfo {
 
+  private static final Log LOG = LogFactory.getLog(VertexInfo.class);
+
   private final String vertexId;
   private final String vertexName;
   private final long finishTime;
@@ -98,9 +103,42 @@ public class VertexInfo extends BaseInfo {
     JSONObject otherInfoNode = jsonObject.getJSONObject(Constants.OTHER_INFO);
     initRequestedTime = otherInfoNode.optLong(Constants.INIT_REQUESTED_TIME);
     startRequestedTime = otherInfoNode.optLong(Constants.START_REQUESTED_TIME);
-    startTime = otherInfoNode.optLong(Constants.START_TIME);
-    initTime = otherInfoNode.optLong(Constants.INIT_TIME);
-    finishTime = otherInfoNode.optLong(Constants.FINISH_TIME);
+
+    long sTime = otherInfoNode.optLong(Constants.START_TIME);
+    long iTime = otherInfoNode.optLong(Constants.INIT_TIME);
+    long eTime = otherInfoNode.optLong(Constants.FINISH_TIME);
+    if (eTime < sTime) {
+      LOG.warn("Vertex has got wrong start/end values. "
+          + "startTime=" + sTime + ", endTime=" + eTime + ". Will check "
+          + "timestamps in DAG started/finished events");
+
+      // Check if events VERTEX_STARTED, VERTEX_FINISHED can be made use of
+      for(Event event : eventList) {
+        switch (HistoryEventType.valueOf(event.getType())) {
+        case VERTEX_INITIALIZED:
+          iTime = event.getAbsoluteTime();
+          break;
+        case VERTEX_STARTED:
+          sTime = event.getAbsoluteTime();
+          break;
+        case VERTEX_FINISHED:
+          eTime = event.getAbsoluteTime();
+          break;
+        default:
+          break;
+        }
+      }
+
+      if (eTime < sTime) {
+        LOG.warn("Vertex has got wrong start/end values in events as well. "
+            + "startTime=" + sTime + ", endTime=" + eTime);
+      }
+    }
+    startTime = sTime;
+    finishTime = eTime;
+    initTime = iTime;
+
+
     diagnostics = otherInfoNode.optString(Constants.DIAGNOSTICS);
     numTasks = otherInfoNode.optInt(Constants.NUM_TASKS);
     failedTasks = otherInfoNode.optInt(Constants.NUM_FAILED_TASKS);
@@ -483,9 +521,7 @@ public class VertexInfo extends BaseInfo {
     }
     Collections.sort(taskInfoList, new Comparator<TaskInfo>() {
       @Override public int compare(TaskInfo o1, TaskInfo o2) {
-        return (o1.getStartTimeInterval() < o2.getStartTimeInterval()) ? -1 :
-            ((o1.getStartTimeInterval() == o2.getStartTimeInterval()) ?
-                0 : 1);
+        return Long.compare(o1.getStartTimeInterval(), o2.getStartTimeInterval());
       }
     });
     return taskInfoList.get(0);
@@ -503,9 +539,7 @@ public class VertexInfo extends BaseInfo {
     }
     Collections.sort(taskInfoList, new Comparator<TaskInfo>() {
       @Override public int compare(TaskInfo o1, TaskInfo o2) {
-        return (o1.getFinishTimeInterval() > o2.getFinishTimeInterval()) ? -1 :
-            ((o1.getStartTimeInterval() == o2.getStartTimeInterval()) ?
-                0 : 1);
+        return -1 * Long.compare(o1.getFinishTimeInterval(), o2.getFinishTimeInterval());
       }
     });
     return taskInfoList.get(0);
@@ -551,8 +585,7 @@ public class VertexInfo extends BaseInfo {
   private Ordering<TaskInfo> orderingOnTimeTaken() {
     return Ordering.from(new Comparator<TaskInfo>() {
       @Override public int compare(TaskInfo o1, TaskInfo o2) {
-        return (o1.getTimeTaken() < o2.getTimeTaken()) ? -1 :
-            ((o1.getTimeTaken() == o2.getTimeTaken()) ? 0 : 1);
+        return Long.compare(o1.getTimeTaken(), o2.getTimeTaken());
       }
     });
   }
@@ -560,8 +593,7 @@ public class VertexInfo extends BaseInfo {
   private Ordering<TaskInfo> orderingOnStartTime() {
     return Ordering.from(new Comparator<TaskInfo>() {
       @Override public int compare(TaskInfo o1, TaskInfo o2) {
-        return (o1.getStartTimeInterval() < o2.getStartTimeInterval()) ? -1 :
-            ((o1.getStartTimeInterval() == o2.getStartTimeInterval()) ? 0 : 1);
+        return Long.compare(o1.getStartTimeInterval(), o2.getStartTimeInterval());
       }
     });
   }
@@ -569,8 +601,7 @@ public class VertexInfo extends BaseInfo {
   private Ordering<TaskAttemptInfo> orderingOnAttemptStartTime() {
     return Ordering.from(new Comparator<TaskAttemptInfo>() {
       @Override public int compare(TaskAttemptInfo o1, TaskAttemptInfo o2) {
-        return (o1.getStartTimeInterval() < o2.getStartTimeInterval()) ? -1 :
-            ((o1.getStartTimeInterval() == o2.getStartTimeInterval()) ? 0 : 1);
+        return Long.compare(o1.getStartTimeInterval(), o2.getStartTimeInterval());
       }
     });
   }
