@@ -117,6 +117,7 @@ public class TestHistoryParser {
   private static String TEZ_BASE_DIR =
       "target" + Path.SEPARATOR + TestHistoryParser.class.getName() + "-tez";
   private static String DOWNLOAD_DIR = TEST_ROOT_DIR + Path.SEPARATOR + "download";
+  private static String yarnTimelineAddress;
 
   @BeforeClass
   public static void setupCluster() throws Exception {
@@ -176,10 +177,12 @@ public class TestHistoryParser {
 
     TezConfiguration tezConf = new TezConfiguration(miniTezCluster.getConfig());
     tezConf.setBoolean(YarnConfiguration.TIMELINE_SERVICE_ENABLED, true);
-    tezConf.set(YarnConfiguration.TIMELINE_SERVICE_WEBAPP_ADDRESS, "0.0.0.0:8188");
+    tezConf.set(YarnConfiguration.TIMELINE_SERVICE_WEBAPP_ADDRESS,
+        miniTezCluster.getConfig().get(YarnConfiguration.TIMELINE_SERVICE_WEBAPP_ADDRESS));
     tezConf.setBoolean(TezConfiguration.TEZ_AM_ALLOW_DISABLED_TIMELINE_DOMAINS, true);
     tezConf.set(TezConfiguration.TEZ_HISTORY_LOGGING_SERVICE_CLASS,
         ATSHistoryLoggingService.class.getName());
+    yarnTimelineAddress = miniTezCluster.getConfig().get(YarnConfiguration.TIMELINE_SERVICE_WEBAPP_ADDRESS);
 
   }
 
@@ -198,7 +201,7 @@ public class TestHistoryParser {
         WordCount.SumProcessor.class.getName(), "WordCount", true);
 
     //Export the data from ATS
-    String[] args = { "--dagId=" + dagId, "--downloadDir=" + DOWNLOAD_DIR };
+    String[] args = { "--dagId=" + dagId, "--downloadDir=" + DOWNLOAD_DIR, "--yarnTimelineAddress=" + yarnTimelineAddress };
 
     int result = ATSImportTool.process(args);
     assertTrue(result == 0);
@@ -207,6 +210,7 @@ public class TestHistoryParser {
     DagInfo dagInfoFromATS = getDagInfo(dagId);
     verifyDagInfo(dagInfoFromATS, true);
     verifyJobSpecificInfo(dagInfoFromATS);
+    checkConfig(dagInfoFromATS);
 
     //Now run with SimpleHistoryLogging
     dagId = runWordCount(WordCount.TokenProcessor.class.getName(),
@@ -241,6 +245,16 @@ public class TestHistoryParser {
     return dagInfo;
   }
 
+  private void checkConfig(DagInfo dagInfo) {
+    assertTrue("DagInfo is " + dagInfo, dagInfo != null);
+    //Check configs
+    assertTrue("DagInfo config size=" + dagInfo.getAppConfig().size(),
+        dagInfo.getAppConfig().size() > 0);
+    //Sample config element
+    assertTrue("DagInfo config=" + dagInfo.getAppConfig(),
+        Integer.parseInt(dagInfo.getAppConfig().get("dfs.replication")) > 0);
+  }
+
   private void verifyJobSpecificInfo(DagInfo dagInfo) {
     //Job specific
     assertTrue(dagInfo.getNumVertices() == 2);
@@ -249,6 +263,7 @@ public class TestHistoryParser {
         WordCount.TokenProcessor.class.getName()));
     assertTrue(dagInfo.getVertex(SUMMATION).getProcessorClassName()
         .equals(WordCount.SumProcessor.class.getName()));
+    assertTrue(dagInfo.getFinishTime() > dagInfo.getStartTime());
     assertTrue(dagInfo.getEdges().size() == 1);
     EdgeInfo edgeInfo = dagInfo.getEdges().iterator().next();
     assertTrue(edgeInfo.getDataMovementType().
@@ -269,6 +284,7 @@ public class TestHistoryParser {
       assertTrue(vertexInfo.getStartRequestedTime() > 0);
       assertTrue(vertexInfo.getStartTime() > 0);
       assertTrue(vertexInfo.getFinishTime() > 0);
+      assertTrue(vertexInfo.getFinishTime() > vertexInfo.getStartTime());
       long finishTime = 0;
       for (TaskInfo taskInfo : vertexInfo.getTasks()) {
         assertTrue(taskInfo.getNumberOfTaskAttempts() == 1);
@@ -280,6 +296,7 @@ public class TestHistoryParser {
         assertTrue(taskInfo.getSuccessfulTaskAttempts().size() > 0);
         assertTrue(taskInfo.getFailedTaskAttempts().size() == 0);
         assertTrue(taskInfo.getKilledTaskAttempts().size() == 0);
+        assertTrue(taskInfo.getFinishTime() > taskInfo.getStartTime());
         List<TaskAttemptInfo> attempts = taskInfo.getTaskAttempts();
         if (vertexInfo.getVertexName().equals(TOKENIZER)) {
           // get the last task to finish and track its successful attempt
@@ -304,6 +321,7 @@ public class TestHistoryParser {
           assertTrue(attemptInfo.getCreationTime() > 0);
           assertTrue(attemptInfo.getAllocationTime() > 0);
           assertTrue(attemptInfo.getStartTime() > 0);
+          assertTrue(attemptInfo.getFinishTime() > attemptInfo.getStartTime());
         }
       }
       assertTrue(vertexInfo.getLastTaskToFinish() != null);
@@ -359,13 +377,15 @@ public class TestHistoryParser {
         .getName(), "WordCount-With-Exception", true);
 
     //Export the data from ATS
-    String[] args = { "--dagId=" + dagId, "--downloadDir=" + DOWNLOAD_DIR };
+    String[] args = { "--dagId=" + dagId, "--downloadDir=" + DOWNLOAD_DIR, "--yarnTimelineAddress=" + yarnTimelineAddress };
 
     int result = ATSImportTool.process(args);
     assertTrue(result == 0);
 
     //Parse ATS data
     DagInfo dagInfo = getDagInfo(dagId);
+    //Applicable for ATS dataset
+    checkConfig(dagInfo);
 
     //Verify DAGInfo. Verifies vertex, task, taskAttempts in recursive manner
     verifyDagInfo(dagInfo, true);
@@ -607,7 +627,6 @@ public class TestHistoryParser {
     TezConfiguration tezConf = new TezConfiguration(miniTezCluster.getConfig());
     if (withTimeline) {
       tezConf.setBoolean(YarnConfiguration.TIMELINE_SERVICE_ENABLED, withTimeline);
-      tezConf.set(YarnConfiguration.TIMELINE_SERVICE_WEBAPP_ADDRESS, "0.0.0.0:8188");
       tezConf.set(TezConfiguration.TEZ_HISTORY_LOGGING_SERVICE_CLASS,
           ATSHistoryLoggingService.class.getName());
     } else {
@@ -693,6 +712,8 @@ public class TestHistoryParser {
       assertTrue(versionInfo.getBuildTime() != null);
     }
 
+    assertTrue(dagInfo.getUserName() != null);
+    assertTrue(!dagInfo.getUserName().isEmpty());
     assertTrue(dagInfo.getStartTime() > 0);
     assertTrue(dagInfo.getFinishTimeInterval() > 0);
     assertTrue(dagInfo.getStartTimeInterval() == 0);
@@ -748,6 +769,7 @@ public class TestHistoryParser {
       assertTrue(vertexInfo.getFirstTaskToStart() != null);
       assertTrue(vertexInfo.getSucceededTasksCount() > 0);
       assertTrue(vertexInfo.getTasks().size() > 0);
+      assertTrue(vertexInfo.getFinishTime() > vertexInfo.getStartTime());
     }
 
     for (TaskInfo taskInfo : vertexInfo.getTasks()) {
@@ -781,6 +803,7 @@ public class TestHistoryParser {
           taskInfo.getStartTimeInterval() > 0 && taskInfo.getStartTime() > taskInfo.getStartTimeInterval());
       assertTrue(taskInfo.getSuccessfulAttemptId() != null);
       assertTrue(taskInfo.getSuccessfulTaskAttempt() != null);
+      assertTrue(taskInfo.getFinishTime() > taskInfo.getStartTime());
     }
     assertTrue(taskInfo.getTaskId() != null);
 
