@@ -46,6 +46,7 @@ import java.util.concurrent.LinkedBlockingQueue;
 
 import org.apache.commons.lang.exception.ExceptionUtils;
 import org.apache.tez.hadoop.shim.HadoopShim;
+import org.apache.tez.runtime.api.TaskFailureType;
 import org.apache.tez.runtime.api.TaskContext;
 import org.apache.tez.runtime.api.impl.TezProcessorContextImpl;
 import org.slf4j.Logger;
@@ -55,6 +56,7 @@ import org.apache.hadoop.conf.Configuration;
 import org.apache.tez.common.CallableWithNdc;
 import org.apache.tez.common.ReflectionUtils;
 import org.apache.tez.common.RunnableWithNdc;
+import org.apache.tez.common.TezExecutors;
 import org.apache.tez.dag.api.InputDescriptor;
 import org.apache.tez.dag.api.OutputDescriptor;
 import org.apache.tez.dag.api.ProcessorDescriptor;
@@ -153,13 +155,19 @@ public class LogicalIOProcessorRuntimeTask extends RuntimeTask {
   private final ExecutionContext ExecutionContext;
   private final long memAvailable;
   private final HadoopShim hadoopShim;
+  private final int maxEventBacklog;
+
+  private final boolean initializeProcessorFirst;
+  private final boolean initializeProcessorIOSerially;
+  private final TezExecutors sharedExecutor;
 
   public LogicalIOProcessorRuntimeTask(TaskSpec taskSpec, int appAttemptNumber,
       Configuration tezConf, String[] localDirs, TezUmbilical tezUmbilical,
       Map<String, ByteBuffer> serviceConsumerMetadata, Map<String, String> envMap,
       Multimap<String, String> startedInputsMap, ObjectRegistry objectRegistry,
       String pid, ExecutionContext ExecutionContext, long memAvailable,
-      boolean updateSysCounters, HadoopShim hadoopShim) throws IOException {
+      boolean updateSysCounters, HadoopShim hadoopShim,
+      TezExecutors sharedExecutor) throws IOException {
     // Note: If adding any fields here, make sure they're cleaned up in the cleanupContext method.
     // TODO Remove jobToken from here post TEZ-421
     super(taskSpec, tezConf, tezUmbilical, pid, updateSysCounters);
@@ -187,8 +195,15 @@ public class LogicalIOProcessorRuntimeTask extends RuntimeTask {
     this.eventsToBeProcessed = new LinkedBlockingQueue<TezEvent>();
     this.state.set(State.NEW);
     this.appAttemptNumber = appAttemptNumber;
+    this.initializeProcessorFirst = tezConf.getBoolean(TezConfiguration.TEZ_TASK_INITIALIZE_PROCESSOR_FIRST,
+        TezConfiguration.TEZ_TASK_INITIALIZE_PROCESSOR_FIRST_DEFAULT);
+    this.initializeProcessorIOSerially = tezConf.getBoolean(TezConfiguration.TEZ_TASK_INITIALIZE_PROCESSOR_IO_SERIALLY,
+        TezConfiguration.TEZ_TASK_INITIALIZE_PROCESSOR_IO_SERIALLY_DEFAULT);
     int numInitializers = numInputs + numOutputs; // Processor is initialized in the main thread.
     numInitializers = (numInitializers == 0 ? 1 : numInitializers);
+    if (initializeProcessorIOSerially) {
+      numInitializers = 1;
+    }
     this.initializerExecutor = Executors.newFixedThreadPool(
         numInitializers,
         new ThreadFactoryBuilder().setDaemon(true)
@@ -203,6 +218,9 @@ public class LogicalIOProcessorRuntimeTask extends RuntimeTask {
     this.ExecutionContext = ExecutionContext;
     this.memAvailable = memAvailable;
     this.hadoopShim = hadoopShim;
+    this.maxEventBacklog = tezConf.getInt(TezConfiguration.TEZ_TASK_MAX_EVENT_BACKLOG,
+        TezConfiguration.TEZ_TASK_MAX_EVENT_BACKLOG_DEFAULT);
+    this.sharedExecutor = sharedExecutor;
   }
 
   /**
@@ -215,6 +233,10 @@ public class LogicalIOProcessorRuntimeTask extends RuntimeTask {
     this.processorContext = createProcessorContext();
     this.processor = createProcessor(processorDescriptor.getClassName(), processorContext);
 
+    if (initializeProcessorFirst || initializeProcessorIOSerially) {
+      // Initialize processor in the current thread.
+      initializeLogicalIOProcessor();
+    }
     int numTasks = 0;
 
     int inputIndex = 0;
@@ -231,9 +253,10 @@ public class LogicalIOProcessorRuntimeTask extends RuntimeTask {
       numTasks++;
     }
 
-    // Initialize processor in the current thread.
-    initializeLogicalIOProcessor();
-
+    if (!(initializeProcessorFirst || initializeProcessorIOSerially)) {
+      // Initialize processor in the current thread.
+      initializeLogicalIOProcessor();
+    }
     int completedTasks = 0;
     while (completedTasks < numTasks) {
       LOG.info("Waiting for " + (numTasks-completedTasks) + " initializers to finish");
@@ -357,31 +380,47 @@ public class LogicalIOProcessorRuntimeTask extends RuntimeTask {
           "Can only run while in RUNNING state. Current: " + this.state);
       this.state.set(State.CLOSED);
 
-      // Close the Processor.
-      processorClosed = true;
-      processor.close();
 
+      List<List<Event>> allCloseInputEvents = Lists.newArrayList();
       // Close the Inputs.
       for (InputSpec inputSpec : inputSpecs) {
         String srcVertexName = inputSpec.getSourceVertexName();
         initializedInputs.remove(srcVertexName);
         List<Event> closeInputEvents = ((InputFrameworkInterface)inputsMap.get(srcVertexName)).close();
-        sendTaskGeneratedEvents(closeInputEvents,
-            EventProducerConsumerType.INPUT, taskSpec.getVertexName(),
-            srcVertexName, taskSpec.getTaskAttemptID());
+        allCloseInputEvents.add(closeInputEvents);
       }
 
+      List<List<Event>> allCloseOutputEvents = Lists.newArrayList();
       // Close the Outputs.
       for (OutputSpec outputSpec : outputSpecs) {
         String destVertexName = outputSpec.getDestinationVertexName();
         initializedOutputs.remove(destVertexName);
         List<Event> closeOutputEvents = ((LogicalOutputFrameworkInterface)outputsMap.get(destVertexName)).close();
-        sendTaskGeneratedEvents(closeOutputEvents,
+        allCloseOutputEvents.add(closeOutputEvents);
+      }
+
+      // Close the Processor.
+      processorClosed = true;
+      processor.close();
+
+      for (int i = 0; i < allCloseInputEvents.size(); i++) {
+        String srcVertexName = inputSpecs.get(i).getSourceVertexName();
+        sendTaskGeneratedEvents(allCloseInputEvents.get(i),
+            EventProducerConsumerType.INPUT, taskSpec.getVertexName(),
+            srcVertexName, taskSpec.getTaskAttemptID());
+      }
+
+      for (int i = 0; i < allCloseOutputEvents.size(); i++) {
+        String destVertexName = outputSpecs.get(i).getDestinationVertexName();
+        sendTaskGeneratedEvents(allCloseOutputEvents.get(i),
             EventProducerConsumerType.OUTPUT, taskSpec.getVertexName(),
             destVertexName, taskSpec.getTaskAttemptID());
       }
+
     } finally {
       setTaskDone();
+      // Clear the interrupt status since the task execution is done.
+      Thread.interrupted();
       if (eventRouterThread != null) {
         eventRouterThread.interrupt();
         LOG.info("Joining on EventRouter");
@@ -574,7 +613,7 @@ public class LogicalIOProcessorRuntimeTask extends RuntimeTask {
         inputSpec.getInputDescriptor().getUserPayload(), this,
         serviceConsumerMetadata, envMap, initialMemoryDistributor,
         inputSpec.getInputDescriptor(), inputMap, inputReadyTracker, objectRegistry,
-        ExecutionContext, memAvailable);
+        ExecutionContext, memAvailable, sharedExecutor);
     return inputContext;
   }
 
@@ -589,7 +628,7 @@ public class LogicalIOProcessorRuntimeTask extends RuntimeTask {
         outputSpec.getOutputDescriptor().getUserPayload(), this,
         serviceConsumerMetadata, envMap, initialMemoryDistributor,
         outputSpec.getOutputDescriptor(), objectRegistry, ExecutionContext,
-        memAvailable);
+        memAvailable, sharedExecutor);
     return outputContext;
   }
 
@@ -600,8 +639,8 @@ public class LogicalIOProcessorRuntimeTask extends RuntimeTask {
         taskSpec.getVertexParallelism(),
         taskSpec.getTaskAttemptID(),
         processorDescriptor.getUserPayload(), this,
-        serviceConsumerMetadata, envMap, initialMemoryDistributor,
-        processorDescriptor, inputReadyTracker, objectRegistry, ExecutionContext, memAvailable);
+        serviceConsumerMetadata, envMap, initialMemoryDistributor, processorDescriptor,
+        inputReadyTracker, objectRegistry, ExecutionContext, memAvailable, sharedExecutor);
     return processorContext;
   }
 
@@ -715,19 +754,27 @@ public class LogicalIOProcessorRuntimeTask extends RuntimeTask {
       }
     } catch (Throwable t) {
       LOG.warn("Failed to handle event", t);
-      setFatalError(t, "Failed to handle event");
+      registerError();
       EventMetaData sourceInfo = new EventMetaData(
           e.getDestinationInfo().getEventGenerator(),
           taskSpec.getVertexName(), e.getDestinationInfo().getEdgeVertexName(),
           getTaskAttemptID());
       setFrameworkCounters();
-      tezUmbilical.signalFatalError(getTaskAttemptID(),
+      // Signal such errors as RETRIABLE. The user code has an option to report this as something
+      // other than retriable before we get control back.
+      // TODO: Don't catch Throwables.
+      tezUmbilical.signalFailure(getTaskAttemptID(), TaskFailureType.NON_FATAL,
           t, ExceptionUtils.getStackTrace(t), sourceInfo);
       return false;
     }
     return true;
   }
 
+  @Override
+  public int getMaxEventsToHandle() {
+    return Math.max(0, maxEventBacklog - eventsToBeProcessed.size());
+  }
+
   @Override
   public synchronized void handleEvents(Collection<TezEvent> events) {
     if (events == null || events.isEmpty()) {
@@ -843,28 +890,6 @@ public class LogicalIOProcessorRuntimeTask extends RuntimeTask {
       LOG.debug("Num of outputs to be closed={}", initializedOutputs.size());
     }
 
-    // Close processor
-    if (!processorClosed && processor != null) {
-      try {
-        processorClosed = true;
-        processor.close();
-        LOG.info("Closed processor for vertex={}, index={}, interruptedStatus={}",
-            processor
-                .getContext().getTaskVertexName(),
-            processor.getContext().getTaskVertexIndex(),
-            Thread.currentThread().isInterrupted());
-        maybeResetInterruptStatus();
-      } catch (InterruptedException ie) {
-        //reset the status
-        LOG.info("Resetting interrupt for processor");
-        Thread.currentThread().interrupt();
-      } catch (Throwable e) {
-        LOG.warn(
-            "Ignoring Exception when closing processor(cleanup). Exception class={}, message={}" +
-                e.getClass().getName(), e.getMessage());
-      }
-    }
-
     // Close the remaining inited Inputs.
     Iterator<Map.Entry<String, LogicalInput>> inputIterator = initializedInputs.entrySet().iterator();
     while (inputIterator.hasNext()) {
@@ -917,6 +942,28 @@ public class LogicalIOProcessorRuntimeTask extends RuntimeTask {
       printThreads();
     }
 
+    // Close processor
+    if (!processorClosed && processor != null) {
+      try {
+        processorClosed = true;
+        processor.close();
+        LOG.info("Closed processor for vertex={}, index={}, interruptedStatus={}",
+            processor
+            .getContext().getTaskVertexName(),
+            processor.getContext().getTaskVertexIndex(),
+            Thread.currentThread().isInterrupted());
+        maybeResetInterruptStatus();
+      } catch (InterruptedException ie) {
+        //reset the status
+        LOG.info("Resetting interrupt for processor");
+        Thread.currentThread().interrupt();
+      } catch (Throwable e) {
+        LOG.warn(
+            "Ignoring Exception when closing processor(cleanup). Exception class={}, message={}" +
+            e.getClass().getName(), e.getMessage());
+      }
+    }
+
     try {
       closeContexts();
       // Cleanup references which may be held by misbehaved tasks.
@@ -1024,4 +1071,10 @@ public class LogicalIOProcessorRuntimeTask extends RuntimeTask {
   public HadoopShim getHadoopShim() {
     return hadoopShim;
   }
+
+  @Private
+  @VisibleForTesting
+  public Configuration getTaskConf() {
+    return tezConf;
+  }
 }
