@@ -34,13 +34,16 @@ import java.util.concurrent.locks.Condition;
 import java.util.concurrent.locks.ReentrantLock;
 
 import org.apache.commons.lang.exception.ExceptionUtils;
+import org.apache.hadoop.util.ShutdownHookManager;
 import org.apache.tez.common.TezTaskUmbilicalProtocol;
 import org.apache.tez.common.counters.TezCounters;
 import org.apache.tez.dag.api.TezException;
 import org.apache.tez.dag.records.TezTaskAttemptID;
 import org.apache.tez.runtime.RuntimeTask;
+import org.apache.tez.runtime.api.*;
 import org.apache.tez.runtime.api.events.TaskAttemptCompletedEvent;
 import org.apache.tez.runtime.api.events.TaskAttemptFailedEvent;
+import org.apache.tez.runtime.api.events.TaskAttemptKilledEvent;
 import org.apache.tez.runtime.api.events.TaskStatusUpdateEvent;
 import org.apache.tez.runtime.api.impl.EventMetaData;
 import org.apache.tez.runtime.api.impl.TaskStatistics;
@@ -116,7 +119,6 @@ public class TaskReporter implements TaskReporterInterface {
   public synchronized void unregisterTask(TezTaskAttemptID taskAttemptID) {
     currentCallable.markComplete();
     currentCallable = null;
-    // KKK Make sure the callable completes before proceeding
   }
 
   @Override
@@ -124,6 +126,10 @@ public class TaskReporter implements TaskReporterInterface {
     heartbeatExecutor.shutdownNow();
   }
 
+  protected boolean isShuttingDown() {
+    return ShutdownHookManager.get().isShutdownInProgress();
+  }
+
   @VisibleForTesting
   static class HeartbeatCallable implements Callable<Boolean> {
 
@@ -184,7 +190,7 @@ public class TaskReporter implements TaskReporterInterface {
     @Override
     public Boolean call() throws Exception {
       // Heartbeat only for active tasks. Errors, etc will be reported directly.
-      while (!task.isTaskDone() && !task.hadFatalError()) {
+      while (!task.isTaskDone() && !task.wasErrorReported()) {
         ResponseWrapper response = heartbeat(null);
 
         if (response.shouldDie) {
@@ -209,7 +215,7 @@ public class TaskReporter implements TaskReporterInterface {
       int pendingEventCount = eventsToSend.size();
       if (pendingEventCount > 0) {
         // This is OK because the pending events will be sent via the succeeded/failed messages.
-        // TaskDone is set before taskSucceeded / taskFailed are sent out - which is what causes the
+        // TaskDone is set before taskSucceeded / taskTerminated are sent out - which is what causes the
         // thread to exit.
         LOG.warn("Exiting TaskReporter thread with pending queue size=" + pendingEventCount);
       }
@@ -235,7 +241,7 @@ public class TaskReporter implements TaskReporterInterface {
       List<TezEvent> events = new ArrayList<TezEvent>();
       eventsToSend.drainTo(events);
 
-      if (!task.isTaskDone() && !task.hadFatalError()) {
+      if (!task.isTaskDone() && !task.wasErrorReported()) {
         boolean sendCounters = false;
         /**
          * Increasing the heartbeat interval can delay the delivery of events. Sending just updated
@@ -254,8 +260,9 @@ public class TaskReporter implements TaskReporterInterface {
       long requestId = requestCounter.incrementAndGet();
       int fromEventId = task.getNextFromEventId();
       int fromPreRoutedEventId = task.getNextPreRoutedEventId();
+      int maxEvents = Math.min(maxEventsToGet, task.getMaxEventsToHandle());
       TezHeartbeatRequest request = new TezHeartbeatRequest(requestId, events, fromPreRoutedEventId,
-          containerIdStr, task.getTaskAttemptID(), fromEventId, maxEventsToGet);
+          containerIdStr, task.getTaskAttemptID(), fromEventId, maxEvents);
       if (LOG.isDebugEnabled()) {
         LOG.debug("Sending heartbeat to AM, request=" + request);
       }
@@ -280,7 +287,7 @@ public class TaskReporter implements TaskReporterInterface {
       // The same umbilical is used by multiple tasks. Problematic in the case where multiple tasks
       // are running using the same umbilical.
       int numEventsReceived = 0;
-      if (task.isTaskDone() || task.hadFatalError()) {
+      if (task.isTaskDone() || task.wasErrorReported()) {
         if (response.getEvents() != null && !response.getEvents().isEmpty()) {
           LOG.info("Current task already complete, Ignoring all events in"
               + " heartbeat response, eventCount=" + response.getEvents().size());
@@ -363,6 +370,8 @@ public class TaskReporter implements TaskReporterInterface {
     /**
      * Sends out final events for task failure.
      * @param taskAttemptID
+     * @param isKilled
+     * @param taskFailureType
      * @param t
      * @param diagnostics
      * @param srcMeta
@@ -372,8 +381,9 @@ public class TaskReporter implements TaskReporterInterface {
      * @throws TezException
      *           indicates an exception somewhere in the AM.
      */
-    private boolean taskFailed(TezTaskAttemptID taskAttemptID, Throwable t, String diagnostics,
-        EventMetaData srcMeta) throws IOException, TezException {
+    private boolean taskTerminated(TezTaskAttemptID taskAttemptID, boolean isKilled, TaskFailureType taskFailureType,
+                                   Throwable t, String diagnostics,
+                                   EventMetaData srcMeta) throws IOException, TezException {
       // Ensure only one final event is ever sent.
       if (!finalEventQueued.getAndSet(true)) {
         List<TezEvent> tezEvents = new ArrayList<TezEvent>();
@@ -382,13 +392,19 @@ public class TaskReporter implements TaskReporterInterface {
         } else {
           diagnostics = diagnostics + ":" + ExceptionUtils.getStackTrace(t);
         }
-        tezEvents.add(new TezEvent(new TaskAttemptFailedEvent(diagnostics),
-            srcMeta == null ? updateEventMetadata : srcMeta));
+        if (isKilled) {
+          tezEvents.add(new TezEvent(new TaskAttemptKilledEvent(diagnostics),
+              srcMeta == null ? updateEventMetadata : srcMeta));
+        } else {
+          tezEvents.add(new TezEvent(new TaskAttemptFailedEvent(diagnostics,
+              taskFailureType),
+              srcMeta == null ? updateEventMetadata : srcMeta));
+        }
         try {
           tezEvents.add(new TezEvent(getStatusUpdateEvent(true), updateEventMetadata));
         } catch (Exception e) {
           // Counter may exceed limitation
-          LOG.warn("Error when get constructing TaskStatusUpdateEvent");
+          LOG.warn("Error when get constructing TaskStatusUpdateEvent. Not sending it out");
         }
         return !heartbeat(tezEvents).shouldDie;
       } else {
@@ -431,9 +447,24 @@ public class TaskReporter implements TaskReporterInterface {
   }
 
   @Override
-  public synchronized boolean taskFailed(TezTaskAttemptID taskAttemptID, Throwable t, String diagnostics,
-      EventMetaData srcMeta) throws IOException, TezException {
-    return currentCallable.taskFailed(taskAttemptID, t, diagnostics, srcMeta);
+  public synchronized boolean taskFailed(TezTaskAttemptID taskAttemptID,
+                                                  TaskFailureType taskFailureType,
+                                                  Throwable t, String diagnostics,
+                                                  EventMetaData srcMeta) throws IOException,
+      TezException {
+    if(!isShuttingDown()) {
+      return currentCallable.taskTerminated(taskAttemptID, false, taskFailureType, t, diagnostics, srcMeta);
+    }
+    return false;
+  }
+
+  @Override
+  public boolean taskKilled(TezTaskAttemptID taskAttemptID, Throwable t, String diagnostics,
+                            EventMetaData srcMeta) throws IOException, TezException {
+    if(!isShuttingDown()) {
+      return currentCallable.taskTerminated(taskAttemptID, true, null, t, diagnostics, srcMeta);
+    }
+    return false;
   }
 
   @Override
