@@ -14,9 +14,7 @@
 
 package org.apache.tez.runtime.task;
 
-import static org.apache.tez.runtime.task.TaskExecutionTestHelpers.createProcessorIOException;
-import static org.apache.tez.runtime.task.TaskExecutionTestHelpers.createProcessorTezException;
-import static org.apache.tez.runtime.task.TaskExecutionTestHelpers.createTaskReporter;
+import static org.apache.tez.runtime.task.TaskExecutionTestHelpers.*;
 import static org.junit.Assert.assertEquals;
 import static org.junit.Assert.assertFalse;
 import static org.junit.Assert.assertNotNull;
@@ -35,6 +33,7 @@ import java.util.concurrent.ExecutorService;
 import java.util.concurrent.Executors;
 import java.util.concurrent.Future;
 import java.util.concurrent.atomic.AtomicBoolean;
+import java.util.concurrent.atomic.AtomicLong;
 import java.util.concurrent.locks.Condition;
 import java.util.concurrent.locks.ReentrantLock;
 
@@ -50,6 +49,8 @@ import org.apache.hadoop.fs.FileSystem;
 import org.apache.hadoop.fs.Path;
 import org.apache.hadoop.security.UserGroupInformation;
 import org.apache.hadoop.yarn.api.records.ApplicationId;
+import org.apache.tez.common.TezExecutors;
+import org.apache.tez.common.TezSharedExecutor;
 import org.apache.tez.common.counters.CounterGroup;
 import org.apache.tez.common.counters.FileSystemCounter;
 import org.apache.tez.common.counters.TaskCounter;
@@ -67,6 +68,7 @@ import org.apache.tez.dag.records.TezVertexID;
 import org.apache.tez.hadoop.shim.DefaultHadoopShim;
 import org.apache.tez.runtime.LogicalIOProcessorRuntimeTask;
 import org.apache.tez.runtime.api.ExecutionContext;
+import org.apache.tez.runtime.api.TaskFailureType;
 import org.apache.tez.runtime.api.ObjectRegistry;
 import org.apache.tez.runtime.api.impl.ExecutionContextImpl;
 import org.apache.tez.runtime.api.impl.InputSpec;
@@ -89,6 +91,9 @@ public class TestTaskExecution2 {
   private static final FileSystem localFs;
   private static final Path workDir;
 
+  private static final String FAILURE_START_STRING = "Error while running task ( failure )";
+  private static final String KILL_START_STRING = "Error while running task ( kill )";
+
   private static final ExecutorService taskExecutor = Executors.newFixedThreadPool(1);
 
   static {
@@ -135,7 +140,7 @@ public class TestTaskExecution2 {
       // Signal the processor to go through
       TestProcessor.signal();
       TaskRunner2Result result = taskRunnerFuture.get();
-      verifyTaskRunnerResult(result, EndReason.SUCCESS, null, false);
+      verifyTaskRunnerResult(result, EndReason.SUCCESS, null, false, null);
       assertNull(taskReporter.currentCallable);
       umbilical.verifyTaskSuccessEvent();
       assertFalse(TestProcessor.wasAborted());
@@ -166,7 +171,7 @@ public class TestTaskExecution2 {
       // Signal the processor to go through
       TestProcessor.signal();
       TaskRunner2Result result = taskRunnerFuture.get();
-      verifyTaskRunnerResult(result, EndReason.SUCCESS, null, false);
+      verifyTaskRunnerResult(result, EndReason.SUCCESS, null, false, null);
       assertNull(taskReporter.currentCallable);
       umbilical.verifyTaskSuccessEvent();
       assertFalse(TestProcessor.wasAborted());
@@ -182,7 +187,7 @@ public class TestTaskExecution2 {
       // Signal the processor to go through
       TestProcessor.signal();
       result = taskRunnerFuture.get();
-      verifyTaskRunnerResult(result, EndReason.SUCCESS, null, false);
+      verifyTaskRunnerResult(result, EndReason.SUCCESS, null, false, null);
       assertNull(taskReporter.currentCallable);
       umbilical.verifyTaskSuccessEvent();
       assertFalse(TestProcessor.wasAborted());
@@ -217,14 +222,15 @@ public class TestTaskExecution2 {
       TestProcessor.awaitStart();
       TestProcessor.signal();
       TaskRunner2Result result = taskRunnerFuture.get();
-      verifyTaskRunnerResult(result, EndReason.TASK_ERROR, createProcessorTezException(), false);
+      verifyTaskRunnerResult(result, EndReason.TASK_ERROR, createProcessorTezException(), false, TaskFailureType.NON_FATAL);
 
       assertNull(taskReporter.currentCallable);
       umbilical.verifyTaskFailedEvent(
-          "Failure while running task",
+          FAILURE_START_STRING,
           TezException.class.getName() + ": " + TezException.class.getSimpleName());
       // Failure detected as a result of fall off from the run method. abort isn't required.
       assertFalse(TestProcessor.wasAborted());
+      assertTrue(taskRunner.task.getCounters().countCounters() != 0);
     } finally {
       executor.shutdownNow();
     }
@@ -253,14 +259,15 @@ public class TestTaskExecution2 {
 
       TaskRunner2Result result = taskRunnerFuture.get();
       verifyTaskRunnerResult(result, EndReason.TASK_ERROR,
-          new TezReflectionException("TezReflectionException"), false);
+          new TezReflectionException("TezReflectionException"), false, TaskFailureType.NON_FATAL);
 
       assertNull(taskReporter.currentCallable);
-      umbilical.verifyTaskFailedEvent("Failure while running task",
+      umbilical.verifyTaskFailedEvent(FAILURE_START_STRING,
           ":org.apache.tez.dag.api.TezReflectionException: "
               + "Unable to load class: NotExitedProcessor");
       // Failure detected as a result of fall off from the run method. abort isn't required.
       assertFalse(TestProcessor.wasAborted());
+      assertTrue(taskRunner.task.getCounters().countCounters() != 0);
     } finally {
       executor.shutdownNow();
     }
@@ -289,15 +296,53 @@ public class TestTaskExecution2 {
       TestProcessor.awaitStart();
       TestProcessor.signal();
       TaskRunner2Result result = taskRunnerFuture.get();
-      verifyTaskRunnerResult(result, EndReason.TASK_ERROR, createProcessorIOException(), false);
+      verifyTaskRunnerResult(result, EndReason.TASK_ERROR, createProcessorIOException(), false, TaskFailureType.NON_FATAL);
 
 
       assertNull(taskReporter.currentCallable);
       umbilical.verifyTaskFailedEvent(
-          "Failure while running task",
+          FAILURE_START_STRING,
           IOException.class.getName() + ": " + IOException.class.getSimpleName());
       // Failure detected as a result of fall off from the run method. abort isn't required.
       assertFalse(TestProcessor.wasAborted());
+      assertTrue(taskRunner.task.getCounters().countCounters() != 0);
+    } finally {
+      executor.shutdownNow();
+    }
+  }
+
+  // test that makes sure errors aren't reported when the container is already failing
+  @Test(timeout = 5000)
+  public void testIgnoreErrorsDuringFailure() throws IOException, InterruptedException, TezException,
+      ExecutionException {
+
+    ListeningExecutorService executor = null;
+    try {
+      ExecutorService rawExecutor = Executors.newFixedThreadPool(1);
+      executor = MoreExecutors.listeningDecorator(rawExecutor);
+      ApplicationId appId = ApplicationId.newInstance(10000, 1);
+      TaskExecutionTestHelpers.TezTaskUmbilicalForTest
+          umbilical = new TaskExecutionTestHelpers.TezTaskUmbilicalForTest();
+
+      TaskReporter taskReporter = new TaskReporter(umbilical, 100, 1000, 100, new AtomicLong(0),
+        createContainerId(appId).toString()) {
+        @Override
+        protected boolean isShuttingDown() {
+          return true;
+        }
+      };
+
+      TezTaskRunner2 taskRunner = createTaskRunner(appId, umbilical, taskReporter, executor,
+          TestProcessor.CONF_THROW_IO_EXCEPTION);
+      // Setup the executor
+
+      taskExecutor.submit(new TaskRunnerCallable2ForTest(taskRunner));
+
+      // Signal the processor to go through
+      TestProcessor.awaitStart();
+      TestProcessor.signal();
+
+      umbilical.verifyNoCompletionEvents();
     } finally {
       executor.shutdownNow();
     }
@@ -330,7 +375,7 @@ public class TestTaskExecution2 {
       TaskRunner2Result result = taskRunnerFuture.get();
       verifyTaskRunnerResult(result, EndReason.COMMUNICATION_FAILURE,
           new IOException("IOException"),
-          TaskExecutionTestHelpers.HEARTBEAT_EXCEPTION_STRING, false);
+          TaskExecutionTestHelpers.HEARTBEAT_EXCEPTION_STRING, false, TaskFailureType.NON_FATAL);
 
       TestProcessor.awaitCompletion();
       assertTrue(TestProcessor.wasInterrupted());
@@ -368,7 +413,7 @@ public class TestTaskExecution2 {
       // Not signaling an actual start to verify task interruption
 
       TaskRunner2Result result = taskRunnerFuture.get();
-      verifyTaskRunnerResult(result, EndReason.CONTAINER_STOP_REQUESTED, null, true);
+      verifyTaskRunnerResult(result, EndReason.CONTAINER_STOP_REQUESTED, null, true, null);
 
 
       TestProcessor.awaitCompletion();
@@ -385,7 +430,7 @@ public class TestTaskExecution2 {
   }
 
   @Test(timeout = 5000)
-  public void testSignalFatalErrorAndLoop() throws IOException, InterruptedException, TezException,
+  public void testSignalDeprecatedFatalErrorAndLoop() throws IOException, InterruptedException, TezException,
       ExecutionException {
 
     ListeningExecutorService executor = null;
@@ -398,7 +443,7 @@ public class TestTaskExecution2 {
       TaskReporter taskReporter = createTaskReporter(appId, umbilical);
 
       TezTaskRunner2 taskRunner = createTaskRunner(appId, umbilical, taskReporter, executor,
-          TestProcessor.CONF_SIGNAL_FATAL_AND_LOOP);
+          TestProcessor.CONF_SIGNAL_DEPRECATEDFATAL_AND_LOOP);
       // Setup the executor
       Future<TaskRunner2Result> taskRunnerFuture =
           taskExecutor.submit(new TaskRunnerCallable2ForTest(taskRunner));
@@ -410,13 +455,13 @@ public class TestTaskExecution2 {
       // The fatal error should have caused an interrupt.
 
       TaskRunner2Result result = taskRunnerFuture.get();
-      verifyTaskRunnerResult(result, EndReason.TASK_ERROR, createProcessorIOException(), false);
+      verifyTaskRunnerResult(result, EndReason.TASK_ERROR, createProcessorIOException(), false, TaskFailureType.NON_FATAL);
 
       TestProcessor.awaitCompletion();
       assertTrue(TestProcessor.wasInterrupted());
       assertNull(taskReporter.currentCallable);
       umbilical.verifyTaskFailedEvent(
-          "Failure while running task",
+          FAILURE_START_STRING,
           IOException.class.getName() + ": " + IOException.class.getSimpleName());
       // Signal fatal error should cause the processor to fail.
       assertTrue(TestProcessor.wasAborted());
@@ -425,6 +470,115 @@ public class TestTaskExecution2 {
     }
   }
 
+  @Test(timeout = 5000)
+  public void testSignalFatalAndThrow() throws IOException, InterruptedException, TezException,
+      ExecutionException {
+
+    ListeningExecutorService executor = null;
+    try {
+      ExecutorService rawExecutor = Executors.newFixedThreadPool(1);
+      executor = MoreExecutors.listeningDecorator(rawExecutor);
+      ApplicationId appId = ApplicationId.newInstance(10000, 1);
+      TaskExecutionTestHelpers.TezTaskUmbilicalForTest
+          umbilical = new TaskExecutionTestHelpers.TezTaskUmbilicalForTest();
+      TaskReporter taskReporter = createTaskReporter(appId, umbilical);
+
+      TezTaskRunner2 taskRunner = createTaskRunner(appId, umbilical, taskReporter, executor,
+          TestProcessor.CONF_SIGNAL_FATAL_AND_THROW);
+      // Setup the executor
+      Future<TaskRunner2Result> taskRunnerFuture =
+          taskExecutor.submit(new TaskRunnerCallable2ForTest(taskRunner));
+      // Signal the processor to go through
+      TestProcessor.awaitStart();
+      TestProcessor.signal();
+
+      TaskRunner2Result result = taskRunnerFuture.get();
+      verifyTaskRunnerResult(result, EndReason.TASK_ERROR, createProcessorIOException(), false, TaskFailureType.FATAL);
+
+      TestProcessor.awaitCompletion();
+      assertNull(taskReporter.currentCallable);
+      umbilical.verifyTaskFailedEvent(
+          FAILURE_START_STRING,
+          IOException.class.getName() + ": " + IOException.class.getSimpleName(), TaskFailureType.FATAL);
+      assertTrue(TestProcessor.wasAborted());
+    } finally {
+      executor.shutdownNow();
+    }
+  }
+
+  @Test(timeout = 5000)
+  public void testSignalNonFatalAndThrow() throws IOException, InterruptedException, TezException,
+      ExecutionException {
+
+    ListeningExecutorService executor = null;
+    try {
+      ExecutorService rawExecutor = Executors.newFixedThreadPool(1);
+      executor = MoreExecutors.listeningDecorator(rawExecutor);
+      ApplicationId appId = ApplicationId.newInstance(10000, 1);
+      TaskExecutionTestHelpers.TezTaskUmbilicalForTest
+          umbilical = new TaskExecutionTestHelpers.TezTaskUmbilicalForTest();
+      TaskReporter taskReporter = createTaskReporter(appId, umbilical);
+
+      TezTaskRunner2 taskRunner = createTaskRunner(appId, umbilical, taskReporter, executor,
+          TestProcessor.CONF_SIGNAL_NON_FATAL_AND_THROW);
+      // Setup the executor
+      Future<TaskRunner2Result> taskRunnerFuture =
+          taskExecutor.submit(new TaskRunnerCallable2ForTest(taskRunner));
+      // Signal the processor to go through
+      TestProcessor.awaitStart();
+      TestProcessor.signal();
+
+      TaskRunner2Result result = taskRunnerFuture.get();
+      verifyTaskRunnerResult(result, EndReason.TASK_ERROR, createProcessorIOException(), false, TaskFailureType.NON_FATAL);
+
+      TestProcessor.awaitCompletion();
+      assertNull(taskReporter.currentCallable);
+      umbilical.verifyTaskFailedEvent(
+          FAILURE_START_STRING,
+          IOException.class.getName() + ": " + IOException.class.getSimpleName(), TaskFailureType.NON_FATAL);
+      assertTrue(TestProcessor.wasAborted());
+    } finally {
+      executor.shutdownNow();
+    }
+  }
+
+  @Test(timeout = 5000)
+  public void testTaskSelfKill() throws IOException, InterruptedException, TezException,
+      ExecutionException {
+
+    ListeningExecutorService executor = null;
+    try {
+      ExecutorService rawExecutor = Executors.newFixedThreadPool(1);
+      executor = MoreExecutors.listeningDecorator(rawExecutor);
+      ApplicationId appId = ApplicationId.newInstance(10000, 1);
+      TaskExecutionTestHelpers.TezTaskUmbilicalForTest
+          umbilical = new TaskExecutionTestHelpers.TezTaskUmbilicalForTest();
+      TaskReporter taskReporter = createTaskReporter(appId, umbilical);
+
+      TezTaskRunner2 taskRunner = createTaskRunner(appId, umbilical, taskReporter, executor,
+          TestProcessor.CONF_SELF_KILL_AND_COMPLETE);
+      // Setup the executor
+      Future<TaskRunner2Result> taskRunnerFuture =
+          taskExecutor.submit(new TaskRunnerCallable2ForTest(taskRunner));
+      // Signal the processor to go through
+      TestProcessor.awaitStart();
+      TestProcessor.signal();
+
+      TaskRunner2Result result = taskRunnerFuture.get();
+      verifyTaskRunnerResult(result, EndReason.TASK_KILL_REQUEST, createProcessorIOException(), false,
+          null);
+
+      TestProcessor.awaitCompletion();
+      assertNull(taskReporter.currentCallable);
+      umbilical.verifyTaskKilledEvent(
+          KILL_START_STRING,
+          IOException.class.getName() + ": " + IOException.class.getSimpleName());
+      assertTrue(TestProcessor.wasAborted());
+    } finally {
+      executor.shutdownNow();
+    }
+  }
+
   @Test(timeout = 5000)
   public void testTaskKilled() throws IOException, InterruptedException, TezException,
       ExecutionException {
@@ -449,7 +603,7 @@ public class TestTaskExecution2 {
       taskRunner.killTask();
 
       TaskRunner2Result result = taskRunnerFuture.get();
-      verifyTaskRunnerResult(result, EndReason.KILL_REQUESTED, null, false);
+      verifyTaskRunnerResult(result, EndReason.KILL_REQUESTED, null, false, null);
 
       TestProcessor.awaitCompletion();
       assertTrue(TestProcessor.wasInterrupted());
@@ -489,7 +643,7 @@ public class TestTaskExecution2 {
 
       taskRunner.killTask();
       TaskRunner2Result result = taskRunnerFuture.get();
-      verifyTaskRunnerResult(result, EndReason.SUCCESS, null, false);
+      verifyTaskRunnerResult(result, EndReason.SUCCESS, null, false, null);
 
       assertFalse(TestProcessor.wasInterrupted());
       assertNull(taskReporter.currentCallable);
@@ -531,15 +685,17 @@ public class TestTaskExecution2 {
 
   private void verifyTaskRunnerResult(TaskRunner2Result taskRunner2Result,
                                       EndReason expectedEndReason, Throwable expectedThrowable,
-                                      boolean wasShutdownRequested) {
+                                      boolean wasShutdownRequested,
+                                      TaskFailureType taskFailureType) {
     verifyTaskRunnerResult(taskRunner2Result, expectedEndReason, expectedThrowable, null,
-        wasShutdownRequested);
+        wasShutdownRequested, taskFailureType);
   }
 
   private void verifyTaskRunnerResult(TaskRunner2Result taskRunner2Result,
                                       EndReason expectedEndReason, Throwable expectedThrowable,
                                       String expectedExceptionMessage,
-                                      boolean wasShutdownRequested) {
+                                      boolean wasShutdownRequested,
+                                      TaskFailureType taskFailureType) {
     assertEquals(expectedEndReason, taskRunner2Result.getEndReason());
     if (expectedThrowable == null) {
       assertNull(taskRunner2Result.getError());
@@ -554,6 +710,7 @@ public class TestTaskExecution2 {
       }
 
     }
+    assertEquals(taskFailureType, taskRunner2Result.getTaskFailureType());
     assertEquals(wasShutdownRequested, taskRunner2Result.isContainerShutdownRequested());
   }
 
@@ -621,21 +778,23 @@ public class TestTaskExecution2 {
         .setUserPayload(UserPayload.create(ByteBuffer.wrap(processorConf)));
     TaskSpec taskSpec =
         new TaskSpec(taskAttemptId, "dagName", "vertexName", -1, processorDescriptor,
-            new ArrayList<InputSpec>(), new ArrayList<OutputSpec>(), null);
+            new ArrayList<InputSpec>(), new ArrayList<OutputSpec>(), null, null);
 
+    TezExecutors sharedExecutor = new TezSharedExecutor(tezConf);
     TezTaskRunner2 taskRunner;
     if (testRunner) {
       taskRunner = new TezTaskRunner2ForTest(tezConf, ugi, localDirs, taskSpec, 1,
           new HashMap<String, ByteBuffer>(), new HashMap<String, String>(),
           HashMultimap.<String, String>create(), taskReporter,
           executor, null, "", new ExecutionContextImpl("localhost"),
-          Runtime.getRuntime().maxMemory(), updateSysCounters);
+          Runtime.getRuntime().maxMemory(), updateSysCounters, sharedExecutor);
     } else {
       taskRunner = new TezTaskRunner2(tezConf, ugi, localDirs, taskSpec, 1,
           new HashMap<String, ByteBuffer>(), new HashMap<String, String>(),
           HashMultimap.<String, String>create(), taskReporter,
           executor, null, "", new ExecutionContextImpl("localhost"),
-          Runtime.getRuntime().maxMemory(), updateSysCounters, new DefaultHadoopShim());
+          Runtime.getRuntime().maxMemory(), updateSysCounters, new DefaultHadoopShim(),
+          sharedExecutor);
     }
 
     return taskRunner;
@@ -660,10 +819,12 @@ public class TestTaskExecution2 {
                                  String pid,
                                  ExecutionContext executionContext,
                                  long memAvailable,
-                                 boolean updateSysCounters) throws IOException {
+                                 boolean updateSysCounters,
+                                 TezExecutors sharedExecutor) throws IOException {
       super(tezConf, ugi, localDirs, taskSpec, appAttemptNumber, serviceConsumerMetadata,
           serviceProviderEnvMap, startedInputsMap, taskReporter, executor, objectRegistry, pid,
-          executionContext, memAvailable, updateSysCounters, new DefaultHadoopShim());
+          executionContext, memAvailable, updateSysCounters, new DefaultHadoopShim(),
+          sharedExecutor);
     }
 
 
