@@ -172,14 +172,29 @@ public class TezRuntimeConfiguration {
       "pipelined.sorter.sort.threads";
   public static final int TEZ_RUNTIME_PIPELINED_SORTER_SORT_THREADS_DEFAULT = 2;
 
+  /**
+   * Integer value. Percentage of buffer to be filled before we spill to disk. Default value is 0,
+   * which will spill for every buffer.
+   */
+  @ConfigurationProperty(type="int")
+  public static final String TEZ_RUNTIME_UNORDERED_PARTITIONED_KVWRITER_BUFFER_MERGE_PERCENT =
+      TEZ_RUNTIME_PREFIX + "unordered-partitioned-kvwriter.buffer-merge-percent";
+  public static final int TEZ_RUNTIME_UNORDERED_PARTITIONED_KVWRITER_BUFFER_MERGE_PERCENT_DEFAULT =
+      0;
+
   /**
    * Report partition statistics (e.g better scheduling in ShuffleVertexManager). TEZ-2496
    * This can be enabled/disabled at vertex level.
+   * {@link org.apache.tez.runtime.library.api.TezRuntimeConfiguration.ReportPartitionStats}
+   * defines the list of values that can be specified.
+   * TODO TEZ-3303 Given ShuffleVertexManager doesn't consume precise stats
+   * yet. So do not set the value to "precise" yet when ShuffleVertexManager is used.
    */
-  @ConfigurationProperty(type = "boolean")
-  public static final String TEZ_RUNTIME_REPORT_PARTITION_STATS = TEZ_RUNTIME_PREFIX +
-      "report.partition.stats";
-  public static final boolean TEZ_RUNTIME_REPORT_PARTITION_STATS_DEFAULT = true;
+  @ConfigurationProperty
+  public static final String TEZ_RUNTIME_REPORT_PARTITION_STATS =
+      TEZ_RUNTIME_PREFIX + "report.partition.stats";
+  public static final String TEZ_RUNTIME_REPORT_PARTITION_STATS_DEFAULT =
+      ReportPartitionStats.MEMORY_OPTIMIZED.getType();
 
   /**
    * Size of the buffer to use if not writing directly to disk.
@@ -223,6 +238,15 @@ public class TezRuntimeConfiguration {
       "shuffle.fetch.failures.limit";
   public static final int TEZ_RUNTIME_SHUFFLE_FETCH_FAILURES_LIMIT_DEFAULT = 5;
 
+  /**
+   * Specifies in milliseconds the maximum delay a penalized host can have before being retried,
+   * defaults to 10 minutes.
+   */
+  @ConfigurationProperty(type = "integer")
+  public static final String TEZ_RUNTIME_SHUFFLE_HOST_PENALTY_TIME_LIMIT_MS = TEZ_RUNTIME_PREFIX +
+      "shuffle.host.penalty.time.limit";
+  public static final int TEZ_RUNTIME_SHUFFLE_HOST_PENALTY_TIME_LIMIT_MS_DEFAULT = 600000;
+
   @Private
   @Unstable
   @ConfigurationProperty(type = "integer")
@@ -373,6 +397,17 @@ public class TezRuntimeConfiguration {
       "shuffle.ssl.enable";
   public static final boolean TEZ_RUNTIME_SHUFFLE_ENABLE_SSL_DEFAULT = false;
 
+  /**
+   * Controls verification of data checksums when fetching data directly to
+   * disk. Enabling verification allows the fetcher to detect corrupted data
+   * and report the failure against the upstream task before the data reaches
+   * the Processor and causes the fetching task to fail.
+   */
+  @ConfigurationProperty(type = "boolean")
+  public static final String TEZ_RUNTIME_SHUFFLE_FETCH_VERIFY_DISK_CHECKSUM =
+      TEZ_RUNTIME_PREFIX + "shuffle.fetch.verify-disk-checksum";
+  public static final boolean TEZ_RUNTIME_SHUFFLE_FETCH_VERIFY_DISK_CHECKSUM_DEFAULT = true;
+
   @ConfigurationProperty(type = "float")
   public static final String TEZ_RUNTIME_SHUFFLE_FETCH_BUFFER_PERCENT = TEZ_RUNTIME_PREFIX +
       "shuffle.fetch.buffer.percent";
@@ -401,6 +436,13 @@ public class TezRuntimeConfiguration {
   public static final boolean TEZ_RUNTIME_SHUFFLE_ENABLE_MEMTOMEM_DEFAULT =
       false;
 
+  @Private
+  @Unstable
+  @ConfigurationProperty(type = "boolean")
+  public static final String TEZ_RUNTIME_SHUFFLE_FETCHER_USE_SHARED_POOL = TEZ_RUNTIME_PREFIX +
+      "shuffle.fetcher.use-shared-pool";
+  public static final boolean TEZ_RUNTIME_SHUFFLE_FETCHER_USE_SHARED_POOL_DEFAULT = false;
+
   @ConfigurationProperty(type = "float")
   public static final String TEZ_RUNTIME_INPUT_POST_MERGE_BUFFER_PERCENT = TEZ_RUNTIME_PREFIX +
       "task.input.post-merge.buffer.percent";
@@ -541,6 +583,7 @@ public class TezRuntimeConfiguration {
     tezRuntimeKeys.add(TEZ_RUNTIME_SHUFFLE_READ_TIMEOUT);
     tezRuntimeKeys.add(TEZ_RUNTIME_SHUFFLE_BUFFER_SIZE);
     tezRuntimeKeys.add(TEZ_RUNTIME_SHUFFLE_ENABLE_SSL);
+    tezRuntimeKeys.add(TEZ_RUNTIME_SHUFFLE_FETCH_VERIFY_DISK_CHECKSUM);
     tezRuntimeKeys.add(TEZ_RUNTIME_SHUFFLE_FETCH_BUFFER_PERCENT);
     tezRuntimeKeys.add(TEZ_RUNTIME_SHUFFLE_MEMORY_LIMIT_PERCENT);
     tezRuntimeKeys.add(TEZ_RUNTIME_SHUFFLE_MERGE_PERCENT);
@@ -573,6 +616,9 @@ public class TezRuntimeConfiguration {
     tezRuntimeKeys.add(TEZ_RUNTIME_CONVERT_USER_PAYLOAD_TO_HISTORY_TEXT);
     tezRuntimeKeys.add(TEZ_RUNTIME_SORTER_CLASS);
     tezRuntimeKeys.add(TEZ_RUNTIME_CLEANUP_FILES_ON_INTERRUPT);
+    tezRuntimeKeys.add(TEZ_RUNTIME_UNORDERED_PARTITIONED_KVWRITER_BUFFER_MERGE_PERCENT);
+    tezRuntimeKeys.add(TEZ_RUNTIME_SHUFFLE_FETCHER_USE_SHARED_POOL);
+    tezRuntimeKeys.add(TEZ_RUNTIME_SHUFFLE_HOST_PENALTY_TIME_LIMIT_MS);
 
     defaultConf.addResource("core-default.xml");
     defaultConf.addResource("core-site.xml");
@@ -623,4 +669,70 @@ public class TezRuntimeConfiguration {
   public static Map<String, String> getOtherConfigDefaults() {
     return Collections.unmodifiableMap(otherConfMap);
   }
+
+  public enum ReportPartitionStats {
+    @Deprecated
+    /**
+     * Don't report partition stats. It is the same as NONE.
+     * It is defined to maintain backward compatibility given
+     * Configuration @link{#TEZ_RUNTIME_REPORT_PARTITION_STATS} used
+     * to be boolean type.
+     */
+    DISABLED("false"),
+
+    @Deprecated
+    /**
+     * Report partition stats. It is the same as MEMORY_OPTIMIZED.
+     * It is defined to maintain backward compatibility given
+     * Configuration @link{#TEZ_RUNTIME_REPORT_PARTITION_STATS} used
+     * to be boolean type.
+     */
+    ENABLED("true"),
+
+    /**
+     * Don't report partition stats.
+     */
+    NONE("none"),
+
+    /**
+     * Report partition stats with less precision to reduce
+     * memory and CPU overhead
+     */
+    MEMORY_OPTIMIZED("memory_optimized"),
+
+    /**
+     * Report precise partition stats in MB.
+     */
+    PRECISE("precise");
+
+    private final String type;
+
+    private ReportPartitionStats(String type) {
+      this.type = type;
+    }
+
+    public final String getType() {
+      return type;
+    }
+
+    public boolean isEnabled() {
+      return !equals(ReportPartitionStats.DISABLED) &&
+              !equals(ReportPartitionStats.NONE);
+    }
+
+    public boolean isPrecise() {
+      return equals(ReportPartitionStats.PRECISE);
+    }
+
+    public static ReportPartitionStats fromString(String type) {
+      if (type != null) {
+        for (ReportPartitionStats b : ReportPartitionStats.values()) {
+          if (type.equalsIgnoreCase(b.type)) {
+            return b;
+          }
+        }
+      }
+      throw new IllegalArgumentException("Invalid type " + type);
+    }
+  }
 }
