@@ -21,6 +21,7 @@ import java.io.DataInputStream;
 import java.io.IOException;
 import java.net.SocketTimeoutException;
 import java.net.URL;
+import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collection;
 import java.util.Iterator;
@@ -29,6 +30,7 @@ import java.util.List;
 import java.util.Map;
 import java.util.concurrent.atomic.AtomicInteger;
 
+import org.apache.hadoop.io.WritableUtils;
 import org.apache.tez.http.BaseHttpConnection;
 import org.apache.tez.http.HttpConnectionParams;
 import org.apache.tez.common.CallableWithNdc;
@@ -59,6 +61,7 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
 
   private final Configuration conf;
   private final boolean localDiskFetchEnabled;
+  private final boolean verifyDiskChecksum;
 
   private final TezCounter connectionErrs;
   private final TezCounter ioErrs;
@@ -68,20 +71,24 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
   private final TezCounter wrongReduceErrs;
   private final FetchedInputAllocatorOrderedGrouped allocator;
   private final ShuffleScheduler scheduler;
-  private final ShuffleClientMetrics metrics;
   private final ExceptionReporter exceptionReporter;
   private final int id;
   private final String logIdentifier;
-  private final String localShuffleHostPort;
+  private final String localShuffleHost;
+  private final int localShufflePort;
+  private final String applicationId;
+ private final int dagId;
   private final MapHost mapHost;
 
-  private final int currentPartition;
+  private final int minPartition;
+  private final int maxPartition;
 
   // Decompression of map-outputs
   private final CompressionCodec codec;
   private final JobTokenSecretManager jobTokenSecretManager;
 
   final HttpConnectionParams httpConnectionParams;
+  private final boolean sslShuffle;
 
   @VisibleForTesting
   volatile boolean stopped = false;
@@ -94,6 +101,7 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
 
   volatile BaseHttpConnection httpConnection;
   private final boolean asyncHttp;
+  private final boolean compositeFetch;
 
 
   // Initiative value is 0, which means it hasn't retried yet.
@@ -102,7 +110,6 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
   public FetcherOrderedGrouped(HttpConnectionParams httpConnectionParams,
                                ShuffleScheduler scheduler,
                                FetchedInputAllocatorOrderedGrouped allocator,
-                               ShuffleClientMetrics metrics,
                                ExceptionReporter exceptionReporter, JobTokenSecretManager jobTokenSecretMgr,
                                boolean ifileReadAhead, int ifileReadAheadLength,
                                CompressionCodec codec,
@@ -118,13 +125,18 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
                                TezCounter wrongMapErrsCounter,
                                TezCounter connectionErrsCounter,
                                TezCounter wrongReduceErrsCounter,
-                               boolean asyncHttp) {
+                               String applicationId,
+                               int dagId,
+                               boolean asyncHttp,
+                               boolean sslShuffle,
+                               boolean verifyDiskChecksum,
+                               boolean compositeFetch) {
     this.scheduler = scheduler;
     this.allocator = allocator;
-    this.metrics = metrics;
     this.exceptionReporter = exceptionReporter;
     this.mapHost = mapHost;
-    this.currentPartition = this.mapHost.getPartitionId();
+    this.minPartition = this.mapHost.getPartitionId();
+    this.maxPartition = this.minPartition + this.mapHost.getPartitionCount() - 1;
     this.id = nextId.incrementAndGet();
     this.jobTokenSecretManager = jobTokenSecretMgr;
 
@@ -134,6 +146,8 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
     this.wrongMapErrs = wrongMapErrsCounter;
     this.connectionErrs = connectionErrsCounter;
     this.wrongReduceErrs = wrongReduceErrsCounter;
+    this.applicationId = applicationId;
+    this.dagId = dagId;
 
     this.ifileReadAhead = ifileReadAhead;
     this.ifileReadAheadLength = ifileReadAheadLength;
@@ -145,9 +159,13 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
       this.codec = null;
     }
     this.conf = conf;
-    this.localShuffleHostPort = localHostname + ":" + String.valueOf(shufflePort);
+    this.localShuffleHost = localHostname;
+    this.localShufflePort = shufflePort;
 
     this.localDiskFetchEnabled = localDiskFetchEnabled;
+    this.sslShuffle = sslShuffle;
+    this.verifyDiskChecksum = verifyDiskChecksum;
+    this.compositeFetch = compositeFetch;
 
     this.logIdentifier = "fetcher [" + srcNameTrimmed + "] #" + id;
   }
@@ -155,10 +173,7 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
   @VisibleForTesting
   protected void fetchNext() throws InterruptedException, IOException {
     try {
-      metrics.threadBusy();
-
-      String hostPort = mapHost.getHostIdentifier();
-      if (localDiskFetchEnabled && hostPort.equals(localShuffleHostPort)) {
+      if (localDiskFetchEnabled && mapHost.getHost().equals(localShuffleHost) && mapHost.getPort() == localShufflePort) {
         setupLocalDiskFetch(mapHost);
       } else {
         // Shuffle
@@ -167,7 +182,6 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
     } finally {
       cleanupCurrentConnection(false);
       scheduler.freeHost(mapHost);
-      metrics.threadFree();
     }
   }
 
@@ -239,7 +253,7 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
     }
     if(LOG.isDebugEnabled()) {
       LOG.debug("Fetcher " + id + " going to fetch from " + host + " for: "
-        + srcAttempts + ", partitionId: " + currentPartition);
+        + srcAttempts + ", partition range: " + minPartition + "-" + maxPartition);
     }
     populateRemainingMap(srcAttempts);
     // Construct the url and connect
@@ -258,11 +272,13 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
       // yet_to_be_fetched list and marking the failed tasks.
       InputAttemptIdentifier[] failedTasks = null;
       while (!remaining.isEmpty() && failedTasks == null) {
+        InputAttemptIdentifier inputAttemptIdentifier =
+            remaining.entrySet().iterator().next().getValue();
         // fail immediately after first failure because we dont know how much to
         // skip for this error in the input stream. So we cannot move on to the 
         // remaining outputs. YARN-1773. Will get to them in the next retry.
         try {
-          failedTasks = copyMapOutput(host, input);
+          failedTasks = copyMapOutput(host, input, inputAttemptIdentifier);
         } catch (FetcherReadTimeoutException e) {
           // Setup connection again if disconnected
           cleanupCurrentConnection(true);
@@ -319,8 +335,9 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
       throws IOException {
     boolean connectSucceeded = false;
     try {
-      URL url = ShuffleUtils.constructInputURL(host.getBaseUrl(), attempts,
-          httpConnectionParams.isKeepAlive());
+      StringBuilder baseURI = ShuffleUtils.constructBaseURIForShuffleHandler(host.getHost(),
+          host.getPort(), host.getPartitionId(), host.getPartitionCount(), applicationId, dagId, sslShuffle);
+      URL url = ShuffleUtils.constructInputURL(baseURI.toString(), attempts, httpConnectionParams.isKeepAlive());
       httpConnection = ShuffleUtils.getHttpConnection(asyncHttp, url, httpConnectionParams,
           logIdentifier, jobTokenSecretManager);
       connectSucceeded = httpConnection.connect();
@@ -385,132 +402,170 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
 
   private static InputAttemptIdentifier[] EMPTY_ATTEMPT_ID_ARRAY = new InputAttemptIdentifier[0];
 
+  private static class MapOutputStat {
+    final InputAttemptIdentifier srcAttemptId;
+    final long decompressedLength;
+    final long compressedLength;
+    final int forReduce;
+
+    MapOutputStat(InputAttemptIdentifier srcAttemptId, long decompressedLength, long compressedLength, int forReduce) {
+      this.srcAttemptId = srcAttemptId;
+      this.decompressedLength = decompressedLength;
+      this.compressedLength = compressedLength;
+      this.forReduce = forReduce;
+    }
+
+    @Override
+    public String toString() {
+      return "id: " + srcAttemptId + ", decompressed length: " + decompressedLength + ", compressed length: " + compressedLength + ", reduce: " + forReduce;
+    }
+  }
+
   protected InputAttemptIdentifier[] copyMapOutput(MapHost host,
-                                DataInputStream input) throws FetcherReadTimeoutException {
+                                DataInputStream input, InputAttemptIdentifier inputAttemptIdentifier) throws FetcherReadTimeoutException {
     MapOutput mapOutput = null;
     InputAttemptIdentifier srcAttemptId = null;
-    long decompressedLength = -1;
-    long compressedLength = -1;
-
+    long decompressedLength = 0;
+    long compressedLength = 0;
     try {
       long startTime = System.currentTimeMillis();
-      int forReduce = -1;
-      //Read the shuffle header
-      try {
-        ShuffleHeader header = new ShuffleHeader();
-        // TODO Review: Multiple header reads in case of status WAIT ? 
-        header.readFields(input);
-        if (!header.mapId.startsWith(InputAttemptIdentifier.PATH_PREFIX)) {
+      int partitionCount = 1;
+
+      if (this.compositeFetch) {
+        // Multiple partitions are fetched
+        partitionCount = WritableUtils.readVInt(input);
+      }
+      ArrayList<MapOutputStat> mapOutputStats = new ArrayList<>(partitionCount);
+      for (int mapOutputIndex = 0; mapOutputIndex < partitionCount; mapOutputIndex++) {
+        MapOutputStat mapOutputStat = null;
+        try {
+          //Read the shuffle header
+          ShuffleHeader header = new ShuffleHeader();
+          // TODO Review: Multiple header reads in case of status WAIT ?
+          header.readFields(input);
+          if (!header.mapId.startsWith(InputAttemptIdentifier.PATH_PREFIX)) {
+            if (!stopped) {
+              badIdErrs.increment(1);
+              LOG.warn("Invalid map id: " + header.mapId + ", expected to start with " +
+                  InputAttemptIdentifier.PATH_PREFIX + ", partition: " + header.forReduce);
+              return new InputAttemptIdentifier[]{getNextRemainingAttempt()};
+            } else {
+              if (LOG.isDebugEnabled()) {
+                LOG.debug("Already shutdown. Ignoring invalid map id error");
+              }
+              return EMPTY_ATTEMPT_ID_ARRAY;
+            }
+          }
+
+          if (header.getCompressedLength() == 0) {
+            // Empty partitions are already accounted for
+            continue;
+          }
+
+          mapOutputStat = new MapOutputStat(scheduler.getIdentifierForFetchedOutput(header.mapId, header.forReduce),
+              header.uncompressedLength,
+              header.compressedLength,
+              header.forReduce);
+          mapOutputStats.add(mapOutputStat);
+        } catch (IllegalArgumentException e) {
           if (!stopped) {
             badIdErrs.increment(1);
-            LOG.warn("Invalid map id: " + header.mapId + ", expected to start with " +
-                InputAttemptIdentifier.PATH_PREFIX + ", partition: " + header.forReduce);
-            return new InputAttemptIdentifier[] {getNextRemainingAttempt()};
+            LOG.warn("Invalid map id ", e);
+            // Don't know which one was bad, so consider this one bad and dont read
+            // the remaining because we dont know where to start reading from. YARN-1773
+            return new InputAttemptIdentifier[]{getNextRemainingAttempt()};
           } else {
             if (LOG.isDebugEnabled()) {
-              LOG.debug("Already shutdown. Ignoring invalid map id error");
+              LOG.debug("Already shutdown. Ignoring invalid map id error. Exception: " +
+                  e.getClass().getName() + ", Message: " + e.getMessage());
             }
             return EMPTY_ATTEMPT_ID_ARRAY;
           }
         }
-        srcAttemptId = 
-            scheduler.getIdentifierForFetchedOutput(header.mapId, header.forReduce);
-        compressedLength = header.compressedLength;
-        decompressedLength = header.uncompressedLength;
-        forReduce = header.forReduce;
-      } catch (IllegalArgumentException e) {
-        if (!stopped) {
-          badIdErrs.increment(1);
-          LOG.warn("Invalid map id ", e);
-          // Don't know which one was bad, so consider this one bad and dont read
-          // the remaining because we dont know where to start reading from. YARN-1773
-          return new InputAttemptIdentifier[] {getNextRemainingAttempt()};
-        } else {
-          if (LOG.isDebugEnabled()) {
-            LOG.debug("Already shutdown. Ignoring invalid map id error. Exception: " +
-                e.getClass().getName() + ", Message: " + e.getMessage());
+
+        // Do some basic sanity verification
+        if (!verifySanity(mapOutputStat.compressedLength, mapOutputStat.decompressedLength, mapOutputStat.forReduce,
+            remaining, mapOutputStat.srcAttemptId)) {
+          if (!stopped) {
+            srcAttemptId = mapOutputStat.srcAttemptId;
+            if (srcAttemptId == null) {
+              srcAttemptId = getNextRemainingAttempt();
+              LOG.warn("Was expecting " + srcAttemptId + " but got null");
+            }
+            assert (srcAttemptId != null);
+            return new InputAttemptIdentifier[]{srcAttemptId};
+          } else {
+            if (LOG.isDebugEnabled()) {
+              LOG.debug("Already stopped. Ignoring verification failure.");
+            }
+            return EMPTY_ATTEMPT_ID_ARRAY;
+          }
+        }
+
+        if (LOG.isDebugEnabled()) {
+          LOG.debug("header: " + mapOutputStat.srcAttemptId + ", len: " + mapOutputStat.compressedLength +
+              ", decomp len: " + mapOutputStat.decompressedLength);
+        }
+      }
+
+      for (MapOutputStat mapOutputStat : mapOutputStats) {
+        // Get the location for the map output - either in-memory or on-disk
+        srcAttemptId = mapOutputStat.srcAttemptId;
+        decompressedLength = mapOutputStat.decompressedLength;
+        compressedLength = mapOutputStat.compressedLength;
+        try {
+          mapOutput = allocator.reserve(srcAttemptId, decompressedLength, compressedLength, id);
+        } catch (IOException e) {
+          if (!stopped) {
+            // Kill the reduce attempt
+            ioErrs.increment(1);
+            scheduler.reportLocalError(e);
+          } else {
+            if (LOG.isDebugEnabled()) {
+              LOG.debug("Already stopped. Ignoring error from merger.reserve");
+            }
           }
           return EMPTY_ATTEMPT_ID_ARRAY;
         }
-      }
 
-      // Do some basic sanity verification
-      if (!verifySanity(compressedLength, decompressedLength, forReduce,
-          remaining, srcAttemptId)) {
-        if (!stopped) {
-          if (srcAttemptId == null) {
-            LOG.warn("Was expecting " + getNextRemainingAttempt() + " but got null");
-            srcAttemptId = getNextRemainingAttempt();
-          }
-          assert (srcAttemptId != null);
-          return new InputAttemptIdentifier[]{srcAttemptId};
-        } else {
-          if (LOG.isDebugEnabled()) {
-            LOG.debug("Already stopped. Ignoring verification failure.");
-          }
+        // Check if we can shuffle *now* ...
+        if (mapOutput.getType() == Type.WAIT) {
+          LOG.info("fetcher#" + id + " - MergerManager returned Status.WAIT ...");
+          //Not an error but wait to process data.
           return EMPTY_ATTEMPT_ID_ARRAY;
         }
-      }
-      
-      if(LOG.isDebugEnabled()) {
-        LOG.debug("header: " + srcAttemptId + ", len: " + compressedLength + 
-            ", decomp len: " + decompressedLength);
-      }
 
-      // Get the location for the map output - either in-memory or on-disk
-      try {
-        mapOutput = allocator.reserve(srcAttemptId, decompressedLength, compressedLength, id);
-      } catch (IOException e) {
-        if (!stopped) {
-          // Kill the reduce attempt
-          ioErrs.increment(1);
-          scheduler.reportLocalError(e);
-        } else {
-          if (LOG.isDebugEnabled()) {
-            LOG.debug("Already stopped. Ignoring error from merger.reserve");
-          }
+        // Go!
+        if (LOG.isDebugEnabled()) {
+          LOG.debug("fetcher#" + id + " about to shuffle output of map " +
+              mapOutput.getAttemptIdentifier() + " decomp: " +
+              decompressedLength + " len: " + compressedLength + " to " + mapOutput.getType());
         }
-        return EMPTY_ATTEMPT_ID_ARRAY;
-      }
-      
-      // Check if we can shuffle *now* ...
-      if (mapOutput.getType() == Type.WAIT) {
-        LOG.info("fetcher#" + id + " - MergerManager returned Status.WAIT ...");
-        //Not an error but wait to process data.
-        return EMPTY_ATTEMPT_ID_ARRAY;
-      } 
-      
-      // Go!
-      if (LOG.isDebugEnabled()) {
-        LOG.debug("fetcher#" + id + " about to shuffle output of map " +
-            mapOutput.getAttemptIdentifier() + " decomp: " +
-            decompressedLength + " len: " + compressedLength + " to " + mapOutput.getType());
-      }
 
-      if (mapOutput.getType() == Type.MEMORY) {
-        ShuffleUtils.shuffleToMemory(mapOutput.getMemory(), input,
-          (int) decompressedLength, (int) compressedLength, codec, ifileReadAhead,
-          ifileReadAheadLength, LOG, mapOutput.getAttemptIdentifier().toString());
-      } else if (mapOutput.getType() == Type.DISK) {
-        ShuffleUtils.shuffleToDisk(mapOutput.getDisk(), host.getHostIdentifier(),
-          input, compressedLength, decompressedLength, LOG, mapOutput.getAttemptIdentifier().toString());
-      } else {
-        throw new IOException("Unknown mapOutput type while fetching shuffle data:" +
-            mapOutput.getType());
+        if (mapOutput.getType() == Type.MEMORY) {
+          ShuffleUtils.shuffleToMemory(mapOutput.getMemory(), input,
+              (int) decompressedLength, (int) compressedLength, codec, ifileReadAhead,
+              ifileReadAheadLength, LOG, mapOutput.getAttemptIdentifier());
+        } else if (mapOutput.getType() == Type.DISK) {
+          ShuffleUtils.shuffleToDisk(mapOutput.getDisk(), host.getHostIdentifier(),
+              input, compressedLength, decompressedLength, LOG,
+              mapOutput.getAttemptIdentifier(),
+              ifileReadAhead, ifileReadAheadLength, verifyDiskChecksum);
+        } else {
+          throw new IOException("Unknown mapOutput type while fetching shuffle data:" +
+              mapOutput.getType());
+        }
+
+        // Inform the shuffle scheduler
+        long endTime = System.currentTimeMillis();
+        // Reset retryStartTime as map task make progress if retried before.
+        retryStartTime = 0;
+
+        scheduler.copySucceeded(srcAttemptId, host, compressedLength, decompressedLength,
+            endTime - startTime, mapOutput, false);
       }
-
-      // Inform the shuffle scheduler
-      long endTime = System.currentTimeMillis();
-      // Reset retryStartTime as map task make progress if retried before.
-      retryStartTime = 0;
-
-      scheduler.copySucceeded(srcAttemptId, host, compressedLength, decompressedLength,
-                              endTime - startTime, mapOutput, false);
-      // Note successful shuffle
-      remaining.remove(srcAttemptId.toString());
-      metrics.successFetch();
-      return null;
-    } catch (IOException ioe) {
+      remaining.remove(inputAttemptIdentifier.toString());
+    } catch(IOException | InternalError ioe) {
       if (stopped) {
         if (LOG.isDebugEnabled()) {
           LOG.debug("Not reporting fetch failure for exception during data copy: ["
@@ -532,23 +587,23 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
       }
       ioErrs.increment(1);
       if (srcAttemptId == null || mapOutput == null) {
-        LOG.info("fetcher#" + id + " failed to read map header" + 
-                 srcAttemptId + " decomp: " + 
-                 decompressedLength + ", " + compressedLength, ioe);
-        if(srcAttemptId == null) {
+        LOG.info("fetcher#" + id + " failed to read map header" +
+            srcAttemptId + " decomp: " +
+            decompressedLength + ", " + compressedLength, ioe);
+        if (srcAttemptId == null) {
           return remaining.values().toArray(new InputAttemptIdentifier[remaining.values().size()]);
         } else {
-          return new InputAttemptIdentifier[] {srcAttemptId};
+          return new InputAttemptIdentifier[]{srcAttemptId};
         }
       }
-      LOG.warn("Failed to shuffle output of " + srcAttemptId + 
-               " from " + host.getHostIdentifier(), ioe); 
+      LOG.warn("Failed to shuffle output of " + srcAttemptId +
+          " from " + host.getHostIdentifier(), ioe);
 
       // Inform the shuffle-scheduler
       mapOutput.abort();
-      metrics.failedFetch();
       return new InputAttemptIdentifier[] {srcAttemptId};
     }
+    return null;
   }
 
   /**
@@ -559,7 +614,7 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
    * @return true to indicate connection retry. false otherwise.
    * @throws IOException
    */
-  private boolean shouldRetry(MapHost host, IOException ioe) {
+  private boolean shouldRetry(MapHost host, Throwable ioe) {
     if (!(ioe instanceof SocketTimeoutException)) {
       return false;
     }
@@ -603,21 +658,13 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
 
     // partitionId verification. Isn't availalbe here because it is encoded into
     // URI
-    if (forReduce != currentPartition) {
+    if (forReduce < minPartition || forReduce > maxPartition) {
       wrongReduceErrs.increment(1);
       LOG.warn(logIdentifier + " data for the wrong partition map: " + srcAttemptId + " len: "
           + compressedLength + " decomp len: " + decompressedLength + " for partition " + forReduce
-          + ", expected partition: " + currentPartition);
+          + ", expected partition range: " + minPartition + "-" + maxPartition);
       return false;
     }
-
-    // Sanity check
-    if (remaining.get(srcAttemptId.toString()) == null) {
-      wrongMapErrs.increment(1);
-      LOG.warn("Invalid map-output! Received output for " + srcAttemptId);
-      return false;
-    }
-    
     return true;
   }
   
@@ -642,7 +689,7 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
 
     if(LOG.isDebugEnabled()) {
       LOG.debug("Fetcher " + id + " going to fetch (local disk) from " + host + " for: "
-          + srcAttempts + ", partitionId: " + currentPartition);
+          + srcAttempts + ", partition range: " + minPartition + "-" + maxPartition);
     }
 
     // List of maps to be fetched yet
@@ -657,37 +704,48 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
         }
         InputAttemptIdentifier srcAttemptId = iter.next();
         MapOutput mapOutput = null;
-        try {
-          long startTime = System.currentTimeMillis();
-          Path filename = getShuffleInputFileName(srcAttemptId.getPathComponent(), null);
+        boolean hasFailures = false;
+        // Fetch partition count number of map outputs (handles auto-reduce case)
+        for (int curPartition = minPartition; curPartition <= maxPartition; curPartition++) {
+          try {
+            long startTime = System.currentTimeMillis();
 
-          TezIndexRecord indexRecord = getIndexRecord(srcAttemptId.getPathComponent(),
-              currentPartition);
-
-          mapOutput = getMapOutputForDirectDiskFetch(srcAttemptId, filename, indexRecord);
-          long endTime = System.currentTimeMillis();
-          scheduler.copySucceeded(srcAttemptId, host, indexRecord.getPartLength(),
-              indexRecord.getRawLength(), (endTime - startTime), mapOutput, true);
-          iter.remove();
-          metrics.successFetch();
-        } catch (IOException e) {
-          if (mapOutput != null) {
-            mapOutput.abort();
-          }
-          if (!stopped) {
-            metrics.failedFetch();
-            ioErrs.increment(1);
-            scheduler.copyFailed(srcAttemptId, host, true, false, true);
-            LOG.warn("Failed to read local disk output of " + srcAttemptId + " from " +
-                host.getHostIdentifier(), e);
-          } else {
-            if (LOG.isDebugEnabled()) {
-              LOG.debug(
-                  "Ignoring fetch error during local disk copy since fetcher has already been stopped");
+            // Partition id is the base partition id plus the relative offset
+            int reduceId = host.getPartitionId() + curPartition - minPartition;
+            srcAttemptId = scheduler.getIdentifierForFetchedOutput(srcAttemptId.getPathComponent(), reduceId);
+            Path filename = getShuffleInputFileName(srcAttemptId.getPathComponent(), null);
+            TezIndexRecord indexRecord = getIndexRecord(srcAttemptId.getPathComponent(), reduceId);
+            if(!indexRecord.hasData()) {
+              continue;
             }
-            return;
+
+            mapOutput = getMapOutputForDirectDiskFetch(srcAttemptId, filename, indexRecord);
+            long endTime = System.currentTimeMillis();
+            scheduler.copySucceeded(srcAttemptId, host, indexRecord.getPartLength(),
+                indexRecord.getRawLength(), (endTime - startTime), mapOutput, true);
+          } catch (IOException | InternalError e) {
+            if (mapOutput != null) {
+              mapOutput.abort();
+            }
+            if (!stopped) {
+              hasFailures = true;
+              ioErrs.increment(1);
+              scheduler.copyFailed(srcAttemptId, host, true, false, true);
+              LOG.warn("Failed to read local disk output of " + srcAttemptId + " from " +
+                  host.getHostIdentifier(), e);
+            } else {
+              if (LOG.isDebugEnabled()) {
+                LOG.debug(
+                    "Ignoring fetch error during local disk copy since fetcher has already been stopped");
+              }
+              return;
+            }
+
           }
         }
+        if (!hasFailures) {
+          iter.remove();
+        }
       }
     } finally {
       putBackRemainingMapOutputs(host);
@@ -701,9 +759,10 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
       throws IOException {
     LocalDirAllocator localDirAllocator = new LocalDirAllocator(TezRuntimeFrameworkConfigs.LOCAL_DIRS);
     suffix = suffix != null ? suffix : "";
-
-    String pathFromLocalDir = Constants.TEZ_RUNTIME_TASK_OUTPUT_DIR + Path.SEPARATOR +
-        pathComponent + Path.SEPARATOR + Constants.TEZ_RUNTIME_TASK_OUTPUT_FILENAME_STRING + suffix;
+    String outputPath = Constants.TEZ_RUNTIME_TASK_OUTPUT_DIR + Path.SEPARATOR +
+        pathComponent + Path.SEPARATOR +
+        Constants.TEZ_RUNTIME_TASK_OUTPUT_FILENAME_STRING + suffix;
+    String pathFromLocalDir = getPathForLocalDir(outputPath);
 
     return localDirAllocator.getLocalPathToRead(pathFromLocalDir.toString(), conf);
   }
@@ -735,5 +794,11 @@ class FetcherOrderedGrouped extends CallableWithNdc<Void> {
     }
   }
 
+  private String getPathForLocalDir(String suffix) {
+    if(ShuffleUtils.isTezShuffleHandler(conf)) {
+      return Constants.DAG_PREFIX + dagId + Path.SEPARATOR + suffix;
+    }
+    return suffix;
+  }
 }
 
