@@ -30,11 +30,13 @@ import org.apache.hadoop.fs.FileSystem;
 import org.apache.hadoop.fs.LocalDirAllocator;
 import org.apache.hadoop.fs.LocalFileSystem;
 import org.apache.hadoop.fs.Path;
+import org.apache.hadoop.io.BoundedByteArrayOutputStream;
 import org.apache.hadoop.io.DataInputBuffer;
 import org.apache.hadoop.io.FileChunk;
 import org.apache.hadoop.io.RawComparator;
 import org.apache.hadoop.io.compress.CompressionCodec;
 import org.apache.hadoop.util.Progressable;
+import org.apache.hadoop.util.Time;
 import org.apache.tez.common.TezUtilsInternal;
 import org.apache.tez.common.counters.TaskCounter;
 import org.apache.tez.common.counters.TezCounter;
@@ -48,10 +50,10 @@ import org.apache.tez.runtime.library.common.combine.Combiner;
 import org.apache.tez.runtime.library.common.sort.impl.IFile;
 import org.apache.tez.runtime.library.common.sort.impl.IFile.Writer;
 import org.apache.tez.runtime.library.common.sort.impl.TezMerger;
+import org.apache.tez.runtime.library.common.sort.impl.TezMerger.DiskSegment;
 import org.apache.tez.runtime.library.common.sort.impl.TezMerger.Segment;
 import org.apache.tez.runtime.library.common.sort.impl.TezRawKeyValueIterator;
 import org.apache.tez.runtime.library.common.task.local.output.TezTaskOutputFiles;
-import org.apache.tez.runtime.library.hadoop.compat.NullProgressable;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
@@ -60,6 +62,7 @@ import java.util.ArrayList;
 import java.util.Collection;
 import java.util.Collections;
 import java.util.Comparator;
+import java.util.Iterator;
 import java.util.List;
 import java.util.Set;
 import java.util.TreeSet;
@@ -92,11 +95,13 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
   };
   private final Combiner combiner;  
   
-  private final Set<MapOutput> inMemoryMergedMapOutputs = 
+  @VisibleForTesting
+  final Set<MapOutput> inMemoryMergedMapOutputs =
     new TreeSet<MapOutput>(new MapOutput.MapOutputComparator());
   private final IntermediateMemoryToMemoryMerger memToMemMerger;
 
-  private final Set<MapOutput> inMemoryMapOutputs = 
+  @VisibleForTesting
+  final Set<MapOutput> inMemoryMapOutputs =
     new TreeSet<MapOutput>(new MapOutput.MapOutputComparator());
   private final InMemoryMerger inMemoryMerger;
 
@@ -143,6 +148,12 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
   private final int ifileReadAheadLength;
   private final int ifileBufferSize;
 
+  // Variables for stats and logging
+  private long lastInMemSegmentLogTime = -1L;
+  private final SegmentStatsTracker statsInMemTotal = new SegmentStatsTracker();
+  private final SegmentStatsTracker statsInMemLastLog = new SegmentStatsTracker();
+
+
   private AtomicInteger mergeFileSequenceId = new AtomicInteger(0);
 
   private final boolean cleanup;
@@ -174,8 +185,10 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
     this.reduceCombineInputCounter = reduceCombineInputCounter;
     this.spilledRecordsCounter = spilledRecordsCounter;
     this.mergedMapOutputsCounter = mergedMapOutputsCounter;
-    this.mapOutputFile = new TezTaskOutputFiles(conf, inputContext.getUniqueIdentifier());
-    
+    this.mapOutputFile = new TezTaskOutputFiles(conf,
+        inputContext.getUniqueIdentifier(),
+        inputContext.getDagIdentifier());
+
     this.localFS = localFS;
     this.rfs = ((LocalFileSystem)localFS).getRaw();
     
@@ -448,22 +461,25 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
 
   @Override
   public synchronized void unreserve(long size) {
-    commitMemory -= size;
     usedMemory -= size;
     if (LOG.isDebugEnabled()) {
-      LOG.debug("Notifying unreserve : commitMemory=" + commitMemory + ", usedMemory=" + usedMemory
+      LOG.debug("Notifying unreserve : size=" + size + ", commitMemory=" + commitMemory + ", usedMemory=" + usedMemory
           + ", mergeThreshold=" + mergeThreshold);
     }
     notifyAll();
   }
 
+  @Override
+  public synchronized void releaseCommittedMemory(long size) {
+    commitMemory -= size;
+    unreserve(size);
+  }
+
+
   @Override
   public synchronized void closeInMemoryFile(MapOutput mapOutput) { 
     inMemoryMapOutputs.add(mapOutput);
-    LOG.info("closeInMemoryFile -> map-output of size: " + mapOutput.getSize()
-          + ", inMemoryMapOutputs.size() -> " + inMemoryMapOutputs.size()
-          + ", commitMemory -> " + commitMemory + ", usedMemory ->" + usedMemory + ", mapOutput=" +
-          mapOutput);
+    trackAndLogCloseInMemoryFile(mapOutput);
 
     commitMemory+= mapOutput.getSize();
 
@@ -482,6 +498,44 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
     }
   }
 
+  private void trackAndLogCloseInMemoryFile(MapOutput mapOutput) {
+    statsInMemTotal.updateStats(mapOutput.getSize());
+
+    if (LOG.isDebugEnabled()) {
+      LOG.debug("closeInMemoryFile -> map-output of size: " + mapOutput.getSize()
+          + ", inMemoryMapOutputs.size() -> " + inMemoryMapOutputs.size()
+          + ", commitMemory -> " + commitMemory + ", usedMemory ->" +
+          usedMemory + ", mapOutput=" +
+          mapOutput);
+    } else {
+      statsInMemLastLog.updateStats(mapOutput.getSize());
+      long now = Time.monotonicNow();
+      if (now > lastInMemSegmentLogTime + 30 * 1000L) {
+        LOG.info(
+            "CloseInMemoryFile. Current state: inMemoryMapOutputs.size={}," +
+                " commitMemory={}," +
+                " usedMemory={}. Since last log:" +
+                " count={}," +
+                " min={}," +
+                " max={}," +
+                " total={}," +
+                " avg={}",
+            inMemoryMapOutputs.size(),
+            commitMemory,
+            usedMemory,
+            statsInMemLastLog.count,
+            statsInMemLastLog.minSize,
+            statsInMemLastLog.maxSize,
+            statsInMemLastLog.size,
+            (statsInMemLastLog.count == 0 ? "nan" :
+                (statsInMemLastLog.size / (double) statsInMemLastLog.count))
+        );
+        statsInMemLastLog.reset();
+        lastInMemSegmentLogTime = now;
+      }
+    }
+  }
+
   private void startMemToDiskMerge() {
     synchronized (inMemoryMerger) {
       if (!inMemoryMerger.isInProgress()) {
@@ -497,9 +551,24 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
   
   public synchronized void closeInMemoryMergedFile(MapOutput mapOutput) {
     inMemoryMergedMapOutputs.add(mapOutput);
-    LOG.info("closeInMemoryMergedFile -> size: " + mapOutput.getSize() +
-             ", inMemoryMergedMapOutputs.size() -> " + 
-             inMemoryMergedMapOutputs.size());
+    if (LOG.isDebugEnabled()) {
+      // This log could be moved to INFO level for a while, after mem-to-mem
+      // merge is production ready.
+      LOG.debug("closeInMemoryMergedFile -> size: " + mapOutput.getSize() +
+          ", inMemoryMergedMapOutputs.size() -> " +
+          inMemoryMergedMapOutputs.size());
+    }
+
+    commitMemory += mapOutput.getSize();
+
+    if (commitMemory >= mergeThreshold) {
+      startMemToDiskMerge();
+    }
+  }
+
+  @Override
+  public FileSystem getLocalFileSystem() {
+    return localFS;
   }
 
   @Override
@@ -516,9 +585,7 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
     }
 
     onDiskMapOutputs.add(file);
-    if (LOG.isDebugEnabled()) {
-      LOG.debug("close onDiskFile=" + file.getPath() + ", len=" + file.getLength());
-    }
+    logCloseOnDiskFile(file);
 
     synchronized (onDiskMerger) {
       if (!onDiskMerger.isInProgress() &&
@@ -528,6 +595,23 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
     }
   }
 
+  private long lastOnDiskSegmentLogTime = -1L;
+  private void logCloseOnDiskFile(FileChunk file) {
+    if (LOG.isDebugEnabled()) {
+      LOG.debug(
+          "close onDiskFile=" + file.getPath() + ", len=" + file.getLength() +
+              ", onDisMapOutputs=" + onDiskMapOutputs.size());
+    } else {
+      long now = Time.monotonicNow();
+      if (now > lastOnDiskSegmentLogTime + 30 * 1000L) {
+        LOG.info(
+            "close onDiskFile. State: NumOnDiskFiles={}. Current: path={}, len={}",
+            onDiskMapOutputs.size(), file.getPath(), file.getLength());
+        lastOnDiskSegmentLogTime = now;
+      }
+    }
+  }
+
   /**
    * Should <b>only</b> be used after the Shuffle phaze is complete, otherwise can
    * return an invalid state since a merge may not be in progress dur to
@@ -540,9 +624,7 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
     return finalMergeComplete;
   }
 
-  public TezRawKeyValueIterator close() throws Throwable {
-    // TODO TEZ-2756. Don't attempt a final merge if close is invoked as a result of a previous
-    // shuffle exception / error.
+  public TezRawKeyValueIterator close(boolean tryFinalMerge) throws Throwable {
     if (!isShutdown.getAndSet(true)) {
       // Wait for on-going merges to complete
       if (memToMemMerger != null) {
@@ -558,18 +640,31 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
       inMemoryMapOutputs.clear();
       List<FileChunk> disk = new ArrayList<FileChunk>(onDiskMapOutputs);
       onDiskMapOutputs.clear();
-      try {
-        TezRawKeyValueIterator kvIter = finalMerge(conf, rfs, memory, disk);
-        this.finalMergeComplete = true;
-        return kvIter;
-      } catch (InterruptedException e) {
-        //Cleanup the disk segments
-        if (cleanup) {
-          cleanup(localFS, disk);
-          cleanup(localFS, onDiskMapOutputs);
+
+      if (statsInMemTotal.count > 0) {
+        LOG.info(
+            "TotalInMemFetchStats: count={}, totalSize={}, min={}, max={}, avg={}",
+            statsInMemTotal.count, statsInMemTotal.size,
+            statsInMemTotal.minSize, statsInMemTotal.maxSize,
+            (statsInMemTotal.size / (float) statsInMemTotal.size));
+      }
+
+      // Don't attempt a final merge if close is invoked as a result of a previous
+      // shuffle exception / error.
+      if (tryFinalMerge) {
+        try {
+          TezRawKeyValueIterator kvIter = finalMerge(conf, rfs, memory, disk);
+          this.finalMergeComplete = true;
+          return kvIter;
+        } catch (InterruptedException e) {
+          //Cleanup the disk segments
+          if (cleanup) {
+            cleanup(localFS, disk);
+            cleanup(localFS, onDiskMapOutputs);
+          }
+          Thread.currentThread().interrupt(); //reset interrupt status
+          throw e;
         }
-        Thread.currentThread().interrupt(); //reset interrupt status
-        throw e;
       }
     }
     return null;
@@ -633,19 +728,64 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
 
       InputAttemptIdentifier dummyMapId = inputs.get(0).getAttemptIdentifier(); 
       List<Segment> inMemorySegments = new ArrayList<Segment>();
-      long mergeOutputSize = 
-        createInMemorySegments(inputs, inMemorySegments, 0);
+
+      MapOutput mergedMapOutputs = null;
+
+      long mergeOutputSize = 0l;
+      //Lock manager so that fetcher threads can not change the mem size
+      synchronized (manager) {
+
+        Iterator<MapOutput> it = inputs.iterator();
+        MapOutput lastAddedMapOutput = null;
+        while(it.hasNext() && !Thread.currentThread().isInterrupted()) {
+          MapOutput mo = it.next();
+          if ((mergeOutputSize + mo.getSize() + manager.getUsedMemory()) > memoryLimit) {
+            //Search for smaller segments that can fit into existing mem
+            if (LOG.isDebugEnabled()) {
+              LOG.debug("Size is greater than usedMemory. "
+                  + "mergeOutputSize=" + mergeOutputSize
+                  + ", moSize=" + mo.getSize()
+                  + ", usedMemory=" + manager.getUsedMemory()
+                  + ", memoryLimit=" + memoryLimit);
+            }
+            continue;
+          } else {
+            mergeOutputSize += mo.getSize();
+            IFile.Reader reader = new InMemoryReader(MergeManager.this,
+                mo.getAttemptIdentifier(), mo.getMemory(), 0, mo.getMemory().length);
+            inMemorySegments.add(new Segment(reader,
+                (mo.isPrimaryMapOutput() ? mergedMapOutputsCounter : null)));
+            lastAddedMapOutput = mo;
+            it.remove();
+            LOG.debug("Added segment for merging. mergeOutputSize=" + mergeOutputSize);
+          }
+        }
+
+        //Add any unused MapOutput back
+        inMemoryMapOutputs.addAll(inputs);
+
+        //Exit early, if 0 or 1 segment is available
+        if (inMemorySegments.size() <= 1) {
+          if (lastAddedMapOutput != null) {
+            inMemoryMapOutputs.add(lastAddedMapOutput);
+          }
+          return;
+        }
+
+        mergedMapOutputs = unconditionalReserve(dummyMapId, mergeOutputSize, false);
+      }
+
       int noInMemorySegments = inMemorySegments.size();
 
-      MapOutput mergedMapOutputs =
-        unconditionalReserve(dummyMapId, mergeOutputSize, false);
-      
-      Writer writer = 
-        new InMemoryWriter(mergedMapOutputs.getArrayStream());
+      Writer writer = new InMemoryWriter(mergedMapOutputs.getMemory());
 
       LOG.info(inputContext.getSourceVertexName() + ": " + "Initiating Memory-to-Memory merge with " + noInMemorySegments +
                " segments of total-size: " + mergeOutputSize);
 
+      if (Thread.currentThread().isInterrupted()) {
+        return; // early exit
+      }
+
       // Nothing will be materialized to disk because the sort factor is being
       // set to the number of in memory segments.
       // TODO Is this doing any combination ?
@@ -662,7 +802,7 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
 
       LOG.info(inputContext.getSourceVertexName() +
                " Memory-to-Memory merge of the " + noInMemorySegments +
-               " files in-memory complete.");
+               " files in-memory complete with mergeOutputSize=" + mergeOutputSize);
 
       // Note the output of the merge
       closeInMemoryMergedFile(mergedMapOutputs);
@@ -717,7 +857,7 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
       srcTaskIdentifier = inputs.get(0).getAttemptIdentifier();
 
       List<Segment> inMemorySegments = new ArrayList<Segment>();
-      long mergeOutputSize = 
+      long mergeOutputSize =
         createInMemorySegments(inputs, inMemorySegments,0);
       int noInMemorySegments = inMemorySegments.size();
 
@@ -726,7 +866,7 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
       // All disk writes done by this merge are overhead - due to the lack of
       // adequate memory to keep all segments in memory.
       outputPath = mapOutputFile.getInputFileForWrite(
-          srcTaskIdentifier.getInputIdentifier().getInputIndex(), srcTaskIdentifier.getSpillEventId(),
+          srcTaskIdentifier.getInputIdentifier(), srcTaskIdentifier.getSpillEventId(),
           mergeOutputSize).suffix(Constants.MERGED_OUTPUT_PREFIX);
 
       Writer writer = null;
@@ -848,7 +988,7 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
         }
         final Path file = fileChunk.getPath();
         approxOutputSize += size;
-        Segment segment = new Segment(rfs, file, offset, size, codec, ifileReadAhead,
+        DiskSegment segment = new DiskSegment(rfs, file, offset, size, codec, ifileReadAhead,
             ifileReadAheadLength, ifileBufferSize, preserve);
         inputSegments.add(segment);
       }
@@ -863,7 +1003,7 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
       if (file0.isLocalFile()) {
         // This is setup the same way a type DISK MapOutput is setup when fetching.
         namePart = mapOutputFile.getSpillFileName(
-            file0.getInputAttemptIdentifier().getInputIdentifier().getInputIndex(),
+            file0.getInputAttemptIdentifier().getInputIdentifier(),
             file0.getInputAttemptIdentifier().getSpillEventId());
       } else {
         namePart = file0.getPath().getName().toString();
@@ -936,10 +1076,11 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
     // closed but not yet present in inMemoryMapOutputs
     long fullSize = 0L;
     for (MapOutput mo : inMemoryMapOutputs) {
-      fullSize += mo.getMemory().length;
+      fullSize += mo.getSize();
     }
+    int inMemoryMapOutputsOffset = 0;
     while((fullSize > leaveBytes) && !Thread.currentThread().isInterrupted()) {
-      MapOutput mo = inMemoryMapOutputs.remove(0);
+      MapOutput mo = inMemoryMapOutputs.get(inMemoryMapOutputsOffset++);
       byte[] data = mo.getMemory();
       long size = data.length;
       totalSize += size;
@@ -947,10 +1088,12 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
       IFile.Reader reader = new InMemoryReader(MergeManager.this, 
                                                    mo.getAttemptIdentifier(),
                                                    data, 0, (int)size);
-      inMemorySegments.add(new Segment(reader, true, 
+      inMemorySegments.add(new Segment(reader,
                                             (mo.isPrimaryMapOutput() ? 
                                             mergedMapOutputsCounter : null)));
     }
+    // Bulk remove removed in-memory map outputs efficiently
+    inMemoryMapOutputs.subList(0, inMemoryMapOutputsOffset).clear();
     return totalSize;
   }
 
@@ -1002,21 +1145,9 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
                                        List<MapOutput> inMemoryMapOutputs,
                                        List<FileChunk> onDiskMapOutputs
                                        ) throws IOException, InterruptedException {
-    LOG.info("finalMerge called with " + 
-             inMemoryMapOutputs.size() + " in-memory map-outputs and " + 
-             onDiskMapOutputs.size() + " on-disk map-outputs");
 
-    if (LOG.isDebugEnabled()) {
-      for (MapOutput inMemoryMapOutput : inMemoryMapOutputs) {
-        LOG.debug("inMemoryOutput=" + inMemoryMapOutput + ", size=" + inMemoryMapOutput
-            .getSize());
-      }
-
-      for (FileChunk onDiskMapOutput : onDiskMapOutputs) {
-        LOG.debug("onDiskMapOutput=" + onDiskMapOutput.getPath() + ", size=" + onDiskMapOutput
-                .getLength());
-      }
-    }
+    logFinalMergeStart(inMemoryMapOutputs, onDiskMapOutputs);
+    StringBuilder finalMergeLog = new StringBuilder();
     
     inputContext.notifyProgress();
 
@@ -1032,7 +1163,7 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
     long inMemToDiskBytes = 0;
     boolean mergePhaseFinished = false;
     if (inMemoryMapOutputs.size() > 0) {
-      int srcTaskId = inMemoryMapOutputs.get(0).getAttemptIdentifier().getInputIdentifier().getInputIndex();
+      int srcTaskId = inMemoryMapOutputs.get(0).getAttemptIdentifier().getInputIdentifier();
       inMemToDiskBytes = createInMemorySegments(inMemoryMapOutputs,
                                                 memDiskSegments,
                                                 this.postMergeMemLimit);
@@ -1081,15 +1212,25 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
         // add to list of final disk outputs.
         onDiskMapOutputs.add(new FileChunk(outputPath, 0, fStatus.getLen()));
 
-        LOG.info("Merged " + numMemDiskSegments + " segments, " +
-                 inMemToDiskBytes + " bytes to disk to satisfy " +
-                 "reduce memory limit. outputPath=" + outputPath);
+        if (LOG.isInfoEnabled()) {
+          finalMergeLog.append("MemMerged: " + numMemDiskSegments + ", " + inMemToDiskBytes);
+          if (LOG.isDebugEnabled()) {
+            LOG.debug("Merged " + numMemDiskSegments + "segments, size=" +
+                inMemToDiskBytes + " to " + outputPath);
+          }
+        }
+
         inMemToDiskBytes = 0;
         memDiskSegments.clear();
       } else if (inMemToDiskBytes != 0) {
-        LOG.info("Keeping " + numMemDiskSegments + " segments, " +
-            inMemToDiskBytes + " bytes in memory for " +
-                 "intermediate, on-disk merge");
+        if (LOG.isInfoEnabled()) {
+          finalMergeLog.append("DelayedMemMerge: " + numMemDiskSegments + ", " + inMemToDiskBytes);
+          if (LOG.isDebugEnabled()) {
+            LOG.debug("Keeping " + numMemDiskSegments + " segments, " +
+                inMemToDiskBytes + " bytes in memory for " +
+                "intermediate, on-disk merge");
+          }
+        }
       }
     }
 
@@ -1100,8 +1241,11 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
     for (FileChunk fileChunk : onDisk) {
       final long fileLength = fileChunk.getLength();
       onDiskBytes += fileLength;
-      LOG.info("Disk file=" + fileChunk.getPath() + ", len=" + fileLength + ", isLocal=" +
-          fileChunk.isLocalFile());
+      if (LOG.isDebugEnabled()) {
+        LOG.debug("Disk file=" + fileChunk.getPath() + ", len=" + fileLength +
+            ", isLocal=" +
+            fileChunk.isLocalFile());
+      }
 
       final Path file = fileChunk.getPath();
       TezCounter counter =
@@ -1109,11 +1253,16 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
 
       final long fileOffset = fileChunk.getOffset();
       final boolean preserve = fileChunk.isLocalFile();
-      diskSegments.add(new Segment(fs, file, fileOffset, fileLength, codec, ifileReadAhead,
+      diskSegments.add(new DiskSegment(fs, file, fileOffset, fileLength, codec, ifileReadAhead,
                                    ifileReadAheadLength, ifileBufferSize, preserve, counter));
     }
-    LOG.info("Merging " + onDisk.length + " files, " +
-             onDiskBytes + " bytes from disk");
+    if (LOG.isInfoEnabled()) {
+      finalMergeLog.append(". DiskSeg: " + onDisk.length + ", " + onDiskBytes);
+      if (LOG.isDebugEnabled()) {
+        LOG.debug("Merging " + onDisk.length + " files, " +
+            onDiskBytes + " bytes from disk");
+      }
+    }
     Collections.sort(diskSegments, new Comparator<Segment>() {
       public int compare(Segment o1, Segment o2) {
         if (o1.getLength() == o2.getLength()) {
@@ -1127,8 +1276,14 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
     List<Segment> finalSegments = new ArrayList<Segment>();
     long inMemBytes = createInMemorySegments(inMemoryMapOutputs, 
                                              finalSegments, 0);
-    LOG.info("Merging " + finalSegments.size() + " segments, " +
-             inMemBytes + " bytes from memory into reduce");
+    if (LOG.isInfoEnabled()) {
+      finalMergeLog.append(". MemSeg: " + finalSegments.size() + ", " + inMemBytes);
+      if (LOG.isDebugEnabled()) {
+        LOG.debug("Merging " + finalSegments.size() + " segments, " +
+            inMemBytes + " bytes from memory into reduce");
+      }
+    }
+
     if (0 != onDiskBytes) {
       final int numInMemSegments = memDiskSegments.size();
       diskSegments.addAll(0, memDiskSegments);
@@ -1142,7 +1297,10 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
         return diskMerge;
       }
       finalSegments.add(new Segment(
-            new RawKVIteratorReader(diskMerge, onDiskBytes), true));
+            new RawKVIteratorReader(diskMerge, onDiskBytes), null));
+    }
+    if (LOG.isInfoEnabled()) {
+      LOG.info(finalMergeLog.toString());
     }
     // This is doing nothing but creating an iterator over the segments.
     return TezMerger.merge(job, fs, keyClass, valueClass,
@@ -1150,4 +1308,75 @@ public class MergeManager implements FetchedInputAllocatorOrderedGrouped {
                  comparator, progressable, spilledRecordsCounter, null,
                  additionalBytesRead, null);
   }
+
+
+  private void logFinalMergeStart(List<MapOutput> inMemoryMapOutputs,
+                                  List<FileChunk> onDiskMapOutputs) {
+    long inMemSegmentSize = 0;
+    for (MapOutput inMemoryMapOutput : inMemoryMapOutputs) {
+      inMemSegmentSize += inMemoryMapOutput.getSize();
+
+      if (LOG.isDebugEnabled()) {
+        LOG.debug("finalMerge: inMemoryOutput=" + inMemoryMapOutput + ", size=" +
+            inMemoryMapOutput.getSize());
+      }
+    }
+    long onDiskSegmentSize = 0;
+    for (FileChunk onDiskMapOutput : onDiskMapOutputs) {
+      onDiskSegmentSize += onDiskMapOutput.getLength();
+
+      if (LOG.isDebugEnabled()) {
+        LOG.debug("finalMerge: onDiskMapOutput=" + onDiskMapOutput.getPath() +
+            ", size=" + onDiskMapOutput.getLength());
+      }
+    }
+
+    LOG.info(
+        "finalMerge with #inMemoryOutputs={}, size={} and #onDiskOutputs={}, size={}",
+        inMemoryMapOutputs.size(), inMemSegmentSize, onDiskMapOutputs.size(),
+        onDiskSegmentSize);
+
+  }
+
+  @VisibleForTesting
+  long getCommitMemory() {
+    return commitMemory;
+  }
+
+  @VisibleForTesting
+  synchronized long getUsedMemory() {
+    return usedMemory;
+  }
+
+  @VisibleForTesting
+  void waitForMemToMemMerge() throws InterruptedException {
+    memToMemMerger.waitForMerge();
+  }
+
+
+
+  private static class SegmentStatsTracker {
+    private long size;
+    private int count;
+    private long minSize;
+    private long maxSize;
+
+    SegmentStatsTracker() {
+      reset();
+    }
+
+    void updateStats(long segSize) {
+      size += segSize;
+      count++;
+      minSize = (segSize < minSize ? segSize : minSize);
+      maxSize = (segSize > maxSize ? segSize : maxSize);
+    }
+
+    void reset() {
+      size = 0L;
+      count = 0;
+      minSize = Long.MAX_VALUE;
+      maxSize = Long.MIN_VALUE;
+    }
+  }
 }
