@@ -46,6 +46,7 @@ import com.google.common.annotations.VisibleForTesting;
 import com.google.common.base.Preconditions;
 import com.google.common.collect.LinkedListMultimap;
 import com.google.common.collect.ListMultimap;
+import com.google.common.collect.Maps;
 import com.google.common.collect.Sets;
 import com.google.common.util.concurrent.FutureCallback;
 import com.google.common.util.concurrent.Futures;
@@ -53,12 +54,13 @@ import com.google.common.util.concurrent.ListenableFuture;
 import com.google.common.util.concurrent.ListeningExecutorService;
 import com.google.common.util.concurrent.MoreExecutors;
 import com.google.common.util.concurrent.ThreadFactoryBuilder;
+
 import org.apache.hadoop.io.compress.CompressionCodec;
-import org.apache.hadoop.security.UserGroupInformation;
+import org.apache.tez.dag.api.TezConfiguration;
 import org.apache.tez.http.HttpConnectionParams;
 import org.apache.tez.common.CallableWithNdc;
 import org.apache.tez.common.security.JobTokenSecretManager;
-import org.apache.tez.dag.api.TezConstants;
+import org.apache.tez.runtime.library.common.CompositeInputAttemptIdentifier;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 import org.apache.hadoop.conf.Configuration;
@@ -71,15 +73,61 @@ import org.apache.tez.runtime.api.InputContext;
 import org.apache.tez.runtime.api.events.InputReadErrorEvent;
 import org.apache.tez.runtime.library.api.TezRuntimeConfiguration;
 import org.apache.tez.runtime.library.common.InputAttemptIdentifier;
-import org.apache.tez.runtime.library.common.InputIdentifier;
 import org.apache.tez.runtime.library.common.TezRuntimeUtils;
 import org.apache.tez.runtime.library.common.shuffle.ShuffleUtils;
+import org.apache.tez.runtime.library.common.shuffle.ShuffleUtils.FetchStatsLogger;
+import org.apache.tez.runtime.library.common.shuffle.HostPort;
+import org.apache.tez.runtime.library.common.shuffle.orderedgrouped.MapHost.HostPortPartition;
 import org.apache.tez.runtime.library.common.shuffle.orderedgrouped.MapOutput.Type;
 
 import com.google.common.collect.Lists;
 
 class ShuffleScheduler {
 
+  public static class PathPartition {
+
+    final String path;
+    final int partition;
+
+    PathPartition(String path, int partition) {
+      this.path = path;
+      this.partition = partition;
+    }
+
+    @Override
+    public int hashCode() {
+      final int prime = 31;
+      int result = 1;
+      result = prime * result + ((path == null) ? 0 : path.hashCode());
+      result = prime * result + partition;
+      return result;
+    }
+
+    @Override
+    public boolean equals(Object obj) {
+      if (this == obj)
+        return true;
+      if (obj == null)
+        return false;
+      if (getClass() != obj.getClass())
+        return false;
+      PathPartition other = (PathPartition) obj;
+      if (path == null) {
+        if (other.path != null)
+          return false;
+      } else if (!path.equals(other.path))
+        return false;
+      if (partition != other.partition)
+        return false;
+      return true;
+    }
+
+    @Override
+    public String toString() {
+      return "PathPartition [path=" + path + ", partition=" + partition + "]";
+    }
+  }
+
   @VisibleForTesting
   enum ShuffleErrors {
     IO_ERROR,
@@ -95,6 +143,9 @@ class ShuffleScheduler {
   private final AtomicLong shuffleStart = new AtomicLong(0);
 
   private static final Logger LOG = LoggerFactory.getLogger(ShuffleScheduler.class);
+  private static final Logger LOG_FETCH = LoggerFactory.getLogger(LOG.getName() + ".fetch");
+  private static final FetchStatsLogger fetchStatsLogger = new FetchStatsLogger(LOG_FETCH, LOG);
+
   static final long INITIAL_PENALTY = 2000L; // 2 seconds
   private static final float PENALTY_GROWTH_RATE = 1.3f;
 
@@ -102,16 +153,16 @@ class ShuffleScheduler {
   private final int numInputs;
   private int numFetchedSpills;
   @VisibleForTesting
-  final Map<String, MapHost> mapLocations = new HashMap<String, MapHost>();
+  final Map<HostPortPartition, MapHost> mapLocations = new HashMap<HostPortPartition, MapHost>();
   //TODO Clean this and other maps at some point
   @VisibleForTesting
-  final ConcurrentMap<String, InputAttemptIdentifier> pathToIdentifierMap
-      = new ConcurrentHashMap<String, InputAttemptIdentifier>();
+  final ConcurrentMap<PathPartition, InputAttemptIdentifier> pathToIdentifierMap
+      = new ConcurrentHashMap<PathPartition, InputAttemptIdentifier>();
 
   //To track shuffleInfo events when finalMerge is disabled in source or pipelined shuffle is
   // enabled in source.
   @VisibleForTesting
-  final Map<InputIdentifier, ShuffleEventInfo> pipelinedShuffleInfoEventsMap;
+  final Map<Integer, ShuffleEventInfo> pipelinedShuffleInfoEventsMap;
 
   @VisibleForTesting
   final Set<MapHost> pendingHosts = new HashSet<MapHost>();
@@ -123,9 +174,8 @@ class ShuffleScheduler {
   private final Referee referee;
   @VisibleForTesting
   final Map<InputAttemptIdentifier, IntWritable> failureCounts = new HashMap<InputAttemptIdentifier,IntWritable>();
-  final Set<String> uniqueHosts = Sets.newHashSet();
-  private final Map<String,IntWritable> hostFailures = 
-    new HashMap<String,IntWritable>();
+  final Set<HostPort> uniqueHosts = Sets.newHashSet();
+  private final Map<HostPort,IntWritable> hostFailures = new HashMap<HostPort,IntWritable>();
   private final InputContext inputContext;
   private final TezCounter shuffledInputsCounter;
   private final TezCounter skippedInputCounter;
@@ -156,7 +206,6 @@ class ShuffleScheduler {
 
   private final HttpConnectionParams httpConnectionParams;
   private final FetchedInputAllocatorOrderedGrouped allocator;
-  private final ShuffleClientMetrics shuffleMetrics;
   private final ExceptionReporter exceptionReporter;
   private final MergeManager mergeManager;
   private final JobTokenSecretManager jobTokenSecretManager;
@@ -167,7 +216,10 @@ class ShuffleScheduler {
   private final boolean localDiskFetchEnabled;
   private final String localHostname;
   private final int shufflePort;
+  private final String applicationId;
+  private final int dagId;
   private final boolean asyncHttp;
+  private final boolean sslShuffle;
 
   private final TezCounter ioErrsCounter;
   private final TezCounter wrongLengthErrsCounter;
@@ -188,8 +240,11 @@ class ShuffleScheduler {
   private final float minReqProgressFraction;
   private final float maxAllowedFailedFetchFraction;
   private final boolean checkFailedFetchSinceLastCompletion;
+  private final boolean verifyDiskChecksum;
+  private final boolean compositeFetch;
 
   private volatile Thread shuffleSchedulerThread = null;
+  private final int maxPenaltyTime;
 
   private long totalBytesShuffledTillNow = 0;
   private final DecimalFormat  mbpsFormat = new DecimalFormat("0.00");
@@ -276,9 +331,13 @@ class ShuffleScheduler {
             TezRuntimeConfiguration
                 .TEZ_RUNTIME_SHUFFLE_FAILED_CHECK_SINCE_LAST_COMPLETION_DEFAULT);
 
+    this.applicationId = inputContext.getApplicationId().toString();
+    this.dagId = inputContext.getDagIdentifier();
     this.localHostname = inputContext.getExecutionContext().getHostName();
+    String auxiliaryService = conf.get(TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID,
+        TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID_DEFAULT);
     final ByteBuffer shuffleMetadata =
-        inputContext.getServiceProviderMetaData(ShuffleUtils.SHUFFLE_HANDLER_SERVICE_ID);
+        inputContext.getServiceProviderMetaData(auxiliaryService);
     this.shufflePort = ShuffleUtils.deserializeShuffleProviderMetaData(shuffleMetadata);
 
     this.referee = new Referee();
@@ -312,19 +371,24 @@ class ShuffleScheduler {
     this.startTime = startTime;
     this.lastProgressTime = startTime;
 
+    this.sslShuffle = conf.getBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_SHUFFLE_ENABLE_SSL,
+        TezRuntimeConfiguration.TEZ_RUNTIME_SHUFFLE_ENABLE_SSL_DEFAULT);
     this.asyncHttp = conf.getBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_SHUFFLE_USE_ASYNC_HTTP, false);
     this.httpConnectionParams = ShuffleUtils.getHttpConnectionParams(conf);
-    this.shuffleMetrics = new ShuffleClientMetrics(inputContext.getDAGName(),
-        inputContext.getTaskVertexName(), inputContext.getTaskIndex(),
-        this.conf, UserGroupInformation.getCurrentUser().getShortUserName());
     SecretKey jobTokenSecret = ShuffleUtils
         .getJobTokenSecretFromTokenBytes(inputContext
-            .getServiceConsumerMetaData(TezConstants.TEZ_SHUFFLE_HANDLER_SERVICE_ID));
+            .getServiceConsumerMetaData(auxiliaryService));
     this.jobTokenSecretManager = new JobTokenSecretManager(jobTokenSecret);
 
-    ExecutorService fetcherRawExecutor = Executors.newFixedThreadPool(numFetchers,
-        new ThreadFactoryBuilder().setDaemon(true)
-            .setNameFormat("Fetcher {" + srcNameTrimmed + "} #%d").build());
+    final ExecutorService fetcherRawExecutor;
+    if (conf.getBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_SHUFFLE_FETCHER_USE_SHARED_POOL,
+        TezRuntimeConfiguration.TEZ_RUNTIME_SHUFFLE_FETCHER_USE_SHARED_POOL_DEFAULT)) {
+      fetcherRawExecutor = inputContext.createTezFrameworkExecutorService(numFetchers,
+          "Fetcher_O {" + srcNameTrimmed + "} #%d");
+    } else {
+      fetcherRawExecutor = Executors.newFixedThreadPool(numFetchers, new ThreadFactoryBuilder()
+          .setDaemon(true).setNameFormat("Fetcher_O {" + srcNameTrimmed + "} #%d").build());
+    }
     this.fetcherExecutor = MoreExecutors.listeningDecorator(fetcherRawExecutor);
 
     this.maxFailedUniqueFetches = Math.min(numberOfInputs, 5);
@@ -337,6 +401,10 @@ class ShuffleScheduler {
         conf.getBoolean(
             TezRuntimeConfiguration.TEZ_RUNTIME_SHUFFLE_NOTIFY_READERROR, 
             TezRuntimeConfiguration.TEZ_RUNTIME_SHUFFLE_NOTIFY_READERROR_DEFAULT);
+    this.verifyDiskChecksum = conf.getBoolean(
+        TezRuntimeConfiguration.TEZ_RUNTIME_SHUFFLE_FETCH_VERIFY_DISK_CHECKSUM,
+        TezRuntimeConfiguration.TEZ_RUNTIME_SHUFFLE_FETCH_VERIFY_DISK_CHECKSUM_DEFAULT);
+
     /**
      * Setting to very high val can lead to Http 400 error. Cap it to 75; every attempt id would
      * be approximately 48 bytes; 48 * 75 = 3600 which should give some room for other info in URL.
@@ -348,8 +416,11 @@ class ShuffleScheduler {
     this.skippedInputCounter = inputContext.getCounters().findCounter(TaskCounter.NUM_SKIPPED_INPUTS);
     this.firstEventReceived = inputContext.getCounters().findCounter(TaskCounter.FIRST_EVENT_RECEIVED);
     this.lastEventReceived = inputContext.getCounters().findCounter(TaskCounter.LAST_EVENT_RECEIVED);
+    this.compositeFetch = ShuffleUtils.isTezShuffleHandler(conf);
+    this.maxPenaltyTime = conf.getInt(TezRuntimeConfiguration.TEZ_RUNTIME_SHUFFLE_HOST_PENALTY_TIME_LIMIT_MS,
+        TezRuntimeConfiguration.TEZ_RUNTIME_SHUFFLE_HOST_PENALTY_TIME_LIMIT_MS_DEFAULT);
 
-    pipelinedShuffleInfoEventsMap = new HashMap<InputIdentifier, ShuffleEventInfo>();
+    pipelinedShuffleInfoEventsMap = Maps.newConcurrentMap();
     LOG.info("ShuffleScheduler running for sourceVertex: "
         + inputContext.getSourceVertexName() + " with configuration: "
         + "maxFetchFailuresBeforeReporting=" + maxFetchFailuresBeforeReporting
@@ -373,35 +444,73 @@ class ShuffleScheduler {
     schedulerCallable.call();
   }
 
-  public void close() throws InterruptedException {
-    if (!isShutdown.getAndSet(true)) {
+  public void close() {
+    try {
+      if (!isShutdown.getAndSet(true)) {
+        try {
+          logProgress();
+        } catch (Exception e) {
+          LOG.warn("Failed log progress while closing, ignoring and continuing shutdown. Message={}",
+              e.getMessage());
+        }
 
-      logProgress();
+        // Notify and interrupt the waiting scheduler thread
+        synchronized (this) {
+          notifyAll();
+        }
+        // Interrupt the ShuffleScheduler thread only if the close is invoked by another thread.
+        // If this is invoked on the same thread, then the shuffleRunner has already complete, and there's
+        // no point interrupting it.
+        // The interrupt is needed to unblock any merges or waits which may be happening, so that the thread can
+        // exit.
+        if (shuffleSchedulerThread != null && !Thread.currentThread()
+            .equals(shuffleSchedulerThread)) {
+          shuffleSchedulerThread.interrupt();
+        }
 
-      // Notify and interrupt the waiting scheduler thread
-      synchronized (this) {
-        notifyAll();
+        // Interrupt the fetchers.
+        for (FetcherOrderedGrouped fetcher : runningFetchers) {
+          try {
+            fetcher.shutDown();
+          } catch (Exception e) {
+            LOG.warn(
+                "Error while shutting down fetcher. Ignoring and continuing shutdown. Message={}",
+                e.getMessage());
+          }
+        }
+
+        // Kill the Referee thread.
+        try {
+          referee.interrupt();
+          referee.join();
+        } catch (InterruptedException e) {
+          LOG.warn(
+              "Interrupted while shutting down referee. Ignoring and continuing shutdown");
+          Thread.currentThread().interrupt();
+        } catch (Exception e) {
+          LOG.warn(
+              "Error while shutting down referee. Ignoring and continuing shutdown. Message={}",
+              e.getMessage());
+        }
       }
-      // Interrupt the ShuffleScheduler thread only if the close is invoked by another thread.
-      // If this is invoked on the same thread, then the shuffleRunner has already complete, and there's
-      // no point interrupting it.
-      // The interrupt is needed to unblock any merges or waits which may be happening, so that the thread can
-      // exit.
-      if (shuffleSchedulerThread != null && !Thread.currentThread().equals(shuffleSchedulerThread)) {
-        shuffleSchedulerThread.interrupt();
+    } finally {
+      long startTime = System.currentTimeMillis();
+      if (!fetcherExecutor.isShutdown()) {
+        // Ensure that fetchers respond to cancel request.
+        fetcherExecutor.shutdownNow();
       }
-
-      // Interrupt the fetchers.
-      for (FetcherOrderedGrouped fetcher : runningFetchers) {
-        fetcher.shutDown();
-      }
-
-      // Kill the Referee thread.
-      referee.interrupt();
-      referee.join();
+      long endTime = System.currentTimeMillis();
+      LOG.info("Shutting down fetchers for input: {}, shutdown timetaken: {} ms, "
+              + "hasFetcherExecutorStopped: {}", srcNameTrimmed,
+          (endTime - startTime), hasFetcherExecutorStopped());
     }
   }
 
+  @VisibleForTesting
+  boolean hasFetcherExecutorStopped() {
+    return fetcherExecutor.isShutdown();
+  }
+
   @VisibleForTesting
   public boolean isShutdown() {
     return isShutdown.get();
@@ -426,10 +535,11 @@ class ShuffleScheduler {
     int finalEventId = -1; //0 indexed
     int attemptNum;
     String id;
+    boolean scheduledForDownload; // whether chunks got scheduled for download (getMapHost)
 
 
     ShuffleEventInfo(InputAttemptIdentifier input) {
-      this.id = input.getInputIdentifier().getInputIndex() + "_" + input.getAttemptNumber();
+      this.id = input.getInputIdentifier() + "_" + input.getAttemptNumber();
       this.eventsProcessed = new BitSet();
       this.attemptNum = input.getAttemptNumber();
     }
@@ -453,7 +563,8 @@ class ShuffleScheduler {
 
     public String toString() {
       return "[eventsProcessed=" + eventsProcessed + ", finalEventId=" + finalEventId
-          +  ", id=" + id + ", attemptNum=" + attemptNum + "]";
+          +  ", id=" + id + ", attemptNum=" + attemptNum
+          + ", scheduledForDownload=" + scheduledForDownload + "]";
     }
   }
 
@@ -467,7 +578,7 @@ class ShuffleScheduler {
                                          ) throws IOException {
 
     inputContext.notifyProgress();
-    if (!isInputFinished(srcAttemptIdentifier.getInputIdentifier().getInputIndex())) {
+    if (!isInputFinished(srcAttemptIdentifier.getInputIdentifier())) {
       if (!isLocalFetch) {
         /**
          * Reset it only when it is a non-local-disk copy.
@@ -478,12 +589,12 @@ class ShuffleScheduler {
 
         failureCounts.remove(srcAttemptIdentifier);
         if (host != null) {
-          hostFailures.remove(host.getHostIdentifier());
+          hostFailures.remove(new HostPort(host.getHost(), host.getPort()));
         }
 
         output.commit();
-        ShuffleUtils.logIndividualFetchComplete(LOG, millis, bytesCompressed,
-            bytesDecompressed, output.getType().toString(), srcAttemptIdentifier);
+        fetchStatsLogger.logIndividualFetchComplete(millis, bytesCompressed, bytesDecompressed,
+            output.getType().toString(), srcAttemptIdentifier);
         if (output.getType() == Type.DISK) {
           bytesShuffledToDisk.increment(bytesCompressed);
         } else if (output.getType() == Type.DISK_DIRECT) {
@@ -505,10 +616,10 @@ class ShuffleScheduler {
        */
       if (!srcAttemptIdentifier.canRetrieveInputInChunks()) {
         remainingMaps.decrementAndGet();
-        setInputFinished(srcAttemptIdentifier.getInputIdentifier().getInputIndex());
+        setInputFinished(srcAttemptIdentifier.getInputIdentifier());
         numFetchedSpills++;
       } else {
-        InputIdentifier inputIdentifier = srcAttemptIdentifier.getInputIdentifier();
+        int inputIdentifier = srcAttemptIdentifier.getInputIdentifier();
         //Allow only one task attempt to proceed.
         if (!validateInputAttemptForPipelinedShuffle(srcAttemptIdentifier)) {
           return;
@@ -533,7 +644,7 @@ class ShuffleScheduler {
         //check if we downloaded all spills pertaining to this InputAttemptIdentifier
         if (eventInfo.isDone()) {
           remainingMaps.decrementAndGet();
-          setInputFinished(inputIdentifier.getInputIndex());
+          setInputFinished(inputIdentifier);
           pipelinedShuffleInfoEventsMap.remove(inputIdentifier);
           if (LOG.isTraceEnabled()) {
             LOG.trace("Removing : " + srcAttemptIdentifier + ", pending: " +
@@ -560,7 +671,7 @@ class ShuffleScheduler {
       if (LOG.isDebugEnabled()) {
         LOG.debug("src task: "
             + TezRuntimeUtils.getTaskAttemptIdentifier(
-                inputContext.getSourceVertexName(), srcAttemptIdentifier.getInputIdentifier().getInputIndex(),
+                inputContext.getSourceVertexName(), srcAttemptIdentifier.getInputIdentifier(),
                 srcAttemptIdentifier.getAttemptNumber()) + " done");
       }
     } else {
@@ -583,12 +694,29 @@ class ShuffleScheduler {
     if (input.canRetrieveInputInChunks()) {
       ShuffleEventInfo eventInfo = pipelinedShuffleInfoEventsMap.get(input.getInputIdentifier());
       if (eventInfo != null && input.getAttemptNumber() != eventInfo.attemptNum) {
-        reportExceptionForInput(new IOException("Previous event already got scheduled for " +
-            input + ". Previous attempt's data could have been already merged "
-            + "to memory/disk outputs.  Failing the fetch early. currentAttemptNum="
-            + eventInfo.attemptNum + ", eventsProcessed=" + eventInfo.eventsProcessed
-            + ", newAttemptNum=" + input.getAttemptNumber()));
-        return false;
+        /*
+         * Check if current attempt has been scheduled for download.
+         * e.g currentAttemptNum=0, eventsProcessed={}, newAttemptNum=1
+         * If nothing is scheduled in current attempt and no events are processed
+         * (i.e copySucceeded), we can ignore current attempt and start processing the new
+         * attempt (e.g LLAP).
+         */
+        if (eventInfo.scheduledForDownload || !eventInfo.eventsProcessed.isEmpty()) {
+          IOException exception = new IOException("Previous event already got scheduled for " +
+              input + ". Previous attempt's data could have been already merged "
+              + "to memory/disk outputs.  Killing (self) this task early."
+              + " currentAttemptNum=" + eventInfo.attemptNum
+              + ", eventsProcessed=" + eventInfo.eventsProcessed
+              + ", scheduledForDownload=" + eventInfo.scheduledForDownload
+              + ", newAttemptNum=" + input.getAttemptNumber());
+          String message = "Killing self as previous attempt data could have been consumed";
+          killSelf(exception, message);
+          return false;
+        }
+        if (LOG.isDebugEnabled()) {
+          LOG.debug("Ignoring current attempt=" + eventInfo.attemptNum + " with eventInfo=" +
+              eventInfo.toString() + "and processing new attempt=" + input.getAttemptNumber());
+        }
       }
 
       if (eventInfo == null) {
@@ -600,9 +728,9 @@ class ShuffleScheduler {
   }
 
   @VisibleForTesting
-  void reportExceptionForInput(Exception exception) {
-    LOG.error(srcNameTrimmed + ": " + "Reporting exception for input", exception);
-    exceptionReporter.reportException(exception);
+  void killSelf(Exception exception, String message) {
+    LOG.error(message, exception);
+    exceptionReporter.killSelf(exception, message);
   }
 
   private final AtomicInteger nextProgressLineEventCount = new AtomicInteger(0);
@@ -679,7 +807,7 @@ class ShuffleScheduler {
       String errorMsg = "Failed " + attemptFailures + " times trying to "
           + "download from " + TezRuntimeUtils.getTaskAttemptIdentifier(
           inputContext.getSourceVertexName(),
-          srcAttempt.getInputIdentifier().getInputIndex(),
+          srcAttempt.getInputIdentifier(),
           srcAttempt.getAttemptNumber()) + ". threshold=" + abortFailureLimit;
       IOException ioe = new IOException(errorMsg);
       // Shuffle knows how to deal with failures post shutdown via the onFailure hook
@@ -692,7 +820,7 @@ class ShuffleScheduler {
   private void penalizeHost(MapHost host, int failures) {
     host.penalize();
 
-    String hostPort = host.getHostIdentifier();
+    HostPort hostPort = new HostPort(host.getHost(), host.getPort());
     // TODO TEZ-922 hostFailures isn't really used for anything apart from
     // hasFailedAcrossNodes().Factor it into error
     // reporting / potential blacklisting of hosts.
@@ -705,7 +833,8 @@ class ShuffleScheduler {
 
     long delay = (long) (INITIAL_PENALTY *
         Math.pow(PENALTY_GROWTH_RATE, failures));
-    penalties.add(new Penalty(host, delay));
+    long penaltyDelay = Math.min(delay, maxPenaltyTime);
+    penalties.add(new Penalty(host, penaltyDelay));
   }
 
   private int getFailureCount(InputAttemptIdentifier srcAttempt) {
@@ -738,15 +867,15 @@ class ShuffleScheduler {
         srcNameTrimmed + ": " + "Reporting fetch failure for InputIdentifier: "
             + srcAttempt + " taskAttemptIdentifier: " + TezRuntimeUtils
             .getTaskAttemptIdentifier(inputContext.getSourceVertexName(),
-                srcAttempt.getInputIdentifier().getInputIndex(),
+                srcAttempt.getInputIdentifier(),
                 srcAttempt.getAttemptNumber()) + " to AM.");
     List<Event> failedEvents = Lists.newArrayListWithCapacity(1);
     failedEvents.add(InputReadErrorEvent.create(
         "Fetch failure for " + TezRuntimeUtils
             .getTaskAttemptIdentifier(inputContext.getSourceVertexName(),
-                srcAttempt.getInputIdentifier().getInputIndex(),
+                srcAttempt.getInputIdentifier(),
                 srcAttempt.getAttemptNumber()) + " to jobtracker.",
-        srcAttempt.getInputIdentifier().getInputIndex(),
+        srcAttempt.getInputIdentifier(),
         srcAttempt.getAttemptNumber()));
 
     inputContext.sendEvents(failedEvents);
@@ -767,7 +896,7 @@ class ShuffleScheduler {
         (int) Math.ceil(numUniqueHosts * hostFailureFraction));
     int total = 0;
     boolean failedAcrossNodes = false;
-    for(String host : uniqueHosts) {
+    for(HostPort host : uniqueHosts) {
       IntWritable failures = hostFailures.get(host);
       if (failures != null && failures.get() > minFailurePerHost) {
         total++;
@@ -927,17 +1056,13 @@ class ShuffleScheduler {
   public synchronized void addKnownMapOutput(String inputHostName,
                                              int port,
                                              int partitionId,
-                                             String hostUrl,
-                                             InputAttemptIdentifier srcAttempt) {
-    String hostPort = (inputHostName + ":" + String.valueOf(port));
-    uniqueHosts.add(hostPort);
-    String identifier = MapHost.createIdentifier(hostPort, partitionId);
-
+                                             CompositeInputAttemptIdentifier srcAttempt) {
+    uniqueHosts.add(new HostPort(inputHostName, port));
+    HostPortPartition identifier = new HostPortPartition(inputHostName, port, partitionId);
 
     MapHost host = mapLocations.get(identifier);
     if (host == null) {
-      host = new MapHost(partitionId, hostPort, hostUrl);
-      assert identifier.equals(host.getIdentifier());
+      host = new MapHost(inputHostName, port, partitionId, srcAttempt.getInputIdentifierCount());
       mapLocations.put(identifier, host);
     }
 
@@ -947,9 +1072,10 @@ class ShuffleScheduler {
     }
 
     host.addKnownMap(srcAttempt);
-    pathToIdentifierMap.put(
-        getIdentifierFromPathAndReduceId(srcAttempt.getPathComponent(),
-            partitionId), srcAttempt);
+    for (int i = 0; i < srcAttempt.getInputIdentifierCount(); i++) {
+      PathPartition pathPartition = new PathPartition(srcAttempt.getPathComponent(), partitionId + i);
+      pathToIdentifierMap.put(pathPartition, srcAttempt.expand(i));
+    }
 
     // Mark the host as pending
     if (host.getState() == MapHost.State.PENDING) {
@@ -958,18 +1084,33 @@ class ShuffleScheduler {
     }
   }
   
-  public synchronized void obsoleteInput(InputAttemptIdentifier srcAttempt) {
+  public void obsoleteInput(InputAttemptIdentifier srcAttempt) {
     // The incoming srcAttempt does not contain a path component.
     LOG.info(srcNameTrimmed + ": " + "Adding obsolete input: " + srcAttempt);
-    if (pipelinedShuffleInfoEventsMap.containsKey(srcAttempt.getInputIdentifier())) {
-      //Pipelined shuffle case (where pipelinedShuffleInfoEventsMap gets populated).
-      //Fail fast here.
-      exceptionReporter.reportException(new IOException(srcAttempt + " is marked as obsoleteInput, but it "
+    ShuffleEventInfo eventInfo = pipelinedShuffleInfoEventsMap.get(srcAttempt.getInputIdentifier());
+
+    //Pipelined shuffle case (where pipelinedShuffleInfoEventsMap gets populated).
+    //Fail fast here.
+    if (eventInfo != null) {
+      // In case this we haven't started downloading it, get rid of it.
+      if (eventInfo.eventsProcessed.isEmpty() && !eventInfo.scheduledForDownload) {
+        // obsoleted anyways; no point tracking if nothing is started
+        pipelinedShuffleInfoEventsMap.remove(srcAttempt.getInputIdentifier());
+        if (LOG.isDebugEnabled()) {
+          LOG.debug("Removing " + eventInfo + " from tracking");
+        }
+        return;
+      }
+      IOException exception = new IOException(srcAttempt + " is marked as obsoleteInput, but it "
           + "exists in shuffleInfoEventMap. Some data could have been already merged "
-          + "to memory/disk outputs.  Failing the fetch early."));
+          + "to memory/disk outputs.  Failing the fetch early. eventInfo:" + eventInfo.toString());
+      String message = "Got obsolete event. Killing self as attempt's data could have been consumed";
+      killSelf(exception, message);
       return;
     }
-    obsoleteInputs.add(srcAttempt);
+    synchronized (this) {
+      obsoleteInputs.add(srcAttempt);
+    }
   }
   
   public synchronized void putBackKnownMapOutput(MapHost host,
@@ -982,7 +1123,7 @@ class ShuffleScheduler {
       if (LOG.isDebugEnabled()) {
         LOG.debug("PendingHosts=" + pendingHosts);
       }
-      wait();
+      waitAndNotifyProgress();
     }
 
     if (!pendingHosts.isEmpty()) {
@@ -1009,12 +1150,23 @@ class ShuffleScheduler {
 
   public InputAttemptIdentifier getIdentifierForFetchedOutput(
       String path, int reduceId) {
-    return pathToIdentifierMap.get(getIdentifierFromPathAndReduceId(path, reduceId));
+    return pathToIdentifierMap.get(new PathPartition(path, reduceId));
   }
-  
-  private boolean inputShouldBeConsumed(InputAttemptIdentifier id) {
-    return (!obsoleteInputs.contains(id) && 
-             !isInputFinished(id.getInputIdentifier().getInputIndex()));
+
+  @VisibleForTesting
+  DelayQueue<Penalty> getPenalties() {
+    return penalties;
+  }
+
+  private synchronized boolean inputShouldBeConsumed(InputAttemptIdentifier id) {
+    boolean isInputFinished = false;
+    if (id instanceof CompositeInputAttemptIdentifier) {
+      CompositeInputAttemptIdentifier cid = (CompositeInputAttemptIdentifier)id;
+      isInputFinished = isInputFinished(cid.getInputIdentifier(), cid.getInputIdentifier() + cid.getInputIdentifierCount());
+    } else {
+      isInputFinished = isInputFinished(id.getInputIdentifier());
+    }
+    return !obsoleteInputs.contains(id) && !isInputFinished;
   }
 
   public synchronized List<InputAttemptIdentifier> getMapsForHost(MapHost host) {
@@ -1029,7 +1181,7 @@ class ShuffleScheduler {
       // This may be removed after TEZ-914
       InputAttemptIdentifier id = listItr.next();
       if (inputShouldBeConsumed(id)) {
-        Integer inputNumber = Integer.valueOf(id.getInputIdentifier().getInputIndex());
+        Integer inputNumber = Integer.valueOf(id.getInputIdentifier());
         List<InputAttemptIdentifier> oldIdList = dedupedList.get(inputNumber);
 
         if (oldIdList == null || oldIdList.isEmpty()) {
@@ -1088,6 +1240,13 @@ class ShuffleScheduler {
         if (includedMaps++ >= maxTaskOutputAtOnce) {
           host.addKnownMap(inputAttemptIdentifier);
         } else {
+          if (inputAttemptIdentifier.canRetrieveInputInChunks()) {
+            ShuffleEventInfo shuffleEventInfo =
+                pipelinedShuffleInfoEventsMap.get(inputAttemptIdentifier.getInputIdentifier());
+            if (shuffleEventInfo != null) {
+              shuffleEventInfo.scheduledForDownload = true;
+            }
+          }
           result.add(inputAttemptIdentifier);
         }
       }
@@ -1130,7 +1289,7 @@ class ShuffleScheduler {
   /**
    * A structure that records the penalty for a host.
    */
-  private static class Penalty implements Delayed {
+  static class Penalty implements Delayed {
     MapHost host;
     private long endTime;
     
@@ -1150,11 +1309,7 @@ class ShuffleScheduler {
     }
     
   }
-  
-  private String getIdentifierFromPathAndReduceId(String path, int reduceId) {
-    return path + "_" + reduceId;
-  }
-  
+
   /**
    * A thread that takes hosts off of the penalty list when the timer expires.
    */
@@ -1193,34 +1348,39 @@ class ShuffleScheduler {
       finishedMaps.set(inputIndex, true);
     }
   }
-  
+
   boolean isInputFinished(int inputIndex) {
     synchronized (finishedMaps) {
       return finishedMaps.get(inputIndex);
     }
   }
 
+  boolean isInputFinished(int inputIndex, int inputEnd) {
+    synchronized (finishedMaps) {
+      return finishedMaps.nextClearBit(inputIndex) > inputEnd;
+    }
+  }
+
   private class ShuffleSchedulerCallable extends CallableWithNdc<Void> {
 
 
     @Override
     protected Void callInternal() throws InterruptedException {
-      outer:
       while (!isShutdown.get() && remainingMaps.get() > 0) {
         synchronized (ShuffleScheduler.this) {
-          if (runningFetchers.size() >= numFetchers || pendingHosts.isEmpty()) {
-            if (remainingMaps.get() > 0) {
-              try {
-                ShuffleScheduler.this.wait();
-              } catch (InterruptedException e) {
-                if (isShutdown.get()) {
-                  LOG.info(srcNameTrimmed + ": " +
-                      "Interrupted while waiting for fetchers to complete and hasBeenShutdown. Breaking out of ShuffleSchedulerCallable loop");
-                  Thread.currentThread().interrupt();
-                  break;
-                } else {
-                  throw e;
-                }
+          while ((runningFetchers.size() >= numFetchers || pendingHosts.isEmpty())
+              && remainingMaps.get() > 0) {
+            try {
+              waitAndNotifyProgress();
+            } catch (InterruptedException e) {
+              if (isShutdown.get()) {
+                LOG.info(srcNameTrimmed + ": " +
+                    "Interrupted while waiting for fetchers to complete" +
+                    "and hasBeenShutdown. Breaking out of ShuffleSchedulerCallable loop");
+                Thread.currentThread().interrupt();
+                break;
+              } else {
+                throw e;
               }
             }
           }
@@ -1275,7 +1435,7 @@ class ShuffleScheduler {
                 count++;
                 if (LOG.isDebugEnabled()) {
                   LOG.debug(srcNameTrimmed + ": " + "Scheduling fetch for inputHost: {}",
-                      mapHost.getIdentifier());
+                      mapHost.getHostIdentifier() + ":" + mapHost.getPartitionId());
                 }
                 FetcherOrderedGrouped fetcherOrderedGrouped = constructFetcherForHost(mapHost);
                 runningFetchers.add(fetcherOrderedGrouped);
@@ -1294,13 +1454,19 @@ class ShuffleScheduler {
     }
   }
 
+  private synchronized void waitAndNotifyProgress() throws InterruptedException {
+      inputContext.notifyProgress();
+      wait(1000);
+  }
+
   @VisibleForTesting
   FetcherOrderedGrouped constructFetcherForHost(MapHost mapHost) {
     return new FetcherOrderedGrouped(httpConnectionParams, ShuffleScheduler.this, allocator,
-        shuffleMetrics, exceptionReporter, jobTokenSecretManager, ifileReadAhead, ifileReadAheadLength,
+        exceptionReporter, jobTokenSecretManager, ifileReadAhead, ifileReadAheadLength,
         codec, conf, localDiskFetchEnabled, localHostname, shufflePort, srcNameTrimmed, mapHost,
         ioErrsCounter, wrongLengthErrsCounter, badIdErrsCounter, wrongMapErrsCounter,
-        connectionErrsCounter, wrongReduceErrsCounter, asyncHttp);
+        connectionErrsCounter, wrongReduceErrsCounter, applicationId, dagId, asyncHttp, sslShuffle,
+        verifyDiskChecksum, compositeFetch);
   }
 
   private class FetchFutureCallback implements FutureCallback<Void> {
