@@ -27,6 +27,7 @@ import java.util.Iterator;
 import java.util.concurrent.atomic.AtomicBoolean;
 
 import com.google.common.annotations.VisibleForTesting;
+
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 import org.apache.hadoop.classification.InterfaceAudience;
@@ -68,6 +69,7 @@ public class IFile {
     (byte) 'F' , (byte) 0};
 
   private static final String INCOMPLETE_READ = "Requested to read %d got %d";
+  private static final String REQ_BUFFER_SIZE_TOO_LARGE = "Size of data %d is greater than the max allowed of %d";
 
   /**
    * <code>IFile.Writer</code> to write out intermediate map-outputs.
@@ -125,11 +127,11 @@ public class IFile {
            writesCounter, serializedBytesCounter);
       ownOutputStream = true;
     }
-    
-    protected Writer(TezCounter writesCounter, TezCounter serializedBytesCounter) {
+
+    protected Writer(TezCounter writesCounter, TezCounter serializedBytesCounter, boolean rle) {
       writtenRecordsCounter = writesCounter;
       serializedUncompressedBytes = serializedBytesCounter;
-      this.rle = false;
+      this.rle = rle;
     }
 
     public Writer(Configuration conf, FSDataOutputStream outputStream,
@@ -186,13 +188,14 @@ public class IFile {
       if (!headerWritten) {
         outputStream.write(HEADER, 0, HEADER.length - 1);
         outputStream.write((compressOutput) ? (byte) 1 : (byte) 0);
-        outputStream.flush();
         headerWritten = true;
       }
     }
 
     public void close() throws IOException {
-      checkState(!closed.getAndSet(true), "Writer was already closed earlier");
+      if (closed.getAndSet(true)) {
+        throw new IOException("Writer was already closed earlier");
+      }
 
       // When IFile writer is created by BackupStore, we do not have
       // Key and Value classes set. So, check before closing the
@@ -212,9 +215,6 @@ public class IFile {
       //account for header bytes
       decompressedBytesWritten += HEADER.length;
 
-      //Flush the stream
-      out.flush();
-
       // Close the underlying stream iff we own it...
       if (ownOutputStream) {
         out.close();
@@ -224,7 +224,7 @@ public class IFile {
           compressedOut.finish();
           compressedOut.resetState();
         }
-        // Write the checksum
+        // Write the checksum and flush the buffer
         checksumOut.finish();
       }
       //header bytes are already included in rawOut
@@ -481,6 +481,13 @@ public class IFile {
     public enum KeyState {NO_KEY, NEW_KEY, SAME_KEY}
 
     private static final int DEFAULT_BUFFER_SIZE = 128*1024;
+    @VisibleForTesting
+    // Not final for testing
+    protected static int MAX_BUFFER_SIZE
+            = Integer.MAX_VALUE - 8;  // The maximum array size is a little less than the
+                                      // max integer value. Trying to create a larger array
+                                      // will result in an OOM exception. The exact value
+                                      // is JVM dependent so setting it to max int - 8 to be safe.
 
     // Count records read from disk
     private long numRecordsRead = 0;
@@ -496,7 +503,7 @@ public class IFile {
 
     protected byte[] buffer = null;
     protected int bufferSize = DEFAULT_BUFFER_SIZE;
-    protected DataInputStream dataIn;
+    protected DataInputStream dataIn = null;
 
     protected int recNo = 1;
     protected int originalKeyLength;
@@ -583,7 +590,9 @@ public class IFile {
         this.in = null;
       }
 
-      this.dataIn = new DataInputStream(this.in);
+      if (in != null) {
+        this.dataIn = new DataInputStream(this.in);
+      }
       this.readRecordsCounter = readsCounter;
       this.bytesReadCounter = bytesReadCounter;
       this.fileLength = length;
@@ -622,6 +631,15 @@ public class IFile {
       }
       try {
         IOUtils.readFully(in, buffer, 0, buffer.length - IFile.HEADER.length);
+        /*
+         * We've gotten the amount of data we were expecting. Verify the
+         * decompressor has nothing more to offer. This action also forces the
+         * decompressor to read any trailing bytes that weren't critical for
+         * decompression, which is necessary to keep the stream in sync.
+         */
+        if (in.read() >= 0) {
+          throw new IOException("Unexpected extra bytes from input stream");
+        }
       } catch (IOException ioe) {
         if(in != null) {
           try {
@@ -641,6 +659,43 @@ public class IFile {
       }
     }
 
+    /**
+     * Read entire IFile content to disk.
+     *
+     * @param out the output stream that will receive the data
+     * @param in the input stream containing the IFile data
+     * @param length the amount of data to read from the input
+     * @return the number of bytes copied
+     * @throws IOException
+     */
+    public static long readToDisk(OutputStream out, InputStream in, long length,
+        boolean ifileReadAhead, int ifileReadAheadLength)
+        throws IOException {
+      final int BYTES_TO_READ = 64 * 1024;
+      byte[] buf = new byte[BYTES_TO_READ];
+
+      // copy the IFile header
+      if (length < HEADER.length) {
+        throw new IOException("Missing IFile header");
+      }
+      IOUtils.readFully(in, buf, 0, HEADER.length);
+      verifyHeaderMagic(buf);
+      out.write(buf, 0, HEADER.length);
+      long bytesLeft = length - HEADER.length;
+      @SuppressWarnings("resource")
+      IFileInputStream ifInput = new IFileInputStream(in, bytesLeft,
+          ifileReadAhead, ifileReadAheadLength);
+      while (bytesLeft > 0) {
+        int n = ifInput.readWithChecksum(buf, 0, (int) Math.min(bytesLeft, BYTES_TO_READ));
+        if (n < 0) {
+          throw new IOException("read past end of stream");
+        }
+        out.write(buf, 0, n);
+        bytesLeft -= n;
+      }
+      return length - bytesLeft;
+    }
+
     public long getLength() {
       return fileLength - checksumIn.getSize();
     }
@@ -701,7 +756,9 @@ public class IFile {
      */
     protected boolean positionToNextRecord(DataInput dIn) throws IOException {
       // Sanity check
-      checkState(!eof, "Reached EOF. Completed reading %d", bytesRead);
+      if (eof) {
+        throw new IOException(String.format("Reached EOF. Completed reading %d", bytesRead));
+      }
       prevKeyLength = currentKeyLength;
 
       if (prevKeyLength == RLE_MARKER) {
@@ -733,6 +790,21 @@ public class IFile {
       return readRawKey(key) != KeyState.NO_KEY;
     }
 
+    private static byte[] createLargerArray(int currentLength) {
+      if (currentLength > MAX_BUFFER_SIZE) {
+        throw new IllegalArgumentException(
+                String.format(REQ_BUFFER_SIZE_TOO_LARGE, currentLength, MAX_BUFFER_SIZE));
+      }
+      int newLength;
+      if (currentLength > (MAX_BUFFER_SIZE - currentLength)) {
+        // possible overflow: if (2*currentLength > MAX_BUFFER_SIZE)
+        newLength = currentLength;
+      } else {
+        newLength = currentLength << 1;
+      }
+      return new byte[newLength];
+    }
+
     public KeyState readRawKey(DataInputBuffer key) throws IOException {
       if (!positionToNextRecord(dataIn)) {
         if (LOG.isDebugEnabled()) {
@@ -749,22 +821,29 @@ public class IFile {
         return KeyState.SAME_KEY;
       }
       if (keyBytes.length < currentKeyLength) {
-        keyBytes = new byte[currentKeyLength << 1];
+        keyBytes = createLargerArray(currentKeyLength);
       }
       int i = readData(keyBytes, 0, currentKeyLength);
-      checkState((i == currentKeyLength), INCOMPLETE_READ, currentKeyLength, i);
+      if (i != currentKeyLength) {
+        throw new IOException(String.format(INCOMPLETE_READ, currentKeyLength, i));
+      }
       key.reset(keyBytes, currentKeyLength);
       bytesRead += currentKeyLength;
       return KeyState.NEW_KEY;
     }
 
     public void nextRawValue(DataInputBuffer value) throws IOException {
-      final byte[] valBytes =
-        ((value.getData().length < currentValueLength) || (value.getData() == keyBytes))
-        ? new byte[currentValueLength << 1]
-        : value.getData();
+      final byte[] valBytes;
+      if ((value.getData().length < currentValueLength) || (value.getData() == keyBytes)) {
+        valBytes = createLargerArray(currentValueLength);
+      } else {
+        valBytes = value.getData();
+      }
+
       int i = readData(valBytes, 0, currentValueLength);
-      checkState((i == currentValueLength), INCOMPLETE_READ, currentValueLength, i);
+      if (i != currentValueLength) {
+        throw new IOException(String.format(INCOMPLETE_READ, currentValueLength, i));
+      }
       value.reset(valBytes, currentValueLength);
 
       // Record the bytes read
@@ -774,14 +853,17 @@ public class IFile {
       ++numRecordsRead;
     }
 
-    public static boolean isCompressedFlagEnabled(InputStream in) throws IOException {
-      byte[] header = new byte[HEADER.length];
-      IOUtils.readFully(in, header, 0, HEADER.length);
-
+    private static void verifyHeaderMagic(byte[] header) throws IOException {
       if (!(header[0] == 'T' && header[1] == 'I'
           && header[2] == 'F')) {
         throw new IOException("Not a valid ifile header");
       }
+    }
+
+    public static boolean isCompressedFlagEnabled(InputStream in) throws IOException {
+      byte[] header = new byte[HEADER.length];
+      IOUtils.readFully(in, header, 0, HEADER.length);
+      verifyHeaderMagic(header);
       return (header[3] == 1);
     }
 
@@ -817,11 +899,4 @@ public class IFile {
     }
   }
 
-  public static void checkState(boolean expression, String format,
-      Object... args) throws IOException {
-    if (!expression) {
-      throw new IOException(String.format(format, args));
-    }
-  }
-
 }
