@@ -17,7 +17,6 @@
 */
 package org.apache.tez.runtime.library.common.sort.impl;
 
-import java.io.DataOutputStream;
 import java.io.IOException;
 import java.io.OutputStream;
 import java.nio.BufferOverflowException;
@@ -31,12 +30,16 @@ import java.util.concurrent.ExecutionException;
 import java.util.concurrent.ExecutorService;
 import java.util.concurrent.Executors;
 import java.util.concurrent.Future;
+import java.util.concurrent.TimeUnit;
+import java.util.zip.Deflater;
 
 import com.google.common.annotations.VisibleForTesting;
 import com.google.common.base.Preconditions;
-import com.google.common.base.Stopwatch;
 import com.google.common.collect.Lists;
 
+import org.apache.hadoop.classification.InterfaceAudience;
+import org.apache.hadoop.fs.permission.FsPermission;
+import org.apache.tez.dag.api.TezConfiguration;
 import org.apache.tez.runtime.library.api.IOInterruptedException;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
@@ -46,22 +49,28 @@ import org.apache.hadoop.fs.Path;
 import org.apache.hadoop.io.DataInputBuffer;
 import org.apache.tez.common.TezUtilsInternal;
 import org.apache.tez.common.CallableWithNdc;
+import org.apache.tez.common.io.NonSyncDataOutputStream;
 import org.apache.tez.runtime.api.Event;
 import org.apache.tez.runtime.library.common.comparator.ProxyComparator;
 import org.apache.hadoop.io.RawComparator;
 import org.apache.hadoop.util.IndexedSortable;
 import org.apache.hadoop.util.IndexedSorter;
 import org.apache.hadoop.util.Progress;
+import org.apache.tez.common.TezCommonUtils;
 import org.apache.tez.runtime.api.OutputContext;
 import org.apache.tez.runtime.library.api.TezRuntimeConfiguration;
 import org.apache.tez.runtime.library.common.ConfigUtils;
 import org.apache.tez.runtime.library.common.shuffle.ShuffleUtils;
 import org.apache.tez.runtime.library.common.sort.impl.IFile.Writer;
+import org.apache.tez.runtime.library.common.sort.impl.TezMerger.DiskSegment;
 import org.apache.tez.runtime.library.common.sort.impl.TezMerger.Segment;
 import org.apache.tez.runtime.library.utils.LocalProgress;
+import org.apache.tez.util.StopWatch;
 
 import com.google.common.util.concurrent.ThreadFactoryBuilder;
 
+import static org.apache.tez.runtime.library.common.sort.impl.TezSpillRecord.SPILL_FILE_PERMS;
+
 @SuppressWarnings({"unchecked", "rawtypes"})
 public class PipelinedSorter extends ExternalSorter {
   
@@ -111,6 +120,8 @@ public class PipelinedSorter extends ExternalSorter {
   private int bufferIndex = -1;
   private final int MIN_BLOCK_SIZE;
   private final boolean lazyAllocateMem;
+  private final Deflater deflater;
+  private final String auxiliaryService;
 
   // TODO Set additional countesr - total bytes written, spills etc.
 
@@ -127,7 +138,8 @@ public class PipelinedSorter extends ExternalSorter {
        * When lazy-allocation is enabled, framework takes care of auto
        * allocating memory on need basis. Desirable block size is set to 256MB
        */
-      MIN_BLOCK_SIZE = 256 << 20; //256 MB
+      //256 MB - 64 bytes. See comment for the 32MB allocation.
+      MIN_BLOCK_SIZE = ((256 << 20) - 64);
     } else {
       int minBlockSize = conf.getInt(TezRuntimeConfiguration
               .TEZ_RUNTIME_PIPELINED_SORTER_MIN_BLOCK_SIZE_IN_MB,
@@ -150,7 +162,8 @@ public class PipelinedSorter extends ExternalSorter {
         .TEZ_RUNTIME_PIPELINED_SHUFFLE_ENABLED_DEFAULT);
 
     pipelinedShuffle = !isFinalMergeEnabled() && confPipelinedShuffle;
-
+    auxiliaryService = conf.get(TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID,
+        TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID_DEFAULT);
     //sanity checks
     final long sortmb = this.availableMemoryMb;
 
@@ -222,6 +235,7 @@ public class PipelinedSorter extends ExternalSorter {
     valSerializer.open(span.out);
     keySerializer.open(span.out);
     minSpillsForCombine = this.conf.getInt(TezRuntimeConfiguration.TEZ_RUNTIME_COMBINE_MIN_SPILLS, 3);
+    deflater = TezCommonUtils.newBestCompressionDeflater();
   }
 
   ByteBuffer allocateSpace() {
@@ -261,7 +275,11 @@ public class PipelinedSorter extends ExternalSorter {
      */
     if (lazyAllocateMem) {
       if (buffers == null || buffers.isEmpty()) {
-        return 32 << 20; //32 MB
+        //32 MB - 64 bytes
+        // These buffers end up occupying 33554456 (32M + 24) bytes.
+        // On large JVMs (64G+), with G1Gc - the region size maxes out at
+        // 32M. Without the -64, this structure would end up using 2 regions.
+        return ((32 << 20) - 64);
       }
     }
 
@@ -301,14 +319,14 @@ public class PipelinedSorter extends ExternalSorter {
 
     if(newSpan == null) {
       //avoid sort/spill of empty span
-      Stopwatch stopWatch = new Stopwatch();
+      StopWatch stopWatch = new StopWatch();
       stopWatch.start();
       // sort in the same thread, do not wait for the thread pool
       merger.add(span.sort(sorter));
       boolean ret = spill(true);
       stopWatch.stop();
       if (LOG.isDebugEnabled()) {
-        LOG.debug(outputContext.getDestinationVertexName() + ": Time taken for spill " + (stopWatch.elapsedMillis()) + " ms");
+        LOG.debug(outputContext.getDestinationVertexName() + ": Time taken for spill " + (stopWatch.now(TimeUnit.MILLISECONDS)) + " ms");
       }
       if (pipelinedShuffle && ret) {
         sendPipelinedShuffleEvents();
@@ -333,6 +351,7 @@ public class PipelinedSorter extends ExternalSorter {
     } else {
       // queue up the sort
       SortTask task = new SortTask(span, sorter);
+      LOG.debug("Submitting span={} for sort", span.toString());
       Future<SpanIterator> future = sortmaster.submit(task);
       merger.add(future);
       span = newSpan;
@@ -345,9 +364,10 @@ public class PipelinedSorter extends ExternalSorter {
   private void sendPipelinedShuffleEvents() throws IOException{
     List<Event> events = Lists.newLinkedList();
     String pathComponent = (outputContext.getUniqueIdentifier() + "_" + (numSpills-1));
-    ShuffleUtils.generateEventOnSpill(events, isFinalMergeEnabled(), false, outputContext,
-        (numSpills - 1), indexCacheList.get(numSpills - 1), partitions, sendEmptyPartitionDetails,
-        pathComponent, partitionStats);
+    ShuffleUtils.generateEventOnSpill(events, isFinalMergeEnabled(), false,
+        outputContext, (numSpills - 1), indexCacheList.get(numSpills - 1),
+        partitions, sendEmptyPartitionDetails, pathComponent, partitionStats,
+        reportDetailedPartitionStats(), auxiliaryService, deflater);
     outputContext.sendEvents(events);
     LOG.info(outputContext.getDestinationVertexName() +
         ": Added spill event for spill (final update=false), spillId=" + (numSpills - 1));
@@ -462,6 +482,9 @@ public class PipelinedSorter extends ExternalSorter {
             * MAP_OUTPUT_INDEX_RECORD_LENGTH);
     spillFilePaths.put(numSpills, filename);
     FSDataOutputStream out = rfs.create(filename, true, 4096);
+    if (!SPILL_FILE_PERMS.equals(SPILL_FILE_PERMS.applyUMask(FsPermission.getUMask(conf)))) {
+      rfs.setPermission(filename, SPILL_FILE_PERMS);
+    }
 
     try {
       LOG.info(outputContext.getDestinationVertexName() + ": Spilling to " + filename.toString() +
@@ -473,8 +496,10 @@ public class PipelinedSorter extends ExternalSorter {
         Writer writer = null;
         try {
           long segmentStart = out.getPos();
-          writer = new Writer(conf, out, keyClass, valClass, codec,
-              spilledRecordsCounter, null, false);
+          if (!sendEmptyPartitionDetails || (i == partition)) {
+            writer = new Writer(conf, out, keyClass, valClass, codec,
+                spilledRecordsCounter, null, false);
+          }
           // we need not check for combiner since its a single record
           if (i == partition) {
             final long recordStart = out.getPos();
@@ -482,16 +507,18 @@ public class PipelinedSorter extends ExternalSorter {
             mapOutputRecordCounter.increment(1);
             mapOutputByteCounter.increment(out.getPos() - recordStart);
           }
-
-          writer.close();
-          adjustSpillCounters(writer.getRawLength(), writer.getCompressedLength());
-
+          long rawLength = 0;
+          long partLength = 0;
+          if (writer != null) {
+            writer.close();
+            rawLength = writer.getRawLength();
+            partLength = writer.getCompressedLength();
+          }
+          adjustSpillCounters(rawLength, partLength);
           // record offsets
           final TezIndexRecord rec =
               new TezIndexRecord(
-                  segmentStart,
-                  writer.getRawLength(),
-                  writer.getCompressedLength());
+                  segmentStart, rawLength, partLength);
           spillRec.putIndex(rec, i);
           writer = null;
         } finally {
@@ -543,6 +570,9 @@ public class PipelinedSorter extends ExternalSorter {
         mapOutputFile.getSpillFileForWrite(numSpills, size);
       spillFilePaths.put(numSpills, filename);
       out = rfs.create(filename, true, 4096);
+      if (!SPILL_FILE_PERMS.equals(SPILL_FILE_PERMS.applyUMask(FsPermission.getUMask(conf)))) {
+        rfs.setPermission(filename, SPILL_FILE_PERMS);
+      }
       LOG.info(outputContext.getDestinationVertexName() + ": Spilling to " + filename.toString());
       for (int i = 0; i < partitions; ++i) {
         if (isThreadInterrupted()) {
@@ -552,28 +582,36 @@ public class PipelinedSorter extends ExternalSorter {
         TezRawKeyValueIterator kvIter = merger.filter(i);
         //write merged output to disk
         long segmentStart = out.getPos();
-        Writer writer =
-          new Writer(conf, out, keyClass, valClass, codec,
+        Writer writer = null;
+        boolean hasNext = kvIter.hasNext();
+        if (hasNext || !sendEmptyPartitionDetails) {
+          writer = new Writer(conf, out, keyClass, valClass, codec,
               spilledRecordsCounter, null, merger.needsRLE());
+        }
         if (combiner == null) {
-          while(kvIter.next()) {
+          while (kvIter.next()) {
             writer.append(kvIter.getKey(), kvIter.getValue());
           }
         } else {          
-          runCombineProcessor(kvIter, writer);
+          if (hasNext) {
+            runCombineProcessor(kvIter, writer);
+          }
         }
+        long rawLength = 0;
+        long partLength = 0;
         //close
-        writer.close();
-        adjustSpillCounters(writer.getRawLength(), writer.getCompressedLength());
+        if (writer != null) {
+          writer.close();
+          rawLength = writer.getRawLength();
+          partLength = writer.getCompressedLength();
+        }
+        adjustSpillCounters(rawLength, partLength);
         // record offsets
-        final TezIndexRecord rec = 
-            new TezIndexRecord(
-                segmentStart,
-                writer.getRawLength(),
-                writer.getCompressedLength());
+        final TezIndexRecord rec =
+            new TezIndexRecord(segmentStart, rawLength, partLength);
         spillRec.putIndex(rec, i);
         if (!isFinalMergeEnabled() && reportPartitionStats()) {
-          partitionStats[i] += writer.getCompressedLength();
+          partitionStats[i] += partLength;
         }
       }
 
@@ -669,7 +707,8 @@ public class PipelinedSorter extends ExternalSorter {
           String pathComponent = (outputContext.getUniqueIdentifier() + "_" + i);
           ShuffleUtils.generateEventOnSpill(events, isFinalMergeEnabled(), isLastEvent,
               outputContext, i, indexCacheList.get(i), partitions,
-              sendEmptyPartitionDetails, pathComponent, partitionStats);
+              sendEmptyPartitionDetails, pathComponent, partitionStats,
+              reportDetailedPartitionStats(), auxiliaryService, deflater);
           LOG.info(outputContext.getDestinationVertexName() + ": Adding spill event for spill (final update=" + isLastEvent + "), spillId=" + i);
         }
         outputContext.sendEvents(events);
@@ -719,22 +758,28 @@ public class PipelinedSorter extends ExternalSorter {
       }
       //The output stream for the final single output file
       FSDataOutputStream finalOut = rfs.create(finalOutputFile, true, 4096);
+      if (!SPILL_FILE_PERMS.equals(SPILL_FILE_PERMS.applyUMask(FsPermission.getUMask(conf)))) {
+        rfs.setPermission(finalOutputFile, SPILL_FILE_PERMS);
+      }
 
       final TezSpillRecord spillRec = new TezSpillRecord(partitions);
 
       for (int parts = 0; parts < partitions; parts++) {
+        boolean shouldWrite = false;
         //create the segments to be merged
         List<Segment> segmentList =
             new ArrayList<Segment>(numSpills);
         for (int i = 0; i < numSpills; i++) {
           Path spillFilename = spillFilePaths.get(i);
           TezIndexRecord indexRecord = indexCacheList.get(i).getIndex(parts);
-
-          Segment s =
-              new Segment(rfs, spillFilename, indexRecord.getStartOffset(),
-                  indexRecord.getPartLength(), codec, ifileReadAhead,
-                  ifileReadAheadLength, ifileBufferSize, true);
-          segmentList.add(i, s);
+          if (indexRecord.hasData() || !sendEmptyPartitionDetails) {
+            shouldWrite = true;
+            DiskSegment s =
+                new DiskSegment(rfs, spillFilename, indexRecord.getStartOffset(),
+                    indexRecord.getPartLength(), codec, ifileReadAhead,
+                    ifileReadAheadLength, ifileBufferSize, true);
+            segmentList.add(s);
+          }
         }
 
         int mergeFactor =
@@ -750,32 +795,35 @@ public class PipelinedSorter extends ExternalSorter {
             (RawComparator) ConfigUtils.getIntermediateOutputKeyComparator(conf),
             progressable, sortSegments, true,
             null, spilledRecordsCounter, additionalSpillBytesRead,
-            null); // Not using any Progress in TezMerger. Should just work.
+            null, merger.needsRLE()); // Not using any Progress in TezMerger. Should just work.
         //write merged output to disk
         long segmentStart = finalOut.getPos();
-        Writer writer =
-            new Writer(conf, finalOut, keyClass, valClass, codec,
-                spilledRecordsCounter, null, merger.needsRLE());
-        if (combiner == null || numSpills < minSpillsForCombine) {
-          TezMerger.writeFile(kvIter, writer, progressable,
-              TezRuntimeConfiguration.TEZ_RUNTIME_RECORDS_BEFORE_PROGRESS_DEFAULT);
-        } else {
-          runCombineProcessor(kvIter, writer);
-        }
+        long rawLength = 0;
+        long partLength = 0;
+        if (shouldWrite) {
+          Writer writer =
+              new Writer(conf, finalOut, keyClass, valClass, codec,
+                  spilledRecordsCounter, null, merger.needsRLE());
+          if (combiner == null || numSpills < minSpillsForCombine) {
+            TezMerger.writeFile(kvIter, writer, progressable,
+                TezRuntimeConfiguration.TEZ_RUNTIME_RECORDS_BEFORE_PROGRESS_DEFAULT);
+          } else {
+            runCombineProcessor(kvIter, writer);
+          }
 
-        //close
-        writer.close();
-        outputBytesWithOverheadCounter.increment(writer.getRawLength());
+          //close
+          writer.close();
+          rawLength = writer.getRawLength();
+          partLength = writer.getCompressedLength();
+        }
+        outputBytesWithOverheadCounter.increment(rawLength);
 
         // record offsets
         final TezIndexRecord rec =
-            new TezIndexRecord(
-                segmentStart,
-                writer.getRawLength(),
-                writer.getCompressedLength());
+            new TezIndexRecord(segmentStart, rawLength, partLength);
         spillRec.putIndex(rec, parts);
         if (reportPartitionStats()) {
-          partitionStats[parts] += writer.getCompressedLength();
+          partitionStats[parts] += partLength;
         }
       }
 
@@ -805,6 +853,7 @@ public class PipelinedSorter extends ExternalSorter {
 
   private interface PartitionedRawKeyValueIterator extends TezRawKeyValueIterator {
     int getPartition();
+    Integer peekPartition();
   }
 
   private static class BufferStreamWrapper extends OutputStream
@@ -859,7 +908,7 @@ public class PipelinedSorter extends ExternalSorter {
     final byte[] rawkvmeta;
     final int kvmetabase;
     final ByteBuffer kvbuffer;
-    final DataOutputStream out;
+    final NonSyncDataOutputStream out;
     final RawComparator comparator;
     final byte[] imeta = new byte[METASIZE];
 
@@ -891,7 +940,7 @@ public class PipelinedSorter extends ExternalSorter {
       kvmeta = kvmetabuffer
                 .order(ByteOrder.nativeOrder())
                .asIntBuffer();
-      out = new DataOutputStream(
+      out = new NonSyncDataOutputStream(
               new BufferStreamWrapper(kvbuffer));
       this.comparator = comparator;
     }
@@ -967,8 +1016,15 @@ public class PipelinedSorter extends ExternalSorter {
           items = 1024*1024;
           perItem = 16;
         }
-        newSpan = new SortSpan(remaining, items, perItem,
-            ConfigUtils.getIntermediateOutputKeyComparator(conf));
+        final RawComparator newComparator = ConfigUtils.getIntermediateOutputKeyComparator(conf);
+        if (this.comparator == newComparator) {
+          LOG.warn("Same comparator used. comparator={}, newComparator={},"
+                  + " hashCode: comparator={}, newComparator={}",
+              this.comparator, newComparator,
+              System.identityHashCode(this.comparator),
+              System.identityHashCode(newComparator));
+        }
+        newSpan = new SortSpan(remaining, items, perItem, newComparator);
         newSpan.index = index+1;
         LOG.info(String.format(outputContext.getDestinationVertexName() + ": " + "New Span%d.length = %d, perItem = %d", newSpan.index, newSpan
             .length(), perItem) + ", counter:" + mapOutputRecordCounter.getValue());
@@ -1078,13 +1134,18 @@ public class PipelinedSorter extends ExternalSorter {
     public boolean next() {
       // caveat: since we use this as a comparable in the merger 
       if(kvindex == maxindex) return false;
-      if(kvindex % 100 == 0) {
-          progress.set((kvindex-maxindex) / (float)maxindex);
-      }
       kvindex += 1;
+      if(kvindex % 100 == 0) {
+        progress.set(1 - ((maxindex - kvindex) / (float) maxindex));
+      }
       return true;
     }
 
+    @Override
+    public boolean hasNext() {
+      return (kvindex == maxindex);
+    }
+
     public void close() {
     }
 
@@ -1102,6 +1163,14 @@ public class PipelinedSorter extends ExternalSorter {
       return partition;
     }
 
+    public Integer peekPartition() {
+      if (!hasNext()) {
+        return null;
+      } else {
+          return kvmeta.get(span.offsetFor(kvindex + 1) + PARTITION);
+      }
+    }
+
     @SuppressWarnings("unused")
     public int size() {
       return (maxindex - kvindex);
@@ -1220,6 +1289,23 @@ public class PipelinedSorter extends ExternalSorter {
       return false;
     }
 
+    @Override
+    public boolean hasNext() throws IOException {
+      if (dirty || iter.hasNext()) {
+        Integer part;
+        if (dirty) {
+          part = iter.getPartition();
+        } else {
+          part = iter.peekPartition();
+        }
+
+        if (part != null) {
+          return (part >>> (32 - partitionBits)) == partition;
+        }
+      }
+      return false;
+    }
+
     public void reset(int partition) {
       this.partition = partition;
     }
@@ -1245,6 +1331,12 @@ public class PipelinedSorter extends ExternalSorter {
     }
   }
 
+  @InterfaceAudience.Private
+  @VisibleForTesting
+  public boolean needsRLE() {
+    return merger.needsRLE();
+  }
+
   private final class SpanMerger implements PartitionedRawKeyValueIterator {
     InputByteBuffer key = new InputByteBuffer();
     InputByteBuffer value = new InputByteBuffer();
@@ -1276,6 +1368,7 @@ public class PipelinedSorter extends ExternalSorter {
     }
 
     public final boolean ready() throws IOException, InterruptedException {
+      int numSpanItr = futures.size();
       try {
         SpanIterator iter = null;
         while(this.futures.size() > 0) {
@@ -1297,8 +1390,11 @@ public class PipelinedSorter extends ExternalSorter {
         LOG.info(outputContext.getDestinationVertexName() + ": " + "Heap = " + sb.toString());
         return true;
       } catch(ExecutionException e) {
-        LOG.info(outputContext.getDestinationVertexName() + ": " + e.toString());
-        return false;
+        LOG.error("Heap size={}, total={}, eq={}, partition={}, gallop={}, totalItr={},"
+                + " futures.size={}, destVertexName={}",
+            heap.size(), total, eq, partition, gallop, numSpanItr, futures.size(),
+            outputContext.getDestinationVertexName(), e);
+        throw new IOException(e);
       }
     }
 
@@ -1349,6 +1445,20 @@ public class PipelinedSorter extends ExternalSorter {
       return false;
     }
 
+    @Override
+    public boolean hasNext() {
+      return peek() != null;
+    }
+
+    public Integer peekPartition() {
+      if (!hasNext()) {
+        return null;
+      } else {
+        SpanIterator peek = peek();
+        return peek.getPartition();
+      }
+    }
+
     public DataInputBuffer getKey() { return key; }
     public DataInputBuffer getValue() { return value; }
     public int getPartition() { return partition; }
