@@ -18,7 +18,6 @@
 
 package org.apache.tez.runtime.library.common.sort.impl.dflt;
 
-import java.io.DataOutputStream;
 import java.io.IOException;
 import java.io.OutputStream;
 import java.nio.ByteBuffer;
@@ -28,10 +27,13 @@ import java.util.ArrayList;
 import java.util.List;
 import java.util.concurrent.locks.Condition;
 import java.util.concurrent.locks.ReentrantLock;
+import java.util.zip.Deflater;
 
 import com.google.common.annotations.VisibleForTesting;
 import com.google.common.collect.Lists;
 import org.apache.commons.lang.exception.ExceptionUtils;
+import org.apache.hadoop.fs.permission.FsPermission;
+import org.apache.tez.dag.api.TezConfiguration;
 import org.apache.tez.runtime.library.api.IOInterruptedException;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
@@ -42,7 +44,9 @@ import org.apache.hadoop.io.DataInputBuffer;
 import org.apache.hadoop.io.RawComparator;
 import org.apache.hadoop.util.IndexedSortable;
 import org.apache.hadoop.util.Progress;
+import org.apache.tez.common.TezCommonUtils;
 import org.apache.tez.common.TezUtilsInternal;
+import org.apache.tez.common.io.NonSyncDataOutputStream;
 import org.apache.tez.runtime.api.Event;
 import org.apache.tez.runtime.api.OutputContext;
 import org.apache.tez.runtime.library.api.TezRuntimeConfiguration;
@@ -55,10 +59,13 @@ import org.apache.tez.runtime.library.common.sort.impl.TezMerger;
 import org.apache.tez.runtime.library.common.sort.impl.TezRawKeyValueIterator;
 import org.apache.tez.runtime.library.common.sort.impl.TezSpillRecord;
 import org.apache.tez.runtime.library.common.sort.impl.IFile.Writer;
+import org.apache.tez.runtime.library.common.sort.impl.TezMerger.DiskSegment;
 import org.apache.tez.runtime.library.common.sort.impl.TezMerger.Segment;
 
 import com.google.common.base.Preconditions;
 
+import static org.apache.tez.runtime.library.common.sort.impl.TezSpillRecord.SPILL_FILE_PERMS;
+
 @SuppressWarnings({"unchecked", "rawtypes"})
 public final class DefaultSorter extends ExternalSorter implements IndexedSortable {
   
@@ -74,7 +81,7 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
   private final static int APPROX_HEADER_LENGTH = 150;
 
   // k/v accounting
-  private final IntBuffer kvmeta; // metadata overlay on backing store
+  private IntBuffer kvmeta; // metadata overlay on backing store
   int kvstart;            // marks origin of spill metadata
   int kvend;              // marks end of spill metadata
   int kvindex;            // marks end of fully serialized records
@@ -87,7 +94,7 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
   int bufvoid;            // marks the point where we should stop
                           // reading at the end of the buffer
 
-  private final byte[] kvbuffer;        // main output buffer
+  private byte[] kvbuffer;        // main output buffer
   private final byte[] b0 = new byte[0];
 
   protected static final int VALSTART = 0;         // val offset in acct
@@ -111,6 +118,8 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
   final BlockingBuffer bb = new BlockingBuffer();
   volatile boolean spillThreadRunning = false;
   final SpillThread spillThread = new SpillThread();
+  private final Deflater deflater;
+  private final String auxiliaryService;
 
   final ArrayList<TezSpillRecord> indexCacheList =
     new ArrayList<TezSpillRecord>();
@@ -126,6 +135,7 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
   public DefaultSorter(OutputContext outputContext, Configuration conf, int numOutputs,
       long initialMemoryAvailable) throws IOException {
     super(outputContext, conf, numOutputs, initialMemoryAvailable);
+    deflater = TezCommonUtils.newBestCompressionDeflater();
     // sanity checks
     final float spillper = this.conf.getFloat(
         TezRuntimeConfiguration.TEZ_RUNTIME_SORT_SPILL_PERCENT,
@@ -148,6 +158,8 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
           TezRuntimeConfiguration.TEZ_RUNTIME_PIPELINED_SHUFFLE_ENABLED + " does not work "
           + "with DefaultSorter. It is supported only with PipelinedSorter.");
     }
+    auxiliaryService = conf.get(TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID,
+        TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID_DEFAULT);
 
     // buffers and accounting
     int maxMemUsage = sortmb << 20;
@@ -473,7 +485,7 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
   /**
    * Inner class managing the spill of serialized records to disk.
    */
-  protected class BlockingBuffer extends DataOutputStream {
+  protected class BlockingBuffer extends NonSyncDataOutputStream {
 
     public BlockingBuffer() {
       super(new Buffer());
@@ -740,6 +752,16 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
     }
   }
 
+  @Override
+  public void close() throws IOException {
+    super.close();
+    kvbuffer = null;
+    kvmeta = null;
+  }
+
+  boolean isClosed() {
+    return kvbuffer == null && kvmeta == null;
+  }
 
   protected class SpillThread extends Thread {
 
@@ -874,6 +896,9 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
           mapOutputFile.getSpillFileForWrite(numSpills, size);
       spillFilePaths.put(numSpills, filename);
       out = rfs.create(filename);
+      if (!SPILL_FILE_PERMS.equals(SPILL_FILE_PERMS.applyUMask(FsPermission.getUMask(conf)))) {
+        rfs.setPermission(filename, SPILL_FILE_PERMS);
+      }
 
       int spindex = mstart;
       final InMemValBytes value = createInMemValBytes();
@@ -882,8 +907,11 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
         IFile.Writer writer = null;
         try {
           long segmentStart = out.getPos();
-          writer = new Writer(conf, out, keyClass, valClass, codec,
-                                    spilledRecordsCounter, null, rle);
+          if (spindex < mend && kvmeta.get(offsetFor(spindex) + PARTITION) == i
+              || !sendEmptyPartitionDetails) {
+            writer = new Writer(conf, out, keyClass, valClass, codec,
+                spilledRecordsCounter, null, rle);
+          }
           if (combiner == null) {
             // spill directly
             DataInputBuffer key = new DataInputBuffer();
@@ -915,21 +943,22 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
               runCombineProcessor(kvIter, writer);
             }
           }
-
+          long rawLength = 0;
+          long partLength = 0;
           // close the writer
-          writer.close();
-          adjustSpillCounters(writer.getRawLength(), writer.getCompressedLength());
+          if (writer != null) {
+            writer.close();
+            rawLength = writer.getRawLength();
+            partLength = writer.getCompressedLength();
+          }
+          adjustSpillCounters(rawLength, partLength);
           // record offsets
           final TezIndexRecord rec =
-              new TezIndexRecord(
-                  segmentStart,
-                  writer.getRawLength(),
-                  writer.getCompressedLength());
+              new TezIndexRecord(segmentStart, rawLength, partLength);
           spillRec.putIndex(rec, i);
-          if (!isFinalMergeEnabled() && reportPartitionStats()) {
-            partitionStats[i] += writer.getCompressedLength();
+          if (!isFinalMergeEnabled() && reportPartitionStats() && writer != null) {
+            partitionStats[i] += partLength;
           }
-
           writer = null;
         } finally {
           if (null != writer) writer.close();
@@ -948,7 +977,8 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
         totalIndexCacheMemory +=
           spillRec.size() * MAP_OUTPUT_INDEX_RECORD_LENGTH;
       }
-      LOG.info(outputContext.getDestinationVertexName() + ": " + "Finished spill " + numSpills);
+      LOG.info(outputContext.getDestinationVertexName() + ": " + "Finished spill " + numSpills
+      + " at " + filename.toString());
       ++numSpills;
       if (!isFinalMergeEnabled()) {
         numShuffleChunks.setValue(numSpills);
@@ -977,6 +1007,9 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
           mapOutputFile.getSpillFileForWrite(numSpills, size);
       spillFilePaths.put(numSpills, filename);
       out = rfs.create(filename);
+      if (!SPILL_FILE_PERMS.equals(SPILL_FILE_PERMS.applyUMask(FsPermission.getUMask(conf)))) {
+        rfs.setPermission(filename, SPILL_FILE_PERMS);
+      }
 
       // we don't run the combiner for a single record
       for (int i = 0; i < partitions; ++i) {
@@ -984,9 +1017,10 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
         try {
           long segmentStart = out.getPos();
           // Create a new codec, don't care!
-          writer = new IFile.Writer(conf, out, keyClass, valClass, codec,
-                                          spilledRecordsCounter, null);
-
+          if (!sendEmptyPartitionDetails || (i == partition)) {
+            writer = new Writer(conf, out, keyClass, valClass, codec,
+                spilledRecordsCounter, null, false);
+          }
           if (i == partition) {
             final long recordStart = out.getPos();
             writer.append(key, value);
@@ -994,16 +1028,17 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
             // compression
             mapOutputByteCounter.increment(out.getPos() - recordStart);
           }
-          writer.close();
-
-          adjustSpillCounters(writer.getRawLength(), writer.getCompressedLength());
+          long rawLength =0;
+          long partLength =0;
+          if (writer != null) {
+            writer.close();
+            rawLength = writer.getRawLength();
+            partLength = writer.getCompressedLength();
+          }
+          adjustSpillCounters(rawLength, partLength);
 
           // record offsets
-          TezIndexRecord rec =
-              new TezIndexRecord(
-                  segmentStart,
-                  writer.getRawLength(),
-                  writer.getCompressedLength());
+          TezIndexRecord rec = new TezIndexRecord(segmentStart, rawLength, partLength);
           spillRec.putIndex(rec, i);
 
           writer = null;
@@ -1097,6 +1132,12 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
       this.end = end;
       current = start - 1;
     }
+
+    @Override
+    public boolean hasNext() throws IOException {
+        return (current + 1) < end;
+    }
+
     public boolean next() throws IOException {
       return ++current < end;
     }
@@ -1132,7 +1173,7 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
     String pathComponent = (outputContext.getUniqueIdentifier() + "_" + index);
     ShuffleUtils.generateEventOnSpill(events, isFinalMergeEnabled(), isLastEvent,
         outputContext, index, spillRecord, partitions, sendEmptyPartitionDetails, pathComponent,
-        partitionStats);
+        partitionStats, reportDetailedPartitionStats(), auxiliaryService, deflater);
 
     LOG.info(outputContext.getDestinationVertexName() + ": " +
         "Adding spill event for spill (final update=" + isLastEvent + "), spillId=" + index);
@@ -1242,26 +1283,30 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
 
     //The output stream for the final single output file
     FSDataOutputStream finalOut = rfs.create(finalOutputFile, true, 4096);
+    if (!SPILL_FILE_PERMS.equals(SPILL_FILE_PERMS.applyUMask(FsPermission.getUMask(conf)))) {
+      rfs.setPermission(finalOutputFile, SPILL_FILE_PERMS);
+    }
 
     if (numSpills == 0) {
       // TODO Change event generation to say there is no data rather than generating a dummy file
       //create dummy files
-
+      long rawLength = 0;
+      long partLength = 0;
       TezSpillRecord sr = new TezSpillRecord(partitions);
       try {
         for (int i = 0; i < partitions; i++) {
           long segmentStart = finalOut.getPos();
-          Writer writer =
-            new Writer(conf, finalOut, keyClass, valClass, codec, null, null);
-          writer.close();
-
+          if (!sendEmptyPartitionDetails) {
+            Writer writer =
+                new Writer(conf, finalOut, keyClass, valClass, codec, null, null);
+            writer.close();
+            rawLength = writer.getRawLength();
+            partLength = writer.getCompressedLength();
+          }
           TezIndexRecord rec =
-              new TezIndexRecord(
-                  segmentStart,
-                  writer.getRawLength(),
-                  writer.getCompressedLength());
+              new TezIndexRecord(segmentStart, rawLength, partLength);
           // Covers the case of multiple spills.
-          outputBytesWithOverheadCounter.increment(writer.getRawLength());
+          outputBytesWithOverheadCounter.increment(rawLength);
           sr.putIndex(rec, i);
         }
         sr.writeToFile(finalIndexFile, conf);
@@ -1280,19 +1325,21 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
     else {
       final TezSpillRecord spillRec = new TezSpillRecord(partitions);
       for (int parts = 0; parts < partitions; parts++) {
+        boolean shouldWrite = false;
         //create the segments to be merged
         List<Segment> segmentList =
-          new ArrayList<Segment>(numSpills);
-        for(int i = 0; i < numSpills; i++) {
+            new ArrayList<Segment>(numSpills);
+        for (int i = 0; i < numSpills; i++) {
           outputContext.notifyProgress();
           TezIndexRecord indexRecord = indexCacheList.get(i).getIndex(parts);
-
-          Segment s =
-            new Segment(rfs, filename[i], indexRecord.getStartOffset(),
-                             indexRecord.getPartLength(), codec, ifileReadAhead,
-                             ifileReadAheadLength, ifileBufferSize, true);
-          segmentList.add(i, s);
-
+          if (indexRecord.hasData() || !sendEmptyPartitionDetails) {
+            shouldWrite = true;
+            DiskSegment s =
+              new DiskSegment(rfs, filename[i], indexRecord.getStartOffset(),
+                               indexRecord.getPartLength(), codec, ifileReadAhead,
+                               ifileReadAheadLength, ifileBufferSize, true);
+            segmentList.add(s);
+          }
           if (LOG.isDebugEnabled()) {
             LOG.debug(outputContext.getDestinationVertexName() + ": "
                 + "TaskIdentifier=" + taskIdentifier + " Partition=" + parts +
@@ -1319,6 +1366,9 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
 
         //write merged output to disk
         long segmentStart = finalOut.getPos();
+        long rawLength = 0;
+        long partLength = 0;
+        if (shouldWrite) {
         Writer writer =
             new Writer(conf, finalOut, keyClass, valClass, codec,
                 spilledRecordsCounter, null);
@@ -1329,17 +1379,16 @@ public final class DefaultSorter extends ExternalSorter implements IndexedSortab
           runCombineProcessor(kvIter, writer);
         }
         writer.close();
-        outputBytesWithOverheadCounter.increment(writer.getRawLength());
-
-        // record offsets
-        final TezIndexRecord rec =
-            new TezIndexRecord(
-                segmentStart,
-                writer.getRawLength(),
-                writer.getCompressedLength());
+        rawLength = writer.getRawLength();
+        partLength = writer.getCompressedLength();
+      }
+      outputBytesWithOverheadCounter.increment(rawLength);
+      // record offsets
+      final TezIndexRecord rec =
+          new TezIndexRecord(segmentStart, rawLength, partLength);
         spillRec.putIndex(rec, parts);
         if (reportPartitionStats()) {
-          partitionStats[parts] += writer.getCompressedLength();
+          partitionStats[parts] += partLength;
         }
       }
       numShuffleChunks.setValue(1); //final merge has happened
