@@ -212,6 +212,13 @@ public class OrderedGroupedKVInputConfig {
       return this;
     }
 
+    @Override
+    public SpecificBuilder setFromConfigurationUnfiltered(
+        Configuration conf) {
+      builder.setFromConfigurationUnfiltered(conf);
+      return this;
+    }
+
     public E done() {
       return edgeBuilder;
     }
@@ -448,6 +455,15 @@ public class OrderedGroupedKVInputConfig {
       return this;
     }
 
+    @SuppressWarnings("unchecked")
+    @Override
+    public Builder setFromConfigurationUnfiltered(Configuration conf) {
+      // Maybe ensure this is the first call ? Otherwise this can end up overriding other parameters
+      Preconditions.checkArgument(conf != null, "Configuration cannot be null");
+      ConfigUtils.mergeConfs(this.conf, conf);
+      return this;
+    }
+
     public Builder setCompression(boolean enabled, @Nullable String compressionCodec,
                                   @Nullable Map<String, String> codecConf) {
       this.conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_COMPRESS, enabled);
