@@ -177,6 +177,13 @@ public class OrderedPartitionedKVOutputConfig {
       return this;
     }
 
+    @Override
+    public SpecificBuilder setFromConfigurationUnfiltered(
+        Configuration conf) {
+      builder.setFromConfigurationUnfiltered(conf);
+      return this;
+    }
+
     public E done() {
       return edgeBuilder;
     }
@@ -374,6 +381,15 @@ public class OrderedPartitionedKVOutputConfig {
       return this;
     }
 
+    @SuppressWarnings("unchecked")
+    @Override
+    public Builder setFromConfigurationUnfiltered(Configuration conf) {
+      // Maybe ensure this is the first call ? Otherwise this can end up overriding other parameters
+      Preconditions.checkArgument(conf != null, "Configuration cannot be null");
+      ConfigUtils.mergeConfs(this.conf, conf);
+      return this;
+    }
+
     /**
      * Set the key comparator class
      *
