@@ -19,15 +19,21 @@
 package org.apache.tez.runtime.library.common.sort.impl;
 
 import java.io.ByteArrayInputStream;
+import java.io.ByteArrayOutputStream;
 import java.io.IOException;
+import java.net.URISyntaxException;
+import java.net.URL;
 import java.util.ArrayList;
+import java.util.Arrays;
 import java.util.LinkedList;
 import java.util.List;
 import java.util.Random;
 
+import org.junit.Assert;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 import org.apache.hadoop.conf.Configuration;
+import org.apache.hadoop.fs.ChecksumException;
 import org.apache.hadoop.fs.FSDataInputStream;
 import org.apache.hadoop.fs.FSDataOutputStream;
 import org.apache.hadoop.fs.FileSystem;
@@ -165,6 +171,114 @@ public class TestIFile {
   }
 
 
+  @Test(timeout = 5000)
+  //test overflow
+  public void testExceedMaxSize() throws IOException {
+    final int oldMaxBufferSize = IFile.Reader.MAX_BUFFER_SIZE;
+
+    Text shortString = new Text("string");
+    Text longString = new Text("A string of length 22.");
+    assertEquals(22, longString.getLength());
+
+    Text readKey = new Text();
+    Text readValue = new Text();
+    DataInputBuffer keyIn = new DataInputBuffer();
+    DataInputBuffer valIn = new DataInputBuffer();
+
+    IFile.Writer writer;
+    IFile.Reader reader;
+    FSDataOutputStream out;
+
+    // Check Key length exceeding MAX_BUFFER_SIZE
+    out = localFs.create(outputPath);
+    writer = new IFile.Writer(defaultConf, out,
+            Text.class, Text.class, null, null, null, false);
+    writer.append(longString, shortString);
+    writer.close();
+
+    out.close();
+
+    // Set this to a smaller value for testing
+    IFile.Reader.MAX_BUFFER_SIZE = 16;
+
+    reader = new IFile.Reader(localFs, outputPath,
+            null, null, null, false, 0, -1);
+
+    try {
+      reader.nextRawKey(keyIn);
+      Assert.fail("Expected IllegalArgumentException to be thrown");
+    } catch (IllegalArgumentException e) {
+      // test passed
+    }
+    reader.close();
+
+    // Check Value length exceeding MAX_BUFFER_SIZE
+    out = localFs.create(outputPath);
+    writer = new IFile.Writer(defaultConf, out,
+            Text.class, Text.class, null, null, null, false);
+    writer.append(shortString, longString);
+    writer.close();
+
+    out.close();
+
+    // Set this to a smaller value for testing
+    IFile.Reader.MAX_BUFFER_SIZE = 16;
+
+    reader = new IFile.Reader(localFs, outputPath,
+            null, null, null, false, 0, -1);
+
+    try {
+      reader.nextRawKey(keyIn);
+      reader.nextRawValue(valIn);
+      Assert.fail("Expected IllegalArgumentException to be thrown");
+    } catch (IllegalArgumentException e) {
+      // test passed
+    }
+    reader.close();
+
+    // Check Key length not getting doubled
+    out = localFs.create(outputPath);
+    writer = new IFile.Writer(defaultConf, out,
+            Text.class, Text.class, null, null, null, false);
+    writer.append(longString, shortString);
+    writer.close();
+
+    out.close();
+
+    // Set this to a smaller value for testing
+    IFile.Reader.MAX_BUFFER_SIZE = 32;
+
+    reader = new IFile.Reader(localFs, outputPath,
+            null, null, null, false, 0, -1);
+
+    reader.nextRawKey(keyIn);
+    assertEquals(longString.getLength() + 1, keyIn.getData().length);
+    reader.close();
+
+    // Check Value length not getting doubled
+    out = localFs.create(outputPath);
+    writer = new IFile.Writer(defaultConf, out,
+            Text.class, Text.class, null, null, null, false);
+    writer.append(shortString, longString);
+    writer.close();
+
+    out.close();
+
+    // Set this to a smaller value for testing
+    IFile.Reader.MAX_BUFFER_SIZE = 32;
+
+    reader = new IFile.Reader(localFs, outputPath,
+            null, null, null, false, 0, -1);
+
+    reader.nextRawKey(keyIn);
+    reader.nextRawValue(valIn);
+    assertEquals(longString.getLength() + 1, valIn.getData().length);
+    reader.close();
+
+    // revert back to original value
+    IFile.Reader.MAX_BUFFER_SIZE = oldMaxBufferSize;
+  }
+
   @Test(timeout = 5000)
   //test with sorted data and repeat keys
   public void testWithRLEMarker() throws IOException {
@@ -199,6 +313,9 @@ public class TestIFile {
     writer = new IFile.Writer(defaultConf, out,
         Text.class, IntWritable.class, codec, null, null, true);
 
+    BoundedByteArrayOutputStream boundedOut = new BoundedByteArrayOutputStream(1024*1024);
+    Writer inMemWriter = new InMemoryWriter(boundedOut, true);
+
     DataInputBuffer kin = new DataInputBuffer();
     kin.reset(kvbuffer, pos, keyLength);
 
@@ -210,6 +327,8 @@ public class TestIFile {
     //Write initial KV pair
     writer.append(kin, vin);
     assertFalse(writer.sameKey);
+    inMemWriter.append(kin, vin);
+    assertFalse(inMemWriter.sameKey);
     pos += (keyLength + valueLength);
 
     //Second key is similar to key1 (RLE should kick in)
@@ -218,6 +337,8 @@ public class TestIFile {
     vin.reset(vout.getData(), vout.getLength());
     writer.append(kin, vin);
     assertTrue(writer.sameKey);
+    inMemWriter.append(kin, vin);
+    assertTrue(inMemWriter.sameKey);
     pos += (keyLength + valueLength);
 
     //Next key (key3) is different (RLE should not kick in)
@@ -226,9 +347,13 @@ public class TestIFile {
     vin.reset(vout.getData(), vout.getLength());
     writer.append(kin, vin);
     assertFalse(writer.sameKey);
+    inMemWriter.append(kin, vin);
+    assertFalse(inMemWriter.sameKey);
 
     writer.close();
     out.close();
+    inMemWriter.close();
+    boundedOut.close();
   }
 
   @Test(timeout = 5000)
@@ -240,6 +365,58 @@ public class TestIFile {
     testWithDataBuffer(sortedData);
   }
 
+  //test concatenated zlib input - as in multiple map outputs during shuffle
+  //This specific input is valid but the decompressor can leave lingering
+  // bytes between segments. If the lingering bytes aren't handled correctly,
+  // the stream will get out-of-sync.
+  @Test(timeout = 5000)
+  public void testConcatenatedZlibPadding()
+      throws IOException, URISyntaxException {
+    byte[] bytes;
+    long compTotal = 0;
+    // Known raw and compressed lengths of input
+    long raws[] = { 2392, 102314, 42576, 31432, 25090 };
+    long compressed[] = { 723, 25396, 10926, 8203, 6665 };
+
+    CompressionCodecFactory codecFactory = new CompressionCodecFactory(new
+        Configuration());
+    codec = codecFactory.getCodecByClassName("org.apache.hadoop.io.compress.DefaultCodec");
+
+    URL url = getClass().getClassLoader()
+        .getResource("TestIFile_concatenated_compressed.bin");
+    assertNotEquals("IFileinput file must exist", null, url);
+    Path p = new Path(url.toURI());
+    FSDataInputStream inStream = localFs.open(p);
+
+    for (int i = 0; i < 5; i++) {
+      bytes = new byte[(int) raws[i]];
+      assertEquals("Compressed stream out-of-sync", inStream.getPos(), compTotal);
+      IFile.Reader.readToMemory(bytes, inStream, (int) compressed[i], codec,
+          false, -1);
+      compTotal += compressed[i];
+
+      // Now read the data
+      InMemoryReader inMemReader = new InMemoryReader(null,
+          new InputAttemptIdentifier(0, 0), bytes, 0, bytes.length);
+
+      DataInputBuffer keyIn = new DataInputBuffer();
+      DataInputBuffer valIn = new DataInputBuffer();
+      Deserializer<Text> keyDeserializer;
+      Deserializer<IntWritable> valDeserializer;
+      SerializationFactory serializationFactory =
+          new SerializationFactory(defaultConf);
+      keyDeserializer = serializationFactory.getDeserializer(Text.class);
+      valDeserializer = serializationFactory.getDeserializer(IntWritable.class);
+      keyDeserializer.open(keyIn);
+      valDeserializer.open(valIn);
+
+      while (inMemReader.nextRawKey(keyIn)) {
+        inMemReader.nextRawValue(valIn);
+      }
+    }
+    inStream.close();
+  }
+
   @Test(timeout = 5000)
   //Test InMemoryWriter
   public void testInMemoryWriter() throws IOException {
@@ -250,25 +427,25 @@ public class TestIFile {
 
     //No RLE, No RepeatKeys, no compression
     writer = new InMemoryWriter(bout);
-    writeTestFileUsingDataBuffer(writer, false, false, data, null);
+    writeTestFileUsingDataBuffer(writer, false, data);
     readUsingInMemoryReader(bout.getBuffer(), data);
 
     //No RLE, RepeatKeys, no compression
     bout.reset();
     writer = new InMemoryWriter(bout);
-    writeTestFileUsingDataBuffer(writer, false, true, data, null);
+    writeTestFileUsingDataBuffer(writer, true, data);
     readUsingInMemoryReader(bout.getBuffer(), data);
 
     //RLE, No RepeatKeys, no compression
     bout.reset();
-    writer = new InMemoryWriter(bout);
-    writeTestFileUsingDataBuffer(writer, true, false, data, null);
+    writer = new InMemoryWriter(bout, true);
+    writeTestFileUsingDataBuffer(writer, false, data);
     readUsingInMemoryReader(bout.getBuffer(), data);
 
     //RLE, RepeatKeys, no compression
     bout.reset();
-    writer = new InMemoryWriter(bout);
-    writeTestFileUsingDataBuffer(writer, true, true, data, null);
+    writer = new InMemoryWriter(bout, true);
+    writeTestFileUsingDataBuffer(writer, true, data);
     readUsingInMemoryReader(bout.getBuffer(), data);
   }
 
@@ -375,6 +552,43 @@ public class TestIFile {
     readAndVerifyData(writer.getRawLength(), writer.getCompressedLength(), data, codec);
   }
 
+  @Test(timeout = 20000)
+  public void testReadToDisk() throws IOException {
+    // verify sending a stream of zeroes generates an error
+    byte[] zeroData = new byte[1000];
+    Arrays.fill(zeroData, (byte) 0);
+    ByteArrayInputStream in = new ByteArrayInputStream(zeroData);
+    try {
+      IFile.Reader.readToDisk(new ByteArrayOutputStream(), in, zeroData.length, false, 0);
+      fail("Exception should have been thrown");
+    } catch (IOException e) {
+    }
+
+    // verify sending same stream of zeroes with a valid IFile header still
+    // generates an error
+    ByteArrayOutputStream baos = new ByteArrayOutputStream();
+    baos.write(IFile.HEADER);
+    baos.write(zeroData);
+    try {
+      IFile.Reader.readToDisk(new ByteArrayOutputStream(),
+          new ByteArrayInputStream(baos.toByteArray()), zeroData.length, false, 0);
+      fail("Exception should have been thrown");
+    } catch (IOException e) {
+      assertTrue(e instanceof ChecksumException);
+    }
+
+    // verify valid data is copied properly
+    List<KVPair> data = KVDataGen.generateTestData(true, 0);
+    Writer writer = writeTestFile(false, false, data, codec);
+    baos.reset();
+    IFile.Reader.readToDisk(baos, localFs.open(outputPath), writer.getCompressedLength(),
+        false, 0);
+    byte[] diskData = baos.toByteArray();
+    Reader reader = new Reader(new ByteArrayInputStream(diskData), diskData.length,
+        codec, null, null, false, 0, 1024);
+    verifyData(reader, data);
+    reader.close();
+  }
 
   /**
    * Test different options (RLE, repeat keys, compression) on reader/writer
@@ -550,13 +764,13 @@ public class TestIFile {
     FSDataOutputStream out = localFs.create(outputPath);
     IFile.Writer writer = new IFile.Writer(defaultConf, out,
         Text.class, IntWritable.class, codec, null, null, rle);
-    writeTestFile(writer, rle, repeatKeys, data, codec);
+    writeTestFile(writer, repeatKeys, data);
     out.close();
     return  writer;
   }
 
-  private Writer writeTestFile(IFile.Writer writer, boolean rle, boolean repeatKeys,
-      List<KVPair> data, CompressionCodec codec) throws IOException {
+  private Writer writeTestFile(IFile.Writer writer, boolean repeatKeys,
+      List<KVPair> data) throws IOException {
     assertNotNull(writer);
 
     Text previousKey = null;
@@ -583,13 +797,13 @@ public class TestIFile {
     FSDataOutputStream out = localFs.create(outputPath);
     IFile.Writer writer = new IFile.Writer(defaultConf, out,
         Text.class, IntWritable.class, codec, null, null, rle);
-    writeTestFileUsingDataBuffer(writer, rle, repeatKeys, data, codec);
+    writeTestFileUsingDataBuffer(writer, repeatKeys, data);
     out.close();
     return writer;
   }
 
-  private Writer writeTestFileUsingDataBuffer(IFile.Writer writer, boolean rle, boolean repeatKeys,
-      List<KVPair> data, CompressionCodec codec) throws IOException {
+  private Writer writeTestFileUsingDataBuffer(Writer writer, boolean repeatKeys,
+      List<KVPair> data) throws IOException {
     DataInputBuffer previousKey = new DataInputBuffer();
     DataInputBuffer key = new DataInputBuffer();
     DataInputBuffer value = new DataInputBuffer();
