@@ -21,7 +21,9 @@ package org.apache.tez.runtime.library.common.sort.impl;
 import com.google.common.collect.Maps;
 import org.apache.commons.lang.RandomStringUtils;
 import org.apache.hadoop.conf.Configuration;
+import org.apache.hadoop.fs.CommonConfigurationKeys;
 import org.apache.hadoop.fs.FileSystem;
+import org.apache.hadoop.fs.LocalDirAllocator;
 import org.apache.hadoop.fs.Path;
 import org.apache.hadoop.io.DataInputBuffer;
 import org.apache.hadoop.io.DataOutputBuffer;
@@ -34,12 +36,17 @@ import org.apache.tez.common.TezRuntimeFrameworkConfigs;
 import org.apache.tez.common.counters.TaskCounter;
 import org.apache.tez.common.counters.TezCounter;
 import org.apache.tez.common.counters.TezCounters;
+import org.apache.tez.dag.api.TezConfiguration;
 import org.apache.tez.runtime.api.Event;
 import org.apache.tez.runtime.api.ExecutionContext;
 import org.apache.tez.runtime.api.OutputContext;
 import org.apache.tez.runtime.api.OutputStatisticsReporter;
+import org.apache.tez.runtime.api.TaskContext;
 import org.apache.tez.runtime.api.impl.ExecutionContextImpl;
 import org.apache.tez.runtime.library.api.TezRuntimeConfiguration;
+import org.apache.tez.runtime.library.api.TezRuntimeConfiguration.ReportPartitionStats;
+import org.apache.tez.runtime.library.common.Constants;
+import org.apache.tez.runtime.library.common.combine.Combiner;
 import org.apache.tez.runtime.library.common.shuffle.ShuffleUtils;
 import org.apache.tez.runtime.library.conf.OrderedPartitionedKVOutputConfig.SorterImpl;
 import org.apache.tez.runtime.library.partitioner.HashPartitioner;
@@ -56,6 +63,7 @@ import java.util.Map;
 import java.util.TreeMap;
 import java.util.UUID;
 
+import static org.junit.Assert.assertFalse;
 import static org.junit.Assert.assertTrue;
 import static org.mockito.Matchers.anyListOf;
 import static org.mockito.Mockito.atLeastOnce;
@@ -65,8 +73,10 @@ import static org.mockito.Mockito.verify;
 import static org.mockito.internal.verification.VerificationModeFactory.times;
 
 public class TestPipelinedSorter {
+  private static Configuration conf;
   private static FileSystem localFs = null;
   private static Path workDir = null;
+  private static LocalDirAllocator dirAllocator;
   private OutputContext outputContext;
 
   private int numOutputs;
@@ -76,13 +86,14 @@ public class TestPipelinedSorter {
   private static TreeMap<String, String> sortedDataMap = Maps.newTreeMap();
 
   static {
-    Configuration conf = getConf();
+    conf = getConf();
     try {
       localFs = FileSystem.getLocal(conf);
       workDir = new Path(
           new Path(System.getProperty("test.build.data", "/tmp")),
           TestPipelinedSorter.class.getName())
           .makeQualified(localFs.getUri(), localFs.getWorkingDirectory());
+      dirAllocator = new LocalDirAllocator(TezRuntimeFrameworkConfigs.LOCAL_DIRS);
     } catch (IOException e) {
       throw new RuntimeException(e);
     }
@@ -95,15 +106,19 @@ public class TestPipelinedSorter {
 
   @Before
   public void setup() throws IOException {
+    conf = getConf();
     ApplicationId appId = ApplicationId.newInstance(10000, 1);
     TezCounters counters = new TezCounters();
     String uniqueId = UUID.randomUUID().toString();
-    this.outputContext = createMockOutputContext(counters, appId, uniqueId);
+    String auxiliaryService = conf.get(TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID,
+        TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID_DEFAULT);
+    this.outputContext = createMockOutputContext(counters, appId, uniqueId, auxiliaryService);
   }
 
   public static Configuration getConf() {
     Configuration conf = new Configuration();
     conf.set("fs.defaultFS", "file:///");
+    conf.set(CommonConfigurationKeys.FS_PERMISSIONS_UMASK_KEY, "077");
     //To enable PipelinedSorter
     conf.set(TezRuntimeConfiguration.TEZ_RUNTIME_SORTER_CLASS, SorterImpl.PIPELINED.name());
 
@@ -132,15 +147,13 @@ public class TestPipelinedSorter {
     //TODO: need to support multiple partition testing later
 
     //# partition, # of keys, size per key, InitialMem, blockSize
-    Configuration conf = getConf();
     conf.setInt(TezRuntimeConfiguration.TEZ_RUNTIME_IO_SORT_MB, 5);
     basicTest(1, 100000, 100, (10 * 1024l * 1024l), 3 << 20);
-
+    verifyOutputPermissions(outputContext.getUniqueIdentifier());
   }
 
   @Test
   public void testWithoutPartitionStats() throws IOException {
-    Configuration conf = getConf();
     conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_REPORT_PARTITION_STATS, false);
     //# partition, # of keys, size per key, InitialMem, blockSize
     basicTest(1, 0, 0, (10 * 1024l * 1024l), 3 << 20);
@@ -149,7 +162,6 @@ public class TestPipelinedSorter {
 
   @Test
   public void testWithEmptyData() throws IOException {
-    Configuration conf = getConf();
     conf.setInt(TezRuntimeConfiguration.TEZ_RUNTIME_IO_SORT_MB, 5);
     //# partition, # of keys, size per key, InitialMem, blockSize
     basicTest(1, 0, 0, (10 * 1024l * 1024l), 3 << 20);
@@ -159,7 +171,6 @@ public class TestPipelinedSorter {
   public void testEmptyDataWithPipelinedShuffle() throws IOException {
     this.numOutputs = 1;
     this.initialAvailableMem = 1 *1024 * 1024;
-    Configuration conf = getConf();
     conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_ENABLE_FINAL_MERGE_IN_OUTPUT, false);
     conf.setInt(TezRuntimeConfiguration
         .TEZ_RUNTIME_PIPELINED_SORTER_MIN_BLOCK_SIZE_IN_MB, 1);
@@ -171,11 +182,65 @@ public class TestPipelinedSorter {
     // final merge is disabled. Final output file would not be populated in this case.
     assertTrue(sorter.finalOutputFile == null);
     TezCounter numShuffleChunks = outputContext.getCounters().findCounter(TaskCounter.SHUFFLE_CHUNK_COUNT);
-    assertTrue(sorter.getNumSpills() == numShuffleChunks.getValue());
+//    assertTrue(sorter.getNumSpills() == numShuffleChunks.getValue());
     conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_ENABLE_FINAL_MERGE_IN_OUTPUT, true);
 
   }
 
+  @Test
+  public void testEmptyPartitionsTwoSpillsNoEmptyEvents() throws Exception {
+    testEmptyPartitionsHelper(2, false);
+  }
+
+  @Test
+  public void testEmptyPartitionsTwoSpillsWithEmptyEvents() throws Exception {
+    testEmptyPartitionsHelper(2, true);
+  }
+
+  @Test
+  public void testEmptyPartitionsNoSpillsNoEmptyEvents() throws Exception {
+    testEmptyPartitionsHelper(0, false);
+  }
+
+  @Test
+  public void testEmptyPartitionsNoSpillsWithEmptyEvents() throws Exception {
+    testEmptyPartitionsHelper(0, true);
+  }
+
+  public void testEmptyPartitionsHelper(int numKeys, boolean sendEmptyPartitionDetails) throws IOException, InterruptedException {
+    int partitions = 50;
+    this.numOutputs = partitions;
+    this.initialAvailableMem = 1 *1024 * 1024;
+    conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_EMPTY_PARTITION_INFO_VIA_EVENTS_ENABLED, sendEmptyPartitionDetails);
+    conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_ENABLE_FINAL_MERGE_IN_OUTPUT, true);
+    conf.setInt(TezRuntimeConfiguration
+        .TEZ_RUNTIME_PIPELINED_SORTER_MIN_BLOCK_SIZE_IN_MB, 1);
+    PipelinedSorter sorter = new PipelinedSorter(this.outputContext, conf, partitions,
+        initialAvailableMem);
+
+    writeData(sorter, numKeys, 1000000);
+    if (numKeys == 0) {
+      assertTrue(sorter.getNumSpills() == 1);
+    } else {
+      assertTrue(sorter.getNumSpills() == numKeys + 1);
+    }
+    verifyCounters(sorter, outputContext);
+    verifyOutputPermissions(outputContext.getUniqueIdentifier());
+    Path indexFile = sorter.getFinalIndexFile();
+    TezSpillRecord spillRecord = new TezSpillRecord(indexFile, conf);
+    for (int i = 0; i < partitions; i++) {
+      TezIndexRecord tezIndexRecord = spillRecord.getIndex(i);
+      if (tezIndexRecord.hasData()) {
+        continue;
+      }
+      if (sendEmptyPartitionDetails) {
+        Assert.assertEquals("Unexpected raw length for " + i + "th partition", 0, tezIndexRecord.getRawLength());
+      } else {
+        Assert.assertEquals("Unexpected raw length for " + i + "th partition", 6, tezIndexRecord.getRawLength());
+      }
+    }
+  }
+
   @Test
   public void basicTestWithSmallBlockSize() throws IOException {
     //3 MB key & 3 MB value, whereas block size is just 3 MB
@@ -203,7 +268,6 @@ public class TestPipelinedSorter {
 
   @Test
   public void testExceedsKVWithMultiplePartitions() throws IOException {
-    Configuration conf = getConf();
     conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_ENABLE_FINAL_MERGE_IN_OUTPUT, true);
     this.numOutputs = 5;
     this.initialAvailableMem = 1 * 1024 * 1024;
@@ -214,13 +278,13 @@ public class TestPipelinedSorter {
 
     writeData(sorter, 100, 1<<20);
     verifyCounters(sorter, outputContext);
+    verifyOutputPermissions(outputContext.getUniqueIdentifier());
   }
 
   @Test
   public void testExceedsKVWithPipelinedShuffle() throws IOException {
     this.numOutputs = 1;
     this.initialAvailableMem = 1 *1024 * 1024;
-    Configuration conf = getConf();
     conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_ENABLE_FINAL_MERGE_IN_OUTPUT, false);
     conf.setInt(TezRuntimeConfiguration
         .TEZ_RUNTIME_PIPELINED_SORTER_MIN_BLOCK_SIZE_IN_MB, 1);
@@ -240,7 +304,6 @@ public class TestPipelinedSorter {
   public void test_TEZ_2602_50mb() throws IOException {
     this.numOutputs = 1;
     this.initialAvailableMem = 1 *1024 * 1024;
-    Configuration conf = getConf();
     conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_ENABLE_FINAL_MERGE_IN_OUTPUT, true);
     conf.setInt(TezRuntimeConfiguration
         .TEZ_RUNTIME_PIPELINED_SORTER_MIN_BLOCK_SIZE_IN_MB, 1);
@@ -257,13 +320,13 @@ public class TestPipelinedSorter {
 
     sorter.flush();
     sorter.close();
+    verifyOutputPermissions(outputContext.getUniqueIdentifier());
   }
 
   //@Test
   public void testLargeDataWithMixedKV() throws IOException {
     this.numOutputs = 1;
     this.initialAvailableMem = 48 *1024 * 1024;
-    Configuration conf = getConf();
     conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_ENABLE_FINAL_MERGE_IN_OUTPUT, true);
     PipelinedSorter sorter = new PipelinedSorter(this.outputContext, conf, numOutputs,
         initialAvailableMem);
@@ -285,6 +348,7 @@ public class TestPipelinedSorter {
 
     sorter.flush();
     sorter.close();
+    verifyOutputPermissions(outputContext.getUniqueIdentifier());
   }
 
 
@@ -321,7 +385,6 @@ public class TestPipelinedSorter {
   @Test
   public void testWithCustomComparator() throws IOException {
     //Test with custom comparator
-    Configuration conf = getConf();
     conf.set(TezRuntimeConfiguration.TEZ_RUNTIME_KEY_COMPARATOR_CLASS,
         CustomComparator.class.getName());
     basicTest(1, 100000, 100, (10 * 1024l * 1024l), 3 << 20);
@@ -331,7 +394,6 @@ public class TestPipelinedSorter {
   public void testWithPipelinedShuffle() throws IOException {
     this.numOutputs = 1;
     this.initialAvailableMem = 5 *1024 * 1024;
-    Configuration conf = getConf();
     conf.setInt(TezRuntimeConfiguration.TEZ_RUNTIME_IO_SORT_MB, 5);
     conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_ENABLE_FINAL_MERGE_IN_OUTPUT, false);
     conf.setInt(TezRuntimeConfiguration
@@ -350,7 +412,6 @@ public class TestPipelinedSorter {
 
   @Test
   public void testCountersWithMultiplePartitions() throws IOException {
-    Configuration conf = getConf();
     conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_ENABLE_FINAL_MERGE_IN_OUTPUT, true);
     this.numOutputs = 5;
     this.initialAvailableMem = 5 * 1024 * 1024;
@@ -361,12 +422,81 @@ public class TestPipelinedSorter {
 
     writeData(sorter, 10000, 100);
     verifyCounters(sorter, outputContext);
+    verifyOutputPermissions(outputContext.getUniqueIdentifier());
+  }
+
+  @Test
+  public void testMultipleSpills() throws IOException {
+    conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_ENABLE_FINAL_MERGE_IN_OUTPUT, true);
+    this.numOutputs = 5;
+    this.initialAvailableMem = 5 * 1024 * 1024;
+    conf.setInt(TezRuntimeConfiguration
+        .TEZ_RUNTIME_PIPELINED_SORTER_MIN_BLOCK_SIZE_IN_MB, 3);
+    PipelinedSorter sorter = new PipelinedSorter(this.outputContext, conf, numOutputs,
+        initialAvailableMem);
+
+    writeData(sorter, 25000, 1000);
+    assertFalse("Expecting needsRLE to be false", sorter.needsRLE());
+    verifyCounters(sorter, outputContext);
+    verifyOutputPermissions(outputContext.getUniqueIdentifier());
+  }
+
+  @Test
+  public void testWithCombiner() throws IOException {
+    conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_ENABLE_FINAL_MERGE_IN_OUTPUT, true);
+    conf.set(TezRuntimeConfiguration.TEZ_RUNTIME_COMBINER_CLASS, DummyCombiner.class.getName());
+    this.numOutputs = 5;
+    this.initialAvailableMem = 5 * 1024 * 1024;
+    conf.setInt(TezRuntimeConfiguration
+            .TEZ_RUNTIME_PIPELINED_SORTER_MIN_BLOCK_SIZE_IN_MB, 3);
+    PipelinedSorter sorter = new PipelinedSorter(this.outputContext, conf, numOutputs,
+            initialAvailableMem);
+
+    writeData(sorter, 1, 20);
+
+    Path outputFile = sorter.finalOutputFile;
+    FileSystem fs = outputFile.getFileSystem(conf);
+    IFile.Reader reader = new IFile.Reader(fs, outputFile, null, null, null, false, -1, 4096);
+    verifyData(reader);
+    reader.close();
+
+    verifyCounters(sorter, outputContext);
+    verifyOutputPermissions(outputContext.getUniqueIdentifier());
+  }
+
+  // for testWithCombiner
+  public static class DummyCombiner implements Combiner {
+    public DummyCombiner(TaskContext ctx) {
+      // do nothing
+    }
+
+    @Override
+    public void combine(TezRawKeyValueIterator rawIter, IFile.Writer writer) throws InterruptedException, IOException {
+      while (rawIter.next()) {
+        writer.append(rawIter.getKey(), rawIter.getValue());
+      }
+    }
+  }
+
+  @Test
+  public void testMultipleSpills_WithRLE() throws IOException {
+    conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_ENABLE_FINAL_MERGE_IN_OUTPUT, true);
+    this.numOutputs = 5;
+    this.initialAvailableMem = 5 * 1024 * 1024;
+    conf.setInt(TezRuntimeConfiguration
+        .TEZ_RUNTIME_PIPELINED_SORTER_MIN_BLOCK_SIZE_IN_MB, 3);
+    PipelinedSorter sorter = new PipelinedSorter(this.outputContext, conf, numOutputs,
+        initialAvailableMem);
+
+    writeSimilarKeys(sorter, 25000, 1000, true);
+    assertTrue("Expecting needsRLE to be true", sorter.needsRLE());
+    verifyCounters(sorter, outputContext);
+    verifyOutputPermissions(outputContext.getUniqueIdentifier());
   }
 
   public void basicTest2(int partitions, int[] numkeys, int[] keysize,
       long initialAvailableMem, int  blockSize) throws IOException {
     this.numOutputs = partitions; // single output
-    Configuration conf = getConf();
     conf.setInt(TezRuntimeConfiguration
         .TEZ_RUNTIME_PIPELINED_SORTER_MIN_BLOCK_SIZE_IN_MB, 100);
     PipelinedSorter sorter = new PipelinedSorter(this.outputContext, conf, numOutputs,
@@ -390,12 +520,12 @@ public class TestPipelinedSorter {
     }
     sorter.flush();
     sorter.close();
+    verifyOutputPermissions(outputContext.getUniqueIdentifier());
   }
 
   public void basicTest(int partitions, int numKeys, int keySize,
       long initialAvailableMem, int minBlockSize) throws IOException {
     this.numOutputs = partitions; // single output
-    Configuration conf = getConf();
     conf.setInt(TezRuntimeConfiguration
         .TEZ_RUNTIME_PIPELINED_SORTER_MIN_BLOCK_SIZE_IN_MB, minBlockSize >> 20);
     PipelinedSorter sorter = new PipelinedSorter(this.outputContext, conf, numOutputs,
@@ -404,20 +534,26 @@ public class TestPipelinedSorter {
     writeData(sorter, numKeys, keySize);
 
     //partition stats;
-    boolean partitionStats = conf.getBoolean(TezRuntimeConfiguration
-        .TEZ_RUNTIME_REPORT_PARTITION_STATS, TezRuntimeConfiguration
-        .TEZ_RUNTIME_REPORT_PARTITION_STATS_DEFAULT);
-    if (partitionStats) {
+    ReportPartitionStats partitionStats =
+        ReportPartitionStats.fromString(conf.get(
+        TezRuntimeConfiguration.TEZ_RUNTIME_REPORT_PARTITION_STATS,
+        TezRuntimeConfiguration.TEZ_RUNTIME_REPORT_PARTITION_STATS_DEFAULT));
+    if (partitionStats.isEnabled()) {
       assertTrue(sorter.getPartitionStats() != null);
     }
 
     verifyCounters(sorter, outputContext);
+    verifyOutputPermissions(outputContext.getUniqueIdentifier());
     Path outputFile = sorter.finalOutputFile;
     FileSystem fs = outputFile.getFileSystem(conf);
-    IFile.Reader reader = new IFile.Reader(fs, outputFile, null, null, null, false, -1, 4096);
+    TezCounter finalOutputBytes =
+        outputContext.getCounters().findCounter(TaskCounter.OUTPUT_BYTES_PHYSICAL);
+    if (finalOutputBytes.getValue() > 0) {
+      IFile.Reader reader = new IFile.Reader(fs, outputFile, null, null, null, false, -1, 4096);
+      verifyData(reader);
+      reader.close();
+    }
     //Verify dataset
-    verifyData(reader);
-    reader.close();
     verify(outputContext, atLeastOnce()).notifyProgress();
   }
 
@@ -448,11 +584,11 @@ public class TestPipelinedSorter {
 
     TezCounter finalOutputBytes =
         context.getCounters().findCounter(TaskCounter.OUTPUT_BYTES_PHYSICAL);
-    assertTrue(finalOutputBytes.getValue() > 0);
+    assertTrue(finalOutputBytes.getValue() >= 0);
 
     TezCounter outputBytesWithOverheadCounter = context.getCounters().findCounter
         (TaskCounter.OUTPUT_BYTES_WITH_OVERHEAD);
-    assertTrue(outputBytesWithOverheadCounter.getValue() > 0);
+    assertTrue(outputBytesWithOverheadCounter.getValue() >= 0);
   }
 
 
@@ -461,7 +597,6 @@ public class TestPipelinedSorter {
   //Its not possible to allocate > 2 GB in test environment.  Carry out basic checks here.
   public void memTest() throws IOException {
     //Verify if > 2 GB can be set via config
-    Configuration conf = getConf();
     conf.setInt(TezRuntimeConfiguration.TEZ_RUNTIME_IO_SORT_MB, 3076);
     long size = ExternalSorter.getInitialMemoryRequirement(conf, 4096 * 1024 * 1024l);
     Assert.assertTrue(size == (3076l << 20));
@@ -546,7 +681,6 @@ public class TestPipelinedSorter {
   //Intentionally not having timeout
   public void test_without_lazyMemAllocation() throws IOException {
     this.numOutputs = 10;
-    Configuration conf = getConf();
 
     //128 MB. Pre-allocate. Request for default block size. Get 1 buffer
     conf.setInt(TezRuntimeConfiguration.TEZ_RUNTIME_IO_SORT_MB, 128);
@@ -587,7 +721,6 @@ public class TestPipelinedSorter {
   //Intentionally not having timeout
   public void test_with_lazyMemAllocation() throws IOException {
     this.numOutputs = 10;
-    Configuration conf = getConf();
 
     //128 MB. Do not pre-allocate.
     // Get 32 MB buffer first and the another buffer with 96 on filling up
@@ -599,13 +732,13 @@ public class TestPipelinedSorter {
         numOutputs, (128l << 20));
     assertTrue("Expected 1 sort buffers. current len=" + sorter.buffers.size(),
         sorter.buffers.size() == 1);
-    assertTrue(sorter.buffers.get(0).capacity() == 32 * 1024 * 1024);
+    assertTrue(sorter.buffers.get(0).capacity() == 32 * 1024 * 1024 - 64);
     writeData(sorter, 100, 1024*1024, false); //100 1 MB KV. Will spill
 
     //Now it should have created 2 buffers, 32 & 96 MB buffers.
     assertTrue(sorter.buffers.size() == 2);
-    assertTrue(sorter.buffers.get(0).capacity() == 32 * 1024 * 1024);
-    assertTrue(sorter.buffers.get(1).capacity() == 96 * 1024 * 1024);
+    assertTrue(sorter.buffers.get(0).capacity() == 32 * 1024 * 1024 - 64);
+    assertTrue(sorter.buffers.get(1).capacity() == 96 * 1024 * 1024 + 64);
     closeSorter(sorter);
     verifyCounters(sorter, outputContext);
 
@@ -617,12 +750,12 @@ public class TestPipelinedSorter {
         .TEZ_RUNTIME_PIPELINED_SORTER_LAZY_ALLOCATE_MEMORY, true);
     sorter = new PipelinedSorter(this.outputContext, conf, numOutputs, (300l << 20));
     assertTrue(sorter.buffers.size() == 1);
-    assertTrue(sorter.buffers.get(0).capacity() == 32 * 1024 * 1024);
+    assertTrue(sorter.buffers.get(0).capacity() == 32 * 1024 * 1024 - 64);
 
     writeData(sorter, 50, 1024*1024, false); //50 1 MB KV to allocate 2nd buf
     assertTrue(sorter.buffers.size() == 2);
-    assertTrue(sorter.buffers.get(0).capacity() == 32 * 1024 * 1024);
-    assertTrue(sorter.buffers.get(1).capacity() == 268 * 1024 * 1024);
+    assertTrue(sorter.buffers.get(0).capacity() == 32 * 1024 * 1024 - 64);
+    assertTrue(sorter.buffers.get(1).capacity() == 268 * 1024 * 1024 + 64);
 
     //48 MB. Do not pre-allocate.
     // Get 32 MB buffer first invariably and proceed with the rest.
@@ -633,13 +766,13 @@ public class TestPipelinedSorter {
         numOutputs, (48l << 20));
     assertTrue("Expected 1 sort buffers. current len=" + sorter.buffers.size(),
         sorter.buffers.size() == 1);
-    assertTrue(sorter.buffers.get(0).capacity() == 32 * 1024 * 1024);
+    assertTrue(sorter.buffers.get(0).capacity() == 32 * 1024 * 1024 - 64);
     writeData(sorter, 20, 1024*1024, false); //100 1 MB KV. Will spill
 
     //Now it should have created 2 buffers, 32 & 96 MB buffers.
     assertTrue(sorter.buffers.size() == 2);
-    assertTrue(sorter.buffers.get(0).capacity() == 32 * 1024 * 1024);
-    assertTrue(sorter.buffers.get(1).capacity() == 16 * 1024 * 1024);
+    assertTrue(sorter.buffers.get(0).capacity() == 32 * 1024 * 1024 - 64);
+    assertTrue(sorter.buffers.get(1).capacity() == 16 * 1024 * 1024 + 64);
     closeSorter(sorter);
   }
 
@@ -647,7 +780,6 @@ public class TestPipelinedSorter {
   //Intentionally not having timeout
   public void testLazyAllocateMem() throws IOException {
     this.numOutputs = 10;
-    Configuration conf = getConf();
     conf.setInt(TezRuntimeConfiguration.TEZ_RUNTIME_IO_SORT_MB, 128);
     conf.setBoolean(TezRuntimeConfiguration
         .TEZ_RUNTIME_PIPELINED_SORTER_LAZY_ALLOCATE_MEMORY, false);
@@ -695,10 +827,40 @@ public class TestPipelinedSorter {
     basicTest(1, 5, (2 << 20), (48 * 1024l * 1024l), 16 << 20);
   }
 
+  private void verifyOutputPermissions(String spillId) throws IOException {
+    String subpath = Constants.TEZ_RUNTIME_TASK_OUTPUT_DIR + "/" + spillId
+        + "/" + Constants.TEZ_RUNTIME_TASK_OUTPUT_FILENAME_STRING;
+    Path outputPath = dirAllocator.getLocalPathToRead(subpath, conf);
+    Path indexPath = dirAllocator.getLocalPathToRead(subpath + Constants.TEZ_RUNTIME_TASK_OUTPUT_INDEX_SUFFIX_STRING, conf);
+    Assert.assertEquals("Incorrect output permissions", (short)0640,
+        localFs.getFileStatus(outputPath).getPermission().toShort());
+    Assert.assertEquals("Incorrect index permissions", (short)0640,
+        localFs.getFileStatus(indexPath).getPermission().toShort());
+  }
+
   private void writeData(ExternalSorter sorter, int numKeys, int keyLen) throws IOException {
     writeData(sorter, numKeys, keyLen, true);
   }
 
+  // duplicate some of the keys
+  private void writeSimilarKeys(ExternalSorter sorter, int numKeys, int keyLen,
+      boolean autoClose) throws IOException {
+    sortedDataMap.clear();
+    String keyStr = RandomStringUtils.randomAlphanumeric(keyLen);
+    for (int i = 0; i < numKeys; i++) {
+      if (i % 4 == 0) {
+        keyStr = RandomStringUtils.randomAlphanumeric(keyLen);
+      }
+      Text key = new Text(keyStr);
+      Text value = new Text(RandomStringUtils.randomAlphanumeric(keyLen));
+      sorter.write(key, value);
+      sortedDataMap.put(key.toString(), value.toString()); //for verifying data later
+    }
+    if (autoClose) {
+      closeSorter(sorter);
+    }
+  }
+
   private void writeData(ExternalSorter sorter, int numKeys, int keyLen,
       boolean autoClose) throws IOException {
     sortedDataMap.clear();
@@ -726,7 +888,6 @@ public class TestPipelinedSorter {
     Text readValue = new Text();
     DataInputBuffer keyIn = new DataInputBuffer();
     DataInputBuffer valIn = new DataInputBuffer();
-    Configuration conf = getConf();
     SerializationFactory serializationFactory = new SerializationFactory(conf);
     Deserializer<Text> keyDeserializer = serializationFactory.getDeserializer(Text.class);
     Deserializer<Text> valDeserializer = serializationFactory.getDeserializer(Text.class);
@@ -751,7 +912,7 @@ public class TestPipelinedSorter {
   }
 
   private static OutputContext createMockOutputContext(TezCounters counters, ApplicationId appId,
-      String uniqueId) throws IOException {
+      String uniqueId, String auxiliaryService) throws IOException {
     OutputContext outputContext = mock(OutputContext.class);
 
     ExecutionContext execContext = new ExecutionContextImpl("localhost");
@@ -759,8 +920,7 @@ public class TestPipelinedSorter {
     DataOutputBuffer serviceProviderMetaData = new DataOutputBuffer();
     serviceProviderMetaData.writeInt(80);
     doReturn(ByteBuffer.wrap(serviceProviderMetaData.getData())).when(outputContext)
-        .getServiceProviderMetaData
-            (ShuffleUtils.SHUFFLE_HANDLER_SERVICE_ID);
+        .getServiceProviderMetaData(auxiliaryService);
 
     doReturn(execContext).when(outputContext).getExecutionContext();
     doReturn(mock(OutputStatisticsReporter.class)).when(outputContext).getStatisticsReporter();
