@@ -18,6 +18,12 @@
 
 package org.apache.tez.runtime.library.common.sort.impl.dflt;
 
+import org.apache.hadoop.fs.LocalDirAllocator;
+import org.apache.tez.runtime.library.api.Partitioner;
+import org.apache.tez.runtime.library.common.Constants;
+import org.apache.tez.runtime.library.common.TezRuntimeUtils;
+import org.junit.Assert;
+import static org.junit.Assert.assertEquals;
 import static org.junit.Assert.assertTrue;
 import static org.junit.Assert.fail;
 import static org.mockito.Matchers.any;
@@ -31,6 +37,8 @@ import static org.mockito.internal.verification.VerificationModeFactory.times;
 
 import java.io.IOException;
 import java.nio.ByteBuffer;
+import java.util.ArrayList;
+import java.util.BitSet;
 import java.util.List;
 import java.util.concurrent.ThreadLocalRandom;
 
@@ -38,6 +46,7 @@ import com.google.protobuf.ByteString;
 import org.apache.commons.lang.RandomStringUtils;
 import org.apache.commons.lang.StringUtils;
 import org.apache.hadoop.conf.Configuration;
+import org.apache.hadoop.fs.CommonConfigurationKeys;
 import org.apache.hadoop.fs.FileSystem;
 import org.apache.hadoop.fs.Path;
 import org.apache.hadoop.io.DataOutputBuffer;
@@ -49,6 +58,9 @@ import org.apache.tez.common.TezUtils;
 import org.apache.tez.common.counters.TaskCounter;
 import org.apache.tez.common.counters.TezCounter;
 import org.apache.tez.common.counters.TezCounters;
+import org.apache.tez.common.TezCommonUtils;
+import org.apache.tez.common.TezUtilsInternal;
+import org.apache.tez.dag.api.TezConfiguration;
 import org.apache.tez.dag.api.UserPayload;
 import org.apache.tez.runtime.api.Event;
 import org.apache.tez.runtime.api.ExecutionContext;
@@ -61,6 +73,8 @@ import org.apache.tez.runtime.library.api.TezRuntimeConfiguration;
 import org.apache.tez.runtime.library.common.MemoryUpdateCallbackHandler;
 import org.apache.tez.runtime.library.common.shuffle.ShuffleUtils;
 import org.apache.tez.runtime.library.common.sort.impl.ExternalSorter;
+import org.apache.tez.runtime.library.common.sort.impl.TezIndexRecord;
+import org.apache.tez.runtime.library.common.sort.impl.TezSpillRecord;
 import org.apache.tez.runtime.library.conf.OrderedPartitionedKVOutputConfig.SorterImpl;
 import org.apache.tez.runtime.library.partitioner.HashPartitioner;
 import org.apache.tez.runtime.library.shuffle.impl.ShuffleUserPayloads;
@@ -75,16 +89,19 @@ import org.mockito.stubbing.Answer;
 
 public class TestDefaultSorter {
 
-  private Configuration conf;
   private static final int PORT = 80;
   private static final String UniqueID = "UUID";
 
   private static FileSystem localFs = null;
   private static Path workingDir = null;
 
+  private Configuration conf;
+  private LocalDirAllocator dirAllocator;
+
   @Before
   public void setup() throws IOException {
     conf = new Configuration();
+    conf.set(CommonConfigurationKeys.FS_PERMISSIONS_UMASK_KEY, "077");
     conf.set(TezRuntimeConfiguration.TEZ_RUNTIME_SORTER_CLASS, SorterImpl.LEGACY.name()); // DefaultSorter
     conf.set("fs.defaultFS", "file:///");
     localFs = FileSystem.getLocal(conf);
@@ -99,6 +116,7 @@ public class TestDefaultSorter {
     conf.set(TezRuntimeConfiguration.TEZ_RUNTIME_PARTITIONER_CLASS,
         HashPartitioner.class.getName());
     conf.setStrings(TezRuntimeFrameworkConfigs.LOCAL_DIRS, localDirs);
+    dirAllocator = new LocalDirAllocator(TezRuntimeFrameworkConfigs.LOCAL_DIRS);
   }
 
   @AfterClass
@@ -253,19 +271,89 @@ public class TestDefaultSorter {
     conf.setLong(TezRuntimeConfiguration.TEZ_RUNTIME_IO_SORT_MB, 1);
     context.requestInitialMemory(ExternalSorter.getInitialMemoryRequirement(conf,
         context.getTotalMemoryAvailableToTask()), handler);
-    DefaultSorter sorter = new DefaultSorter(context, conf, 5, handler.getMemoryAssigned());
+    SorterWrapper sorterWrapper = new SorterWrapper(context, conf, 5, handler.getMemoryAssigned());
+    DefaultSorter sorter = sorterWrapper.getSorter();
 
     //Write 1000 keys each of size 1000, (> 1 spill should happen)
     try {
-      writeData(sorter, 1000, 1000);
+      Text[] keys = generateData(1000, 1000);
+      Text[] values = generateData(1000, 1000);
+      for (int i = 0; i < keys.length; i++) {
+        sorterWrapper.writeKeyValue(keys[i], values[i]);
+      }
+      sorterWrapper.close();
       assertTrue(sorter.getNumSpills() > 2);
       verifyCounters(sorter, context);
     } catch(IOException ioe) {
       fail(ioe.getMessage());
     }
+
+    verifyOutputPermissions(context.getUniqueIdentifier());
   }
 
   @Test(timeout = 30000)
+  public void testEmptyCaseFileLengths() throws IOException {
+    testEmptyCaseFileLengthsHelper(50, new String[] {"a", "b"}, new String[] {"1", "2"});
+    testEmptyCaseFileLengthsHelper(50, new String[] {"a", "a"}, new String[] {"1", "2"});
+    testEmptyCaseFileLengthsHelper(50, new String[] {"aaa", "bbb", "aaa"}, new String[] {"1", "2", "3"});
+    testEmptyCaseFileLengthsHelper(1, new String[] {"abcdefghij"}, new String[] {"1234567890"});
+  }
+
+  public void testEmptyCaseFileLengthsHelper(int numPartitions, String[] keys, String[] values)
+      throws IOException {
+    OutputContext context = createTezOutputContext();
+
+    MemoryUpdateCallbackHandler handler = new MemoryUpdateCallbackHandler();
+    conf.setInt(TezRuntimeConfiguration.TEZ_RUNTIME_IO_SORT_MB, 1);
+    context.requestInitialMemory(ExternalSorter.getInitialMemoryRequirement(conf,
+        context.getTotalMemoryAvailableToTask()), handler);
+    String auxService = conf.get(TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID, TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID_DEFAULT);
+    SorterWrapper sorterWrapper = new SorterWrapper(context, conf, numPartitions, handler.getMemoryAssigned());
+    DefaultSorter sorter = sorterWrapper.getSorter();
+    assertEquals("Key and Values must have the same number of elements", keys.length, values.length);
+    BitSet keyRLEs = new BitSet(keys.length);
+    for (int i = 0; i < keys.length; i++) {
+      boolean isRLE = sorterWrapper.writeKeyValue(new Text(keys[i]), new Text(values[i]));
+      keyRLEs.set(i, isRLE);
+    }
+    sorterWrapper.close();
+
+    List<Event> events = new ArrayList<>();
+    String pathComponent = (context.getUniqueIdentifier() + "_" + 0);
+    ShuffleUtils.generateEventOnSpill(events, true, true, context, 0,
+        sorter.indexCacheList.get(0), 0, true, pathComponent, sorter.getPartitionStats(),
+        sorter.reportDetailedPartitionStats(), auxService, TezCommonUtils.newBestCompressionDeflater());
+
+    CompositeDataMovementEvent compositeDataMovementEvent =
+        (CompositeDataMovementEvent) events.get(1);
+    ByteBuffer bb = compositeDataMovementEvent.getUserPayload();
+    ShuffleUserPayloads.DataMovementEventPayloadProto shufflePayload =
+        ShuffleUserPayloads.DataMovementEventPayloadProto.parseFrom(ByteString.copyFrom(bb));
+
+    if (shufflePayload.hasEmptyPartitions()) {
+      byte[] emptyPartitionsBytesString =
+          TezCommonUtils.decompressByteStringToByteArray(
+              shufflePayload.getEmptyPartitions());
+      BitSet emptyPartitionBitSet = TezUtilsInternal.fromByteArray(emptyPartitionsBytesString);
+      Assert.assertEquals("Number of empty partitions did not match!",
+          emptyPartitionBitSet.cardinality(), sorterWrapper.getEmptyPartitionsCount());
+    } else {
+      Assert.assertEquals(sorterWrapper.getEmptyPartitionsCount(), 0);
+    }
+    // Each non-empty partition adds 4 bytes for header, 2 bytes for EOF_MARKER, 4 bytes for checksum
+    int expectedFileOutLength = sorterWrapper.getNonEmptyPartitionsCount() * 10;
+    for (int i = 0; i < keys.length; i++) {
+      // Each Record adds 1 byte for key length, 1 byte Text overhead (length), key.length bytes for key
+      expectedFileOutLength += keys[i].length() + 2;
+      // Each Record adds 1 byte for value length, 1 byte Text overhead (length), value.length bytes for value
+      expectedFileOutLength += values[i].length() + 2;
+    }
+    assertEquals("Unexpected Output File Size!", localFs.getFileStatus(sorter.getFinalOutputFile()).getLen(), expectedFileOutLength);
+    assertEquals(sorter.getNumSpills(), 1);
+    verifyCounters(sorter, context);
+  }
+
+  @Test
   public void testWithEmptyData() throws IOException {
     OutputContext context = createTezOutputContext();
 
@@ -279,6 +367,7 @@ public class TestDefaultSorter {
     try {
       sorter.flush();
       sorter.close();
+      assertTrue(sorter.isClosed());
       assertTrue(sorter.getFinalOutputFile().getParent().getName().equalsIgnoreCase(UniqueID));
       verifyCounters(sorter, context);
     } catch(Exception e) {
@@ -301,6 +390,7 @@ public class TestDefaultSorter {
     try {
       sorter.flush();
       sorter.close();
+      assertTrue(sorter.isClosed());
       assertTrue(sorter.getFinalOutputFile().getParent().getName().equalsIgnoreCase(UniqueID +
           "_0"));
       verifyCounters(sorter, context);
@@ -309,6 +399,70 @@ public class TestDefaultSorter {
     }
   }
 
+  @Test
+  public void testEmptyPartitions() throws Exception {
+    testEmptyPartitionsHelper(2, false);
+    testEmptyPartitionsHelper(2, true);
+    testEmptyPartitionsHelper(0, true);
+    testEmptyPartitionsHelper(0, true);
+  }
+
+  public void testEmptyPartitionsHelper(int numKeys, boolean sendEmptyPartitionDetails) throws IOException {
+    OutputContext context = createTezOutputContext();
+    conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_EMPTY_PARTITION_INFO_VIA_EVENTS_ENABLED, sendEmptyPartitionDetails);
+    conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_ENABLE_FINAL_MERGE_IN_OUTPUT, true);
+    conf.setLong(TezRuntimeConfiguration.TEZ_RUNTIME_IO_SORT_MB, 1);
+    MemoryUpdateCallbackHandler handler = new MemoryUpdateCallbackHandler();
+    context.requestInitialMemory(ExternalSorter.getInitialMemoryRequirement(conf,
+        context.getTotalMemoryAvailableToTask()), handler);
+    int partitions = 50;
+    SorterWrapper sorterWrapper = new SorterWrapper(context, conf, partitions, handler.getMemoryAssigned());
+    DefaultSorter sorter = sorterWrapper.getSorter();
+
+    Text[] keys = generateData(numKeys, 1000000);
+    Text[] values = generateData(numKeys, 1000000);
+    for (int i = 0; i < keys.length; i++) {
+      sorterWrapper.writeKeyValue(keys[i], values[i]);
+    }
+    sorterWrapper.close();
+    if (numKeys == 0) {
+      assertTrue(sorter.getNumSpills() == 1);
+    } else {
+      assertTrue(sorter.getNumSpills() == numKeys);
+    }
+    verifyCounters(sorter, context);
+    verifyOutputPermissions(context.getUniqueIdentifier());
+    if (sorter.indexCacheList.size() != 0) {
+      for (int i = 0; i < sorter.getNumSpills(); i++) {
+        TezSpillRecord record = sorter.indexCacheList.get(i);
+        for (int j = 0; j < partitions; j++) {
+          TezIndexRecord tezIndexRecord = record.getIndex(j);
+          if (tezIndexRecord.hasData()) {
+            continue;
+          }
+          if (sendEmptyPartitionDetails) {
+            Assert.assertEquals("Unexpected raw length for " + i + "th partition", 0, tezIndexRecord.getRawLength());
+          } else {
+            Assert.assertEquals("", tezIndexRecord.getRawLength(), 6);
+          }
+        }
+      }
+    }
+    Path indexFile = sorter.getFinalIndexFile();
+    TezSpillRecord spillRecord = new TezSpillRecord(indexFile, conf);
+    for (int i = 0; i < partitions; i++) {
+      TezIndexRecord tezIndexRecord = spillRecord.getIndex(i);
+      if (tezIndexRecord.hasData()) {
+        continue;
+      }
+      if (sendEmptyPartitionDetails) {
+        Assert.assertEquals("Unexpected raw length for " + i + "th partition", 0, tezIndexRecord.getRawLength());
+      } else {
+        Assert.assertEquals("Unexpected raw length for " + i + "th partition", 6, tezIndexRecord.getRawLength());
+      }
+    }
+  }
+
   void testPartitionStats(boolean withStats) throws IOException {
     conf.setBoolean(TezRuntimeConfiguration.TEZ_RUNTIME_REPORT_PARTITION_STATS, withStats);
     OutputContext context = createTezOutputContext();
@@ -318,9 +472,15 @@ public class TestDefaultSorter {
     MemoryUpdateCallbackHandler handler = new MemoryUpdateCallbackHandler();
     context.requestInitialMemory(ExternalSorter.getInitialMemoryRequirement(conf,
         context.getTotalMemoryAvailableToTask()), handler);
-    DefaultSorter sorter = new DefaultSorter(context, conf, 1, handler.getMemoryAssigned());
+    SorterWrapper sorterWrapper = new SorterWrapper(context, conf, 1, handler.getMemoryAssigned());
+    DefaultSorter sorter = sorterWrapper.getSorter();
 
-    writeData(sorter, 1000, 10);
+    Text[] keys = generateData(1000, 10);
+    Text[] values = generateData(1000, 10);
+    for (int i = 0; i < keys.length; i++) {
+      sorterWrapper.writeKeyValue(keys[i], values[i]);
+    }
+    sorterWrapper.close();
     assertTrue(sorter.getNumSpills() == 1);
     verifyCounters(sorter, context);
 
@@ -351,9 +511,16 @@ public class TestDefaultSorter {
     MemoryUpdateCallbackHandler handler = new MemoryUpdateCallbackHandler();
     context.requestInitialMemory(ExternalSorter.getInitialMemoryRequirement(conf,
         context.getTotalMemoryAvailableToTask()), handler);
-    DefaultSorter sorter = new DefaultSorter(context, conf, 1, handler.getMemoryAssigned());
 
-    writeData(sorter, 1000, 10);
+    SorterWrapper sorterWrapper = new SorterWrapper(context, conf, 1, handler.getMemoryAssigned());
+    DefaultSorter sorter = sorterWrapper.getSorter();
+
+    Text[] keys = generateData(1000, 10);
+    Text[] values = generateData(1000, 10);
+    for (int i = 0; i < keys.length; i++) {
+      sorterWrapper.writeKeyValue(keys[i], values[i]);
+    }
+    sorterWrapper.close();
     assertTrue(sorter.getNumSpills() == 1);
     ArgumentCaptor<List> eventCaptor = ArgumentCaptor.forClass(List.class);
     verify(context, times(1)).sendEvents(eventCaptor.capture());
@@ -364,6 +531,7 @@ public class TestDefaultSorter {
         ShuffleUserPayloads.DataMovementEventPayloadProto shufflePayload = ShuffleUserPayloads
             .DataMovementEventPayloadProto.parseFrom(ByteString.copyFrom(cdme.getUserPayload()));
         assertTrue(shufflePayload.getPathComponent().equalsIgnoreCase(UniqueID + "_0"));
+        verifyOutputPermissions(shufflePayload.getPathComponent());
       }
     }
 
@@ -381,9 +549,16 @@ public class TestDefaultSorter {
     MemoryUpdateCallbackHandler handler = new MemoryUpdateCallbackHandler();
     context.requestInitialMemory(ExternalSorter.getInitialMemoryRequirement(conf,
         context.getTotalMemoryAvailableToTask()), handler);
-    DefaultSorter sorter = new DefaultSorter(context, conf, 1, handler.getMemoryAssigned());
+    SorterWrapper sorterWrapper = new SorterWrapper(context, conf, 1, handler.getMemoryAssigned());
+    DefaultSorter sorter = sorterWrapper.getSorter();
+
+    Text[] keys = generateData(10000, 1000);
+    Text[] values = generateData(10000, 1000);
+    for (int i = 0; i < keys.length; i++) {
+      sorterWrapper.writeKeyValue(keys[i], values[i]);
+    }
+    sorterWrapper.close();
 
-    writeData(sorter, 10000, 1000);
     int spillCount = sorter.getNumSpills();
     ArgumentCaptor<List> eventCaptor = ArgumentCaptor.forClass(List.class);
     verify(context, times(1)).sendEvents(eventCaptor.capture());
@@ -395,6 +570,7 @@ public class TestDefaultSorter {
         ShuffleUserPayloads.DataMovementEventPayloadProto shufflePayload = ShuffleUserPayloads
             .DataMovementEventPayloadProto.parseFrom(ByteString.copyFrom(cdme.getUserPayload()));
         assertTrue(shufflePayload.getPathComponent().equalsIgnoreCase(UniqueID + "_" + spillIndex));
+        verifyOutputPermissions(shufflePayload.getPathComponent());
         spillIndex++;
       }
     }
@@ -402,6 +578,17 @@ public class TestDefaultSorter {
     verifyCounters(sorter, context);
   }
 
+  private void verifyOutputPermissions(String spillId) throws IOException {
+    String subpath = Constants.TEZ_RUNTIME_TASK_OUTPUT_DIR + "/" + spillId
+        + "/" + Constants.TEZ_RUNTIME_TASK_OUTPUT_FILENAME_STRING;
+    Path outputPath = dirAllocator.getLocalPathToRead(subpath, conf);
+    Path indexPath = dirAllocator.getLocalPathToRead(subpath + Constants.TEZ_RUNTIME_TASK_OUTPUT_INDEX_SUFFIX_STRING, conf);
+    Assert.assertEquals("Incorrect output permissions", (short)0640,
+        localFs.getFileStatus(outputPath).getPermission().toShort());
+    Assert.assertEquals("Incorrect index permissions", (short)0640,
+        localFs.getFileStatus(indexPath).getPermission().toShort());
+  }
+
   private void verifyCounters(DefaultSorter sorter, OutputContext context) {
     TezCounter numShuffleChunks = context.getCounters().findCounter(TaskCounter.SHUFFLE_CHUNK_COUNT);
     TezCounter additionalSpills = context.getCounters().findCounter(TaskCounter.ADDITIONAL_SPILL_COUNT);
@@ -425,22 +612,65 @@ public class TestDefaultSorter {
     }
 
     TezCounter finalOutputBytes = context.getCounters().findCounter(TaskCounter.OUTPUT_BYTES_PHYSICAL);
-    assertTrue(finalOutputBytes.getValue() > 0);
+    assertTrue(finalOutputBytes.getValue() >= 0);
 
     TezCounter outputBytesWithOverheadCounter = context.getCounters().findCounter
         (TaskCounter.OUTPUT_BYTES_WITH_OVERHEAD);
-    assertTrue(outputBytesWithOverheadCounter.getValue() > 0);
+    assertTrue(outputBytesWithOverheadCounter.getValue() >= 0);
     verify(context, atLeastOnce()).notifyProgress();
   }
 
-  private void writeData(ExternalSorter sorter, int numKeys, int keyLen) throws IOException {
-    for (int i = 0; i < numKeys; i++) {
-      Text key = new Text(RandomStringUtils.randomAlphanumeric(keyLen));
-      Text value = new Text(RandomStringUtils.randomAlphanumeric(keyLen));
-      sorter.write(key, value);
+  private static class SorterWrapper {
+
+    private final DefaultSorter sorter;
+    private final Partitioner partitioner;
+    private final BitSet nonEmptyPartitions;
+    private final Object[] lastKeys;
+    private final int numPartitions;
+
+
+    public SorterWrapper(OutputContext context, Configuration conf, int numPartitions, long memoryAssigned) throws IOException {
+      sorter = new DefaultSorter(context, conf, numPartitions, memoryAssigned);
+      partitioner = TezRuntimeUtils.instantiatePartitioner(conf);
+      nonEmptyPartitions = new BitSet(numPartitions);
+      lastKeys = new Object[numPartitions];
+      this.numPartitions = numPartitions;
     }
-    sorter.flush();
-    sorter.close();
+
+    public boolean writeKeyValue(Object key, Object value) throws IOException {
+      int partition = partitioner.getPartition(key, value, this.numPartitions);
+      nonEmptyPartitions.set(partition);
+      sorter.write(key, value);
+
+      boolean isRLE = key.equals(lastKeys[partition]);
+      lastKeys[partition] = key;
+      return isRLE;
+    }
+
+    public int getNonEmptyPartitionsCount() {
+      return nonEmptyPartitions.cardinality();
+    }
+
+    public int getEmptyPartitionsCount() {
+      return numPartitions - nonEmptyPartitions.cardinality();
+    }
+
+    public void close () throws IOException {
+      sorter.flush();
+      sorter.close();
+    }
+
+    public DefaultSorter getSorter() {
+      return sorter;
+    }
+  }
+
+  private static Text[] generateData(int numKeys, int keyLen) {
+    Text[] ret = new Text[numKeys];
+    for (int i = 0; i < numKeys; i++) {
+      ret[i] = new Text(RandomStringUtils.randomAlphanumeric(keyLen));
+    }
+    return ret;
   }
 
   private OutputContext createTezOutputContext() throws IOException {
@@ -463,7 +693,8 @@ public class TestDefaultSorter {
     doReturn("v1").when(context).getDestinationVertexName();
     doReturn(ByteBuffer.wrap(serviceProviderMetaData.getData())).when(context)
         .getServiceProviderMetaData
-            (ShuffleUtils.SHUFFLE_HANDLER_SERVICE_ID);
+            (conf.get(TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID,
+                TezConfiguration.TEZ_AM_SHUFFLE_AUXILIARY_SERVICE_ID_DEFAULT));
     doAnswer(new Answer() {
       @Override public Object answer(InvocationOnMock invocation) throws Throwable {
         long requestedSize = (Long) invocation.getArguments()[0];
