@@ -19,8 +19,17 @@
 package org.apache.tez.test;
 
 import java.io.IOException;
+import java.util.Arrays;
 import java.util.Random;
 
+import org.apache.tez.common.TezUtils;
+import org.apache.tez.dag.api.EdgeManagerPluginDescriptor;
+import org.apache.tez.dag.api.ProcessorDescriptor;
+import org.apache.tez.dag.api.UserPayload;
+import org.apache.tez.dag.api.VertexManagerPluginDescriptor;
+import org.apache.tez.runtime.library.cartesianproduct.CartesianProductConfig;
+import org.apache.tez.runtime.library.cartesianproduct.CartesianProductEdgeManager;
+import org.apache.tez.runtime.library.cartesianproduct.CartesianProductVertexManager;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
@@ -49,6 +58,8 @@ import org.apache.tez.test.dag.ThreeLevelsFailingDAG;
 import org.apache.tez.test.dag.TwoLevelsFailingDAG;
 import org.junit.AfterClass;
 import org.junit.Assert;
+import org.junit.Ignore;
+import org.junit.Before;
 import org.junit.BeforeClass;
 import org.junit.Test;
 
@@ -62,7 +73,8 @@ public class TestFaultTolerance {
   protected static MiniDFSCluster dfsCluster;
   
   private static TezClient tezSession = null;
-  
+  private static TezConfiguration tezConf;
+
   @BeforeClass
   public static void setup() throws Exception {
     LOG.info("Starting mini clusters");
@@ -87,13 +99,14 @@ public class TestFaultTolerance {
           .valueOf(new Random().nextInt(100000))));
       TezClientUtils.ensureStagingDirExists(conf, remoteStagingDir);
       
-      TezConfiguration tezConf = new TezConfiguration(miniTezCluster.getConfig());
+      tezConf = new TezConfiguration(miniTezCluster.getConfig());
       tezConf.set(TezConfiguration.TEZ_AM_STAGING_DIR,
           remoteStagingDir.toString());
       tezConf.setBoolean(TezConfiguration.TEZ_AM_NODE_BLACKLISTING_ENABLED, false);
       tezConf.setDouble(TezConfiguration.TEZ_TASK_MAX_ALLOWED_OUTPUT_FAILURES_FRACTION, 0.4);
       tezConf.setInt(TezConfiguration.TEZ_AM_MAX_ALLOWED_TIME_FOR_TASK_READ_ERROR_SEC, 3);
       tezConf.setInt(TezConfiguration.TEZ_TASK_AM_HEARTBEAT_INTERVAL_MS, 100);
+      tezConf.setLong(TezConfiguration.TEZ_AM_SLEEP_TIME_BEFORE_EXIT_MILLIS, 500);
 
       tezSession = TezClient.create("TestFaultTolerance", tezConf, true);
       tezSession.start();
@@ -116,6 +129,11 @@ public class TestFaultTolerance {
     }
   }
 
+  @Before
+  public void checkSessionStatus() {
+    // TODO restart session if it crashed due to some test error
+  }
+
   void runDAGAndVerify(DAG dag, DAGStatus.State finalState) throws Exception {
     runDAGAndVerify(dag, finalState, -1);
   }
@@ -150,7 +168,40 @@ public class TestFaultTolerance {
       Assert.assertTrue(Joiner.on(":").join(dagStatus.getDiagnostics()).contains(diagnostics));
     }
   }
-  
+
+  @Test (timeout=600000)
+  public void testSessionStopped() throws Exception {
+    Configuration testConf = new Configuration(false);
+
+    testConf.setBoolean(TestProcessor.getVertexConfName(
+        TestProcessor.TEZ_FAILING_PROCESSOR_DO_FAIL, "v1"), true);
+    testConf.set(TestProcessor.getVertexConfName(
+        TestProcessor.TEZ_FAILING_PROCESSOR_FAILING_TASK_INDEX, "v1"), "0");
+    testConf.setInt(TestProcessor.getVertexConfName(
+        TestProcessor.TEZ_FAILING_PROCESSOR_FAILING_UPTO_TASK_ATTEMPT, "v1"), 0);
+
+    // verify value at v2 task1
+    testConf.set(TestProcessor.getVertexConfName(
+        TestProcessor.TEZ_FAILING_PROCESSOR_VERIFY_TASK_INDEX, "v2"), "1");
+
+    testConf.setInt(TestProcessor.getVertexConfName(
+        TestProcessor.TEZ_FAILING_PROCESSOR_VERIFY_VALUE, "v2", 1), 4);
+    DAG dag = SimpleTestDAG.createDAG("testBasicTaskFailure", testConf);
+    tezSession.waitTillReady();
+
+    DAGClient dagClient = tezSession.submitDAG(dag);
+    dagClient.waitForCompletion();
+    // kill the session now
+    tezSession.stop();
+
+    // Check if killing DAG does not throw any exception
+    dagClient.tryKillDAG();
+
+    // restart the session for rest of the tests
+    tezSession = TezClient.create("TestFaultTolerance", tezConf, true);
+    tezSession.start();
+  }
+
   @Test (timeout=60000)
   public void testBasicSuccessScatterGather() throws Exception {
     DAG dag = SimpleTestDAG.createDAG("testBasicSuccessScatterGather", null);
@@ -754,6 +805,7 @@ public class TestFaultTolerance {
     runDAGAndVerify(dag, DAGStatus.State.SUCCEEDED);
   }
   
+  @Ignore
   @Test (timeout=240000)
   public void testRandomFailingInputs() throws Exception {
     Configuration testConf = new Configuration(false);
@@ -766,8 +818,9 @@ public class TestFaultTolerance {
   @Test (timeout=240000)
   public void testNoProgress() throws Exception {
     Configuration testConf = new Configuration(false);
-    testConf.setInt(TestProcessor.TEZ_FAILING_PROCESSOR_SLEEP_MS, 1000*100); // long sleep
     testConf.setInt(SimpleTestDAG.TEZ_SIMPLE_DAG_NUM_TASKS, 1);
+    testConf.setLong(TestProcessor.getVertexConfName(
+        TestProcessor.TEZ_FAILING_PROCESSOR_SLEEP_MS, "v1"), 1000*100); // long sleep
     DAG dag = SimpleTestDAG.createDAG(testConf);
     Vertex hung = dag.getVertex("v1");
     hung.setConf(TezConfiguration.TEZ_TASK_PROGRESS_STUCK_INTERVAL_MS, Long.toString(1000));
@@ -776,5 +829,71 @@ public class TestFaultTolerance {
     // dag will fail with 2 attempts failing from vertex v1
     runDAGAndVerify(dag, DAGStatus.State.FAILED, 2, "no progress");
   }
-  
+
+  /**
+   * In unpartitioned cartesian product, failure fraction should be #unique failure/#consumer that
+   * depends on the src task. Here we test a 2x2 cartesian product and let 4th destination task fail.
+   * The failure fraction limit is configured to be 0.25. So the failure fraction should be 1/2,
+   * not 1/4.
+   * @throws Exception
+   */
+  @Test
+  public void testCartesianProduct() throws Exception {
+    Configuration dagConf = new Configuration();
+    dagConf.setDouble(TezConfiguration.TEZ_TASK_MAX_ALLOWED_OUTPUT_FAILURES_FRACTION, 0.25);
+    DAG dag = DAG.create("dag");
+
+    Configuration vertexConf = new Configuration();
+    vertexConf.setInt(TestProcessor.getVertexConfName(
+      TestProcessor.TEZ_FAILING_PROCESSOR_VERIFY_TASK_INDEX, "v3"), 3);
+    vertexConf.setInt(TestProcessor.getVertexConfName(
+      TestProcessor.TEZ_FAILING_PROCESSOR_VERIFY_VALUE, "v3"), 5);
+    UserPayload vertexPayload = TezUtils.createUserPayloadFromConf(vertexConf);
+    ProcessorDescriptor processorDescriptor =
+      ProcessorDescriptor.create(TestProcessor.class.getName()).setUserPayload(vertexPayload);
+    Vertex v1 = Vertex.create("v1", processorDescriptor, 2);
+    Vertex v2 = Vertex.create("v2", processorDescriptor, 2);
+    Vertex v3 = Vertex.create("v3", processorDescriptor);
+
+    String[] sourceVertices = {"v1", "v2"};
+    CartesianProductConfig cartesianProductConfig =
+      new CartesianProductConfig(Arrays.asList(sourceVertices));
+    TezConfiguration tezConf = new TezConfiguration();
+    tezConf.setInt(CartesianProductVertexManager.TEZ_CARTESIAN_PRODUCT_NUM_PARTITIONS, 1);
+    tezConf.setBoolean(CartesianProductVertexManager.TEZ_CARTESIAN_PRODUCT_ENABLE_GROUPING, false);
+    UserPayload cartesianProductPayload =
+      cartesianProductConfig.toUserPayload(tezConf);
+
+    v3.setVertexManagerPlugin(
+      VertexManagerPluginDescriptor.create(CartesianProductVertexManager.class.getName())
+        .setUserPayload(cartesianProductPayload));
+
+    EdgeManagerPluginDescriptor edgeManagerPluginDescriptor =
+      EdgeManagerPluginDescriptor.create(CartesianProductEdgeManager.class.getName())
+        .setUserPayload(cartesianProductPayload);
+
+    Configuration inputConf = new Configuration();
+    inputConf.setBoolean(TestInput.getVertexConfName(
+      TestInput.TEZ_FAILING_INPUT_DO_FAIL, "v3"), true);
+    inputConf.setInt(TestInput.getVertexConfName(
+      TestInput.TEZ_FAILING_INPUT_FAILING_TASK_INDEX, "v3"), 3);
+    inputConf.setInt(TestInput.getVertexConfName(
+      TestInput.TEZ_FAILING_INPUT_FAILING_TASK_ATTEMPT, "v3"), 0);
+    inputConf.setInt(TestInput.getVertexConfName(
+      TestInput.TEZ_FAILING_INPUT_FAILING_INPUT_INDEX, "v3"), 0);
+    inputConf.setInt(TestInput.getVertexConfName(
+      TestInput.TEZ_FAILING_INPUT_FAILING_UPTO_INPUT_ATTEMPT, "v3"), 0);
+    UserPayload inputPayload = TezUtils.createUserPayloadFromConf(inputConf);
+    EdgeProperty edgeProperty =
+      EdgeProperty.create(edgeManagerPluginDescriptor, DataMovementType.CUSTOM,
+        DataSourceType.PERSISTED, SchedulingType.SEQUENTIAL, TestOutput.getOutputDesc(null),
+        TestInput.getInputDesc(inputPayload));
+    Edge e1 = Edge.create(v1, v3, edgeProperty);
+    Edge e2 = Edge.create(v2, v3, edgeProperty);
+    dag.addVertex(v1).addVertex(v2).addVertex(v3);
+    dag.addEdge(e1).addEdge(e2);
+
+    // run dag
+    runDAGAndVerify(dag, DAGStatus.State.SUCCEEDED);
+  }
 }
